///////////////////////////////////////////////////////////////////////
//  Copyright (C) ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/CaseInsensitiveString.h,v 1.2 2004/09/07 06:08:57 me133 Exp $
//
//    File name:	CaseInsensitiveString.h
//   Created By:	Marc Eaddy
//   Created on:	3/31/99 2:04:38 PM
//
//      Purpose:	Provide a case-insensitive string.
//
//  Description:	The TypeFuncLessTString typedef defined below can
//					be passed as a template parameter to an STL map
//					so that key lookups will be case-insensitive.  Here
//					is an example that maps case-insensitive strings to
//					variants.
//
//					map<tstring, _variant_t, TypeFuncLessTString >
//
//					So myMap["Marc"] would be the same as myMap["MARC"].
//					This is useful when dealing with XML attributes
//					since XML is case-insensitive.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: CaseInsensitiveString.h,v $
 * Revision 1.2  2004/09/07 06:08:57  me133
 * o Second checkin to CVS
 * 
 * 
 * 7     11/17/00 3:32p Marce
 * o If you don't include the implementation of operator() in the header
 * file, it never gets called!
 * 
 * 6     11/10/00 1:21p Marce
 * o Added comments
 * 
 * 2     3/31/99 7:43p Marce
 * o Optimizations: _bstr_t -> tstring
 */

#ifndef INCLUDED_CASE_INSENSITIVE_STRING_H
#define INCLUDED_CASE_INSENSITIVE_STRING_H

#include <string>		// STL basic_string<>
#include <tchar.h>		// TCHAR
#include <functional>	// STL less<>

// Type defines

typedef std::basic_string<TCHAR>	tstring;
typedef std::less<tstring>			TypeFuncLessTString;

bool TypeFuncLessTString::operator()(const tstring& lhs, const tstring& rhs) const
{
	return _tcsicmp(lhs.c_str(), rhs.c_str()) < 0;
}

#endif // INCLUDED_CASE_INSENSITIVE_STRING_H