///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ComRegistryHelpers.cpp,v 1.1.1.1 2004/08/16 01:46:59 me133 Exp $
//
//    File name:	ComRegistryHelpers.cpp
//   Created By:	Marc Eaddy
//   Created On:	11/17/2000 4:10:01 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ComRegistryHelpers.cpp,v $
 * Revision 1.1.1.1  2004/08/16 01:46:59  me133
 * o Initial checkin
 * 
 * 
 * 2     11/17/00 5:09p Marce
 */

#include "ComRegistryHelpers.h"	// Me

#include <string>				// STD string

#include <stddef.h>				// Needed by atlconv.h
#include <tchar.h>				// _T()
#include <crtdbg.h>				// _ASSERTE()
#include <atlconv.h>			// USES_CONVERSION
#include <shlobj.h>

using namespace std;

typedef basic_string<TCHAR>		tstring;

//*********************************************************************
// FUNCTION		:	GetInterfaceNameForIID
//					
// DESCRIPTION	:	
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   riid			- IID to map
//   pbstrIfaceName	- buf size
//					
// COMMENTS		:	Thanks to Don Box for this function.
//					
//*********************************************************************
HRESULT GetInterfaceNameForIID(	REFIID riid,
							/*[out]*/ BSTR* pbstrIfaceName)
{
	HRESULT hr = S_OK;

	if (!pbstrIfaceName) return E_POINTER;
	
	// Initialize out parameters
	*pbstrIfaceName = 0;

	USES_CONVERSION;

	LPOLESTR pszIID = 0;

	hr = ::StringFromCLSID(riid, &pszIID);
	if (FAILED(hr)) return hr;
	
	TCHAR szIface[255];
	long nSize = sizeof szIface / sizeof *szIface;

	// Step 1:	Try the Interface (IID) key

	HKEY hkey;

	long nSuccess = ::RegOpenKeyEx(	HKEY_CLASSES_ROOT,
									_T("Interface"),
									0, 
									KEY_READ, 
									&hkey);
	if (nSuccess != ERROR_SUCCESS)
	{
		goto cleanup;
	}

	// Read value at corresponding key
	nSuccess = ::RegQueryValue(	hkey, 
								OLE2CT(pszIID), 
								szIface, 
								&nSize);
	::RegCloseKey(hkey);

	_ASSERTE(nSize <= sizeof szIface / sizeof *szIface);

	if (nSuccess == ERROR_SUCCESS)
	{
		(*pbstrIfaceName) = ::SysAllocString(szIface);
		goto cleanup;
	}

	// Step 2:	Try the CLSID (IID) key

	nSuccess = ::RegOpenKeyEx(	HKEY_CLASSES_ROOT,
								_T("CLSID"),
								0, 
								KEY_READ, 
								&hkey);
	if (nSuccess != ERROR_SUCCESS)
	{
		goto cleanup;
	}

	// Read value at corresponding key
	nSuccess = ::RegQueryValue(	hkey, 
								OLE2CT(pszIID), 
								szIface, 
								&nSize);
	::RegCloseKey(hkey);

	if (nSuccess != ERROR_SUCCESS)
	{
		(*pbstrIfaceName) = ::SysAllocString(pszIID);
		goto cleanup;
	}

cleanup:

	::CoTaskMemFree(pszIID);
	pszIID = 0;

	return nSuccess == ERROR_SUCCESS ? S_OK : HRESULT_FROM_WIN32(nSuccess);
}


HRESULT GetPathForCLSID(/*[in]*/  const CLSID& clsid, 
						/*[out]*/ TCHAR* szFilePath, 
						/*[in]*/  long nMaxLen)
{
	HRESULT hr = S_OK;
	LPOLESTR pszIID = 0;

	::StringFromCLSID(clsid, &pszIID);

	tstring strCLSID = __T("CLSID\\");
	strCLSID += OLE2CT(pszIID);
	strCLSID += __T("\\InprocServer32");

	HKEY hkCLSID = NULL;

	hr = ::RegOpenKeyEx(	HKEY_CLASSES_ROOT, 
							strCLSID.c_str(), 
							0, 
							KEY_READ, 
							&hkCLSID);
	if (hr != ERROR_SUCCESS)
	{
		return HRESULT_FROM_WIN32(hr);
	}

	DWORD dwSize = nMaxLen;
	DWORD dwType = 0;

	hr = ::RegQueryValueEx(	hkCLSID, 
							NULL, 
							NULL, 
							&dwType, 
							(unsigned char*)szFilePath, 
							&dwSize);
	if (hr != ERROR_SUCCESS)
	{
		RegCloseKey(hkCLSID); // Close registry key
		return HRESULT_FROM_WIN32(hr);
	}

	// Convert the path to a long file path
	hr = ::MyGetLongPathName(szFilePath, szFilePath, _MAX_PATH);
	if (FAILED(hr)) return hr;

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	MyGetLongPathName
//					
// DESCRIPTION	:	Converts a short path (8.3) name into a long one.
//					
// RETURNS		:	S_OK - Success
//   HRESULT 		Failure code - Otherwise
//					
// PARAMETERS	:	
//   szShortPath	
//   szLongPath		- This can be the same as szShortPath
//   nLength	
//					
// COMMENTS		:	Written 1997 by Jeffrey Richter.  Windows NT
//					does not support GetLongPathName(), only Windows
//					2000.
//					
//*********************************************************************
HRESULT MyGetLongPathName(const TCHAR* szShortPath, TCHAR* szLongPath, size_t nLength)
{
    LPSHELLFOLDER psfDesktop = NULL;
    ULONG chEaten = 0;
    LPITEMIDLIST pidlShellItem = NULL;
	
    // Get the Desktop's shell folder interface
    HRESULT hr = SHGetDesktopFolder(&psfDesktop);
	
    // Request an ID list (relative to the desktop) for the short pathname
    hr = psfDesktop->ParseDisplayName(	NULL, 
										NULL, 
										const_cast<TCHAR*>(szShortPath), 
										&chEaten, 
										&pidlShellItem, 
										NULL);

    psfDesktop->Release();  // Release the desktop's IShellFolder
	psfDesktop = 0;

	if (FAILED(hr)) return hr;

	// We did get an ID list, convert it to a long pathname
	SHGetPathFromIDListW(pidlShellItem, szLongPath);
	
	// Free the ID list allocated by ParseDisplayName
	LPMALLOC pMalloc = NULL;
	SHGetMalloc(&pMalloc);
	pMalloc->Free(pidlShellItem);
	pMalloc->Release();
	pMalloc = 0;
	
	return S_OK;
}
