///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2001 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/GenericXmlHelpers.cpp,v 1.1 2004/09/07 06:10:48 me133 Exp $
//
//  File name:	GenericXmlHelpers.cpp
//  Created By:	Jason Yu
//  Created On:	9/26/2001 2:21:46 PM
//
//  Purpose:	Supports loading exchange-group mapping xml file
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: GenericXmlHelpers.cpp,v $
 * Revision 1.1  2004/09/07 06:10:48  me133
 * o Initial checkin to CVS
 * 
 * 
 * 7     8/16/02 8:09p Zakirh
 * o Clear the Variants inside the loop.
 * 
 * 6     2/25/02 10:51a Jason
 * Check if there are duplicate items in group.xml file and log a warning
 * message when this happens.
 * 
 * 5     12/28/01 9:52a Jason
 * Added GetGroupNamesByGroupType().
 * 
 * 4     10/17/01 12:06p Jason
 * Do some error checking in GetGroupNameByItemName().
 * 
 * 3     10/08/01 12:01p Jason
 * Put in some error checking.
 * 
 * 2     10/03/01 4:48p Jason
 * 1. Changed the function prototype of GetGroupNameByItemName().
 * 2. Make this module more generic.
 * 
 * 1     9/26/01 5:00p Jason
 */


#include "StdAfx.h"
#include "GenericXmlHelpers.h"

#include "ILXLogHelper.h"

#pragma warning(disable:4503)	// decorated name truncated

#include <map>
#include <list>
#include <set>

#ifndef INCLUDED_MSXML_TYPE_LIBRARY
#define INCLUDED_MSXML_TYPE_LIBRARY
#import "msxml.dll" rename_namespace("MSXML"), named_guids, raw_interfaces_only, raw_native_types
#endif // INCLUDED_MSXML_TYPE_LIBRARY

static const TCHAR* const szGROUP_FILE_NAME = _T("group.xml");
static const TCHAR* const szGROUPS          = _T("GROUPS");
static const TCHAR* const szCONTROLS        = _T("CONTROLS");
static const TCHAR* const szGROUP	        = _T("GROUP");
static const TCHAR* const szID  	        = _T("ID");
static const TCHAR* const szTYPE	        = _T("TYPE");
static const TCHAR* const szITEM	        = _T("ITEM");
static const TCHAR* const szNAME	        = _T("NAME");

static const TCHAR* const szLOGFILE_NAME = _T("group.log");

static const int nSEARCH_STRING_MAX_SIZE = 1000;

static const CComBSTR bstrDefaultID = _T("-1");

using namespace MSXML;

typedef std::map<CComBSTR, CComBSTR> Item2GroupMapType;
typedef std::map<CComBSTR, Item2GroupMapType*> ID2GroupMapType;
typedef std::map<short, ID2GroupMapType*> ID2GroupListType;
typedef std::list<short> TypeListIDs;

ID2GroupListType _mapID2GroupList;
ID2GroupMapType _mapGroups;

HRESULT PreLoadGroupInfo(IXMLDOMDocumentPtr spiXmlDoc);
HRESULT LoadGroupItems(short controlID, IXMLDOMDocumentPtr spiXmlDoc);
void DelimitedString(const TCHAR* szBuffer, TypeListIDs* pList, TCHAR cDelimiter);


HRESULT LoadGroupInfoToMap(short controlID, BSTR bstrPath)
{
    HRESULT hr = S_OK;

    // Check if the group info for the specified control is loaded or not
    if (!_mapID2GroupList.empty())
    {
        ID2GroupListType::iterator it = _mapID2GroupList.find(controlID);

        if (it != _mapID2GroupList.end())
        {
            if ((*it).second != NULL && !((*it).second)->empty())
                return S_OK;
        }
        else    // Use default setting
        {
            return S_OK;
        }
    }

    CComBSTR bstrFullGroupFileName(bstrPath);
    if (bstrFullGroupFileName.Length() != 0)
        bstrFullGroupFileName += _T("\\");

    bstrFullGroupFileName += szGROUP_FILE_NAME;

    IXMLDOMDocumentPtr spiGroupXmlDoc = 0;

    // Step 1: Create the XML Document 
    hr = spiGroupXmlDoc.CreateInstance(MSXML::CLSID_DOMDocument);
    if (FAILED(hr))
        return hr;

    hr = spiGroupXmlDoc->put_async(VARIANT_FALSE);
    if (FAILED(hr)) 
        return hr;

    // Step 2: Load and parse the XML file
    VARIANT_BOOL bIsSuccessful = VARIANT_FALSE;
    _variant_t varFileName = bstrFullGroupFileName;
    hr = spiGroupXmlDoc->load(varFileName, &bIsSuccessful);
    if (FAILED(hr)) 
        return hr;

    if (bIsSuccessful == VARIANT_FALSE)
    {
        spiGroupXmlDoc = 0;
        return hr; //should handle error here
    }

    if (_mapID2GroupList.empty())
        hr = PreLoadGroupInfo(spiGroupXmlDoc);

    hr = LoadGroupItems(controlID, spiGroupXmlDoc);

    return hr;
}


HRESULT PreLoadGroupInfo(IXMLDOMDocumentPtr spiXmlDoc)
{
    HRESULT hr = S_OK;

	TCHAR szXslSearch[nSEARCH_STRING_MAX_SIZE];
    _stprintf(szXslSearch, L"/%s/%s", szGROUPS, szCONTROLS);

    // Retrieve all the control nodes
    IXMLDOMNodeListPtr spiNodeList = 0; 
    hr = spiXmlDoc->selectNodes(szXslSearch, &spiNodeList);
    if (FAILED(hr))
        return S_OK;

    long nControlNodes = 0;
    hr = spiNodeList->get_length(&nControlNodes);
    if (FAILED(hr))
        return S_OK;

    IXMLDOMNodePtr spiNode = 0;
    IXMLDOMNamedNodeMapPtr spiNodeAttrMap = 0;
    IXMLDOMNodePtr spiControlIDNode = 0;
    CComVariant varControlIDs;
    TypeListIDs listIDs;

    // Allocate one Item2GroupMapType map for each CONTROLS node
    for (int i = 0; i < nControlNodes; i++)
    {
        hr = spiNodeList->get_item(i, &spiNode);
        if (FAILED(hr))
            return S_OK;

        hr = spiNode->get_attributes(&spiNodeAttrMap);
        if (FAILED(hr))
            return S_OK;

        hr = spiNodeAttrMap->getNamedItem(CComBSTR(szID), &spiControlIDNode);
        spiControlIDNode->get_nodeValue(&varControlIDs);

        _ASSERTE(varControlIDs.vt == VT_BSTR || varControlIDs.vt == VT_EMPTY);
        if (varControlIDs.vt != VT_BSTR)
            continue;

        CComBSTR bstrIDs(varControlIDs.bstrVal);

        if (bstrIDs == bstrDefaultID)
        {
            _mapID2GroupList[_ttol(bstrIDs)] = &_mapGroups;            
            continue;
        }

        listIDs.clear();
        DelimitedString(bstrIDs, &listIDs, _T(','));
        if (listIDs.empty())
            continue;

        // This allocated memory will not be released until GridLayout dll is unloaded.
        ID2GroupMapType* pMap = new ID2GroupMapType;
        TypeListIDs::const_iterator it = listIDs.begin();
        for ( ; it != listIDs.end(); ++it)
        {
            _mapID2GroupList[(*it)] = pMap;
        }
    }

    return hr;
}


HRESULT LoadGroupItems(short controlID, IXMLDOMDocumentPtr spiXmlDoc)
{
    HRESULT hr = S_OK;
	TCHAR szXslSearch[nSEARCH_STRING_MAX_SIZE];
    IXMLDOMNodePtr spiControlsNode = 0;
    Item2GroupMapType* pGroupMap = 0;
    bool bUseDefaultSetting = false;
    bool bLoggingInited = false;
    bool bShutdownLogging = false;

    ID2GroupListType::iterator it = _mapID2GroupList.find(controlID);
    if (it == _mapID2GroupList.end())   // Use default setting
    {
        // If it's already loaded, just return
        if (!_mapGroups.empty())
            return hr;

        _stprintf(szXslSearch, 
            L"/%s/%s[@%s = \"%s\"]",
            szGROUPS,		// "GROUPS" 
            szCONTROLS,		// "CONTROLS"
            szID,		    // "ID"
            bstrDefaultID);

        hr = spiXmlDoc->selectSingleNode(szXslSearch, &spiControlsNode);
        if (FAILED(hr))
            return E_FAIL;

        bUseDefaultSetting = true;
    }
    else    // Find the specific CONTROLS node
    {
        _stprintf(szXslSearch, L"/%s/%s", szGROUPS, szCONTROLS);

        IXMLDOMNodeListPtr spiControlNodes = 0;
        hr = spiXmlDoc->selectNodes(szXslSearch, &spiControlNodes);
        if (FAILED(hr))
            return S_OK;

        long nControlNodes = 0;
        hr = spiControlNodes->get_length(&nControlNodes);
        if (FAILED(hr))
            return S_OK;

        IXMLDOMNodePtr spiNode = 0;
        IXMLDOMNamedNodeMapPtr  spiNodeAttrMap = 0;
        IXMLDOMNodePtr spiControlIDNode = 0;
        CComVariant varControlIDs;
        TypeListIDs listIDs;

        for (int i = 0; i < nControlNodes; i++)
        {
            hr = spiControlNodes->get_item(i, &spiNode);
            if (FAILED(hr))
                return S_OK;

            hr = spiNode->get_attributes(&spiNodeAttrMap);
            if (FAILED(hr))
                return S_OK;

            hr = spiNodeAttrMap->getNamedItem(CComBSTR(szID), &spiControlIDNode);
            spiControlIDNode->get_nodeValue(&varControlIDs);
            if (varControlIDs.vt != VT_BSTR || 
                CComBSTR(varControlIDs.bstrVal) == bstrDefaultID)
                continue;

            DelimitedString(varControlIDs.bstrVal, &listIDs, _T(','));
            if (listIDs.empty())
                continue;

            TypeListIDs::const_iterator it = listIDs.begin();
            for ( ; it != listIDs.end(); ++it)
            {
                if (controlID == (*it))
                    break;
            }

            if (it != listIDs.end())
            {
                spiControlsNode = spiNode;
                break;
            }
        }

        if (spiControlsNode == NULL)
            return E_FAIL;
    }

    _stprintf(szXslSearch, L"%s", szGROUP);

    // Retrieve all the group nodes for the specified type
    IXMLDOMNodeListPtr spiNodeList = 0; 
    hr = spiControlsNode->selectNodes(szXslSearch, &spiNodeList);
    if (FAILED(hr))
        return S_OK;

    long nGroupNodes = 0;
    hr = spiNodeList->get_length(&nGroupNodes);
    if (FAILED(hr))
        return S_OK;

    IXMLDOMNodePtr spiNode = 0;
    IXMLDOMNamedNodeMapPtr  spiNodeAttrMap = 0;
    IXMLDOMNodePtr spiGroupTypeNode = 0;
    CComVariant varGroupType;
    IXMLDOMNodePtr spiGroupNameNode = 0;
    CComVariant varGroupName;
    IXMLDOMNodeListPtr spiItemNodeList = 0;
    IXMLDOMNodePtr spiItemNode = 0;
    IXMLDOMNamedNodeMapPtr  spiItemNodeAttrMap = 0;
    IXMLDOMNodePtr spiItemNameNode = 0;
    CComVariant varItemName;

    //Loop through the matching GROUP node
    for (int i = 0; i < nGroupNodes; i++)
    {
		//clear all variants before use.
		varGroupType.Clear();
		varGroupName.Clear();
		varItemName.Clear();

        hr = spiNodeList->get_item(i, &spiNode);
        if (FAILED(hr))
            return S_OK;

        hr = spiNode->get_attributes(&spiNodeAttrMap);
        if (FAILED(hr))
            return S_OK;

        hr = spiNodeAttrMap->getNamedItem(CComBSTR(szTYPE), &spiGroupTypeNode);
        if (FAILED(hr) || spiGroupTypeNode == NULL)
            continue;

        spiGroupTypeNode->get_nodeValue(&varGroupType);
        if (varGroupType.vt != VT_BSTR)
            continue;

        CComBSTR bstrGroupType(varGroupType.bstrVal);

        if (bUseDefaultSetting)
        {
            if (_mapGroups[bstrGroupType] == NULL)
                pGroupMap = _mapGroups[bstrGroupType] = new Item2GroupMapType;
        }
        else
        {
            if ((*_mapID2GroupList[controlID])[bstrGroupType] == NULL)
                pGroupMap = (*_mapID2GroupList[controlID])[bstrGroupType] = new Item2GroupMapType;
        }

        hr = spiNodeAttrMap->getNamedItem(CComBSTR(szNAME), &spiGroupNameNode);
        if (spiGroupNameNode == NULL)
            continue;

        spiGroupNameNode->get_nodeValue(&varGroupName);
        if (varGroupName.vt != VT_BSTR)
            continue;

        CComBSTR bstrGroup(varGroupName.bstrVal);

        _stprintf(szXslSearch, L"%s", szITEM);
        //Get the item nodes.
        hr = spiNode->selectNodes(szXslSearch,&spiItemNodeList);
        if (FAILED(hr))
            return S_OK;

        long nItemNodes = 0;
        hr = spiItemNodeList->get_length(&nItemNodes);
        if (FAILED(hr))
            return S_OK;

        // Loop through each item node
        for (int j = 0; j < nItemNodes; j++)
        {
            hr = spiItemNodeList->get_item(j, &spiItemNode);
            if (FAILED(hr))
                return S_OK;

            spiItemNode->get_attributes(&spiItemNodeAttrMap);
            if (spiItemNodeAttrMap == NULL)
                continue;

            spiItemNodeAttrMap->getNamedItem(CComBSTR(szNAME), &spiItemNameNode);
            if (spiItemNameNode == NULL)
                continue;

            spiItemNameNode->get_nodeValue(&varItemName);
            if (varItemName.vt == VT_BSTR)
            {
                CComBSTR bstrItem(varItemName.bstrVal);

                if (pGroupMap->count(bstrItem))
                {
                    // Initialize ILX Logger
                    if (!bLoggingInited && !GetLog().isInitialized())
                    {
                        hr = GetLog().initialize(szLOGFILE_NAME);
                        _ASSERTE(SUCCEEDED(hr));

                        CComBSTR LogFileName(szLOGFILE_NAME);
                        hr = GetLog().useLogFile(LogFileName);
                        _ASSERTE(SUCCEEDED(hr));

                        bLoggingInited = bShutdownLogging = true;
                    }

                    CComBSTR bstrMessage(L"Duplicate ");
                    bstrMessage += bstrGroupType;
                    bstrMessage += L": ";
                    bstrMessage += bstrItem;
                    bstrMessage += L" (";
                    bstrMessage += bstrGroup;
                    bstrMessage += L")";
                    ILXLOGMSG(bstrMessage, ILXLOG_WARNING);
                }

                (*pGroupMap)[bstrItem] = bstrGroup;
            }
        }
    }

    if (bShutdownLogging)
        GetLog().shutdown();

    return hr;
}

HRESULT GetGroupNameByItemName(short controlID, BSTR bstrType, BSTR bstrItem, BSTR* pbstrGroup)
{
    HRESULT hr = S_OK;

    if (_mapID2GroupList.empty())
        return E_FAIL;

    Item2GroupMapType* pGroupMap;

    ID2GroupListType::iterator it = _mapID2GroupList.find(controlID);
    if (it != _mapID2GroupList.end())
    {
        pGroupMap = (*(*it).second)[bstrType];
        if (pGroupMap == NULL)
        {
            *pbstrGroup = NULL;
            return E_FAIL;
        }
    }
    else
        pGroupMap = _mapGroups[bstrType];

    if (!pGroupMap)
    {
        *pbstrGroup = NULL;
        return E_FAIL;
    }

    Item2GroupMapType::iterator itItem = pGroupMap->find(bstrItem);
    if (itItem != pGroupMap->end())
    {
       CComBSTR bstrGroupName((*itItem).second);
       *pbstrGroup = bstrGroupName.Detach();
    }
    else
       *pbstrGroup = NULL;

    return hr;
}


HRESULT GetGroupNamesByGroupType(short controlID, BSTR bstrType, BSTR* pbstrGroups)
{
    HRESULT hr = S_OK;

    if (_mapID2GroupList.empty())
        return E_FAIL;

    Item2GroupMapType* pGroupMap;

    ID2GroupListType::iterator it = _mapID2GroupList.find(controlID);
    if (it != _mapID2GroupList.end())
    {
        pGroupMap = (*(*it).second)[bstrType];
        if (pGroupMap == NULL)
        {
            *pbstrGroups = NULL;
            return E_FAIL;
        }
    }
    else
        pGroupMap = _mapGroups[bstrType];

    if (!pGroupMap)
    {
        *pbstrGroups = NULL;
        return E_FAIL;
    }

    std::set<_bstr_t> setNames;
    Item2GroupMapType::iterator itItem = pGroupMap->begin();
    for ( ; itItem != pGroupMap->end(); itItem++)
    {
        _bstr_t bstrGroupName((*itItem).second);
        setNames.insert(bstrGroupName);
    }

    CComBSTR bstrGroups;
    bstrGroups.Empty();

    std::set<_bstr_t>::iterator itName = setNames.begin();
    for ( ; itName != setNames.end(); itName++)
    {
        if (bstrGroups)
            bstrGroups += _T(";");

        bstrGroups += (BSTR)(*itName);
    }

    *pbstrGroups = bstrGroups.Detach();

    return hr;
}


void DelimitedString(const TCHAR* szBuffer, TypeListIDs* pList, TCHAR cDelimiter)
{
	_ASSERTE(szBuffer);
	_ASSERTE(pList);

	TypeListIDs& list = *pList;	// Shortcut

	CComBSTR bstrBuffer	= _T("");
	TCHAR szOneChar[2]	= { 0 };

    USES_CONVERSION;

	for (int i = 0; i < _tcslen(szBuffer); ++i)
	{
		if (szBuffer[i] == cDelimiter)
		{
			list.push_back(_ttoi(bstrBuffer));
			bstrBuffer = _T("");
		}
		else
		{
			szOneChar[0] = szBuffer[i];
			bstrBuffer += szOneChar;
		}
	}

	if (bstrBuffer.Length() > 0)
		list.push_back(_ttoi(bstrBuffer));
}

