///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2001 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/GetErrorMessage.cpp,v 1.2 2004/09/03 16:47:12 me133 Exp $
//
//    File name:	GetErrorMessage.cpp
//   Created By:	Marc Eaddy
//   Created On:	02/27/2001 2:05:58 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: GetErrorMessage.cpp,v $
 * Revision 1.2  2004/09/03 16:47:12  me133
 * o Fixed compile errors
 *
 * Revision 1.1.1.1  2004/08/16 01:46:59  me133
 * o Initial checkin
 * 
 * 
 * 2     2/27/01 2:25p Marce
 * o Moved function into separate file so it can be shared
 */

#include "StdAfx.h"

#include "GetErrorMessage.h"

// HELPER FUNCTIONS

//*********************************************************************
// FUNCTION		:	GetErrorMessage
//					
// DESCRIPTION	:	Looks up the message for the specified HRESULT
//					in the list of resource modules.
//					
// PARAMETERS	:	
//   hrError	
//   dwLanguageId	
//   listResourceModules	
//   pstrMessage	
//					
// COMMENTS		:	First tries to find the HRESULT in the list of
//					resource modules.  If that fails then it tries to
//					get the HRESULT message from the system.  If we
//					still can't find the message then we say "Message
//					not found".
//					
//*********************************************************************
void GetErrorMessage(	HRESULT hrError, 
						DWORD dwLanguageId, 
						const TypeListModules& listResourceModules,
						TCHAR* szErrorMsg,
						long nMaxErrorMsgLen)
{
	// Verify parameters
	_ASSERTE(szErrorMsg);

	TCHAR* pszBuffer = 0;
	bool bGotMessage = false;
	long nUnexpectedError = 0;

	DWORD dwFormat = 0;
	long nLastError = ERROR_MR_MID_NOT_FOUND;

	const DWORD dwMODULE_FLAGS =		FORMAT_MESSAGE_FROM_HMODULE | 
										FORMAT_MESSAGE_ALLOCATE_BUFFER |
										FORMAT_MESSAGE_IGNORE_INSERTS |
										FORMAT_MESSAGE_MAX_WIDTH_MASK;
	
	const DWORD dwSYSTEM_FLAGS =		FORMAT_MESSAGE_FROM_SYSTEM |
										FORMAT_MESSAGE_ALLOCATE_BUFFER |
										FORMAT_MESSAGE_IGNORE_INSERTS |
										FORMAT_MESSAGE_MAX_WIDTH_MASK;

	const DWORD dwSYSTEM_ARY_FLAGS =	FORMAT_MESSAGE_FROM_SYSTEM |
										FORMAT_MESSAGE_ALLOCATE_BUFFER |
										FORMAT_MESSAGE_ARGUMENT_ARRAY |
										FORMAT_MESSAGE_MAX_WIDTH_MASK;

	// Step 1:	Look for the HRESULT in the list of modules

	for(TypeListModules::const_iterator it = listResourceModules.begin();
		it != listResourceModules.end();
		++it)
	{
		HMODULE hModule = (*it).first;
		_ASSERTE(hModule);

		// Load the error message from rdcerrors.dll

		dwFormat = ::FormatMessage(	dwMODULE_FLAGS,
									hModule,
									hrError,
									dwLanguageId,
									(LPTSTR) &pszBuffer,
									0,
									0);	// No argument list
		if (dwFormat > 0)
		{
			bGotMessage = true; // We found the message!
			break;
		}

		// It is normal for the message to not be found in a
		// module.  However, we should remember other errors.
		
		nLastError = ::GetLastError();
		
		if (nLastError != ERROR_MR_MID_NOT_FOUND &&
			nLastError != ERROR_RESOURCE_TYPE_NOT_FOUND &&
			nLastError != ERROR_RESOURCE_NAME_NOT_FOUND &&
			nLastError != ERROR_RESOURCE_DATA_NOT_FOUND)
		{
			_ASSERTE(!"Error retrieving message for HRESULT");
			nUnexpectedError = nLastError;
			break;
		}
	}

	// Step 2:	If unsuccessful, see if the HRESULT refers to a
	//			system error

	if (!bGotMessage && nUnexpectedError == 0)
	{

		if (pszBuffer)
		{
			::LocalFree(pszBuffer); // Just in case
			pszBuffer = 0;
		}

		dwFormat = ::FormatMessage(	dwSYSTEM_FLAGS,
									0,
									hrError,
									dwLanguageId,
									(LPTSTR) &pszBuffer,
									0,
									0);	// No argument list
		if (dwFormat > 0)
		{
			bGotMessage = true;
		}
		else
		{
			// It is normal for the message to not be found.
			// However, we should remember other errors.
			nLastError = ::GetLastError();
			
			if (nLastError != ERROR_MR_MID_NOT_FOUND &&
				nLastError != ERROR_RESOURCE_TYPE_NOT_FOUND &&
				nLastError != ERROR_RESOURCE_NAME_NOT_FOUND &&
				nLastError != ERROR_RESOURCE_DATA_NOT_FOUND)
			{
				_ASSERTE(!"Error retrieving message for HRESULT");
				nUnexpectedError = nLastError;
			}
		}
	}

	// Step 3:	If we cannot retrieve the message at all then find
	//			out why the retrieval failed

	if (!bGotMessage)
	{
		// Failed to retrieve the error message

		if (pszBuffer)
		{
			::LocalFree(pszBuffer); // Just in case
			pszBuffer = 0;
		}

		// Get the error message for why we could not retrieve the module
		// error message and output it to the debug output

		if (nUnexpectedError == 0)
		{
			// By default we assume that none of the modules defined
			// the error message
			nLastError = ERROR_MR_MID_NOT_FOUND;
		}
		else
		{
			// Otherwise, some other type of error occurred (who knows!)
			nLastError = nUnexpectedError;
		}

		// Here is an example of what the message will look like:
		//
		// "The system cannot find message for message number 0xC4010003
		// in message file for application."

		if (nLastError != ERROR_MR_MID_NOT_FOUND)
		{
			// Get the error message

			dwFormat = ::FormatMessage(	dwSYSTEM_ARY_FLAGS,
										0,
										nLastError,
										dwLanguageId,
										(LPTSTR) &pszBuffer,
										0,
										0);
			
			// Sanity Check: This should never fail because we are
			// retrieving a known system message!
			_ASSERTE(dwFormat != 0);
		}
		else
		{
			// Force the "[Unknown error code]" message to be used
			bGotMessage = true;
		}
	}

	// Step 4:	Truncate the error message if needed

	if (pszBuffer && (int) _tcslen(pszBuffer) > nMaxErrorMsgLen - 50)
	{
		// Truncate the error message
		_tcscpy(pszBuffer + nMaxErrorMsgLen - 50, _T("[...]"));
	}

	// Step 5:	Format the error message

	if (!bGotMessage)
	{
		_stprintf(	szErrorMsg, 
					_T("%s"), 
					pszBuffer ? pszBuffer : _T("Unknown"));
	}
	else
	{
		_stprintf(	szErrorMsg, 
					_T("%s (0x%08X)"), 
					pszBuffer ? pszBuffer : _T("[Unknown error code]"), 
					hrError);
	}

	if (pszBuffer)
	{
		::LocalFree(pszBuffer); // Free the buffer
		pszBuffer = 0;
	}
}
