///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXCrashFinder/ILXCrashFinder.h,v 1.1 2004/09/07 06:08:22 me133 Exp $
//
//    File name:	ILXCrashFinder.h
//   Created By:	Zakir Hussain
//   Created On:	7/17/2002 1:28:05 PM
//
//      Purpose:	To perform a stack trace when exception occurs.
//
//  Description:	
//					This DLL is based on the book "Debugging Applications"
//					by John Robbins. Most of the code seen here have been
//					lifted from the BugSlayerUtil.DLL 
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ILXCrashFinder.h,v $
 * Revision 1.1  2004/09/07 06:08:22  me133
 * o Initial checkin to CVS
 * 
 * 
 * 11    8/05/02 3:31p Zakirh
 * o now uses DbgHelp.dll instead of imageHlp.dll
 * o included all the review feedback
 * o Added support to detect older and newer pdb 
 * o code cleanup
 * 
 * 10    7/26/02 2:24p Zakirh
 * o Removed unused error codes.
 * 
 * 9     7/25/02 3:58p Marce
 * o Fixed bug where exception pointer was being modified by _StackWalk()
 * call.  This prevented us from calling GetExceptionInfo() twice with the
 * same ptr.
 * o Made argument const
 * 
 * 8     7/25/02 3:01p Marce
 * o Made argument const
 * 
 * 7     7/25/02 2:02p Zakirh
 * o Added comments and code cleanup.
 * 
 * 6     7/24/02 6:32p Marce
 * o Comment change
 * 
 * 5     7/24/02 2:48p Zakirh
 * o Code CleanUp and New structure format as per marc's request.
 * 
 * 4     7/18/02 2:39p Zakirh
 * added address to the Stack Frame.
 * 
 * 3     7/17/02 9:51p Zakirh
 * o changed the structure name to SIlxStackFrames from SIlxStackFrame
 * and removed Inc from the preprocessor directive.
 * 
 * 2     7/17/02 7:00p Zakirh
 * o New API Definition now uses Structure.
 * 
 * 1     7/17/02 5:04p Zakirh
 * back to project.
 */


#ifndef __ILXCrashFinder_H__
#define __ILXCrashFinder_H__

// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the ILXCRASHFINDER_EXPORTS
// symbol defined on the command line. this symbol should not be defined on any project
// that uses this DLL. This way any other project whose source files include this file see 
// ILXCRASHFINDER_API functions as being imported from a DLL, wheras this DLL sees symbols
// defined with this macro as being exported.
#ifdef ILXCRASHFINDER_EXPORTS
#define ILXCRASHFINDER_API __declspec(dllexport)
#else
#define ILXCRASHFINDER_API __declspec(dllimport)
#endif

//constants
#define	MAX_FUNCTION_NAME_LEN		MAX_PATH*4
//end constants

//CUSTOM ERROR CODES RETURNED 

#define			DBG_HLP_DLL_NOTFOUND					31310
#define			SYM_INIT_FAILED							31311
#define			STACKWALK_FAILED						31312

// END CUSTOM ERROR CODES 
	

/*
	SAMPLE CODE TO CALL THIS API

	Client code would like some what like this.

	try
	{
		*((char *)0) = 0; //will cause access violation
	}
	catch (unsigned code, EXCEPTION_POINTERS* pInfo)
	{
		//making this static will allocate memory on the heap rather than on the stack
		static PILXStackFrame paryStackFrames[10]; 
		GetExceptionInfo( pInfo, _T("c:\\3.8\\bin"), paryStackFrames, 10 );
  	}

	o make judicious use of stack length don't set to huge values like 50 or 100 
	o optimal value will be in the range 10 - 20.
	o in case of stack coruption ( check the code value ) set the stack size to small
	value and log the exception and get out as quickly as possible so that at least
	the exception is logged.

	for more detailed and exact client code look at ILXLOG::LogCrash methods 


*/

//ideally this method must be invoked from your exception handler 
//thats when you can get EXECPTION_POINTERS structure 

typedef enum
{
	enumPdbInfoNotAvailable = -999,//info not available or was a MFC or ystem DLL
	enumPdbInfoSame			= 0, //pdb is in sync with the DLL
	enumPdbInfoNewer		= 1,//pdb is newer than the dll loaded
	enumPdbInfoOlder		= -1//pdb is older than the dll loaded

}PDBINFO;

//this structure is used by the API.
typedef struct SIlxStackFrames
{
	//constructor initialisations 
	SIlxStackFrames() :
		lLineNumber(0),
		lColNumber(0),
		dwAddress(0),
		enumPdbInfo(enumPdbInfoNotAvailable)
	{	
		//set all strings to null
		szModuleName[0]		= TCHAR('\0');
		szFilePath[0]		= TCHAR('\0');
		szFunctionName[0]	= TCHAR('\0');
	}

	DWORD   dwAddress;								//Stack Address
	TCHAR   szModuleName[MAX_PATH];					//Module Name
	TCHAR   szFilePath[MAX_PATH];					//File Path 
	TCHAR   szFunctionName[MAX_FUNCTION_NAME_LEN];	//Function Name apporx 1024
	long	lLineNumber;							//Line number
	long	lColNumber;								//Column Number
	PDBINFO enumPdbInfo;							//info about pdb 

}ILXStackFrame, *PILXStackFrame;

ILXCRASHFINDER_API HRESULT GetExceptionInfo (	
	/*[in]*/     const EXCEPTION_POINTERS*	pExceptionInfo,//Execption Information 
	/*[in]*/     const TCHAR*				szSearchPath,//Search Path Cannot be NULL 
	/*[in,out]*/ PILXStackFrame				paryStackFrames,//array of stack frames
	/*[in]*/     long						nStackFrames);//Stack Depth 

#endif // __ILXCrashFinder_H__