///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/BeepDestination.cpp,v 1.2 2004/09/07 06:06:03 me133 Exp $
//
//    File name:	BeepDestination.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/05/2000 1:34:26 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: BeepDestination.cpp,v $
 * Revision 1.2  2004/09/07 06:06:03  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 9     7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 6     1/30/01 10:51a Marce
 * o Fixed compile errors for Debug build
 * 
 * 4     10/25/00 4:37p Marce
 * o Comments
 * 
 * 3     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 2     10/05/00 1:34p Marce
 * o Added Beep destination
 */

#include "StdAfx.h"

#include "BeepDestination.h"	// Me

#include "ILXLog2.h"			// ILXLOG__SUCCESS, etc.

// Constants

static const TCHAR* const szSUB_TYPE				= _T("Beep");
static const TCHAR* const szBEEP_TYPE_ATTR			= _T("BEEP_TYPE");
static const wchar_t* const szBEEP_TYPE_ASTERISK	= L"asterisk";
static const wchar_t* const szBEEP_TYPE_EXCLAMATION	= L"exclamation";
static const wchar_t* const szBEEP_TYPE_HAND		= L"hand";
static const wchar_t* const szBEEP_TYPE_OK			= L"ok";
static const wchar_t* const szBEEP_TYPE_QUESTION	= L"question";
static const wchar_t* const szBEEP_TYPE_DEFAULT		= L"default";
static const wchar_t* const szBEEP_TYPE_SEVERITY	= L"severity";

static const long nBASED_ON_SEVERITY = 0xBEEFBEEF;

////////////////////////////////////////
// LOG FILE DESTINATION IMPLEMENTATION
////////////////////////////////////////

CBeepDestination::CBeepDestination()
	: CLogDestination(szSUB_TYPE)
	, m_nBeepType(nBASED_ON_SEVERITY)
{ }

CBeepDestination::~CBeepDestination()
{ }

CXmlElement* CBeepDestination::clone() const
{
	return new CBeepDestination(); // We create a new object not a clone
}

//*********************************************************************
// FUNCTION		:	CBeepDestination::initialize
//					
// DESCRIPTION	:	Initializes the beep type member variable based on
//					the BEEP_TYPE XML attribute.
//					
// RETURNS		:	
//   HRESULT 		
//					
//*********************************************************************
HRESULT CBeepDestination::initialize()
{
	HRESULT hr = S_OK;

	// Invariants: Only call initialize() once
	_ASSERTE(!m_bIsInitialized);

	// Step 1:	Initialize member variables

	const _variant_t& varBeepType = m_mapAttributes[szBEEP_TYPE_ATTR];
	if (varBeepType.vt != VT_BSTR || !*varBeepType.bstrVal)
	{
		m_nBeepType = nBASED_ON_SEVERITY; // Default
	}
	else
	{
		if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_ASTERISK) == 0)
		{
			m_nBeepType = MB_ICONASTERISK;
		}
		else if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_EXCLAMATION) == 0)
		{
			m_nBeepType = MB_ICONEXCLAMATION;
		}
		else if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_HAND) == 0)
		{
			m_nBeepType = MB_ICONHAND;
		}
		else if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_OK) == 0)
		{
			m_nBeepType = MB_OK;
		}
		else if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_QUESTION) == 0)
		{
			m_nBeepType = MB_ICONQUESTION;
		}
		else if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_DEFAULT) == 0)
		{
			m_nBeepType = -1;
		}
		else if (wcsicmp(varBeepType.bstrVal, szBEEP_TYPE_SEVERITY) == 0)
		{
			m_nBeepType = nBASED_ON_SEVERITY;
		}
		else
		{
			_ASSERTE(!"Unrecognized beep type!");
			m_nBeepType = nBASED_ON_SEVERITY; // Default
		}
	}

	m_bIsInitialized = true;

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CBeepDestination::log
//					
// DESCRIPTION	:	Pushes the concept of "log" to a whole other
//					level!
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	See header file.
//					
//*********************************************************************
HRESULT CBeepDestination::log(	const TCHAR* szSubsystem, 
								const TCHAR* szPath,
								const TCHAR* szFilename, 
								long nLine, 
								long nThreadId, 
								long hrErrorCode,
								SIlxStackFrames* pvIlxStackFrames,
								long nNumStackFrames,
								const TCHAR* szMessage, 
								long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions
	if (!m_bIsInitialized)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	// Step 1:	Figure out what beep to use

	long nBeepType = m_nBeepType;

	if (nBeepType == nBASED_ON_SEVERITY)
	{
		if (nSeverity == ILXLOG__SUCCESS)
		{
			nBeepType = MB_OK;
		}
		else if (nSeverity == ILXLOG__INFO)
		{
			nBeepType = MB_OK;
		}
		else if (nSeverity == ILXLOG__WARNING)
		{
			nBeepType = MB_ICONEXCLAMATION;
		}
		else if (nSeverity == ILXLOG__ERROR)
		{
			nBeepType = MB_ICONEXCLAMATION;
		}
		else
		{
			_ASSERTE(!"Unknown severity");
			nBeepType = MB_ICONEXCLAMATION;
		}
	}

	// Step 2:	Beep!

	BOOL bReturn = MessageBeep(nBeepType);
	_ASSERTE(bReturn);

	return S_OK;
}
