///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/BeepDestination.h,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	BeepDestination.h
//   Created By:	Marc Eaddy
//   Created On:	10/05/2000 1:34:35 PM
//
//      Purpose:	Beeps when log() is called.
//
//  Description:	This is helpful when you want to be notified when
//					certain messages (e.g., errors) are logged but you
//					don't want to keep seeing message boxes.
//
//   Xml Format:	
//					Example:
//
//					<DEST TYPE="Beep" BEEP_TYPE="default"/>
//
//					Attributes:                                            
//                                                           
//					BEEP_TYPE - [Optional] One of "asterisk", "exclamation",
//						"hand", "ok", "question", "default" or "severity".
//						Plays the standards Windows sounds as described in
//						the MessageBeep() documentation.  However, "severity"
//						(the default) causes the sound to be based on the
//						message severity.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: BeepDestination.h,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 8     7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 4     10/25/00 4:34p Marce
 * o Comments
 * 
 * 3     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 2     10/05/00 1:34p Marce
 * o Added Beep destination
 */

#ifndef INCLUDED_BEEP_DESTINATION_H
#define INCLUDED_BEEP_DESTINATION_H

#include "LogDestination.h"	// Parent

//*****************************
// CBeepDestination
//*****************************

class CBeepDestination : public CLogDestination
{
public:

	CBeepDestination();
	virtual ~CBeepDestination();

public:

	virtual HRESULT initialize();
		//
		// See LogDestination.h for a description
		//

	virtual HRESULT log(const TCHAR* szSubsystem, 
						const TCHAR* szPath,
						const TCHAR* szFilename, 
						long nLine, 
						long nThreadId, 
						long hrErrorCode,
						SIlxStackFrames* pvIlxStackFrames,
						long nNumStackFrames,
						const TCHAR* szMessage, 
						long nSeverity);
		//
		// See LogDestination.h for a description
		//

protected:
	// Overrides CXmlElement
	virtual CXmlElement* clone() const;
		//
		// Clones this element.
		//

private: // NOT IMPLEMENTED

	CBeepDestination(const CBeepDestination&);
	CBeepDestination& operator=(const CBeepDestination&);

private:	// DATA
	long	m_nBeepType;
};

#endif // INCLUDED_BEEP_DESTINATION_H