///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/DebugOutDestination.cpp,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	DebugOutDestination.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/04/2000 2:29:21 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: DebugOutDestination.cpp,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 12    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 9     11/21/01 4:12p Zakirh
 * merged in with marc's hopewell changes....
 * 
 * 8     1/30/01 7:48p Marce
 * o Now uses OutputDebugString() so that debug messages aren't lost just
 * because IlxLog.dll is a release version
 * 
 * 6     10/25/00 4:41p Marce
 * o Comments
 * 
 * 5     10/25/00 3:41p Marce
 * o Added GetSeverity()
 * 
 * 4     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 3     10/05/00 1:35p Marce
 * o Removed unneeded initialize() method
 * 
 * 2     10/04/00 5:38p Marce
 * o Implemented log filtering
 */

#include "StdAfx.h"

#include "DebugOutDestination.h"	// Me

#include "HelperFunctions.h"		// GetSeverity()

// Constants

static const TCHAR* const szSUB_TYPE = _T("DebugOut");

static const long nONE_LOG_LINE_MAX_LENGTH	= 4096;		// Characters

////////////////////////////////////////
// LOG FILE DESTINATION IMPLEMENTATION
////////////////////////////////////////

CDebugOutDestination::CDebugOutDestination()
	: CLogDestination(szSUB_TYPE)
{ }

CDebugOutDestination::~CDebugOutDestination()
{ }

CXmlElement* CDebugOutDestination::clone() const
{
	return new CDebugOutDestination(); // We create a new object not a clone
}

//*********************************************************************
// FUNCTION		:	CDebugOutDestination::log
//					
// DESCRIPTION	:	Sends message to debug output window.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	See header file.
//					
//*********************************************************************
HRESULT CDebugOutDestination::log(	const TCHAR* szSubsystem, 
									const TCHAR* szPath,
									const TCHAR* szFilename, 
									long nLine, 
									long nThreadId, 
									long hrErrorCode,
									SIlxStackFrames* pvIlxStackFrames,
									long nNumStackFrames,
									const TCHAR* szMessage, 
									long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions
	if (!m_bIsInitialized)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	const TCHAR* szSeverity = ::GetSeverity(nSeverity);
	_ASSERTE(szSeverity && *szSeverity);

	// Format the trace message to allow the developer to double-click on the
	// message in the debug window and go straight to the line in source code.
	//
	// Looks like:
	// D:\PROJECTS\ActiveILX\IlxScripting\IlxScriptingImpl.cpp(1524) : INFO -
	//		ILXScripting: Starting event dispatcher thread.
	//

	TCHAR szMsg[nONE_LOG_LINE_MAX_LENGTH + 20]; // Add 20 for " (TID: %d)\n"
	TCHAR szFirstPart[nONE_LOG_LINE_MAX_LENGTH+1];

	_sntprintf(szFirstPart, nONE_LOG_LINE_MAX_LENGTH, _T("%s(%d) : %s - %s "), 
		szPath,
		nLine,
		szSeverity,
		szSubsystem);

	szFirstPart[nONE_LOG_LINE_MAX_LENGTH] = 0;

	long nLenLeft = nONE_LOG_LINE_MAX_LENGTH - _tcslen(szFirstPart);

	if (nLenLeft > 0)
	{
		_tcsncat(szFirstPart, szMessage, nLenLeft);
		szFirstPart[nONE_LOG_LINE_MAX_LENGTH] = 0;
	}

	_stprintf(szMsg, _T("%s (TID: %d)\n"), 
		szFirstPart,
		nThreadId);

	OutputDebugString(szMsg);

	return S_OK;
}
