///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/DebugOutDestination.h,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	DebugOutDestination.h
//   Created By:	Marc Eaddy
//   Created On:	10/04/2000 2:27:28 PM
//
//      Purpose:	Outputs the log message in the debug window.
//
//  Description:	The log message is formatted so that you can double-
//					click on it in the debug window and it will take to
//					the appropriate source file and line.  Here is an
//					example:
//
//					D:\PROJECTS\ActiveILX\ILXScripting\ScriptingSite.cpp(576) :
//						INFO - ILXScripting Adding object to script namespace: 
//						"logger" (TID: 65)
//
//   Xml Format:	<DEST TYPE="DebugOut"/>
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: DebugOutDestination.h,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 9     7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 5     10/25/00 4:41p Marce
 * o Comments
 * 
 * 4     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 3     10/05/00 1:35p Marce
 * o Removed unneeded initialize() method
 * 
 * 2     10/04/00 5:38p Marce
 * o Implemented log filtering
 */

#ifndef INCLUDED_DEBUG_OUT_DESTINATION_H
#define INCLUDED_DEBUG_OUT_DESTINATION_H

#include "LogDestination.h"	// Parent

//*****************************
// CDebugOutDestination
//*****************************

class CDebugOutDestination : public CLogDestination
{
public:

	CDebugOutDestination();
	virtual ~CDebugOutDestination();

public:

	virtual HRESULT log(	const TCHAR* szSubsystem, 
							const TCHAR* szPath,
							const TCHAR* szFilename, 
							long nLine, 
							long nThreadId, 
							long hrErrorCode,
							SIlxStackFrames* pvIlxStackFrames,
							long nNumStackFrames,
							const TCHAR* szMessage, 
							long nSeverity);
		//
		// See LogDestination.h for a description.
		//

protected:
	// Overrides CXmlElement
	virtual CXmlElement* clone() const;
		//
		// Clones this element.
		//

private: // NOT IMPLEMENTED

	CDebugOutDestination(const CDebugOutDestination&);
	CDebugOutDestination& operator=(const CDebugOutDestination&);
};

#endif // INCLUDED_DEBUG_OUT_DESTINATION_H