///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/ErrorDialog.cpp,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	ErrorDialog.cpp
//   Created By:	Marc Eaddy
//   Created on:	04/20/2000 4:32:04 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ErrorDialog.cpp,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 22    8/30/02 9:15a Petert
 * Added email functionality
 * 
 * 21    8/22/02 5:27p Marce
 * o Changed "Thomson ONE" -> "ActiveILX"
 * 
 * 20    7/26/02 3:45p Zakirh
 * o set the max per line buffer length consistent with rest of the code.
 * 
 * 19    7/18/02 6:42p Zakirh
 * o New format specifications.
 * 
 * 18    7/18/02 2:53p Zakirh
 * o fixed the scrolling problem and added address.
 * 
 * 17    7/18/02 12:08a Zakirh
 * integration with crashfinder.
 * 
 * 16    7/17/02 3:49p Marce
 * o Now passes stack frames to error dialog
 * 
 * 12    11/02/00 11:15a Marce
 * o Minor change in wording
 * 
 * 11    11/02/00 11:14a Marce
 * o Changed error dialog wording
 * 
 * 10    11/02/00 11:08a Marce
 * o Error dialog is now topmost window
 * 
 * 9     10/13/00 7:17p Marce
 * o Now displays the correct severity icon
 * 
 * 8     10/13/00 6:18p Marce
 * o Title is more informative
 * 
 * 7     10/13/00 6:12p Marce
 * o Now error dialog is System Modal to prevent the freezing problem
 * 
 * 6     10/04/00 5:40p Marce
 * o Added extra error checking
 * 
 * 5     8/14/00 5:07p Marce
 * o Changed enum names so they would be visible from script
 * 
 * 4     6/22/00 7:51p Marce
 * o Fixed display problems
 * 
 * 3     4/26/00 8:59p Marce
 * o Added the Debug button
 * 
 * 2     4/26/00 1:41a Marce
 * o Now displays severity icon (does not work) and subsystem
 * 
 * 1     4/24/00 11:27a Marce
 */

#include "StdAfx.h"			 // Pch 
#include "ErrorDialog.h"	 // CErrorDialog
#include "IlxLog2.h"		 // ILXLOG__ERROR
#include "HelperFunctions.h" // GetILXDirectoryFromRegistry

using namespace std;

// Constants

//static const bool bHIDE_DETAILS = false;
//static const bool bSHOW_DETAILS = true;

//static const CComBSTR bstrSHOW_DETAILS = L"More Details >>";
//static const CComBSTR bstrHIDE_DETAILS = L"Hide Details <<";
static const long nMAX_BUFFER_LENGTH   = 4096;
static const TCHAR* szLogViwer_SendMail	= _T("ILXLogViewer.exe email");
static const wchar_t* strKeysz_Install_Directory =L"Install Directory";

////////////////////////////////
// ERROR DIALOG IMPLEMENTATION
////////////////////////////////

CErrorDialog::CErrorDialog()
	: m_strErrMsg()
	//, m_strDetailMsg()
	, m_strSubsystem()
	, m_nSeverity(ILXLOG__ERROR)
	//, m_bShowDetails(true)
	, m_bDebugBtnClicked(false)
	//, m_nWidth(-1)
	//, m_nFullHeight(-1)
	//, m_pIlxStackFrames(NULL)
	//, m_nNumStackFrames(0)
{ }

CErrorDialog::~CErrorDialog()
{ }

/*
void CErrorDialog::init(const TCHAR*		szErrMsg, 
						SIlxStackFrames*	pvIlxStackFrames,
						long				nNumStackFrames,
						const TCHAR*		szSubsystem, 
						long				nSeverity)
{
	_ASSERTE(szErrMsg);
	_ASSERTE(szSubsystem);

	m_strErrMsg				=	szErrMsg;
	m_strDetailMsg			=	_T("No Details");
	m_strSubsystem			=	szSubsystem;
	m_nSeverity				=	nSeverity;
	m_nNumStackFrames		=	nNumStackFrames;
	m_pIlxStackFrames		=	pvIlxStackFrames;
}
*/

void CErrorDialog::init(const TCHAR*		szErrMsg, 
						 const TCHAR*		szSubsystem, 
						 long				nSeverity)
{
	_ASSERTE(szErrMsg);
	_ASSERTE(szSubsystem);

	m_strErrMsg				=	szErrMsg;
	m_strSubsystem			=	szSubsystem;
	m_nSeverity				=	nSeverity;
}



LRESULT CErrorDialog::OnInitDialog(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
	USES_CONVERSION;

	CWindow wndErrMsg(GetDlgItem(IDC_ERR_MSG));
	//CWindow wndDetailMsg(GetDlgItem(IDC_DETAIL_MSG)); //list box 
	//CWindow wndDetailBtn(GetDlgItem(IDC_DETAIL_BTN)); //button 
	CWindow wndSeverityIcon(GetDlgItem(IDC_SEVERITY_ICON));

	#ifdef _DEBUG
	CWindow wndDebugBtn(GetDlgItem(IDC_DEBUG_BTN));
	wndDebugBtn.ShowWindow(SW_SHOW);
	#endif

	const TCHAR* szSeverityHdr = 0;
	const TCHAR* szSeverityTitle = 0;

	switch (m_nSeverity)
	{
	case ILXLOG__ERROR: 
		szSeverityHdr = _T("An error");
		szSeverityTitle = _T("Error");
		break;
	case ILXLOG__WARNING:
		szSeverityHdr = _T("A warning");
		szSeverityTitle = _T("Warning");
		break;
	case ILXLOG__INFO:
		szSeverityHdr = _T("An information message");
		szSeverityTitle = _T("Information");
		break;
	case ILXLOG__SUCCESS:
		szSeverityHdr = _T("A success message");
		szSeverityTitle = _T("Success");
		break;
	default:
		_ASSERTE(!"Unknown severity");
		szSeverityHdr = _T("Unknown severity");
		szSeverityTitle = _T("Unknown severity");
	};

	static const tstring strMSG = _T(" was reported by ");
	static const tstring strTITLE = _T("Thomson ONE ");

	tstring strHeader = szSeverityHdr + strMSG + _T("Thomson ONE") + _T(":");

	CWindow wndErrHdr(GetDlgItem(IDC_ERR_HEADER));
	wndErrHdr.SetWindowText(strHeader.c_str());

	tstring strTitle = strTITLE + szSeverityTitle;
	SetWindowText(strTitle.c_str());

	wndErrMsg.SetWindowText(m_strErrMsg.c_str());

	//wndDetailMsg.SetWindowText(m_strDetailMsg.c_str());
	//RefreshListBox();

	//if ( !m_pIlxStackFrames )
	//{
		//if no stack frame pointer hide the Details button
		//wndDetailBtn.ShowWindow( SW_HIDE );	
	//}
	//else
	//{
		//display the "Details" text on the button 
		//wndDetailBtn.SetWindowText(OLE2CT(bstrSHOW_DETAILS));

		//also display the text in the list box
	//}

	//RECT rcWnd;
	//BOOL bRet = GetClientRect(&rcWnd);
	//_ASSERTE(bRet);

	//m_nFullHeight = rcWnd.bottom;
	//m_nWidth = rcWnd.right;

	//if (m_strDetailMsg.size() == 0)
	//{
	//	wndDetailBtn.EnableWindow(FALSE);
	//}

	//m_bShowDetails = true; // The details are not hidden yet
	//hideOrShowDetails(bHIDE_DETAILS);

    wndSeverityIcon.GetWindowRect(&m_rectSeverityIcon);
    wndSeverityIcon.DestroyWindow();
    
	ScreenToClient(&m_rectSeverityIcon);

	BOOL bRet = ::SetWindowPos(m_hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOSIZE|SWP_NOMOVE);
	_ASSERTE(bRet);

	return 1;  // Let the system set the focus
}


LRESULT CErrorDialog::OnPaint(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
	const TCHAR* pIcon = 0;

	switch (m_nSeverity)
	{
	case ILXLOG__ERROR: 
		pIcon = IDI_ERROR;
		break;
	case ILXLOG__WARNING:
		pIcon = IDI_WARNING;
		break;
	case ILXLOG__INFO:
		pIcon = IDI_INFORMATION;
		break;
	case ILXLOG__SUCCESS:
		pIcon = IDI_INFORMATION;
		break;
	default:
		_ASSERTE(!"Unknown severity");
		pIcon = 0;
		return 1;
	};

	HICON hIcon = ::LoadIcon(0, pIcon);

	if (!hIcon)
	{
		_ASSERTE(!"Failed to load icon");
		return 1;
	}

	PAINTSTRUCT paintstru;
	HDC hdc = ::BeginPaint(m_hWnd, &paintstru);
	_ASSERTE(hdc);

	int nRet = ::DrawIcon(	hdc, 
							m_rectSeverityIcon.left, 
							m_rectSeverityIcon.top, 
							hIcon);
	if (nRet == 0)
	{
		long nLastError = ::GetLastError();
		_ASSERTE(!"DrawIcon() failed!");
	}

	::EndPaint(m_hWnd, &paintstru);

	return 1;
}

LRESULT CErrorDialog::OnOK(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled)
{
	EndDialog(wID);
	return 0;
}

LRESULT CErrorDialog::OnDebugBtnClick(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled)
{
	m_bDebugBtnClicked = true;
	EndDialog(wID);
	return 0;
}

bool CErrorDialog::debugBtnClicked() const
{
	return m_bDebugBtnClicked;
}

LRESULT CErrorDialog::OnDetailBtnClick(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled)
{
	//CWindow wndDetailBtn(GetDlgItem(IDC_DETAIL_BTN));

	//Email Functionality 
	BOOL bRet = ::SetWindowPos(m_hWnd, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOSIZE|SWP_NOMOVE);
	_ASSERTE(bRet);
	if(!LaunchLogviewer())
	{
		_ASSERTE(!"Failed to Launch LogViewer");
		return 1;
	}
	//PETER TAKE OVER HERE 

	return 0;
}

bool CErrorDialog::LaunchLogviewer() const
{
	USES_CONVERSION;

	wchar_t szInstallDirectory[_MAX_PATH+1];
	*szInstallDirectory = _T('\0');
	HRESULT hr  = ::GetILXDirectoryFromRegistry( W2CT(strKeysz_Install_Directory),
										szInstallDirectory,
										_MAX_PATH);
	if (FAILED(hr)) return false;

	_bstr_t bstrLogViewerSendMail = szInstallDirectory;

	if (bstrLogViewerSendMail.length() > 0)
	{
		bstrLogViewerSendMail += _T("\\");
	}
	
	bstrLogViewerSendMail += szLogViwer_SendMail;

	STARTUPINFO sui;
    ZeroMemory(&sui, sizeof(sui));
    sui.cb = sizeof (STARTUPINFO);

    PROCESS_INFORMATION pi;
    ZeroMemory(&pi, sizeof(pi));

	BOOL fRet = ::CreateProcess(NULL,
								bstrLogViewerSendMail,
								NULL,
								NULL,
								FALSE,
								0,
								NULL,
								NULL,
								&sui,
								&pi);
	if (!fRet)
    	return false;
    else
	{
		_ASSERTE(pi.hProcess);
		_ASSERTE(pi.hThread);

		::CloseHandle(pi.hProcess);
		::CloseHandle(pi.hThread);
		return true;
	}
}

/*
LRESULT CErrorDialog::OnDetailBtnClick(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled)
{
	CWindow wndDetailBtn(GetDlgItem(IDC_DETAIL_BTN));

	CComBSTR bstrText;

	BOOL bRet = wndDetailBtn.GetWindowText(&bstrText);
	_ASSERTE(bRet);

	if (wcscmp(bstrText, bstrSHOW_DETAILS) == 0)
	{
		hideOrShowDetails(bSHOW_DETAILS);
	}
	else
	{
		hideOrShowDetails(bHIDE_DETAILS);
	}

	return 0;
}

void CErrorDialog::hideOrShowDetails(bool bShow)
{
	USES_CONVERSION;

	if (m_bShowDetails == bShow)
	{
		return;
	}

	CWindow wndDetailMsg(GetDlgItem(IDC_DETAIL_MSG));
	RECT rectDetails;
	wndDetailMsg.GetWindowRect(&rectDetails);

	POINT ptDetails = { rectDetails.left, rectDetails.top };
	ScreenToClient(&ptDetails);

	CWindow wndDetailBtn(GetDlgItem(IDC_DETAIL_BTN));

	if (bShow)
	{
		BOOL bRet = ResizeClient(m_nWidth, m_nFullHeight);
		_ASSERTE(bRet);
		wndDetailBtn.SetWindowText(OLE2CT(bstrHIDE_DETAILS));
		wndDetailMsg.ShowWindow( SW_SHOW );
		RefreshListBox();
	}
	else // Hide
	{
		BOOL bRet = ResizeClient(m_nWidth, ptDetails.y);
		_ASSERTE(bRet);
		wndDetailBtn.SetWindowText(OLE2CT(bstrSHOW_DETAILS));
		wndDetailMsg.ShowWindow( SW_HIDE );
	}

	m_bShowDetails = bShow;
}

void CErrorDialog::RefreshListBox()
{
	CWindow wndDetailsMsg(GetDlgItem(IDC_DETAIL_MSG)); //list box 

	//first clear the list box
	::SendMessage( wndDetailsMsg.m_hWnd , LB_RESETCONTENT, (WPARAM)0, (LPARAM)0 );

	//now add the strings
	TCHAR szBuffer[nMAX_BUFFER_LENGTH+1];
	for ( int nIndex =0; nIndex<m_nNumStackFrames; nIndex++ )
	{
		//empty the buffer
		::ZeroMemory( szBuffer, sizeof( szBuffer ) );

		BOOL bRet = FormatStackString2( &m_pIlxStackFrames[nIndex], szBuffer, nMAX_BUFFER_LENGTH);
		if ( bRet && ( _tcslen( szBuffer ) > 0 ) )
		{
			//send the message to the list box
			long lResult =  ::SendMessage( wndDetailsMsg.m_hWnd, LB_INSERTSTRING, (WPARAM)nIndex, (LPARAM)(LPCTSTR)szBuffer );

		}
	}
	::SendMessage( wndDetailsMsg.m_hWnd, LB_SETHORIZONTALEXTENT, (WPARAM)nMAX_BUFFER_LENGTH, (LPARAM)0 );
}
*/
