///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/ErrorDialog.h,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	ErrorDialog.h
//   Created By:	Marc Eaddy
//   Created on:	04/20/2000 4:34:04 PM
//
//      Purpose:	Displays an error message
//
//  Description:	Displays an error message with the severity and
//					subsystem.  Also displays a Details section which
//					can contain additional error information (such as
//					the log history).
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ErrorDialog.h,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 12    8/30/02 9:14a Petert
 * Added email functionality
 * 
 * 11    7/18/02 12:08a Zakirh
 * integration with crashfinder.
 * 
 * 10    7/17/02 3:49p Marce
 * o Now passes stack frames to error dialog
 * 
 * 6     10/25/00 4:57p Marce
 * o Comments
 * 
 * 5     10/13/00 7:17p Marce
 * o Now displays the correct severity icon
 * 
 * 4     6/22/00 7:51p Marce
 * o Fixed display problems
 * 
 * 3     4/26/00 8:59p Marce
 * o Added the Debug button
 * 
 * 2     4/26/00 1:41a Marce
 * o Now displays severity icon (does not work) and subsystem
 * 
 * 1     4/24/00 11:27a Marce
 */

#ifndef __ERRORDIALOG_H_
#define __ERRORDIALOG_H_

#include "resource.h"       // main symbols
#include <atlhost.h>

#include <string>			// STD string

// Forward declarations
struct SIlxStackFrames;

// Type defines
typedef std::basic_string<TCHAR>	tstring;

//*****************************
// CErrorDialog
//*****************************

class CErrorDialog : 
	public CAxDialogImpl<CErrorDialog>
{
public:
	CErrorDialog();
	virtual ~CErrorDialog();

	enum { IDD = IDD_ERRORDIALOG };

BEGIN_MSG_MAP(CErrorDialog)
	MESSAGE_HANDLER(WM_INITDIALOG, OnInitDialog)
	MESSAGE_HANDLER(WM_PAINT, OnPaint)
	COMMAND_ID_HANDLER(IDOK, OnOK)
	COMMAND_ID_HANDLER(IDCANCEL, OnOK)
	COMMAND_HANDLER(IDC_DETAIL_BTN, BN_CLICKED, OnDetailBtnClick)
	COMMAND_HANDLER(IDC_DETAIL_BTN, BN_DOUBLECLICKED, OnDetailBtnClick)
	COMMAND_HANDLER(IDC_DEBUG_BTN, BN_CLICKED, OnDebugBtnClick)
	COMMAND_HANDLER(IDC_DEBUG_BTN, BN_DOUBLECLICKED, OnDebugBtnClick)
END_MSG_MAP()
// Handler prototypes:
//  LRESULT MessageHandler(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
//  LRESULT CommandHandler(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled);
//  LRESULT NotifyHandler(int idCtrl, LPNMHDR pnmh, BOOL& bHandled);

	LRESULT OnInitDialog(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
	LRESULT OnPaint(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled);
	LRESULT OnOK(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled);
	LRESULT OnDetailBtnClick(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled);
	LRESULT OnDebugBtnClick(WORD wNotifyCode, WORD wID, HWND hWndCtl, BOOL& bHandled);

public:
	//void init(	const TCHAR* szErrMsg, 
	//			SIlxStackFrames* pvIlxStackFrames,
	//			long nNumStackFrames,
	//			const TCHAR* szSubsystem, 
	//			long nSeverity);

	void init(	const TCHAR* szErrMsg, 
				const TCHAR* szSubsystem, 
				long nSeverity);

	bool debugBtnClicked() const;

private: // Helper functions

	//void hideOrShowDetails(bool bShow);
	//void RefreshListBox();

	bool LaunchLogviewer() const;

private: // DATA
	tstring				m_strErrMsg;		// Ex: "File not found"
	//tstring			m_strDetailMsg;	// Ex: Last 20 lines of the log file
	tstring				m_strSubsystem;		// Ex: "T 1"Thomson One 
	long				m_nSeverity;		// One of
	//bool				m_bShowDetails;
	bool				m_bDebugBtnClicked;
	//long				m_nWidth;
	//long				m_nFullHeight;
	RECT				m_rectSeverityIcon;
	//SIlxStackFrames *	m_pIlxStackFrames;
	//long				m_nNumStackFrames;
};

#endif //__ERRORDIALOG_H_
