///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/HelperFunctions.cpp,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	HelperFunctions.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/04/2000 4:17:48 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: HelperFunctions.cpp,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 18    8/30/02 9:15a Petert
 * Added email functionality
 * 
 * 17    8/20/02 4:40p Zakirh
 * o Made the susbsystem T1 if it is ActiveILX.
 * 
 * 16    8/15/02 4:55p Zakirh
 * o FIxed the bug where the same address was getting displayed.
 * o Removed temp buffer.
 * 
 * 15    8/05/02 4:09p Zakirh
 * o logs the address only if it is greater than 0.
 * 
 * 14    7/26/02 3:48p Zakirh
 * o now uses _sntprintf instead of stprintf.
 * o more severe checking on stack overflow 
 * o null terminate the strings.
 * 
 * 13    7/25/02 3:54p Marce
 * o Added a space after "Line"
 * 
 * 12    7/24/02 2:49p Zakirh
 * o Log changes after CrashFinder API change memory is now in stack
 * rather than in heap.
 * 
 * 11    7/19/02 5:38p Zakirh
 * o fixed some format problems.
 * 
 * 10    7/18/02 6:42p Zakirh
 * o change in format methods.
 * 
 * 9     7/18/02 2:53p Zakirh
 * o minor bug fixes.
 * 
 * 8     7/18/02 12:08a Zakirh
 * integration with crashfinder.
 * 
 * 5     2/12/01 12:13p Marce
 * o Moved GetHRESULTSeverity() into common helper functions
 * 
 * 3     10/25/00 3:41p Marce
 * o Added GetSeverity()
 * 
 * 2     10/04/00 5:38p Marce
 * o Implemented log filtering
 */

#include "StdAfx.h"

#include "HelperFunctions.h"	// Me

#include "ILXLog2.h"			// ILXLOG__SUCCESS, etc.

#include "ILXLogConstants.h"	// szRDC_REGISTRY_PATH

static const long DECIMAL_CONV	= 10; //for _ltoT conversions

static const tstring strActiveILX	= _T("ActiveILX");
static const tstring strThomsonOne	= _T("T1");

void ExtractFileAndSubsystem(	const TCHAR* szSubsystem,
								const TCHAR* szPath, 
								tstring* pstrFile, 
								tstring* pstrSubsystem)
{
	_ASSERTE(szPath);
	_ASSERTE(pstrFile);
	_ASSERTE(pstrSubsystem);

	tstring& strFile = *pstrFile;
	tstring& strSubsystem = *pstrSubsystem;

	if (szSubsystem)
	{
		strSubsystem = szSubsystem;
	}

	if (szPath && *szPath)
	{
		strFile = szPath;

		long nLastSlash = strFile.rfind(_T('\\'));
		if (nLastSlash != tstring::npos && nLastSlash - 1 > 0)
		{
			long nSecondToLastSlash = strFile.rfind(_T('\\'), nLastSlash - 1);
			if (nSecondToLastSlash != tstring::npos)
			{
				if (strSubsystem.empty())
				{
					strSubsystem.assign(	strFile.begin() + nSecondToLastSlash + 1, 
											strFile.begin() + nLastSlash);

					//If Subsystem is equal to ActiveILX
					if ( strSubsystem == strActiveILX )
					{
						//then manually set the subsystem to T1 --> For Thomson One 
						strSubsystem = strThomsonOne;
					}
				}

				strFile.erase(strFile.begin(), strFile.begin() + nLastSlash + 1);
			}
		}
	}
}

const TCHAR* GetSeverity(long nSeverity)
{
	const TCHAR* szSeverity = 0;

	if (nSeverity == ILXLOG__SUCCESS)
	{
		szSeverity = _T("SUCCESS");
	}
	else if (nSeverity == ILXLOG__INFO)
	{
		szSeverity = _T("INFO");
	}
	else if (nSeverity == ILXLOG__WARNING)
	{
		szSeverity = _T("WARNING");
	}
	else if (nSeverity == ILXLOG__ERROR)
	{
		szSeverity = _T("ERROR");
	}
	else
	{
		_ASSERTE(!"Unknown severity");
		szSeverity = _T("UNKNOWN SEVERITY");
	}

	return szSeverity;
}


//*********************************************************************
// FUNCTION		:	GetHRESULTSeverity
//					
// DESCRIPTION	:	Converts the HRESULT severity into a
//					ILXLOG_SEVERITIES enumeration value.
//					
//*********************************************************************
long GetHRESULTSeverity(HRESULT hr)
{
	unsigned long nHR = hr;

	if (HRESULT_FACILITY(hr) == 0) 
	{
		// We have an old OLE error code that only uses 1 bit for severity
		return ( (nHR >> 31) & 0x1 ) == 0 ? ILXLOG__SUCCESS : ILXLOG__ERROR;
	}
	else
	{
		// Severity bits 31 and 30
		return nHR >> 30; 
	}
}

//*********************************************************************
// FUNCTION		:	GetILXDirectoryFromRegistry()
//					
// DESCRIPTION	:	Reads a directory from the registry.
//					
// RETURNS		:	
//   HRESULT		
//					
//*********************************************************************
HRESULT GetILXDirectoryFromRegistry(const TCHAR* szDirectoryName,
									wchar_t* szDirectory,
									long nMaxLen)
{
	USES_CONVERSION;

	HRESULT hr = S_OK;

	HKEY hkResult = 0; 

	// Step 1:	Open the root RDC key

	long lRet = ::RegOpenKeyExW(HKEY_CURRENT_USER, 
								szRDC_REGISTRY_PATH, 
								NULL, 
								KEY_ALL_ACCESS, 
								&hkResult); 

	if (ERROR_SUCCESS != lRet)
	{
		return HRESULT_FROM_WIN32(::GetLastError());
	}

	// Step 3:	Get the path to the Shared RDC directory
	
	szDirectory[0] = L'\0';
	DWORD dwType = REG_SZ;
	ULONG cbValue = nMaxLen;
	lRet = RegQueryValueExW(	hkResult, 
								T2CW(szDirectoryName), 
								NULL, 
								&dwType,  
								(unsigned char*) szDirectory, 
								&cbValue); 

	if (ERROR_SUCCESS == lRet) 
	{
		// Step 4:	Add "rdcerrors.dll" to our resource list
		
		// Remove trailing slash

		long nFilePathLen = wcslen(szDirectory);
		_ASSERTE(nFilePathLen >= 0);

		if (nFilePathLen > 0)
		{
			wchar_t& cLastChar = szDirectory[nFilePathLen-1];
			if (cLastChar == L'\\' || cLastChar == L'/')
			{
				cLastChar = L'\0';
			}
		}
	}

	RegCloseKey(hkResult);
	
	return S_OK;
}

BOOL FormatStackString1( PILXStackFrame pStackFrame, TCHAR *stString, const long lLength )
{
	if ( !pStackFrame || !stString )
	{
		return FALSE;
	}

	//Initialise the string first 
	stString[0] = TCHAR('\0');

	//For example it should like some like this 
	//in ActiveILX.exe function CServicewrapper::CServiceWrapper().  or
	//in ActiveILX.exe Function CServicewrapper::CServiceWrapper() (Line 79, Column 9).
	int			nCount		= 0;
	int			nTempCount	= 0;
	
	//get the module name length 	
	if ( _tcslen( pStackFrame->szModuleName ) > 0 )
	{
		nTempCount = _sntprintf( stString+nCount, lLength, pStackFrame->szModuleName );

		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		nCount += nTempCount;
	}

	//get the function name length 
	if (  _tcslen( pStackFrame->szFunctionName ) > 0 )
	{
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T(" in function %s"), pStackFrame->szFunctionName );

		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		nCount += nTempCount;
	}

	//if we have line number info 
	if ( pStackFrame->lLineNumber > 0 )
	{
		//now line number 
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T(" Line %d, "), pStackFrame->lLineNumber );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		nCount += nTempCount;

		//now column number
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T("Col %d"), pStackFrame->lColNumber );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		nCount += nTempCount;
	}

	//Add a period
	_sntprintf( stString+nCount, lLength - nCount, _T("%c"), TCHAR('.') );

	//make sure to NULL TERMINATE THE string 
	stString[ lLength-1] = TCHAR('\0' );

	return TRUE;
}

BOOL FormatStackString2( PILXStackFrame pStackFrame, TCHAR *stString, const long lLength )
{
	if ( !pStackFrame || !stString )
	{
		return FALSE;
	}

	//Initialise the string first 
	stString[0] = TCHAR('\0');

	//For example it should like some like this 
	//0012A49c Activeilx.exe, CILXToolBarMgr::OnTooNotify (Line 1990, Col 89), c:\phase5\haha.cpp

	int nCount = 0;
	int nTempCount =0;
	//Address
	if ( pStackFrame->dwAddress > 0 )
	{
		nTempCount = _sntprintf( stString+nCount, lLength, _T("0x%08X"), pStackFrame->dwAddress );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}
	}

	//increment our counter
	nCount += nTempCount;

	//now Module Name
	if ( _tcslen( pStackFrame->szModuleName ) > 0 )
	{
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T(" %s"), pStackFrame->szModuleName );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}
		
		//increment our counter
		nCount += nTempCount;
	}

	//Function Name
	if ( _tcslen( pStackFrame->szFunctionName ) > 0 )
	{
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T(", %s"), pStackFrame->szFunctionName );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		//increment our counter
		nCount += nTempCount;
		
	}

	//if we have line number info 
	if ( pStackFrame->lLineNumber > 0 )
	{
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T(", Line %d, "), pStackFrame->lLineNumber );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		//increment our counter
		nCount += nTempCount;
		
		//now column number
		nTempCount = _sntprintf( stString+nCount, lLength - nCount, _T("Col %d"), pStackFrame->lColNumber );
		//check for buffer overwrite
		if ( nTempCount < 0 )
		{
			//thats it truncate it ...no point returning failure
			//make sure to NULL TERMINATE THE string 
			stString[ lLength-1] = TCHAR('\0' );
			return TRUE;
		}

		//increment our counter
		nCount += nTempCount;

	}

	//src file Path 
	if ( _tcslen(pStackFrame->szFilePath) > 0 )
	{
		_sntprintf( stString+nCount, lLength - nCount, _T(", %s"), pStackFrame->szFilePath );
	}
	
	//make sure to NULL TERMINATE THE string 
	stString[ lLength-1] = TCHAR('\0' );

	return TRUE;
}
