///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/ILXLogHelper.cpp,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	ILXLogHelper.cpp
//   Created By:	Marc Eaddy
//   Created on:	04/24/2000 12:16:08 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ILXLogHelper.cpp,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 42    8/22/02 5:27p Marce
 * o Changed "Thomson ONE" -> "ActiveILX"
 * 
 * 41    7/30/02 4:22p Marce
 * o Made some local variables static so that the function doesn't require
 * a large stack which can be a problem if we get a stack overrun
 * exception
 * 
 * 40    7/26/02 6:20p Marce
 * o Comments
 * 
 * 39    7/25/02 3:58p Marce
 * o No longer forces the modal dialog to be displayed when a crash occurs
 * o Now uses shared constants file
 * 
 * 38    7/22/02 11:39a Marce
 * o Argument should be const
 * 
 * 37    7/19/02 8:03p Zakirh
 * o ActiveIlx to ActiveILX.
 * 
 * 36    7/19/02 5:38p Zakirh
 * o now uses the SearchPath for CrashFinder.
 * 
 * 35    7/18/02 12:08a Zakirh
 * integration with crashfinder.
 * 
 * 34    7/17/02 3:40p Marce
 * o Finalized how to pass exception info
 * 
 * 33    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 30    10/25/01 6:52p Marce
 * o Moved "ILXLog is shutting down" message into ILXLog::shutdown()
 * 
 * 29    2/12/01 12:14p Marce
 * o Versioned the interface for Phase3 by calling it CILXLogHelper2
 * 
 * 28    1/31/01 2:58p Marce
 * o Fixed BC warning
 * 
 * 27    1/30/01 5:03p Vinodinis
 * o Fixed assertion that occurs only in VB.  shutdown() was not cleaning
 * up everything properly.
 * 
 * 26    1/30/01 12:08p Marce
 * o Added API versioning to prevent DLL mismatches that cause undefined
 * runtime behavior
 * o Add CoInit and CoUninit since we are creating COM objects. This was
 * breaking for clients that did not do this prior to using the log
 * helper.
 * 
 * 25    1/30/01 10:51a Marce
 * o Fixed compile errors for Debug build
 * 
 * 24    11/17/00 6:01p Marce
 * o addReourceModule() and addResouceFile() now return an HRESULT to
 * provide error info
 * o Add ILXLOG0 and other XXX0 macros to allow you to specify 0 arguments
 * and put severity before the message for consistency.
 * 
 * 23    11/15/00 2:33p Marce
 * o Removed the original LogHR().  Renamed LogHR2() to LogHR().
 * o Removed the overloaded log() method that did not use a severity
 * o Changed the severity param to use the enum for better type checking
 * o Now supports message formatting
 * o Added more comments
 * 
 * 21    10/05/00 4:47p Marce
 * o Fixed possible deadlock
 * o Now ensures that useLogFile() and loadConfig() are not BOTH called
 * 
 * 20    10/05/00 1:47p Marce
 * o Clarified parameter names
 * 
 * 19    10/04/00 5:39p Marce
 * o Removed Initialize() and load()
 * o Added LoadConfig() and UseLogFile()
 * o Implemented log filtering
 * 
 * 18    9/22/00 12:37p Marce
 * o Now includes thread id in the log message
 * 
 * 17    8/22/00 10:49a Laijunx
 * o Fixed non-unicode compile-time bug
 * 
 * 16    6/06/00 6:31p Marce
 * o Fixed memory leak
 * 
 * 15    5/31/00 4:47p Marce
 * o Added overloaded log() function that allows you to specify the
 * subsystem explicitly
 * 
 * 14    5/20/00 8:51p Marce
 * o Changed "Log" method to "LogMsg" because it was getting confused for
 * the logarithm function in VB Script!
 * 
 * 13    5/13/00 12:49a Marce
 * o Added the ability to pop-up a message box with a string message
 * 
 * 12    5/05/00 1:59p Marce
 * o Now ILXLogHelper is part of ILXLog.dll
 * 
 * 11    5/04/00 1:58p Marce
 * o Now IILXLog aggregates the FTM so we don't need to perform
 * marshalling
 * 
 * 10    5/03/00 12:55p Marce
 * o Made msgBox() more robust by having it pop-up a standard message box
 * itself if an error occurs
 * 
 * 9     5/03/00 12:25p Marce
 * o assert logic was wrong
 * 
 * 8     5/02/00 12:40p Marce
 * o We can leave the critical section earlier
 * 
 * 7     5/01/00 11:32a Marce
 * o Now logs source file and line number
 * 
 * 6     4/25/00 9:51p Marce
 * o Now the message box only takes an HRESULT
 * 
 * 5     4/25/00 8:31p Marce
 * o Added support for including resource modules to search when resolving
 * HRESULTs
 * o Changed "Module" to "Subsystem" to avoid confusion with resource
 * modules
 * 
 * 4     4/25/00 1:33p Marce
 * o Now uses 4 severities instead of 30 levels
 * o Separated out error dialog popup from normal logging
 * 
 * 3     4/24/00 4:09p Marce
 * o Fixed bugs caused by not marshalling the IILXLog object properly
 * 
 * 2     4/24/00 2:05p Marce
 * 
 * 1     4/24/00 12:17p Marce
 */

#include "StdAfx.h"

#include "ILXLogHelper.h"	// Me

#include "ILXLog2.h"		// IILXLog

#include "ILXLogImpl.h"		// CILXLog

#include "MsgBoxDestination.h"	// CMsgBoxDestination

#include "HelperFunctions.h"	// IlxCrashFinder

#include "ILXLogConstants.h"	// nMAX_ERROR_MSG_LEN

// Constants

static bool g_bWarningMessageDisplayed  =	false;
static const TCHAR* szCRASH_SUBSYSTEM	=	_T("T1");	

////////////////////////////////////
// ILX LOG HELPER3 IMPLEMENATATION
////////////////////////////////////

// Static member initialization

CILXLogHelper3 CILXLogHelper3::s_singleton;

CILXLogHelper3& CILXLogHelper3::GetLogSafe(long nIlxLogApiVersion)
{
	if (ILX_LOG_API_VERSION != nIlxLogApiVersion &&
		!g_bWarningMessageDisplayed)
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("The version of ILXLog.dll installed on this machine ")
					_T("is not compatible with this program.\n")
					_T("It is recommended that you close this program.\n")
					_T("[Installed: %d, Needed: %d]"),
					ILX_LOG_API_VERSION,
					nIlxLogApiVersion);

		OutputDebugString(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		
		g_bWarningMessageDisplayed = true;
	}

	return s_singleton;
}

CILXLogHelper3* CILXLogHelper3::GetLogSafePtr(long nIlxLogApiVersion)
{
	if (ILX_LOG_API_VERSION == nIlxLogApiVersion)
	{
		return &s_singleton;
	}
	else if (!g_bWarningMessageDisplayed)
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("The version of ILXLog.dll installed on this machine ")
					_T("is not compatible with this program.\n")
					_T("It is recommended that you close this program.\n")
					_T("[Installed: %d, Needed: %d]"),
					ILX_LOG_API_VERSION,
					nIlxLogApiVersion);

		OutputDebugString(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		
		g_bWarningMessageDisplayed = true;

		return 0;
	}
	else
	{
		return 0;
	}
}

CILXLogHelper3::CILXLogHelper3()
	: CILXLogHelper2()
{ }

void CILXLogHelper3::vlogCrash(	/*[in]*/ const char* szPath, 
								/*[in]*/ long nLine,
								/*[in]*/ EXCEPTION_POINTERS* pExceptionInfo,
								/*[in]*/ const TCHAR* szSearchPath,
								/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity,
								/*[in]*/ const TCHAR* szFormat,
								/*[in]*/ ...)
{
	HRESULT hr = S_OK;

	// Validate arguments
	if (!szFormat || _tcslen(szFormat) > nMAX_ERROR_MSG_LEN)
	{
		_ASSERTE(!"Log message format string is invalid.");
		return;
	}

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);
		
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Exception Address: 0x%08X, Severity: %d)\". ")
					_T("Logging has been disabled.\n"),
			szFormat,
			pExceptionInfo,
			eSeverity);
		return;
	}

	// Step 1:	Create the formatted message

	static TCHAR szMessage[nMAX_ERROR_MSG_LEN];

	va_list args;
	va_start(args, szFormat);

	// Create the formatted message
	int nCount = _vsntprintf(szMessage, nMAX_ERROR_MSG_LEN-1, szFormat, args);

	// Make sure string is null terminated.  snprintf() doesn't
	// do this!
	szMessage[nMAX_ERROR_MSG_LEN-1] = _T('\0');

	va_end(args);

	// Verify that the buffer was not overrun by snprintf()
	if (nCount < 0)
	{
		USES_CONVERSION;

		AtlTrace(	_T("WARNING: ILXLogHelper - Log Message was truncated: \"%s ")
					_T("(Error Code: 0x%08X, Severity: %d, File: %s, Line: %d)\"."),
					szFormat,
					pExceptionInfo,
					eSeverity,
					A2CT(szPath),
					nLine);
	}

	// Step 2:	Send the crash info to the other (non-dialog) log destinations
	//			(e.g. log file, debug output)

	hr = m_piIlxLog->LogCrash(	CComBSTR(szCRASH_SUBSYSTEM), 
								CComBSTR(szPath),
								nLine,
								::GetCurrentThreadId(),
								reinterpret_cast<DWORD>(pExceptionInfo),
								szSearchPath ? CComBSTR(szSearchPath) : 0,
								CComBSTR(nCount >= 0 ? szMessage : szFormat),
								eSeverity);	
	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Crash message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("IILXLog::LogCrash() failed with error 0x%08X.\n"),
			szMessage,
			pExceptionInfo,
			eSeverity,
			hr);
	}

	// Step 3:	Convert pExceptionInfo to SIlxStackFrames (use IlxCrashFinder
	//			API)

	static ILXStackFrame aryIlxStackFrames[nMAX_STACK_FRAMES];

	// Call GetExceptionInfo() in ILXCrashFinder.DLL
	hr = GetExceptionInfo(pExceptionInfo, szSearchPath, aryIlxStackFrames, nMAX_STACK_FRAMES);

	// Must be *after* calling GetExceptionInfo() because IMAGEHLP.DLL is not
	// thread safe
	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Crash message box not displayed: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("GetExceptionInfo() failed with error 0x%08X.\n"),
			szMessage,
			pExceptionInfo,
			eSeverity,
			hr);
		return;
	}

	// ----------------------------------------------------------------------
	// Now that we've left the critical section we can safely popup a modal
	// error dialog box with the crash info.
	// ----------------------------------------------------------------------

	// Step 4:	Append exception info from first stack frame to main message

	TCHAR szBuffer[nMAX_ERROR_BUFFER_LENGTH];
	BOOL bRet = FormatStackString1( &aryIlxStackFrames[0], szBuffer, nMAX_ERROR_BUFFER_LENGTH);
	if ( !bRet )
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Crash message box not displayed: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("FormatStackString1() failed.\n"),
			szMessage,
			pExceptionInfo,
			eSeverity);
		return;
	}

	USES_CONVERSION;

	// Create the final error message

	tstring strFullErrorMsg = szMessage + strHYPHEN + szBuffer;

	// Step 5:	Extract the filename and subsystem name from the path

	tstring strFile;
	tstring strSubsystem;

	::ExtractFileAndSubsystem(szCRASH_SUBSYSTEM, A2CT(szPath), &strFile, &strSubsystem);

	// Step 6:	Create a modal dialog box to display the crash info
	//
	//			WE MUST NOT DO THIS IN THE CRITICAL SECTION!

	static CMsgBoxDestination msgBox;

	if (!msgBox.isInitialized())
	{
		msgBox.initialize();
		msgBox.setIsModal(true);
	}

	msgBox.log(	strSubsystem.c_str(), 
				A2CT(szPath),
				strFile.c_str(), 
				nLine, 
				::GetCurrentThreadId(),
				nINVALID_HRESULT,
				aryIlxStackFrames,
				nMAX_STACK_FRAMES,
				strFullErrorMsg.c_str(), 
				eSeverity);
}

///////////////////////////////////
// ILX LOG HELPER2 IMPLEMENTATION
///////////////////////////////////

CILXLogHelper2& CILXLogHelper2::GetLogSafe(long nIlxLogApiVersion)
{
	if (ILX_LOG_API_VERSION != nIlxLogApiVersion &&
		!g_bWarningMessageDisplayed)
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("The version of ILXLog.dll installed on this machine ")
					_T("is not compatible with this program.\n")
					_T("It is recommended that you close this program.\n")
					_T("[Installed: %d, Needed: %d]"),
					ILX_LOG_API_VERSION,
					nIlxLogApiVersion);

		OutputDebugString(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		
		g_bWarningMessageDisplayed = true;
	}

	return CILXLogHelper3::s_singleton;
}

CILXLogHelper2* CILXLogHelper2::GetLogSafePtr(long nIlxLogApiVersion)
{
	if (ILX_LOG_API_VERSION == nIlxLogApiVersion)
	{
		return &CILXLogHelper3::s_singleton;
	}
	else if (!g_bWarningMessageDisplayed)
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("The version of ILXLog.dll installed on this machine ")
					_T("is not compatible with this program.\n")
					_T("It is recommended that you close this program.\n")
					_T("[Installed: %d, Needed: %d]"),
					ILX_LOG_API_VERSION,
					nIlxLogApiVersion);

		OutputDebugString(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		
		g_bWarningMessageDisplayed = true;

		return 0;
	}
	else
	{
		return 0;
	}
}

CILXLogHelper2::CILXLogHelper2()
	: m_bstrSubsystem()
	, m_piIlxLog(0)
	, m_csLogging()
	, m_bIsInitialized(false)
{ 
	::InitializeCriticalSection(&m_csLogging);
}

CILXLogHelper2::~CILXLogHelper2()
{ 
	//_ASSERTE(!m_piIlxLog || !"Must call shutdown() first");
	
	if (m_bstrSubsystem)
	{
		::SysFreeString(m_bstrSubsystem);
		m_bstrSubsystem = 0;
	}

	::DeleteCriticalSection(&m_csLogging);

	m_bIsInitialized = false;
}

HRESULT CILXLogHelper2::initialize(	const TCHAR* szSubsystem, 
									HMODULE hResourceModule)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);
	_ASSERTE(!m_piIlxLog);

	// Verify parameters
	_ASSERTE(szSubsystem && *szSubsystem);
	_ASSERTE(!m_bstrSubsystem);

	USES_CONVERSION;

	// Initialize member variables

	::SysFreeString(m_bstrSubsystem);
	m_bstrSubsystem = ::SysAllocString(T2COLE(szSubsystem));

	// Initialize COM
	hr = ::CoInitializeEx(0, COINIT_APARTMENTTHREADED);
	if (FAILED(hr))
	{
		::LeaveCriticalSection(&m_csLogging);
		_ASSERTE(!"Failed to initialize COM!");
		AtlTrace(_T("ERROR: Failed to initialize COM! - 0x%08X\n"), hr);
		return hr;
	}

	CComObject<CILXLog>* pILXLog = 0;
	hr = CComObject<CILXLog>::CreateInstance(&pILXLog);
	if (FAILED(hr))
	{
		::LeaveCriticalSection(&m_csLogging);

		_ASSERTE(!"Could not initialize IlxScripting logging!");
		AtlTrace(_T("ERROR: Could not initialize IlxScripting logging - 0x%08X!\n"),
			hr);
		return hr;
	}

	hr = pILXLog->QueryInterface(&m_piIlxLog);
	if (FAILED(hr))
	{
		::LeaveCriticalSection(&m_csLogging);

		_ASSERTE(!"Could not initialize IlxScripting logging!");
		AtlTrace(_T("ERROR: Could not initialize IlxScripting logging - 0x%08X!\n"),
			hr);
		return hr;
	}

	::LeaveCriticalSection(&m_csLogging);

	addResourceModule(hResourceModule);

	return S_OK;
}

HRESULT CILXLogHelper2::useLogFile(	const TCHAR* szLogFilename, 
									long nMaxFilesize)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);
	_ASSERTE(m_piIlxLog || !"Must call initialize() first.");
	_ASSERTE(!m_bIsInitialized || !"useLogFile() or loadConfig() already called.");

	// Verify parameters
	_ASSERTE(szLogFilename && *szLogFilename);

	USES_CONVERSION;

	// Initialize member variables

	hr = m_piIlxLog->UseLogFile(	CComBSTR(szLogFilename),
									nMaxFilesize);
	if (FAILED(hr))
	{
		m_piIlxLog->Release();
		m_piIlxLog = 0;

		::LeaveCriticalSection(&m_csLogging);

		_ASSERTE(!"Could not initialize IlxScripting logging!");
		AtlTrace(_T("ERROR: Could not initialize IlxScripting logging - 0x%08X!\n"),
			hr);

		return FAILED(hr) ? hr : E_UNEXPECTED;
	}

	::LeaveCriticalSection(&m_csLogging);

	m_bIsInitialized = true;

	return S_OK;
}

HRESULT CILXLogHelper2::loadConfig(MSXML::IXMLDOMNode* piLogConfigNode)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);
	_ASSERTE(m_piIlxLog || !"Must call initialize() first.");
	_ASSERTE(!m_bIsInitialized || !"useLogFile() or loadConfig() already called.");

	// Verify parameters
	_ASSERTE(piLogConfigNode);
	if (!piLogConfigNode) 
	{
		::LeaveCriticalSection(&m_csLogging);
		_ASSERTE(false);
		return E_POINTER;
	}

	// Initialize member variables

	hr = m_piIlxLog->LoadConfig(piLogConfigNode);
	if (FAILED(hr))
	{
		::LeaveCriticalSection(&m_csLogging);

		_ASSERTE(!"Could not initialize IlxScripting logging!");
		AtlTrace(_T("ERROR: Could not initialize IlxScripting logging - 0x%08X!\n"),
			hr);

		return FAILED(hr) ? hr : E_UNEXPECTED;
	}

	::LeaveCriticalSection(&m_csLogging);

	m_bIsInitialized = true;

	return S_OK;
}

bool CILXLogHelper2::isInitialized() const
{
	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	return m_piIlxLog != 0 && m_bIsInitialized;
}

void CILXLogHelper2::shutdown()
{
	::EnterCriticalSection(&m_csLogging);

	HRESULT hr = S_OK;

	if (m_piIlxLog)
	{
		hr = m_piIlxLog->LogMsg(0, 
								CComBSTR(__FILE__),
								__LINE__,
								::GetCurrentThreadId(),
								CComBSTR(L"ILXLog is shutting down."), 
								ILXLOG_INFO);	
		if (FAILED(hr))
		{
			AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Severity: %d)\". ")
						_T("IILXLog::Log() failed with error 0x%08X.\n"),
				_T("ILXLog is shutting down."),
				ILXLOG_INFO,
				hr);
		}

		m_piIlxLog->Release();
		m_piIlxLog = 0;
	}

	if (m_bstrSubsystem)
	{
		::SysFreeString(m_bstrSubsystem);
		m_bstrSubsystem = 0;
	}

	m_bIsInitialized = false;

	::LeaveCriticalSection(&m_csLogging);

	::CoUninitialize();
}

HRESULT CILXLogHelper2::addResourceModule(HMODULE hModule)
{
	HRESULT hr = S_OK;

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	TCHAR szFilename[_MAX_PATH] = { 0 };

	DWORD dwResult = ::GetModuleFileName(	hModule, 
											szFilename,
											_MAX_PATH);
	_ASSERTE(dwResult < _MAX_PATH);

	if (dwResult == 0)
	{
		_ASSERTE(!"Failed to retrieve module file name");
		return HRESULT_FROM_WIN32(GetLastError());
	}

	return addResourceFile(szFilename);
}

HRESULT CILXLogHelper2::addResourceFile(const TCHAR* szFilename)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);
		_ASSERTE(!"Must call addModule() after initialize() or attach()");
		return E_UNEXPECTED;
	}

	hr = m_piIlxLog->AddResource(CComBSTR(szFilename));	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Module not added: %s.  ")
					_T("IILXLog::Log() failed with error 0x%08X.\n"),
			szFilename,
			hr);
		return hr;
	}

	return S_OK;
}

HRESULT CILXLogHelper2::getILXLogUnk(IUnknown** ppiIlxLogUnk)
{
	HRESULT hr = S_OK;

	if (!ppiIlxLogUnk) return E_POINTER;
	if (*ppiIlxLogUnk) return E_INVALIDARG; // Don't forget to Release() first

	::EnterCriticalSection(&m_csLogging);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);
		_ASSERTE(false);
		return E_UNEXPECTED;
	}

	hr = m_piIlxLog->QueryInterface(	IID_IUnknown,
										reinterpret_cast<void**>(ppiIlxLogUnk));

	::LeaveCriticalSection(&m_csLogging);

	return hr;
}

void CILXLogHelper2::log(	/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ const TCHAR* szMessage, 
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);

		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Severity: %d)\". ")
					_T("Logging has been disabled.\n"),
			szMessage,
			eSeverity);

		return;
	}

	hr = m_piIlxLog->LogMsg(0, 
							CComBSTR(szPath),
							nLine,
							::GetCurrentThreadId(),
							CComBSTR(szMessage), 
							eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Severity: %d)\". ")
					_T("IILXLog::Log() failed with error 0x%08X.\n"),
			szMessage,
			eSeverity,
			hr);
	}
}

HRESULT CILXLogHelper2::log(	/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ HRESULT hrErrorCode, 
							/*[in]*/ const TCHAR* szMessage, 
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity)
{
	return log(0, szPath, nLine, hrErrorCode, szMessage, eSeverity);
}

HRESULT CILXLogHelper2::log(	/*[in]*/ const char* szSubsystem, 
							/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ HRESULT hrErrorCode, 
							/*[in]*/ const TCHAR* szMessage, 
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);
		
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("Logging has been disabled.\n"),
			szMessage,
			hrErrorCode,
			eSeverity);

		return hrErrorCode;
	}

	hr = m_piIlxLog->LogHR(	szSubsystem ? CComBSTR(szSubsystem) : 0, 
							CComBSTR(szPath),
							nLine,
							::GetCurrentThreadId(),
							hrErrorCode, 
							CComBSTR(szMessage), 
							eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("IILXLog::LogHR() failed with error 0x%08X.\n"),
			szMessage,
			hrErrorCode,
			eSeverity,
			hr);
	}

	return hrErrorCode;
}

void CILXLogHelper2::vlog(	/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity,
							/*[in]*/ const TCHAR* szFormat, 
							...)
{
	HRESULT hr = S_OK;

	// Validate arguments
	if (!szFormat || _tcslen(szFormat) > nMAX_ERROR_MSG_LEN)
	{
		_ASSERTE(!"Log message format string is invalid.");
		return;
	}

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);

		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Severity: %d)\". ")
					_T("Logging has been disabled.\n"),
			szFormat,
			eSeverity);

		return;
	}

	TCHAR szMessage[nMAX_ERROR_MSG_LEN];

	va_list args;
	va_start(args, szFormat);

	// Create the formatted message
	int nCount = _vsntprintf(szMessage, nMAX_ERROR_MSG_LEN-1, szFormat, args);

	// Make sure string is null terminated.  snprintf() doesn't
	// do this!
	szMessage[nMAX_ERROR_MSG_LEN-1] = _T('\0');

	va_end(args);

	// Verify that the buffer was not overrun by snprintf()
	if (nCount < 0)
	{
		USES_CONVERSION;

		AtlTrace(	_T("WARNING: ILXLogHelper - Log Message was truncated: \"%s ")
					_T("(Severity: %d, File: %s, Line: %d)\"."),
					szFormat,
					eSeverity,
					A2CT(szPath),
					nLine);
	}

	hr = m_piIlxLog->LogMsg(0, 
							CComBSTR(szPath),
							nLine,
							::GetCurrentThreadId(),
							CComBSTR(szMessage), 
							eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Severity: %d)\". ")
					_T("IILXLog::Log() failed with error 0x%08X.\n"),
			szMessage,
			eSeverity,
			hr);
	}
}

HRESULT CILXLogHelper2::vlog(/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ HRESULT hrErrorCode, 
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity,
							/*[in]*/ const TCHAR* szFormat,
							/*[in]*/ ...)
{
	HRESULT hr = S_OK;

	// Validate arguments
	if (!szFormat || _tcslen(szFormat) > nMAX_ERROR_MSG_LEN)
	{
		_ASSERTE(!"Log message format string is invalid.");
		return E_INVALIDARG;
	}

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);
		
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("Logging has been disabled.\n"),
			szFormat,
			hrErrorCode,
			eSeverity);

		return hrErrorCode;
	}

	TCHAR szMessage[nMAX_ERROR_MSG_LEN];

	va_list args;
	va_start(args, szFormat);

	// Create the formatted message
	int nCount = _vsntprintf(szMessage, nMAX_ERROR_MSG_LEN-1, szFormat, args);

	// Make sure string is null terminated.  snprintf() doesn't
	// do this!
	szMessage[nMAX_ERROR_MSG_LEN-1] = _T('\0');

	va_end(args);

	// Verify that the buffer was not overrun by snprintf()
	if (nCount < 0)
	{
		USES_CONVERSION;

		AtlTrace(	_T("WARNING: ILXLogHelper - Log Message was truncated: \"%s ")
					_T("(Error Code: 0x%08X, Severity: %d, File: %s, Line: %d)\"."),
					szFormat,
					hrErrorCode,
					eSeverity,
					A2CT(szPath),
					nLine);
	}

	hr = m_piIlxLog->LogHR(	0, 
							CComBSTR(szPath),
							nLine,
							::GetCurrentThreadId(),
							hrErrorCode, 
							CComBSTR(nCount >= 0 ? szMessage : szFormat),
							eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("IILXLog::LogHR() failed with error 0x%08X.\n"),
			szMessage,
			hrErrorCode,
			eSeverity,
			hr);
	}

	return hrErrorCode;
}

HRESULT CILXLogHelper2::vlog(/*[in]*/ const char* szSubsystem, 
							/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ HRESULT hrErrorCode, 
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity,
							/*[in]*/ const TCHAR* szFormat,
							/*[in]*/ ...)
{
	HRESULT hr = S_OK;

	// Validate arguments
	if (!szFormat || _tcslen(szFormat) > nMAX_ERROR_MSG_LEN)
	{
		_ASSERTE(!"Log message format string is invalid.");
		return E_INVALIDARG;
	}

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);
		
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("Logging has been disabled.\n"),
			szFormat,
			hrErrorCode,
			eSeverity);

		return hrErrorCode;
	}

	TCHAR szMessage[nMAX_ERROR_MSG_LEN];

	va_list args;
	va_start(args, szFormat);

	// Create the formatted message
	int nCount = _vsntprintf(szMessage, nMAX_ERROR_MSG_LEN-1, szFormat, args);

	// Make sure string is null terminated.  snprintf() doesn't
	// do this!
	szMessage[nMAX_ERROR_MSG_LEN-1] = _T('\0');

	va_end(args);

	// Verify that the buffer was not overrun by snprintf()
	if (nCount < 0)
	{
		USES_CONVERSION;

		AtlTrace(	_T("WARNING: ILXLogHelper - Log Message was truncated: \"%s ")
					_T("(Error Code: 0x%08X, Severity: %d, File: %s, Line: %d)\"."),
					szFormat,
					hrErrorCode,
					eSeverity,
					A2CT(szPath),
					nLine);
	}

	hr = m_piIlxLog->LogHR(	szSubsystem ? CComBSTR(szSubsystem) : 0, 
							CComBSTR(szPath),
							nLine,
							::GetCurrentThreadId(),
							hrErrorCode, 
							CComBSTR(nCount >= 0 ? szMessage : szFormat),
							eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		AtlTrace(	_T("WARNING: ILXLogHelper - Message not logged: \"%s (Error Code: 0x%08X, Severity: %d)\". ")
					_T("IILXLog::LogHR() failed with error 0x%08X.\n"),
			szMessage,
			hrErrorCode,
			eSeverity,
			hr);
	}

	return hrErrorCode;
}

void CILXLogHelper2::msgBox(	/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ HRESULT hrErrorCode)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);

		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("WARNING: ILXLogHelper - Message not logged: \"(Error Code: 0x%08X)\". ")
					_T("Logging has been disabled.\n"),
					hrErrorCode);

		AtlTrace(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		return;
	}

	_ASSERTE(m_bstrSubsystem);

	hr = m_piIlxLog->MsgBox(	m_bstrSubsystem, 
								CComBSTR(szPath),
								nLine,
								::GetCurrentThreadId(),
								hrErrorCode);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("WARNING: ILXLogHelper - Message not logged: \"(Error Code: 0x%08X)\". ")
					_T("IILXLog::MsgBox() failed with error 0x%08X.\n"),
					hrErrorCode,
					hr);

		AtlTrace(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
	}
}

void CILXLogHelper2::msgBox(	/*[in]*/ const char* szPath, 
								/*[in]*/ long nLine,
								/*[in]*/ const TCHAR* szMessage, 
								/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity)
{
	HRESULT hr = S_OK;

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);

		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("WARNING: ILXLogHelper - Message not logged: \"%s\" (Severity: %d). ")
					_T("Logging has been disabled.\n"),
					szMessage,
					eSeverity);

		AtlTrace(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		return;
	}

	_ASSERTE(m_bstrSubsystem);

	hr = m_piIlxLog->MsgBox2(	m_bstrSubsystem, 
								CComBSTR(szPath),
								nLine,
								::GetCurrentThreadId(),
								CComBSTR(szMessage),
								eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("WARNING: ILXLogHelper - Message not logged: \"%s\" (Severity: %d). ")
					_T("IILXLog::MsgBox() failed with error 0x%08X.\n"),
					szMessage,
					eSeverity,
					hr);

		AtlTrace(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
	}
}

void CILXLogHelper2::vmsgBox(/*[in]*/ const char* szPath, 
							/*[in]*/ long nLine,
							/*[in]*/ ILXLOGHELPER_SEVERITY eSeverity,
							/*[in]*/ const TCHAR* szFormat, 
							/*[in]*/ ...)
{
	HRESULT hr = S_OK;

	// Validate arguments
	if (!szFormat || _tcslen(szFormat) > nMAX_ERROR_MSG_LEN)
	{
		_ASSERTE(!"Log message format string is invalid.");
		return;
	}

	::EnterCriticalSection(&m_csLogging);

	// Pre-conditions
	_ASSERTE(GetLogSafePtr() == this);

	if (!m_piIlxLog)
	{
		::LeaveCriticalSection(&m_csLogging);

		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("WARNING: ILXLogHelper - Message not logged: \"%s\" (Severity: %d). ")
					_T("Logging has been disabled.\n"),
					szFormat,
					eSeverity);

		AtlTrace(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
		return;
	}

	_ASSERTE(m_bstrSubsystem);

	TCHAR szMessage[nMAX_ERROR_MSG_LEN];

	va_list args;
	va_start(args, szFormat);

	// Create the formatted message
	int nCount = _vsntprintf(szMessage, nMAX_ERROR_MSG_LEN-1, szFormat, args);

	// Make sure string is null terminated.  snprintf() doesn't
	// do this!
	szMessage[nMAX_ERROR_MSG_LEN-1] = _T('\0');

	va_end(args);

	// Verify that the buffer was not overrun by snprintf()
	if (nCount < 0)
	{
		USES_CONVERSION;

		AtlTrace(	_T("WARNING: ILXLogHelper - Log Message was truncated: \"%s ")
					_T("(Severity: %d, File: %s, Line: %d)\"."),
					szFormat,
					eSeverity,
					A2CT(szPath),
					nLine);
	}

	hr = m_piIlxLog->MsgBox2(	m_bstrSubsystem, 
								CComBSTR(szPath),
								nLine,
								::GetCurrentThreadId(),
								CComBSTR(nCount >= 0 ? szMessage : szFormat),
								eSeverity);	

	::LeaveCriticalSection(&m_csLogging);

	if (FAILED(hr))
	{
		TCHAR szMsg[1000];
		_stprintf(	szMsg, 
					_T("WARNING: ILXLogHelper - Message not logged: \"%s\" (Severity: %d). ")
					_T("IILXLog::MsgBox() failed with error 0x%08X.\n"),
					szMessage,
					eSeverity,
					hr);

		AtlTrace(szMsg);

		MessageBox(	0,
					szMsg,
					_T("Thomson ONE Error"),
					MB_OK | MB_ICONERROR);
	}
}
