///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2001 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/ILXLogHelperOld.cpp,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	ILXLogHelperOld.cpp
//   Created By:	Marc Eaddy
//   Created On:	02/12/2001 11:12:59 AM
//
//      Purpose:	Support ILXLog Phase2 interface for backwards compatibility
//
//  Description:	This file exports and implements the Phase 2 version
//					of the CILXLogHelper interface.  This is so that
//					Phase3 ILXLog is backwards compatible with Phase 2
//					ActiveILX and RDC components.
//
//					To achieve backwards compatability for a C++
//					DLL interface, the DLL must export the exact same
//					exports (data, methods, functions) of the previous
//					version.  However, we aren't going to provide a header
//					file for this old interface because we want newly
//					compiled projects to use the new version.  This is
//					why this private CPP file contains the class declaration
//					and definition.
//
//					If you compare the exports of this DLL (dumpbin /exports)
//					to the exports of the Phase 2 version you will see that
//					all of the Phase 2 exports are present.  This makes
//					ILXLog.DLL backwards compatible.
//
//					This implementation simply forwards all calls to the
//					new interface.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ILXLogHelperOld.cpp,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 5     7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 2     2/12/01 12:14p Marce
 * o Now backwards compatible with Phase 2
 */

#include "StdAfx.h"

#include "ILXLogHelper.h"		// We forward all calls to this class

#include "HelperFunctions.h"	// GetHRESULTSeverity()

//*****************************
// CILXLogHelper
//*****************************

class __declspec(dllexport) CILXLogHelper
{
public:
	HRESULT initialize(	const TCHAR* szSubsystem, HMODULE hResourceModule)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->initialize(szSubsystem, hResourceModule) : E_FAIL;
	}

	HRESULT loadConfig(MSXML::IXMLDOMNode* piLogConfigNode)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->loadConfig(piLogConfigNode) : E_FAIL;
	}

	HRESULT useLogFile(	const TCHAR* szLogFilename, long nMaxFileSize)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->useLogFile(szLogFilename, nMaxFileSize) : E_FAIL;
	}

	bool isInitialized() const
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->isInitialized() : false;
	}

	void shutdown()
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		if (pLog) pLog->shutdown();
	}

	void addResourceModule(HMODULE hModule)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		if (pLog) pLog->addResourceModule(hModule);
	}

	void addResourceFile(const TCHAR* szFilename)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		if (pLog) pLog->addResourceFile(szFilename);
	}

	void log(const char* szPath, long nLine, const TCHAR* szMessage, long nSeverity)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		if (pLog) pLog->log(szPath, nLine, szMessage, (ILXLOGHELPER_SEVERITY) nSeverity);
	}

	HRESULT log(const char* szPath, long nLine, HRESULT hr, const TCHAR* szMessage)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->log(szPath, nLine, hr, szMessage, 
			(ILXLOGHELPER_SEVERITY) GetHRESULTSeverity(hr)) : E_FAIL;
	}

	HRESULT log(const char* szPath, long nLine, HRESULT hr, 
				const TCHAR* szMessage, long nSeverity)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->log(szPath, nLine, hr, szMessage, (ILXLOGHELPER_SEVERITY) nSeverity) : E_FAIL;
	}

	HRESULT log(const char* szSubsystem, const char* szPath, long nLine, HRESULT hr, 
				const TCHAR* szMessage, long nSeverity)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? 
			pLog->log(szSubsystem, szPath, nLine, hr, szMessage, (ILXLOGHELPER_SEVERITY) nSeverity) 
			: E_FAIL;
	}

	void msgBox(const char* szPath, long nLine, HRESULT hr)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		if (pLog) pLog->msgBox(szPath, nLine, hr);
	}

	void msgBox(const char* szPath, long nLine, const TCHAR* szMessage, long nSeverity)
	{
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		if (pLog) pLog->msgBox(szPath, nLine, szMessage, 
			(ILXLOGHELPER_SEVERITY) nSeverity);
	}

	HRESULT getILXLogUnk(IUnknown** ppiIlxLogUnk)
	{
		if (!ppiIlxLogUnk) return E_POINTER;
		*ppiIlxLogUnk = 0;
		CILXLogHelper3* pLog = CILXLogHelper3::GetLogSafePtr();
		return pLog ? pLog->getILXLogUnk(ppiIlxLogUnk) : E_FAIL;
	}

	~CILXLogHelper() { }

private:
	CILXLogHelper() 
		: m_bstrSubsystem(0)
		, m_piIlxLog(0)
		, m_csLogging()
		, m_bIsInitialized(false) 
	{ }

private: // NOT IMPLEMENTED
	CILXLogHelper& operator=(const CILXLogHelper&);
	CILXLogHelper(const CILXLogHelper&);

private: // DATA
	BSTR				m_bstrSubsystem;
	IILXLog*			m_piIlxLog;
	CRITICAL_SECTION	m_csLogging;
	bool				m_bIsInitialized;
	static CILXLogHelper s_singleton;
};

// Don't forget to define the singleton object (even though we never use it)
CILXLogHelper CILXLogHelper::s_singleton;
