///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/ILXLogImpl.cpp,v 1.3 2004/09/23 05:51:13 me133 Exp $
//
//    File name:	ILXLogImpl.cpp
//   Created By:	Marc Eaddy
//   Created on:	
//
//      Purpose:	Implementation of CILXLog
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ILXLogImpl.cpp,v $
 * Revision 1.3  2004/09/23 05:51:13  me133
 * o Now prints INFO and SUCCESS messages
 * o Now supports verbose output
 *
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 81    8/30/02 9:15a Petert
 * Added email functionality
 * 
 * 80    8/02/02 6:07p Marce
 * o Now provides better log messages if crash finder fails
 * 
 * 79    7/29/02 5:17p Marce
 * o Fixed bug where ILXLog message box was not appearing when ILXMSGBOX()
 * macro was used
 * 
 * 78    7/25/02 3:58p Marce
 * o No longer forces the modal dialog to be displayed when a crash occurs
 * o Now uses shared constants file
 * 
 * 77    7/25/02 2:51p Marce
 * o Now that we use CrashFinder we no longer display the last 20 lines of
 * the log file in the Details section of the Error dialog box
 * 
 * 76    7/24/02 2:49p Zakirh
 * o Log changes after CrashFinder API change memory is now in stack
 * rather than in heap.
 * 
 * 75    7/22/02 7:32p Marce
 * o Now sends log messages to log files before displaying error dialogs.
 * This ensures that the message is logged no matter what else happens.
 * 
 * 74    7/19/02 5:37p Zakirh
 * o Fixed MLK.
 * 
 * 73    7/18/02 6:42p Zakirh
 * o new format of logging.
 * 
 * 72    7/18/02 2:53p Zakirh
 * o minor bug fixes.
 * 
 * 71    7/18/02 12:08a Zakirh
 * integration with crashfinder.
 * 
 * 70    7/17/02 3:40p Marce
 * o Finalized how to pass exception info
 * 
 * 69    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 68    7/17/02 12:56p Marce
 * o Now handles the case when Log and SharedRDC directories have trailing
 * slashes
 * 
 * 65    5/01/01 4:10p Marce
 * o Added StdOut destination
 * 
 * 64    3/27/01 2:06p Marce
 * Now supports MODAL attribute for forcing a modal dialog box
 * 
 * 63    2/27/01 2:38p Marce
 * o Moved GetErrorMessage() into separate common file
 * 
 * 62    2/12/01 12:13p Marce
 * o Moved GetHRESULTSeverity() into common helper functions
 * 
 * 61    1/30/01 12:09p Marce
 * o Now ignores when the DLL does not contain a DATA section
 * 
 * 60    1/30/01 10:51a Marce
 * o Fixed compile errors for Debug build
 * 
 * 59    1/10/01 2:28p Marce
 * o Now useLogFile() does not display a MsgBox for errors.  It beeps
 * instead.
 * 
 * 58    12/19/00 3:32p Marce
 * o Now the MsgBox() and MsgBox2() error dialog is modal
 * 
 * 57    11/15/00 2:31p Marce
 * o Removed the original LogHR().  Renamed LogHR2() to LogHR().
 * 
 * 56    11/10/00 4:39p Marce
 * o Changed function name for clarity
 * 
 * 55    11/10/00 4:05p Marce
 * o Added comments
 * o Made GetErrorMessage() operate directly on string to eliminate
 * unnecessary allcoation
 * 
 * 53    10/17/00 3:02p Marce
 * o Thread id is allowed to be -1 to indicate that we should use the
 * current thread id
 * 
 * 52    10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 51    10/10/00 4:27p Marce
 * o Need to check before initializing
 * 
 * 50    10/10/00 2:27p Marce
 * o Fixed bug when calling MsgBox() when the "MsgBox" destination has not
 * been initialized by a log route (e.g., release configuration)
 * 
 * 49    10/05/00 1:56p Marce
 * o Cosmetic
 * 
 * 48    10/05/00 1:48p Marce
 * o Clarified parameter names
 * o Now gives the correct subsystem and file to CLogConfig and
 * destination objects
 * 
 * 47    10/05/00 1:35p Marce
 * o Added Beep destination
 * 
 * 46    10/04/00 5:39p Marce
 * o Removed Initialize() and load()
 * o Added LoadConfig() and UseLogFile()
 * o Implemented log filtering
 * 
 * 45    10/03/00 6:33p Marce
 * o Added log configuration classes
 * 
 * 44    10/03/00 2:19p Marce
 * o Cosmetic
 * 
 * 43    9/27/00 5:12p Marce
 * o Added funny comment!! Ha ha
 * 
 * 42    9/25/00 1:44p Marce
 * o Added comment
 * 
 * 41    9/25/00 1:41p Marce
 * o Changed WriteLogHeader() to CreateLogFile() which combines log file
 * backup, creation, and writing the header
 * o Added two resources for locating errors
 * 
 * 40    9/22/00 12:37p Marce
 * o Now includes thread id in the log message
 * 
 * 39    8/29/00 12:22p Marce
 * o Previously when the log file was backed up it would start empty.  Now
 * it outputs the log header again and a log message stating where the
 * backup file is.
 * o Fixed a possible bug that would occur when the file was backed-up
 * since the seek position array was not reinitialized properly.
 * 
 * 38    8/14/00 5:07p Marce
 * o Changed enum names so they would be visible from script
 * 
 * 37    7/03/00 1:47p Marce
 * o Updated log file format version number.  RDC is considered to be
 * version 1.0.
 * 
 * 36    6/22/00 8:46p Marce
 * o Forgot to increment the line count
 * 
 * 35    6/22/00 8:45p Marce
 * o Now reports the log file format version number
 * 
 * 34    6/22/00 8:36p Marce
 * o Now uses the "Log Directory" from the registry
 * 
 * 33    6/07/00 8:06p Marce
 * o Changed trace format
 * 
 * 32    6/07/00 7:46p Marce
 * o Opps, fixed a bug in the trace format
 * 
 * 31    6/07/00 7:42p Marce
 * o Now the log messages that we trace are double-clickable (takes you to
 * the source file and line)
 * 
 * 30    6/07/00 3:49p Marce
 * o Fixed BoundsChecker warning
 * 
 * 29    6/01/00 9:18p Marce
 * 
 * 28    5/20/00 10:53p Marce
 * o Now allows the subsystem name to be empty
 * 
 * 27    5/20/00 10:13p Marce
 * o Now allows nLine to be negative (doesn't display it)
 * 
 * 26    5/20/00 8:51p Marce
 * o Changed "Log" method to "LogMsg" because it was getting confused for
 * the logarithm function in VB Script!
 * 
 * 25    5/13/00 8:16p Marce
 * o Added OnLog event.  Don't let this fool you, it doesn't work! (see
 * ReadMe)
 * 
 * 24    5/13/00 1:05a Marce
 * o Fixed message box multiple lines bug
 * 
 * 23    5/13/00 1:01a Marce
 * o Debug button works better
 * 
 * 22    5/13/00 12:49a Marce
 * o Added the ability to pop-up a message box with a string message
 * 
 * 21    5/12/00 11:41p Marce
 * o Fixed bug where the number of lines was not being counted correctly
 * 
 * 20    5/08/00 3:40p Marce
 * o Added configuration method
 * 
 * 19    5/08/00 11:38a Garys
 * o Fixed syntax error that prevented system errors from being logged
 * 
 * 18    5/05/00 1:58p Marce
 * o Removed unused code
 * 
 * 17    5/05/00 1:58p Marce
 * o Now parses the filename to determine the subsystem
 * 
 * 16    5/05/00 9:03a Marce
 * o Made sure that the last byte of the log string is 0
 * o Modified trace statement to trace more info
 * o Small optimization to fprintf statement
 * 
 * 15    5/04/00 1:58p Marce
 * o Now aggregates the FTM so we can be called from any apartment or
 * thread.  However, I don't know if the event log handle has thread
 * affinity.
 * 
 * 14    5/04/00 1:17p Marce
 * o Improved logfile format
 * 
 * 13    5/03/00 1:12p Marce
 * o Fixed message when HRESULT is not found
 * 
 * 12    5/03/00 12:47p Marce
 * 
 * 11    5/01/00 11:32a Marce
 * o Now logs source file and line number
 * 
 * 10    4/26/00 8:59p Marce
 * o Added the Debug button
 * 
 * 9     4/26/00 1:42a Marce
 * o Fixed log history bug
 * 
 * 8     4/25/00 11:24p Marce
 * o Now includes the log file in the detail section of the error dialog
 * 
 * 7     4/25/00 9:51p Marce
 * o Now the message box only takes an HRESULT
 * 
 * 6     4/25/00 8:58p Marce
 * o Fixed bug where severity was not extracted correctly from the HRESULT
 * o Fixed bug where message was never being found
 * 
 * 5     4/25/00 8:31p Marce
 * o Added support for including resource modules to search when resolving
 * HRESULTs
 * o Changed "Module" to "Subsystem" to avoid confusion with resource
 * modules
 * 
 * 4     4/25/00 1:33p Marce
 * o Now uses 4 severities instead of 30 levels
 * o Separated out error dialog popup from normal logging
 * 
 * 3     4/24/00 4:09p Marce
 * o Cosmetic changes to the log format
 * 
 * 2     4/24/00 2:05p Marce
 * 
 * 1     4/24/00 11:27a Marce
 */

#include "StdAfx.h"

#include "ILXLogImpl.h"				// Me

#include "LogConfig.h"				// CLogConfig

// Log Destinations
#include "LogFileDestination.h"		// CLogFileDestination
#include "MsgBoxDestination.h"		// CMsgBoxDestination
#include "DebugOutDestination.h"	// CDebugOutDestination
#include "BeepDestination.h"		// CBeepDestination
#include "StdOutDestination.h"		// CStdOutDestination

#include "HelperFunctions.h"		// ExtractFileAndSubsystem()

#include "XmlLogHelper.h"			// LogXmlError()

#include "GetErrorMessage.h"		// GetErrorMessage()

#include "ILXLogConstants.h"		// nMAX_ERROR_MSG_LEN

using namespace MSXML;

// Constants

static const wchar_t* szSHARED_RDC_REG_KEY	= L"SharedRDC Directory";
static const wchar_t* szLANG_REG_KEY		= L"Lang";

// Some DLLs we search for message compiled resources
static const TCHAR* szRDC_ERRORS_FILE		= _T("rdcerrors.dll");
static const CComBSTR bstrURLMON_FILE		= _T("urlmon.dll");
static const CComBSTR bstrMSXML_FILE		= _T("msxml.dll");

static const CComBSTR bstrILX_LOG_MODULE	=  L"ILXLOG";
static const wchar_t* const szILX_LOG_MODULE =  L"ILXLOG";
static const TCHAR* const szXML_TAG_NAME	= _T("ILXLOG");
static const TCHAR* const szCONFIG_TO_USE_ATTR = _T("CONFIG_TO_USE");
static CComBSTR bstrNAME_ATTR				= _T("NAME");

//////////////////////////////
// CILXLOGMANAGER IMPLEMENTATION
//////////////////////////////

// Static member initialization

CILXLog::CILXLog()
	: CXmlElement(szXML_TAG_NAME)
	, m_listResourceModules()
	, m_nLanguageId(0)
	, m_pUnkMarshaler(0)
	, m_pLogConfig(new CLogConfig)
	, m_setAllDestinations()
	, m_bInitCalled(false)
{ 
	addChildPrototype(new CLogConfig);
	
	// Add prototypes for all the CLogDestination subclasses
	addChildPrototype(new CLogFileDestination);
	addChildPrototype(new CMsgBoxDestination);
	addChildPrototype(new CDebugOutDestination);
	addChildPrototype(new CBeepDestination);
	addChildPrototype(new CStdOutDestination(true));
	addChildPrototype(new CStdOutDestination(false));
}

CILXLog::~CILXLog()
{ 
	for(TypeListModules::iterator it = m_listResourceModules.begin();
		it != m_listResourceModules.end();
		++it)
	{
		HMODULE& hModule = (*it).first;
		_ASSERTE(hModule);

		bool bShouldFreeLibrary = (*it).second;

		if (bShouldFreeLibrary)
		{
			::FreeLibrary(hModule);
		}

		hModule = NULL;
	}
} 

HRESULT CILXLog::FinalConstruct()
{
	return ::CoCreateFreeThreadedMarshaler(	GetControllingUnknown(), 
											&m_pUnkMarshaler.p);
}

void CILXLog::FinalRelease()
{
	m_pUnkMarshaler.Release();
}

CXmlElement* CILXLog::clone() const
{
	_ASSERTE(false);
	return 0;
}

//*********************************************************************
// FUNCTION		:	CILXLog::load
//					
// DESCRIPTION	:	Loads XML configuration information.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   piNode	
//					
// COMMENTS		:	The XML like this:
//					
//	<ilxlog FORMAT_VERSION="1.1" FILE_VERSION="1.4" CONFIG_TO_USE="DEBUG">
//		<dest TYPE="MsgBox"/>
//		<dest TYPE="DebugOut"/>
//		<config NAME="DEBUG">
//			<route SEVERITY="ERROR,WARNING" DEST="MsgBox,DebugOut"/>
//		</config>
//		<config NAME="RELEASE">
//			<route SEVERITY="ERROR,WARNING" DEST="MsgBox"/>
//		</config>
//	</ilxlog>
//
//					The top-level <ilxlog> element contains <dest> and
//					<config> elements.  However, we only use one
//					configuration at a time.  The CONFIG_TO_USE attribute
//					selects which configuration we will use.
//
//*********************************************************************
HRESULT CILXLog::load(MSXML::IXMLDOMNode* piNode)
{
	HRESULT hr = S_OK;

	if (!piNode) return E_POINTER;

	_ASSERTE(m_setAllDestinations.size() == 0);

	// Step 1:	Load all the attributes and children elements

	hr = CXmlElement::load(piNode);
//	VERIFY_HR_ASSERT(hr, _T("Failed to load log XML entry."), ILXLOG_ERROR); 
	_ASSERTE(SUCCEEDED(hr));
	if (FAILED(hr)) return hr;

	// Step 2:	Get the name of the configuration to use

	const _variant_t& varConfigToUseAttr = m_mapAttributes[szCONFIG_TO_USE_ATTR];
	if (varConfigToUseAttr.vt != VT_BSTR || !varConfigToUseAttr.bstrVal ||
		!*varConfigToUseAttr.bstrVal)
	{
		_ASSERTE(!"CONFIG_TO_USE attribute not found.");
		return E_FAIL;
	}

	// Step 3:	Find the corresponding <config> child

	for(TypeVecXmlElementPtrs::iterator it = getChildren().begin();
		it != getChildren().end();
		++it)
	{
		CXmlElement* pXmlElement = *it;
		_ASSERTE(pXmlElement);

		// Is this a <config> element?  Is it the one we want?
		CLogConfig* pLogConfig = dynamic_cast<CLogConfig*>(pXmlElement);
		if (!pLogConfig || 
			wcsicmp(pLogConfig->getNameAttr(), varConfigToUseAttr.bstrVal) != 0)
		{
			continue;
		}

		// Found the config element to use
		m_pLogConfig = pLogConfig;
		break;
	}

	// Post-condition: Sanity check
	if (!m_pLogConfig)
	{
		_ASSERTE(!"Could not find the specified log configuration.");
		return E_FAIL;
	}

	// Step 4:	Initialize the log destination children
	return Initialize();
}

//*********************************************************************
// FUNCTION		:	CILXLog::Initialize
//					
// DESCRIPTION	:	Initializes the log destination children and
//					adds some default resource modules.
//					
// RETURNS		:	
//   HRESULT 		
//					
//*********************************************************************
HRESULT CILXLog::Initialize()
{
	HRESULT hr = S_OK;

	// Invariants: Only call Initialize() once
	if (m_bInitCalled)
	{
		_ASSERTE(!"CILXLog::Initialize() has already called");
		return E_UNEXPECTED;
	}

	// Sanity check: load() or useLogFile() must have been called
	if (!m_pLogConfig)
	{
		_ASSERTE(!"Could not find the specified log configuration.");
		return E_FAIL;
	}

	CMsgBoxDestination* pMsgBoxDestination = new CMsgBoxDestination;
	if (!pMsgBoxDestination) return E_OUTOFMEMORY;

	pMsgBoxDestination->setIsModal(true);

	// Add msg box destination to master set of destinations because
	// ILXMSGBOX() will use it
	insertChild(pMsgBoxDestination);

	// Step 1:	Initialize the master set of all log destinations

	for(TypeVecXmlElementPtrs::iterator it_dest = getChildren().begin();
		it_dest != getChildren().end();
		++it_dest)
	{
		CXmlElement* pXmlElement = *it_dest;
		_ASSERTE(pXmlElement);

		// Is this a log destination element?
		CLogDestination* pLogDestination = dynamic_cast<CLogDestination*>(pXmlElement);
		if (!pLogDestination)
		{
			continue;
		}

		m_setAllDestinations.insert(pLogDestination);
	}

	// Step 2:	Initialize the log routes stored inside the
	//			CLogConfig object with the log destinations.

	m_pLogConfig->initDestinations(&m_setAllDestinations);

	// Step 3:	Read the log directory from registry
	
	hr = LoadRegistrySettings();
	//_ASSERTE(SUCCEEDED(hr));

	// Step 4:	Add resources that we depend on which can also
	//			generate errors

	// ONLY ADD DLLS THAT ILXLOG.DLL DEPENDS ON DIRECTLY!!
	// IF IT IS NOT LISTED IN "DUMPBIN /IMPORTS ILXLOG.DLL"
	// THEN DON'T ADD IT! - ME

	hr = AddResource(bstrURLMON_FILE);	// Some MSXML errors
	_ASSERTE(SUCCEEDED(hr));

	hr = AddResource(bstrMSXML_FILE);	// Some MSXML errors
	_ASSERTE(SUCCEEDED(hr));

	m_bInitCalled = true;

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CILXLog::LoadConfig
//					
// DESCRIPTION	:	COM interface version of load() above.
//					
// RETURNS		:	
//   STDMETHODIMP 		
//					
// PARAMETERS	:	
//   piunkXmlNode	
//					
// COMMENTS		:	IUnknown is used instead of IXMLDOMNode to prevent
//					the type library from being dependent on MSXML.
//					Client is only dependent if they call this method.
//					
//*********************************************************************
STDMETHODIMP CILXLog::LoadConfig(/*[in]*/ IUnknown* piunkXmlNode)
{
	HRESULT hr = S_OK;

	if (!piunkXmlNode) return E_POINTER;

	CComPtr<MSXML::IXMLDOMNode> piXmlNode = 0;

	hr = piunkXmlNode->QueryInterface(MSXML::IID_IXMLDOMNode, (void**) &piXmlNode);
	if (FAILED(hr)) return hr;

	return load(piXmlNode);
}

//*********************************************************************
// FUNCTION		:	CILXLog::UseLogFile
//					
// DESCRIPTION	:	Provides a hardcoded log configuration that uses
//					a named log file.
//					
// RETURNS		:	
//   STDMETHODIMP 		
//					
// PARAMETERS	:	
//   bstrLogFilename	
//   nMaxFileSize	
//					
// COMMENTS		:	
//					
//*********************************************************************
STDMETHODIMP CILXLog::UseLogFile(BSTR bstrLogFilename, long nMaxFileSize)
{
	HRESULT hr = S_OK;

	// Verify arguments
	if (!bstrLogFilename || !*bstrLogFilename) return E_INVALIDARG;

	// Invariants: Only call Initialize() once
	if (m_bInitCalled)
	{
		_ASSERTE(!"CILXLog::Initialize() has already called");
		return E_UNEXPECTED;
	}

	USES_CONVERSION;

	TCHAR szMaxSize[100];
	_stprintf(szMaxSize, _T("%d"), nMaxFileSize);

	// Step 1:	Create a default XML log configuration
	//
	//			The configuration sends error messages to the error
	//			dialog box, a log file with the specified filename,
	//			and the debug output.  Warnings are sent to the same
	//			places except the error dialog.

	tstring strXML;

	strXML += _T("<?xml version=\"1.0\"?>\n");
	strXML += _T("<ilxlog CONFIG_TO_USE=\"DEFAULT\">\n");
	strXML += _T("  <dest TYPE=\"Beep\"/>\n");
	strXML += _T("  <dest TYPE=\"DebugOut\"/>\n");
	strXML += _T("  <dest TYPE=\"StdOut\"/>\n");
	strXML += _T("  <dest TYPE=\"StdOutVerbose\"/>\n");
	strXML += _T("  <dest TYPE=\"LogFile\" FILE=\"");
	strXML += OLE2CT(bstrLogFilename);
	strXML += _T("\" MAX_SIZE=\"");
	strXML += szMaxSize;
	strXML += _T("\"/>\n");
	strXML += _T("  <config NAME=\"DEFAULT\">\n");
	strXML += _T("	   <route SEVERITY=\"SUCCESS,INFO\" DEST=\"LogFile,StdOut,DebugOut\"/>\n");
	strXML += _T("	   <route SEVERITY=\"ERROR,WARNING\" DEST=\"Beep,LogFile,StdOutVerbose,DebugOut\"/>\n");
	strXML += _T("  </config>\n");
	strXML += _T("</ilxlog>\n");

	// Step 2:	Parse the XML

	CComPtr<MSXML::IXMLDOMDocument> spiXmlDocument;

	hr = spiXmlDocument.CoCreateInstance(MSXML::CLSID_DOMDocument);
	if(FAILED(hr))
	{
		_ASSERTE(!"Failed to create XML parser.");
		return hr;
	}

	VARIANT_BOOL bIsSuccessful = VARIANT_FALSE;
	hr = spiXmlDocument->loadXML(CComBSTR(strXML.c_str()), &bIsSuccessful);
	if (FAILED(hr))
	{
		_ASSERTE(!"Failed to load default log settings.");
		return hr;
	}
	else if (bIsSuccessful != VARIANT_TRUE)
	{
		hr = ::LogXmlError(spiXmlDocument);
		_ASSERTE(!"Failed to load default log settings.");
		return hr;
	}

	CComPtr<MSXML::IXMLDOMElement> spiDocumentElement;
	hr = spiXmlDocument->get_documentElement(&spiDocumentElement);
	if (hr != S_OK || !spiDocumentElement)
	{
		_ASSERTE(!"Failed to get the document object.");
		return FAILED(hr) ? hr : E_FAIL;
	}

	// Step 3:	Load the static configuration

	return load(spiDocumentElement);
}

//*********************************************************************
// FUNCTION		:	CILXLog::AddResource
//					
// DESCRIPTION	:	Adds a module to be used to resolve HRESULTs into
//					string messages.
//					
// RETURNS		:	
//   STDMETHODIMP 		
//					
// PARAMETERS	:	
//   bstrFilename	
//					
//*********************************************************************
STDMETHODIMP CILXLog::AddResource(/*[in]*/ BSTR bstrFilename)
{
	USES_CONVERSION;

	const TCHAR* szFilename = OLE2CT(bstrFilename);

	HMODULE hModule = NULL;

	// Step 1:	See if the module is already loaded
	//
	//			This will happen if someone tries to add
	//			"msvcrt.dll" for example because this module
	//			is always loaded at startup.

	hModule = ::GetModuleHandle(szFilename);
	if (hModule)
	{
		// False means we don't need to release the module
		AddResource(hModule, false);
		return S_OK;
	}

	long nLastError = ::GetLastError();

	// If the module is not already loaded it is not an error.
	// However, we need to handle other types of errors.
	if (nLastError != ERROR_MOD_NOT_FOUND)
	{
		_ASSERTE(!"Unexpected GetModuleHandle() error");

		_bstr_t bstrMsg = _T("Unable to load the '");
		bstrMsg += szFilename;
		bstrMsg += _T("' resource file.");

		if (m_bInitCalled)
		{
			LogHR(	bstrILX_LOG_MODULE, 
					CComBSTR(__FILE__), 
					__LINE__, 
					::GetCurrentThreadId(),
					nLastError, 
					bstrMsg, 
					ILXLOG__ERROR);
		}
		else
		{
			bstrMsg += _T("\n");
			AtlTrace(static_cast<const TCHAR*>(bstrMsg));
		}
	}

	// Step 2:	Load the module explicitly

	hModule = ::LoadLibraryEx(	szFilename,
								NULL,
								LOAD_LIBRARY_AS_DATAFILE);
	if (hModule)
	{
		AddResource(hModule, true);
		return S_OK;
	}

	nLastError = ::GetLastError();

	_bstr_t bstrMsg = _T("Unable to load the '");
	bstrMsg += szFilename;
	bstrMsg += _T("' resource file.");

	if (m_bInitCalled)
	{
		LogHR(	bstrILX_LOG_MODULE, 
				CComBSTR(__FILE__), 
				__LINE__, 
				::GetCurrentThreadId(),
				nLastError, 
				bstrMsg, 
				ILXLOG__ERROR);
	}
	else
	{
		bstrMsg += _T("\n");
		AtlTrace(static_cast<const TCHAR*>(bstrMsg));
	}

	return E_FAIL;
}

//*********************************************************************
// FUNCTION		:	CILXLog::AddResource
//					
// DESCRIPTION	:	Adds the module to our list of resources if it is
//					not already there.
//					
// PARAMETERS	:	
//   hModule	
//   bShouldFreeLibrary	
//					
//*********************************************************************
void CILXLog::AddResource(HMODULE hModule, bool bShouldFreeLibrary)
{
	for(TypeListModules::iterator it = m_listResourceModules.begin();
		it != m_listResourceModules.end();
		++it)
	{
		if ((*it).first == hModule)
		{
			return; // Module is already in the list
		}
	}

	m_listResourceModules.push_back(TypeModuleAndBool(hModule, bShouldFreeLibrary));
}

//*********************************************************************
// FUNCTION		:	CILXLog::LogImpl
//					
// DESCRIPTION	:	Handles all logging requests.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   szSubsystem	Logical subsystem.  If 0, the path is used.
//   szPath			Path to the source file.
//   nLine			Line number in source file.
//   nThreadId		Thread id of the thread that is logging the message.
//   hrErrorCode	HRESULT for the message.  nINVALID_HRESULT means
//					no HRESULT is used.
//   szMessage		User message for the error.
//   nSeverity		Severity of the message.
//   bMsgBox		Whether or not we must display the error dialog box.
//					
// COMMENTS		:	All log requests get funneled through this function.
//					
//*********************************************************************
HRESULT CILXLog::LogImpl(	/*[in]*/ const TCHAR* szSubsystem, 
							/*[in]*/ const TCHAR* szPath,
							/*[in]*/ long nLine,
							/*[in]*/ long nThreadId,
							/*[in]*/ long hrErrorCode,
							/*[in]*/ SIlxStackFrames* pIlxStackFrames,
							/*[in]*/ long nNumStackFrames,
							/*[in]*/ const TCHAR* szMessage, 
							/*[in]*/ long nSeverity,
							/*[in]*/ bool bMsgBox)
{
	HRESULT hr = S_OK;

	// Pre-conditions

	if (!m_bInitCalled)
	{
		_ASSERTE(!"Must call load() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	if (!m_pLogConfig)
	{
		_ASSERTE(m_pLogConfig);
		return E_FAIL;
	}

	// Initialize thread id if needed
	if (nThreadId == -1)
	{
		nThreadId = ::GetCurrentThreadId();
	}

	// Step 1:	Extract the filename and subsystem name from the path

	tstring strFile;
	tstring strSubsystem;

	::ExtractFileAndSubsystem(szSubsystem, szPath, &strFile, &strSubsystem);

	// Step 2:	Retrieve the set of destinations for this message
	//
	//			This queries each route object to determine if it
	//			is a route for the message.  If so, the route's
	//			destinations are added to the set.  One message
	//			may have multiple routes and multiple destinations.

	TypeSetDestinations setLogDestinations;
	m_pLogConfig->getDestinations(	strSubsystem.c_str(),
									strFile.c_str(), 
									nLine, 
									nThreadId,
									hrErrorCode,
									szMessage, 
									nSeverity,
									&setLogDestinations);
	
	// Step 3:	Add the message box destination if required
	//
	//			This is true when the MsgBox() or MsgBox2() methods
	//			are used.

	CMsgBoxDestination* pMsgBox = 0;
	bool bResetModalFlagToFalse = false;

	if (bMsgBox)
	{
		// Step 3a:	First see if we already have a message box in our
		//			destination set
		for(TypeSetDestinations::iterator it = setLogDestinations.begin();
			it != setLogDestinations.end();
			++it)
		{
			CLogDestination* pLogDestination = *it;
			_ASSERTE(pLogDestination);

			pMsgBox = dynamic_cast<CMsgBoxDestination*>(pLogDestination);
			if (pMsgBox)
			{
				// The MsgBox pointer will be stored in pMsgBox
				break;
			}
		}

		// Step 3b:	Otherwise, add the first message box destination from our
		//			master list of destinations
		for(it = m_setAllDestinations.begin();
			!pMsgBox && it != m_setAllDestinations.end();
			++it)
		{
			CLogDestination* pLogDestination = *it;
			_ASSERTE(pLogDestination);

			pMsgBox = dynamic_cast<CMsgBoxDestination*>(pLogDestination);
			if (pMsgBox)
			{
				// Add the message box to our destination set
				setLogDestinations.insert(pMsgBox);
				break;
			}
		}

		_ASSERTE(pMsgBox);

		if (pMsgBox)
		{
			// Initialize the desintation if it hasn't been initialized
			if (!pMsgBox->isInitialized())
			{
				pMsgBox->initialize();
			}

			if (!pMsgBox->isModal())
			{
				// The dialog should be modal when MsgBox() or MsgBox2()
				// are called.  Otherwise, the dialog runs in a separate
				// thread.
				pMsgBox->setIsModal(true);
				bResetModalFlagToFalse = true;
			}

			_ASSERTE(pMsgBox->isModal());
		}
	}

	if (setLogDestinations.size() == 0)
	{
		_ASSERTE(!bMsgBox); // There should always be one dest for MsgBox()

		return S_FALSE;		// Message was not logged (no route specified)
	}

	// Step 4:	Send the message to each log file destination first
	//
	//			This ensures that the log message is saved to some
	//			persistent storage before the user sees an error 
	//			dialog box (which may hang or the user might kill).

	TypeSetDestinations::iterator it = setLogDestinations.begin();
	while (it != setLogDestinations.end())
	{
		CLogDestination* pLogDestination = *it;
		_ASSERTE(pLogDestination);

		// Only send message to log file
		if (dynamic_cast<CLogFileDestination*>(pLogDestination) != 0)
		{
			hr = pLogDestination->log(	strSubsystem.c_str(), 
										szPath,
										strFile.c_str(), 
										nLine, 
										nThreadId,
										hrErrorCode,
										pIlxStackFrames,
										nNumStackFrames,
										szMessage, 
										nSeverity);
			_ASSERTE(SUCCEEDED(hr));

			it = setLogDestinations.erase(it);
		}
		else
		{
			++it;
		}
	}

	// Step 5:	Send the message to the other destinations

	for(it = setLogDestinations.begin();
		it != setLogDestinations.end();
		++it)
	{
		CLogDestination* pLogDestination = *it;
		_ASSERTE(pLogDestination);

		hr = pLogDestination->log(	strSubsystem.c_str(), 
									szPath,
									strFile.c_str(), 
									nLine, 
									nThreadId,
									hrErrorCode,
									pIlxStackFrames,
									nNumStackFrames,
									szMessage, 
									nSeverity);
		_ASSERTE(SUCCEEDED(hr));
	}

	// Reset modal flag
	if (bMsgBox && pMsgBox && bResetModalFlagToFalse)
	{
		_ASSERTE(pMsgBox->isModal());
		pMsgBox->setIsModal(false);
	}

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CILXLog::LogMsg
//					
// DESCRIPTION	:	Calls LogImpl().  See ILXLogHelper.h for
//					documentation.
//					
//*********************************************************************
STDMETHODIMP CILXLog::LogMsg(	/*[in]*/ BSTR bstrSubsystem, 
								/*[in]*/ BSTR bstrPath, 
								/*[in]*/ long nLine, 
								/*[in]*/ long nThreadId,
								/*[in]*/ BSTR bstrMessage, 
								/*[in]*/ long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions

	if (!m_bInitCalled)
	{
		_ASSERTE(!"Must call load() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!bstrMessage) return E_INVALIDARG;
	
	USES_CONVERSION;

	const TCHAR* szSubsystem	= OLE2CT(bstrSubsystem);
	const TCHAR* szMessage		= OLE2CT(bstrMessage);
	const TCHAR* szPath			= OLE2CT(bstrPath);

	return LogImpl(	szSubsystem, 
					szPath, 
					nLine, 
					nThreadId, 
					nINVALID_HRESULT,
					0,
					0,
					szMessage, 
					nSeverity, 
					false);
}

//*********************************************************************
// FUNCTION		:	CILXLog::LogCrash
//					
// DESCRIPTION	:	Calls LogImpl().  See ILXLogHelper.h for
//					documentation.
//					
//*********************************************************************
STDMETHODIMP CILXLog::LogCrash(	/*[in]*/ BSTR bstrSubsystem, 
								/*[in]*/ BSTR bstrPath,
								/*[in]*/ long nLine,
								/*[in]*/ long nThreadId,
								/*[in]*/ DWORD pExceptionInfoPtr, 
								/*[in]*/ BSTR bstrSearchPath,
								/*[in]*/ BSTR bstrMessage, 
								/*[in]*/ long nSeverity)
{
	USES_CONVERSION;

	HRESULT hr = S_OK;

	// Pre-conditions

	if (!m_bInitCalled)
	{
		_ASSERTE(!"Must call load() first");
		return E_UNEXPECTED;
	}

	// Verify arguments
	if (!bstrMessage || !pExceptionInfoPtr ) 
		return E_INVALIDARG;
	
	
	EXCEPTION_POINTERS* pExceptionInfo = 0;
	pExceptionInfo = reinterpret_cast<EXCEPTION_POINTERS*>(pExceptionInfoPtr);
	if (!pExceptionInfo) 
		return E_INVALIDARG;

	// Step 1:	Convert pExceptionInfo to SIlxStackFrames (use IlxCrashFinder
	//			API)

	// Use statics because we might have gotten an exception due to a low
	// memory condition
	static TCHAR szBuffer[nMAX_ERROR_BUFFER_LENGTH];
	szBuffer[0] = _T('\0');
	
	//setup the parameters for the call
	static ILXStackFrame aryIlxStackFrames[nMAX_STACK_FRAMES];

	hr = GetExceptionInfo(pExceptionInfo, OLE2CT(bstrSearchPath), aryIlxStackFrames, nMAX_STACK_FRAMES);
	if (SUCCEEDED(hr))
	{
		// Step 2:	Append exception info from first stack frame to log message
		//			(see LogHR)
		//
		//			tstring strFullErrorMsg = szMessage + strHYPHEN + szErrorMsg;
		//
		BOOL bRet = FormatStackString1( &aryIlxStackFrames[0], szBuffer, nMAX_ERROR_BUFFER_LENGTH );
		if ( !bRet )
		{
			_tcscpy(szBuffer, _T("ILXCrashFinder failed to locate the crash."));
		}
	}
	else
	{
		_stprintf(szBuffer, _T("ILXCrashFinder failed to locate the crash (0x%08X)."),
			hr);
		_ASSERTE( !"ILXCrashFinder failed to locate the crash." );
	}
	
	const TCHAR* szSubsystem	= OLE2CT(bstrSubsystem);
	const TCHAR* szMessage		= OLE2CT(bstrMessage);
	const TCHAR* szPath			= OLE2CT(bstrPath);

	tstring strFullErrorMsg = szMessage + strHYPHEN + szBuffer;

	return LogImpl(	szSubsystem, 
					szPath, 
					nLine, 
					nThreadId, 
					nINVALID_HRESULT,
					aryIlxStackFrames,
					nMAX_STACK_FRAMES,
					strFullErrorMsg.c_str(),
					nSeverity, 
					false);
}

//*********************************************************************
// FUNCTION		:	CILXLog::LogHR
//					
// DESCRIPTION	:	Calls LogImpl().  See ILXLogHelper.h for
//					documentation.
//					
//*********************************************************************
STDMETHODIMP CILXLog::LogHR(	/*[in]*/ BSTR bstrSubsystem, 
								/*[in]*/ BSTR bstrPath, 
								/*[in]*/ long nLine, 
								/*[in]*/ long nThreadId,
								/*[in]*/ long hrErrorCode, 
								/*[in]*/ BSTR bstrMessage,
								/*[in]*/ long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions

	if (!m_bInitCalled)
	{
		_ASSERTE(!"Must call load() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!bstrMessage) return E_INVALIDARG;

	// Step 1:	Convert HRESULT to string message

	TCHAR szErrorMsg[nMAX_ERROR_BUFFER_LENGTH];

	::GetErrorMessage(	hrErrorCode, 
						m_nLanguageId, 
						m_listResourceModules,
						szErrorMsg,
						nMAX_ERROR_BUFFER_LENGTH);

	// Step 2:	Call LogImpl()

	USES_CONVERSION;

	const TCHAR* szSubsystem	= OLE2CT(bstrSubsystem);
	const TCHAR* szMessage		= OLE2CT(bstrMessage);
	const TCHAR* szPath			= OLE2CT(bstrPath);

	tstring strFullErrorMsg = szMessage + strHYPHEN + szErrorMsg;

	return LogImpl(	szSubsystem, 
					szPath, 
					nLine, 
					nThreadId, 
					hrErrorCode, 
					0,
					0,
					strFullErrorMsg.c_str(), 
					nSeverity,
					false);
}

//*********************************************************************
// FUNCTION		:	CILXLog::MsgBox
//					
// DESCRIPTION	:	Calls LogImpl().  See ILXLogHelper.h for
//					documentation.
//					
//*********************************************************************
STDMETHODIMP CILXLog::MsgBox(	/*[in]*/ BSTR bstrSubsystem, 
								/*[in]*/ BSTR bstrPath, 
								/*[in]*/ long nLine, 
								/*[in]*/ long nThreadId,
								/*[in]*/ long hrErrorCode)
{
	HRESULT hr = S_OK;

	// Pre-conditions

	if (!m_bInitCalled)
	{
		_ASSERTE(!"Must call load() first");
		return E_UNEXPECTED;
	}

	long nSeverity = ::GetHRESULTSeverity(hrErrorCode);

	// Step 1:	Convert HRESULT to string message

	TCHAR szErrorMsg[nMAX_ERROR_BUFFER_LENGTH];

	::GetErrorMessage(	hrErrorCode, 
						m_nLanguageId, 
						m_listResourceModules,
						szErrorMsg,
						nMAX_ERROR_BUFFER_LENGTH);

	// Step 2:	Cal LogImpl()

	USES_CONVERSION;

	const TCHAR* szSubsystem	= OLE2CT(bstrSubsystem);
	const TCHAR* szPath			= OLE2CT(bstrPath);

	return LogImpl(	szSubsystem, 
					szPath, 
					nLine, 
					nThreadId, 
					hrErrorCode, 
					0,
					0,
					szErrorMsg, 
					nSeverity,
					true); // Use MsgBox destination
}

//*********************************************************************
// FUNCTION		:	CILXLog::MsgBox2
//					
// DESCRIPTION	:	Calls LogImpl().  See ILXLogHelper.h for
//					documentation.
//					
//*********************************************************************
STDMETHODIMP CILXLog::MsgBox2(	/*[in]*/ BSTR bstrSubsystem, 
								/*[in]*/ BSTR bstrPath, 
								/*[in]*/ long nLine, 
								/*[in]*/ long nThreadId,
								/*[in]*/ BSTR bstrMessage,
								/*[in]*/ long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions

	if (!m_bInitCalled)
	{
		_ASSERTE(!"Must call load() first");
		return E_UNEXPECTED;
	}

	USES_CONVERSION;

	const TCHAR* szSubsystem		= OLE2CT(bstrSubsystem);
	const TCHAR* szPath				= OLE2CT(bstrPath);
	const TCHAR* szDetailedErrorMsg	= OLE2CT(bstrMessage);

	return LogImpl(	szSubsystem, 
					szPath, 
					nLine, 
					nThreadId,
					nINVALID_HRESULT, 
					0,
					0,
					szDetailedErrorMsg, 
					nSeverity,
					true); // Use MsgBox destination
}

//*********************************************************************
// FUNCTION		:	CILXLog::Fire_OnLog
//					
// DESCRIPTION	:	NOT USED
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   bstrSubsystem	
//   bstrPath	
//   nLine	
//   nThreadId	
//   hrErrorCode	
//   bstrMessage	
//   nSeverity	
//					
// COMMENTS		:	We cannot fire events because we are aggregating
//					the free-threaded marshaller.  This means that
//					the event won't be marshalled correctly.  We need
//					to use the Global Interface Table (GIT) instead
//					so that we are thread-safe.
//					
//*********************************************************************
HRESULT CILXLog::Fire_OnLog(BSTR bstrSubsystem, 
							BSTR bstrPath, 
							long nLine, 
							long nThreadId,
							long hrErrorCode, 
							BSTR bstrMessage,
							long nSeverity)
{
	HRESULT hr = S_OK;
	HRESULT hrRet = S_OK;

	// Step 1:	Store the arguments in a VARIANT array

	CComVariant varArgs[] = 
	{
		// Remember it must be in reverse order
		nSeverity,
		bstrMessage,
		hrErrorCode, 
		nThreadId,
		nLine, 
		bstrPath, 
		bstrSubsystem
	};

	DISPPARAMS disp = 
	{ 
		reinterpret_cast<VARIANT*>(&varArgs), 
		NULL, 
		sizeof varArgs / sizeof *varArgs, 
		0 
	};

	// Step 2:	Fire the event to all the connected sinks

	int nConnections = m_vec.GetSize();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; ++nConnectionIndex)
	{
		Lock();
		CComPtr<IUnknown> sp = m_vec.GetAt(nConnectionIndex);
		Unlock();

		IDispatch* piDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (!piDispatch)
		{
			_ASSERTE(!"Log event sink is null!");
			continue;
		}

		// Step 3:	Invoke/fire the event handler

		CComVariant varResult;
		varResult.Clear();
		varResult.scode = S_OK;

		hr = piDispatch->Invoke(	DISPID_ILXLOG_ON_LOG, 
									IID_NULL, 
									LOCALE_USER_DEFAULT, 
									DISPATCH_METHOD, 
									&disp, 
									&varResult, 
									NULL, 
									NULL);
		
		// Deal with failure (log it and continue)
		if (FAILED(hr))
		{
			// Only keep track of the first error
			hrRet = FAILED(hrRet) ? hrRet : hr;
			AtlTrace(_T("Failed to fire the OnLog event.\n"));
		}
		
		// Deal with success
		else if (hrRet == S_OK)
		{
			// Only keep track of the first non-success code returned by the
			// event handler
			hrRet = varResult.scode;
		}
	}

	return hrRet;
}

//*********************************************************************
// FUNCTION		:	CILXLog::LoadRegistrySettings
//					
// DESCRIPTION	:	Reads the log directory from the registry.
//					Also, adds the rdcerrors.dll module to the list
//					of message resource modules.
//					
// RETURNS		:	
//   HRESULT		
//					
//*********************************************************************
HRESULT CILXLog::LoadRegistrySettings()
{
	USES_CONVERSION;

	HRESULT hr = S_OK;

	HKEY hkResult = 0; 
	ULONG cbValue = 10;
	DWORD dwType = REG_SZ;

	// Step 1:	Open the root RDC key

	long lRet = ::RegOpenKeyExW(HKEY_CURRENT_USER, 
								szRDC_REGISTRY_PATH, 
								NULL, 
								KEY_ALL_ACCESS, 
								&hkResult); 

	if (ERROR_SUCCESS != lRet)
	{
		return HRESULT_FROM_WIN32(::GetLastError());
	}

	// Step 2:	Get the Language id

	dwType = REG_DWORD;
	cbValue = sizeof(DWORD);
	lRet = RegQueryValueExW(	hkResult, 
								szLANG_REG_KEY,
								NULL, 
								&dwType, 
								(unsigned char*) &m_nLanguageId, 
								&cbValue); 
	
	if (ERROR_SUCCESS != lRet) 
	{
		m_nLanguageId = 0; //failed
	}

	RegCloseKey(hkResult);

	wchar_t szRdcDirectory[_MAX_PATH+1];
	*szRdcDirectory = _T('\0');
	hr = ::GetILXDirectoryFromRegistry(	W2CT(szSHARED_RDC_REG_KEY),
										szRdcDirectory,
										_MAX_PATH);

	_bstr_t bstrRdcErrorsFile = szRdcDirectory;

	if (bstrRdcErrorsFile.length() > 0)
	{
		bstrRdcErrorsFile += _T("\\");
	}
	
	bstrRdcErrorsFile += szRDC_ERRORS_FILE;
	
	hr = AddResource(bstrRdcErrorsFile);
	if (FAILED(hr)) return hr;

	return S_OK;
}
