///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/ILXLogImpl.h,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	ILXLogImpl.h
//   Created By:	Marc Eaddy
//   Created on:	Wed, April 18, 2000
//
//      Purpose:	Logs messages to user customizable log destinations
//
//  Description:	Implements the IlxLog coclass and IILXLog interface
//					to transparently log messages to multiple destinations.
//
//					Basically, this class converts HRESULTs into message
//					strings and then routes the message to the appropriate
//					LogDestinations.
//
//        Usage:	The only client of this COM object should be the 
//					CILXLogHelper singleton class.  Not intended to be
//					used by anyone else. (see ILXLogHelper.h).  Currently,
//					IlxScripting uses the COM object directly but this is
//					not safe because the object aggregates the free threaded
//					marshaller.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: ILXLogImpl.h,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 30    7/19/02 5:35p Zakirh
 * o LogCrash added a xtra paramter.
 * 
 * 29    7/17/02 3:40p Marce
 * o Finalized how to pass exception info
 * 
 * 28    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 25    11/15/00 2:31p Marce
 * o Removed the original LogHR().  Renamed LogHR2() to LogHR().
 * 
 * 24    11/10/00 4:04p Marce
 * o Cosmetic
 * 
 * 22    11/07/00 5:33p Marce
 * o Added comments
 * 
 * 21    10/05/00 1:47p Marce
 * o Clarified parameter names
 * 
 * 20    10/04/00 5:39p Marce
 * o Removed Initialize() and load()
 * o Added LoadConfig() and UseLogFile()
 * o Implemented log filtering
 * 
 * 19    10/03/00 6:33p Marce
 * o Added log configuration classes
 * 
 * 18    10/03/00 2:19p Marce
 * o Comment
 * 
 * 17    9/25/00 1:41p Marce
 * o Changed WriteLogHeader() to CreateLogFile() which combines log file
 * backup, creation, and writing the header
 * o Added two resources for locating errors
 * 
 * 16    9/22/00 12:37p Marce
 * o Now includes thread id in the log message
 * 
 * 15    8/29/00 12:22p Marce
 * o Previously when the log file was backed up it would start empty.  Now
 * it outputs the log header again and a log message stating where the
 * backup file is.
 * o Fixed a possible bug that would occur when the file was backed-up
 * since the seek position array was not reinitialized properly.
 * 
 * 14    6/22/00 8:36p Marce
 * o Now uses the "Log Directory" from the registry
 * 
 * 13    5/20/00 8:51p Marce
 * o Changed "Log" method to "LogMsg" because it was getting confused for
 * the logarithm function in VB Script!
 * 
 * 12    5/13/00 8:16p Marce
 * o Added OnLog event.  Don't let this fool you, it doesn't work! (see
 * ReadMe)
 * 
 * 11    5/13/00 12:49a Marce
 * o Added the ability to pop-up a message box with a string message
 * 
 * 10    5/08/00 3:40p Marce
 * o Added configuration method
 * 
 * 9     5/04/00 1:58p Marce
 * o Now aggregates the FTM so we can be called from any apartment or
 * thread.  However, I don't know if the event log handle has thread
 * affinity.
 * 
 * 8     5/01/00 11:32a Marce
 * o Now logs source file and line number
 * 
 * 7     4/25/00 11:24p Marce
 * o Now includes the log file in the detail section of the error dialog
 * 
 * 6     4/25/00 9:51p Marce
 * o Now the message box only takes an HRESULT
 * 
 * 5     4/25/00 8:31p Marce
 * o Added support for including resource modules to search when resolving
 * HRESULTs
 * o Changed "Module" to "Subsystem" to avoid confusion with resource
 * modules
 * 
 * 4     4/25/00 2:50p Marce
 * o Added comments
 * 
 * 3     4/25/00 1:33p Marce
 * o Now uses 4 severities instead of 30 levels
 * o Separated out error dialog popup from normal logging
 * 
 * 2     4/24/00 11:47a Marce
 * 
 * 1     4/24/00 11:27a Marce
 */

#ifndef INCLUDED_ILX_LOG_H
#define INCLUDED_ILX_LOG_H

#include "XmlElement.h"	// XML Persistence

#include "ILXLog2.h"	// IIDs, ifaces, LIBIDs, etc.

#include "resource.h"   // main symbols

#include <atlctl.h>		// ATL IObjectSafetyImpl<>

#include <list>			// STL list<>
#include <vector>		// STL vector<>
#include <set>			// STL set<>
#include <string>		// basic_string

// Forward declarations

class CLogConfig;
class CLogDestination;
struct SIlxStackFrames;

// Type defines

typedef std::basic_string<TCHAR>		tstring;
typedef std::pair<HMODULE, bool>		TypeModuleAndBool;
typedef std::list<TypeModuleAndBool>	TypeListModules;
typedef std::set<CLogDestination*>		TypeSetDestinations;

//*****************************
// CILXLog
//*****************************

class ATL_NO_VTABLE CILXLog : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<CILXLog, &CLSID_ILXLog>,
    public IObjectSafetyImpl<CILXLog, INTERFACESAFE_FOR_UNTRUSTED_CALLER | INTERFACESAFE_FOR_UNTRUSTED_DATA>,
	public IDispatchImpl<IILXLog, &IID_IILXLog, &LIBID_ILXLOGLib>,
	public IConnectionPointContainerImpl<CILXLog>,
	public IConnectionPointImpl<CILXLog, &DIID_DILXLogEvents, CComDynamicUnkArray>,
	public IProvideClassInfo2Impl<&CLSID_ILXLog, &DIID_DILXLogEvents, &LIBID_ILXLOGLib>,
	public CXmlElement
{
public:
	CILXLog();
	virtual ~CILXLog();
	HRESULT FinalConstruct();
	void FinalRelease();

DECLARE_REGISTRY_RESOURCEID(IDR_ILXLOG)

DECLARE_PROTECT_FINAL_CONSTRUCT()

DECLARE_GET_CONTROLLING_UNKNOWN()

BEGIN_COM_MAP(CILXLog)
	COM_INTERFACE_ENTRY(IILXLog)
	COM_INTERFACE_ENTRY(IDispatch)
    COM_INTERFACE_ENTRY(IObjectSafety)
	COM_INTERFACE_ENTRY(IConnectionPointContainer)
	COM_INTERFACE_ENTRY(IProvideClassInfo)
	COM_INTERFACE_ENTRY(IProvideClassInfo2)

	// We aggregate the free-threaded marshaller.  That means that
	// calls to IILXLog are never marshalled.
	COM_INTERFACE_ENTRY_AGGREGATE(IID_IMarshal, m_pUnkMarshaler.p)
END_COM_MAP()

BEGIN_CONNECTION_POINT_MAP(CILXLog)
	CONNECTION_POINT_ENTRY(DIID_DILXLogEvents)
END_CONNECTION_POINT_MAP()

	// 
	// CILXLog Methods
	// 

public:
	//--------------------------
	// XML Persistence Methods
	//--------------------------

	HRESULT load(MSXML::IXMLDOMNode* piNode);
		//
		// Loads the event table and scripts from the node.
		//
		// PARAMS:
		//
		//	piNode - Must be a valid SCRIPTING element.
		//

	CXmlElement* clone() const;

public:

	// 
	// IILXLog Methods
	//
	// (see ILXLogHelper.h for descriptions)
	// 

	STDMETHOD(LoadConfig)(/*[in]*/ IUnknown* piXmlNode);

	STDMETHOD(AddResource)(	/*[in]*/ BSTR bstrFilename);

	STDMETHOD(LogMsg)(		/*[in]*/ BSTR bstrSubsystem, 
							/*[in]*/ BSTR bstrPath, 
							/*[in]*/ long nLine, 
						    /*[in]*/ long nThreadId, 
							/*[in]*/ BSTR bstrMessage, 
							/*[in]*/ long nSeverity);

	STDMETHOD(LogHR)(		/*[in]*/ BSTR bstrSubsystem, 
							/*[in]*/ BSTR bstrPath, 
							/*[in]*/ long nLine, 
							/*[in]*/ long nThreadId, 
							/*[in]*/ long hrErrorCode, 
							/*[in]*/ BSTR bstrMessage,
							/*[in]*/ long nSeverity);

	STDMETHOD(MsgBox)(		/*[in]*/ BSTR bstrSubsystem, 
							/*[in]*/ BSTR bstrPath, 
							/*[in]*/ long nLine, 
							/*[in]*/ long nThreadId, 
							/*[in]*/ long hrErrorCode);

	STDMETHOD(MsgBox2)(		/*[in]*/ BSTR bstrSubsystem, 
							/*[in]*/ BSTR bstrPath, 
							/*[in]*/ long nLine, 
							/*[in]*/ long nThreadId, 
							/*[in]*/ BSTR bstrMessage,
							/*[in]*/ long nSeverity);

	STDMETHOD(UseLogFile)(	/*[in]*/ BSTR bstrLogFilename, 
							/*[in]*/ long nMaxFileSize);

	STDMETHOD(LogCrash)(	/*[in]*/ BSTR bstrSubsystem, 
							/*[in]*/ BSTR bstrPath,
							/*[in]*/ long nLine,
							/*[in]*/ long nThreadId,
							/*[in]*/ DWORD pExceptionInfoPtr, 
							/*[in]*/ BSTR bstrSearchPath,
							/*[in]*/ BSTR bstrMessage, 
							/*[in]*/ long nSeverity);

private: // HELPER FUNCTIONS

	HRESULT LoadRegistrySettings();
		//
		// Retrieves the log directory from the registry
		//

	HRESULT LogImpl(/*[in]*/ const TCHAR* szSubsystem, 
					/*[in]*/ const TCHAR* szPath, 
					/*[in]*/ long nLine, 
					/*[in]*/ long nThreadId, 
					/*[in]*/ long hrErrorCode, 
					/*[in]*/ SIlxStackFrames* pvIlxStackFrames,
					/*[in]*/ long nNumStackFrames,
					/*[in]*/ const TCHAR* szMessage, 
					/*[in]*/ long nSeverity,
					/*[in]*/ bool bMsgBox);
		//
		// Calls BackupFile() and then LogImplNoBackup()
		//

	void AddResource(HMODULE hModule, bool bShouldFreeLibrary);

	HRESULT Initialize();

	HRESULT Fire_OnLog(		/*[in]*/ BSTR bstrSubsystem, 
							/*[in]*/ BSTR bstrPath, 
							/*[in]*/ long nLine, 
							/*[in]*/ long nThreadId, 
							/*[in]*/ long hrErrorCode, 
							/*[in]*/ BSTR bstrMessage,
							/*[in]*/ long nSeverity);
		//
		// This does not work and is not used!
		//

private: // NOT IMPLEMENTED
	CILXLog(const CILXLog&);				// COPY CTOR
	CILXLog& operator=(const CILXLog&);		// ASSIGNMENT

public: // DATA
	TypeListModules		m_listResourceModules;
		//
		// List of modules (DLLs/EXEs) with message resources to use
		// when resolving HRESULTs.
		//
	
	long				m_nLanguageId;		// Language to use (English)

	CComPtr<IUnknown>	m_pUnkMarshaler;	// For aggregating FTM

	CLogConfig*			m_pLogConfig;		// Message routing information

	TypeSetDestinations	m_setAllDestinations;
		//
		// Set of all possible log destinations.
		//

	bool				m_bInitCalled;		// Has init been called yet?
};

#endif //INCLUDED_ILX_LOG_H
