///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/LogConfig.cpp,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	LogConfig.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/03/2000 3:13:36 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: LogConfig.cpp,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 6     11/10/00 4:12p Marce
 * o Removed unused header
 * 
 * 5     11/10/00 4:11p Marce
 * o Added comments
 * 
 * 3     10/04/00 5:38p Marce
 * o Implemented log filtering
 * 
 * 2     10/03/00 6:33p Marce
 * o Added log configuration classes
 */

#include "StdAfx.h"

#include "LogConfig.h"				// Me

#include "LogRoute.h"				// CLogRoute

// Constants

static const TCHAR* const szNAME_ATTR	= _T("NAME");

//////////////////////////////
// LOG CONFIG IMPLEMENTATION
//////////////////////////////

// Initialize static members

const tstring CLogConfig::s_strXML_TAG_NAME = _T("CONFIG");

CLogConfig::CLogConfig()
	: CXmlElement(s_strXML_TAG_NAME.c_str())
{ 
	addChildPrototype(new CLogRoute);
}

CLogConfig::~CLogConfig()
{ }

CXmlElement* CLogConfig::clone() const
{
	return new CLogConfig(); // Create new object not a clone
}

//*********************************************************************
// FUNCTION		:	CLogConfig::initDestinations
//					
// DESCRIPTION	:	Initialize all the routes with their destinations.
//					
// PARAMETERS	:	
//   plistDestinations	
//					
// COMMENTS		:	Must be called after load().
//					
//*********************************************************************
void CLogConfig::initDestinations(TypeSetDestinations* plistDestinations)
{
	for(TypeVecXmlElementPtrs::iterator it = getChildren().begin(); it != getChildren().end(); ++it)
	{
		CXmlElement* pXmlElement = *it;
		_ASSERTE(pXmlElement);

		CLogRoute* pLogRoute = dynamic_cast<CLogRoute*>(pXmlElement);
		if (pLogRoute)
		{
			pLogRoute->initDestinations(plistDestinations);
		}
	}
}

//*********************************************************************
// FUNCTION		:	CLogConfig::getDestinations
//					
// DESCRIPTION	:	Retrieves all the destinations for this message.
//					
// PARAMETERS	:	
//   szSubsystem	
//   szFile	
//   nLine	
//   nThreadId	
//   hrErrorCode	
//   szMessage	
//   nSeverity	
//   psetDestinations	
//					
// COMMENTS		:	
//					
//*********************************************************************
void CLogConfig::getDestinations(	/*[in]*/ const TCHAR* szSubsystem, 
									/*[in]*/ const TCHAR* szFile, 
									/*[in]*/ long nLine, 
									/*[in]*/ long nThreadId, 
									/*[in]*/ long hrErrorCode, 
									/*[in]*/ const TCHAR* szMessage, 
									/*[in]*/ long nSeverity,
									/*[out]*/ TypeSetDestinations* psetDestinations)
{
	// Verify arguments
	_ASSERTE(psetDestinations);
	_ASSERTE(psetDestinations->size() == 0);
	_ASSERTE(szMessage);
	
	for(TypeVecXmlElementPtrs::iterator it = getChildren().begin();
		it != getChildren().end();
		++it)
	{
		CXmlElement* pChild = *it;
		_ASSERTE(pChild);

		CLogRoute* pRoute = dynamic_cast<CLogRoute*>(pChild);

		if (pRoute && pRoute->matches(	szSubsystem, 
										szFile, 
										nLine, 
										nThreadId, 
										hrErrorCode, 
										szMessage, 
										nSeverity) )
		{
			TypeSetDestinations& destinations = pRoute->getDestinations();

			for(TypeSetDestinations::iterator it_dest = destinations.begin();
				it_dest != destinations.end();
				++it_dest)
			{
				psetDestinations->insert(*it_dest);
			}
		}
	}
}
