///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/LogConfig.h,v 1.2 2004/09/07 06:06:04 me133 Exp $
//
//    File name:	LogConfig.h
//   Created By:	Marc Eaddy
//   Created On:	10/03/2000 3:11:02 PM
//
//      Purpose:	Collection of CLogRoutes
//
//  Description:	Determines which log destinations a message should
//					be sent to by asking each of its child log route
//					objects.
//
//					Log configuration objects are persisted to XML as
//					follows:
//
//					<CONFIG NAME="name">
//						<ROUTE .../> (zero or more)
//					</CONFIG>
//
//					Here is an example:
//
//					<config NAME="RELEASE">
//						<route SEVERITY="ERROR,WARNING" DEST="FiniteLogFile"/>
//					</config>
//
//					The NAME attribute must match the CONFIG_TO_USE
//					attribute of the parent <ILXLOG> element.  You can
//					have multiple configurations in the log settings
//					file.  The CONFIG_TO_USE attribute is used to
//					determine which config to use.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: LogConfig.h,v $
 * Revision 1.2  2004/09/07 06:06:04  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 4     11/08/00 11:22a Marce
 * o Comments
 * 
 * 3     10/04/00 5:38p Marce
 * o Implemented log filtering
 * 
 * 2     10/03/00 6:33p Marce
 * o Added log configuration classes
 */

#ifndef INCLUDED_LOG_CONFIG_H
#define INCLUDED_LOG_CONFIG_H

#include "XmlElement.h"	// Parent

#include <set>			// STL set<>

// Forward declarations
class CLogDestination;

// Type defines

typedef std::set<CLogDestination*>	TypeSetDestinations;

//*****************************
// CLogConfig
//*****************************

class CLogConfig : public CXmlElement
{
public:

	CLogConfig();
	virtual ~CLogConfig();

public:

	void initDestinations(TypeSetDestinations* plistDestinations);
		// Initializes all the route objects with the list of all
		// the log destinations.  Must be called before 
		// getDestinations().

	void getDestinations(	/*[in]*/ const TCHAR* szSubsystem, 
							/*[in]*/ const TCHAR* szFile, 
							/*[in]*/ long nLine, 
							/*[in]*/ long nThreadId, 
							/*[in]*/ long hrErrorCode, 
							/*[in]*/ const TCHAR* szMessage, 
							/*[in]*/ long nSeverity,
							/*[out]*/ TypeSetDestinations* psetDestinations);
		//
		// Returns a list of destinations for this message.
		//
		// Each log route object is queried to determine if it
		// is a route for the message.  If it is, the destinations
		// assigned to the log route are added to the set of
		// destinations that is returned.
		//
		// Ex: Suppose we have the following two routes:
		//
		// <route SEVERITY="ERROR,WARNING" DEST="InfiniteLogFile,DebugOut"/>
		// <route SEVERITY="INFO,SUCCESS" DEST="DebugOut"/>
		//
		// In the case of an error or warning message, the destinations
		// returned would be for the InfiniteLogFile and DebugOut.  For
		// an info or success message, the destination returned would be
		// for DebugOut.
		//

protected:
	// Overrides CXmlElement
	virtual CXmlElement* clone() const;
		//
		// Clones this element.
		//

private: // NOT IMPLEMENTED
	CLogConfig(const CLogConfig&);
	CLogConfig& operator=(const CLogConfig&);

public:
	static const tstring s_strXML_TAG_NAME;
};

#endif // INCLUDED_LOG_CONFIG_H