///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/LogFileDestination.cpp,v 1.2 2004/09/07 06:06:05 me133 Exp $
//
//    File name:	LogFileDestination.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/03/2000 1:34:59 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: LogFileDestination.cpp,v $
 * Revision 1.2  2004/09/07 06:06:05  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 29    8/21/02 8:35p Zakirh
 * o Fixed Empty lines in the log file.
 * 
 * 28    7/26/02 6:01p Marce
 * o Fixed compiler warning
 * 
 * 27    7/26/02 3:48p Zakirh
 * o now passes one xtra char to the format function for null termination.
 * 
 * 26    7/22/02 1:34p Marce
 * o Reformatted call stack
 * 
 * 25    7/19/02 5:34p Zakirh
 * o Added logging of Stack Frames.
 * 
 * 24    7/17/02 4:32p Marce
 * o Fixed bug that prevents IlxLog from working
 * 
 * 23    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 22    7/17/02 12:56p Marce
 * o Now handles the case when Log and SharedRDC directories have trailing
 * slashes
 * 
 * 20    4/25/02 2:56a Marce
 * o Now ThreadId is 5 digits so that all the messages line up properly
 * 
 * 18    11/21/01 4:17p Zakirh
 * merged in with changes made by marce from hopewell......
 * 
 * 17    1/30/01 10:51a Marce
 * o Fixed compile errors for Debug build
 * 
 * 15    10/25/00 4:27p Marce
 * o Comments
 * 
 * 14    10/25/00 3:56p Marce
 * o Added comments and cleaned up some code
 * 
 * 13    10/25/00 2:54p Marce
 * o You can now specify -1 to use the default max log file size
 * 
 * 12    10/17/00 3:01p Marce
 * o Removed assertions.  These seem to happen at a bad time and caused
 * the app to hang.
 * 
 * 11    10/12/00 7:14p Marce
 * o Added critical section checking for getLogHistory()
 * 
 * 10    10/12/00 5:32p Marce
 * o Added some diagnostics (assertions)
 * 
 * 9     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 8     10/10/00 2:36p Marce
 * o Minor changes to assert statements
 * 
 * 7     10/05/00 6:39p Marce
 * o Checking wrong condition
 * 
 * 6     10/05/00 6:13p Marce
 * o fprintf is not needed since we are only writing a string (not a
 * formatted string)
 * 
 * 5     10/05/00 5:26p Garys
 * o Now retrieves the error code when failing to write to the log file
 * 
 * 4     10/05/00 1:48p Marce
 * o Now gives the correct subsystem and file to CLogConfig and
 * destination objects
 * 
 * 3     10/04/00 5:38p Marce
 * o Implemented log filtering
 * 
 * 2     10/03/00 6:33p Marce
 * o Added log configuration classes
 */

#include "StdAfx.h"

#include "LogFileDestination.h"	// Me

#include "ILXLog2.h"			// ILXLOG__SUCCESS, etc.

#include "HelperFunctions.h"	// GetSeverity()

#include <io.h>					// _open()
#include <sys/stat.h>			// _fstat()
#include <fcntl.h>				// _O_WRONLY

using namespace std;

// Constants

static const double nLOG_FILE_FORMAT_VER	= 2.2;
	// Change this number if you change the printf format in
	// logImplNoBackup().  Be careful!  The LogViewerCtrl will
	// only parse your log file if the format is one it 
	// recognizes.

static const long nDEFAULT_MAX_FILE_SIZE	= 65536;	// Bytes
static const long nNO_MAX_SIZE				= 0;		// Do not wrap file
static const long nUSE_DEFAULT_MAX_SIZE		= -1;		// Use nDEFAULT_MAX_FILE_SIZE
static const long nONE_LOG_LINE_MAX_LENGTH	= 4096;		// Characters
static const long nMAX_LOG_HISTORY			= 20;		// Lines

static const wchar_t* szRDC_REGISTRY_PATH	= L"Software\\ILX\\RDC";
static const wchar_t* szLOG_DIR_REG_KEY		= L"Log Directory";

static const wchar_t* const szILX_LOG_MODULE =  L"ILXLOG";

static const TCHAR* const szSUB_TYPE		= _T("LogFile");
static const TCHAR* const szBACKUP_EXT		= _T(".bak");

static const TCHAR* const szFILE_ATTR		= _T("FILE");
static const TCHAR* const szMAX_SIZE_ATTR	= _T("MAX_SIZE");
static const TCHAR* const szBACKUP_ATTR		= _T("SAVE_BACKUP");
static const wchar_t* const szUSE_DEFAULT	= L"default";
static const TCHAR szNewLine				= _T('\n');

// Helper functions
HRESULT GetLogDirectory(tstring* pstrLogDirectory);

////////////////////////////////////////
// LOG FILE DESTINATION IMPLEMENTATION
////////////////////////////////////////

CLogFileDestination::CLogFileDestination()
	: CLogDestination(szSUB_TYPE)
	, m_fileLog(0)
	, m_nMaxFileSize(nNO_MAX_SIZE)
	, m_nLogLineCount(0)
	, m_vecLogHistorySeekPos(nMAX_LOG_HISTORY)
	, m_strLogFile()
	, m_stStartTime()
	, m_bSaveBackup(true)
{ }

CLogFileDestination::~CLogFileDestination()
{ 
	if( m_fileLog )
	{
		fclose(m_fileLog);
		m_fileLog = NULL;
	}
}

CXmlElement* CLogFileDestination::clone() const
{
	return new CLogFileDestination(); // We create a new object not a clone
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::initialize
//					
// DESCRIPTION	:	Creates the log file and backups up the old log
//					file if necessary.
//					
// RETURNS		:	
//   HRESULT 		
//					
// COMMENTS		:	
//					
//*********************************************************************
HRESULT CLogFileDestination::initialize()
{
	HRESULT hr = S_OK;

	// Invariants: Only call initialize() once
	_ASSERTE(!m_fileLog);
	_ASSERTE(!m_bIsInitialized);

	// Step 1:	Initialize member variables

	// Get the log file name from the XML attributes read in by load()
	// unless setLogFile() has already been called.
	if (m_strLogFile.length() == 0)
	{
		const _variant_t& varFileName = m_mapAttributes[szFILE_ATTR];
		if (varFileName.vt != VT_BSTR || !*varFileName.bstrVal)
		{
			_ASSERTE(false);
			// Log message
			return E_FAIL;
		}
		else
		{
			m_strLogFile = static_cast<_bstr_t>(varFileName);
		}
	}

	// Get the max log file size from the XML attributes read in by load()
	// unless setMaxSize() has already been called.
	if (m_nMaxFileSize == nNO_MAX_SIZE)
	{
		const _variant_t& varMaxSize = m_mapAttributes[szMAX_SIZE_ATTR];
		if (varMaxSize.vt != VT_BSTR || !*varMaxSize.bstrVal)
		{
			// Max size not specified so use infinite
			m_nMaxFileSize = nNO_MAX_SIZE;
		}
		else if (wcsicmp(varMaxSize.bstrVal, szUSE_DEFAULT) == 0)
		{
			// Max size is "default" so use default size
			m_nMaxFileSize = nDEFAULT_MAX_FILE_SIZE;
		}
		else
		{
			// Use the specified size
			m_nMaxFileSize = varMaxSize;
		}
	}

	// Get the backup flag from the XML attributes read in by load().
	const _variant_t& varBackup = m_mapAttributes[szBACKUP_ATTR];

	if (varBackup.vt != VT_BSTR || 
		!*varBackup.bstrVal ||
		wcsicmp(varBackup.bstrVal, szUSE_DEFAULT) == 0)
	{
		// Default to true if the BACKUP attribute is missing, empty,
		// or says "default"
		m_bSaveBackup = true;
	}
	else
	{
		// Otherwise, use the value ("true" or "false") specified
		m_bSaveBackup = varBackup;
	}

	// Get the local time used when outputing the log file header
	::GetLocalTime( &m_stStartTime ); 

	// Step 2:	Read the log directory from the registry
	
	tstring strLogDirectory;
	hr = ::GetLogDirectory(&strLogDirectory);
	_ASSERTE(SUCCEEDED(hr));

	// Create the full path to the log file
	if (!strLogDirectory.empty())
	{
		m_strLogFile = strLogDirectory + _T("\\") + m_strLogFile;
	}

	// Step 3:	Create the new log file, backing up the old
	//			log file if necessary
	
	hr = createLogFile(); 
	if (FAILED(hr)) return hr;

	// We are now initialized!
	m_bIsInitialized = true;

	// Post-conditions
	_ASSERTE(m_fileLog && m_bIsInitialized);

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::setLogFile
//					
// DESCRIPTION	:	Sets the log file name.
//					
// PARAMETERS	:	
//   szFile	
//					
// COMMENTS		:	Should only be used if load() is not called.
//					
//*********************************************************************
void CLogFileDestination::setLogFile(const TCHAR* szFile)
{
	_ASSERTE(szFile && *szFile);

	m_strLogFile = szFile;
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::setMaxSize
//					
// DESCRIPTION	:	Sets the max log file size.
//					
// RETURNS		:	
//   void 		
//					
// PARAMETERS	:	
//   nMaxSize	
//					
// COMMENTS		:	Should only be used if load() is not called.
//					
//*********************************************************************
void CLogFileDestination::setMaxSize(long nMaxSize)
{
	m_nMaxFileSize = nMaxSize;

	if (m_nMaxFileSize == nUSE_DEFAULT_MAX_SIZE)
	{
		m_nMaxFileSize = nDEFAULT_MAX_FILE_SIZE;
	}
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::log
//					
// DESCRIPTION	:	Backs up the log file if necessary and then calls
//					logNoBackup().
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	See header file for info.
//					
//*********************************************************************
HRESULT CLogFileDestination::log(	const TCHAR* szSubsystem, 
									const TCHAR* szPath,
									const TCHAR* szFilename, 
									long nLine, 
									long nThreadId, 
									long hrErrorCode,
									SIlxStackFrames* pvIlxStackFrames,
									long nNumStackFrames,
									const TCHAR* szMessage, 
									long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions
	if (!m_bIsInitialized)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_UNEXPECTED;
	}

	if (!m_fileLog)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_FAIL;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	// Backup the file if necessary
	
	hr = backupFile();
	_ASSERTE(SUCCEEDED(hr));
	
	// Copy the message 
	tstring strFullErrorMsg = szMessage;

	if ( pvIlxStackFrames ) //if valid stack frames pointer
	{
		// add stack frames info to the message
		for ( int nIndex = 0; nIndex < nNumStackFrames; nIndex++)
		{
			if (nIndex == 0)
			{
				strFullErrorMsg += _T("\n\n\tCall Stack:\n");
			}

			TCHAR szBuffer[nONE_LOG_LINE_MAX_LENGTH+1];
			*szBuffer = _T('\0');

			// Format the stack frame
			BOOL bRet = FormatStackString2( &pvIlxStackFrames[nIndex], 
											szBuffer, 
											nONE_LOG_LINE_MAX_LENGTH);
			_ASSERTE(bRet);
			if ( szBuffer[0] != TCHAR('\0') ) //add a tab and new line only if the buffer has something to be displayed.
			{
				strFullErrorMsg += _T('\t');
				strFullErrorMsg += szBuffer;	// add the buffer
				strFullErrorMsg += szNewLine;	// append a new Line
			}
		}
	}

	// Write the message to the log file

	return logNoBackup(szSubsystem, szFilename, nLine, nThreadId, strFullErrorMsg.c_str(), nSeverity);
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::createLogFile
//					
// DESCRIPTION	:	
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   pstrBackupFilename	
//					
// COMMENTS		:	
//					
//*********************************************************************
HRESULT CLogFileDestination::createLogFile(tstring* pstrBackupFilename)
{
	// Close the log file if it is already open

	if (m_fileLog)
	{
		fclose(m_fileLog);
		m_fileLog = 0;
	}

	// Initialize out parameters
	if (pstrBackupFilename)
	{
		*pstrBackupFilename = _T("");
	}

	// Step 1:	Backup the log file

	if (m_bSaveBackup)
	{
		// See if the log file exists
		int lFile = _topen(m_strLogFile.c_str(), _O_WRONLY );
		if( -1 != lFile )
		{
			_close( lFile );

			// Create the backup file name
			tstring strBackupFilename = m_strLogFile + szBACKUP_EXT;
			
			// Remove the backup file
			_tremove(strBackupFilename.c_str());
			
			// Rename the log file to the backup file
			if (_trename(m_strLogFile.c_str(), strBackupFilename.c_str()) == -1 )
			{
				_ASSERTE(!"Failed to rename log file.  Backup failed.");
			}

			if (pstrBackupFilename)
			{
				*pstrBackupFilename = strBackupFilename.c_str();
			}
		}
	}

	// Step 2:	Create the new log file

	m_fileLog = _tfopen(m_strLogFile.c_str(), _T("w+"));
	if (!m_fileLog)
	{
		return HRESULT_FROM_WIN32(::GetLastError());
	}

	// Step 3:	Write the log file header

	long nBytes = _ftprintf(m_fileLog,
		_T("\n**  Started %s on %02ld/%02ld %2ld:%02ld:%02ld **\n")
							_T("**  Log File Format Version: %.1f **\n"),
							m_strLogFile.c_str(),
							m_stStartTime.wMonth,
							m_stStartTime.wDay,
							m_stStartTime.wHour,
							m_stStartTime.wMinute,
							m_stStartTime.wSecond,
							nLOG_FILE_FORMAT_VER);
	if (nBytes <= 0)
	{
		long nLastError = ::GetLastError();
		_ASSERTE(!"ftprintf() failed!");
	}

	::fflush(m_fileLog); // Do not buffer the message

	// Reset log line count to 3 (the header is three lines)
	m_nLogLineCount = 3;

	// Clear the log history seek position array since we are
	// starting a brand new file
	fill(m_vecLogHistorySeekPos.begin(), m_vecLogHistorySeekPos.end(), 0);

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::logNoBackup
//					
// DESCRIPTION	:	Actually logs the message.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	See header file.
//					
// COMMENTS		:	Do not change the log file message format (fprintf)
//					without also changing the format version number!
//					Be careful because if you change the version number
//					the LogViewerCtrl may not be able to parse the log
//					file.  You have to coordinate version changes.
//					
//*********************************************************************
HRESULT CLogFileDestination::logNoBackup(	const TCHAR* szSubsystem, 
											const TCHAR* szFilename,
											long nLine,
											long nThreadId,
											const TCHAR* szMessage, 
											long nSeverity)
{
	// Pre-conditions

	if (!m_fileLog)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_FAIL;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	const TCHAR* szSeverity = ::GetSeverity(nSeverity);
	_ASSERTE(szSeverity && *szSeverity);

	// Determine the source filename

	tstring strFile = szFilename;
	tstring strSubsystem = szSubsystem;
	tstring strFileAndLine;

	// Ex: "ActiveILX: "
	if (!strSubsystem.empty())
	{
		strSubsystem += _T(": ");
	}

	// Ex: " [File: ActiveILX.cpp, Line: 1234]"
	// Ex: " [File: ActiveILX.cpp]"
	if (!strFile.empty())
	{
		strFileAndLine += _T(" [File: ") + strFile;

		if (nLine >= 0)
		{
			TCHAR szLineBuf[50];
			_stprintf(szLineBuf, _T(", Line: %d]"), nLine);
			strFileAndLine += szLineBuf;
		}
		else
		{
			strFileAndLine += _T("]");
		}
	}

	SYSTEMTIME SystemTime;
	::GetLocalTime( &SystemTime ); 

	TCHAR acLogString[nONE_LOG_LINE_MAX_LENGTH+1];
	_sntprintf(acLogString, 
		nONE_LOG_LINE_MAX_LENGTH, 
		_T("%02ld/%02ld %2ld:%02ld:%02ld:%03ld %5d - %s - %s%s%s\n"), 
		SystemTime.wMonth,
		SystemTime.wDay,
		SystemTime.wHour,
		SystemTime.wMinute, 
		SystemTime.wSecond,
		SystemTime.wMilliseconds,
		nThreadId,
		szSeverity,
		strSubsystem.c_str(), 
		szMessage,
		strFileAndLine.c_str());
	
	// If the message is too big, sprintf does not NULL terminate!
	// So we have to do it ourselves.
	acLogString[nONE_LOG_LINE_MAX_LENGTH] = _T('\0');

	// Remember the seek position right before we write the message
	m_vecLogHistorySeekPos[m_nLogLineCount%nMAX_LOG_HISTORY] = ::ftell(m_fileLog);

	// Output the message
	long nRet = _fputts(acLogString, m_fileLog);
	if (nRet == EOF)
	{
		long nLastError = ::GetLastError();
		// Don't assert because we are in a critical section and this
		// can cause a deadlock
		//_ASSERTE(!"fputts() failed!"); 
	}

	::fflush(m_fileLog); // Output the message immediately

	// Count the number of lines of the message
	TCHAR* pPos = acLogString;
	while (pPos = _tcschr(pPos, _T('\n')))
	{
		++m_nLogLineCount; // Update the log line count
		++pPos; // Skip past newline
	}
	
	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::backupFile
//					
// DESCRIPTION	:	Backs up the log file if it is too big.
//					
//*********************************************************************
HRESULT CLogFileDestination::backupFile()
{
	HRESULT hr = S_OK;

	if (!m_fileLog) 
	{
		return E_FAIL;
	}

	// Step 1:	Check to see if the log file size is too big

	if (m_nMaxFileSize == nNO_MAX_SIZE  ||
		ftell(m_fileLog) <= m_nMaxFileSize)
	{
		return S_FALSE; // No backup needed
	}

	// --- Log file size is too big ---

	// Step 2:	Create a new log file and backup the old one

	tstring strBackupFilename;
	hr = createLogFile(&strBackupFilename);
	if (FAILED(hr)) return hr;

	// Step 3:	Output a log message explaining that the log file
	//			has been restarted

	USES_CONVERSION;

	tstring strMsg;
	strMsg = _T("Maximum log file size exceeded.  The original log file has been copied to ");
	strMsg += strBackupFilename.c_str();
	strMsg += _T(".");

	// LogImplNoBackup() is safe to call because it will not call backupFile()
	// again (which would lead to an infinite loop)
	return logNoBackup(	W2CT(szILX_LOG_MODULE), 
						A2T(__FILE__), 
						__LINE__, 
						::GetCurrentThreadId(),
						strMsg.c_str(), 
						ILXLOG__INFO);
}

//*********************************************************************
// FUNCTION		:	CLogFileDestination::getLogHistory
//					
// DESCRIPTION	:	Retrieves the last 20 lines of the log file.
//					
// PARAMETERS	:	
//   pstrHistory	
//					
// COMMENTS		:	This is used by the error dialog (CMsgBoxDestination)
//					to populate the "Details" section.  This requires
//					CMsgBoxDestination to hold a pointer to a
//					CLogFileDestination.
//					
//*********************************************************************
void CLogFileDestination::getLogHistory(tstring* pstrHistory)
{
	// Pre-conditions
	_ASSERTE(m_fileLog);

	// Verify parameters
	_ASSERTE(pstrHistory);

	// Step 1:	Remember the current position

	long nCurPos = ::ftell(m_fileLog);

	if (nCurPos == -1)
	{
		long nLastError = ::GetLastError();
		_ASSERTE(!"ftell() failed!");
		return;
	}

	// Step 2:	Get the position to start reading the log history

	long nNewPosition = m_vecLogHistorySeekPos[ m_nLogLineCount % nMAX_LOG_HISTORY ];
	if (nNewPosition < 0 || nNewPosition > nCurPos)
	{
		_ASSERTE(false);
		return;
	}

	// Step 3:	Read the log history

	int nSuccess = ::fseek(m_fileLog, nNewPosition, SEEK_SET);
	if (nSuccess != 0)
	{
		long nLastError = ::GetLastError();
		_ASSERTE(!"fseek() failed!");
		return;
	}

	TCHAR szBuffer[nONE_LOG_LINE_MAX_LENGTH+1];

	long nLinesToRead = min(nMAX_LOG_HISTORY, m_nLogLineCount);

	for(int i = 0; i < nLinesToRead; ++i)
	{
		if (!::_fgetts(szBuffer, nONE_LOG_LINE_MAX_LENGTH, m_fileLog))
		{
			long nLastError = ::GetLastError();
			if (ferror(m_fileLog) || feof(m_fileLog))
			{
				//_ASSERTE(!"fgets() failed!");
			}
			
			break;
		}

		szBuffer[nONE_LOG_LINE_MAX_LENGTH] = _T('\0');

		(*pstrHistory) += szBuffer;
	}

	// Step 4:	Restore the old position

	nSuccess = ::fseek(m_fileLog, nCurPos, SEEK_SET);
	_ASSERTE(nSuccess == 0);
}

// HELPER FUNCTIONS

//*********************************************************************
// FUNCTION		:	GetLogDirectory
//					
// DESCRIPTION	:	Retrieves the RDC Log Directory from the registry.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   pstrLogDirectory	
//					
//*********************************************************************
HRESULT GetLogDirectory(tstring* pstrLogDirectory)
{
	HRESULT hr = S_OK;

	// Verify arguments
	_ASSERTE(pstrLogDirectory);
	
	// Initialize out params
	*pstrLogDirectory = _T("");

	HKEY hkResult = 0; 
	ULONG cbValue = 10;
	DWORD dwType = REG_SZ;

	// Step 1:	Open the root RDC key

	long lRet = ::RegOpenKeyExW(HKEY_CURRENT_USER, 
								szRDC_REGISTRY_PATH, 
								NULL, 
								KEY_ALL_ACCESS, 
								&hkResult); 

	if (ERROR_SUCCESS != lRet)
	{
		return HRESULT_FROM_WIN32(::GetLastError());
	}

	// Step 3:	Get the path to the Shared RDC directory
	
	wchar_t awcFilePath[_MAX_PATH];

	dwType = REG_SZ;
	cbValue = _MAX_PATH;
	lRet = RegQueryValueExW(	hkResult, 
								szLOG_DIR_REG_KEY, 
								NULL, 
								&dwType,  
								(unsigned char*) awcFilePath, 
								&cbValue); 

	USES_CONVERSION;

	if (ERROR_SUCCESS == lRet && *awcFilePath) 
	{
		long nFilePathLen = wcslen(awcFilePath);
		_ASSERTE(nFilePathLen >= 0);

		if (nFilePathLen > 0)
		{
			wchar_t& cLastChar = awcFilePath[nFilePathLen-1];
			if (cLastChar == L'\\' || cLastChar == L'/')
			{
				cLastChar = L'\0';
			}
		}

		*pstrLogDirectory = W2CT(awcFilePath);
	}

	RegCloseKey(hkResult);
	
	return S_OK;
}

