///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/MsgBoxDestination.cpp,v 1.2 2004/09/07 06:06:05 me133 Exp $
//
//    File name:	MsgBoxDestination.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/04/2000 1:58:42 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: MsgBoxDestination.cpp,v $
 * Revision 1.2  2004/09/07 06:06:05  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 26    8/30/02 9:15a Petert
 * Added email functionality
 * 
 * 25    7/30/02 1:41p Marce
 * o If you set IsModal then the number of instances must be 1
 * 
 * 24    7/25/02 2:51p Marce
 * o Now that we use CrashFinder we no longer display the last 20 lines of
 * the log file in the Details section of the Error dialog box
 * 
 * 23    7/17/02 3:49p Marce
 * o Now passes stack frames to error dialog
 * 
 * 22    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 19    4/17/01 6:20p Marce
 * o Added comment about handling DoModal() returning -1
 * 
 * 18    4/17/01 1:10p Marce
 * o Now checks for -1 after calling DoModal() because this is the return
 * code we get when we force the dialog to close when handling logoff.
 * 
 * 17    3/27/01 2:10p Marce
 * o Fixed compiler warning
 * 
 * 16    3/27/01 2:08p Marce
 * o Cosmetic
 * 
 * 15    3/27/01 2:06p Marce
 * Now supports MODAL attribute for forcing a modal dialog box
 * 
 * 14    1/30/01 10:51a Marce
 * o Fixed compile errors for Debug build
 * 
 * 13    12/19/00 3:29p Marce
 * o Now supports modal and modeless operation
 * 
 * 12    11/10/00 5:37p Marce
 * o Now limits the number of error dialogs that will be displayed at one
 * time
 * 
 * 11    11/10/00 5:18p Marce
 * o Now shows dialog box in separate thread
 * 
 * 10    11/10/00 3:15p Marce
 * o Added isLogFileDestinationSet()
 * 
 * 8     10/25/00 4:57p Marce
 * o Fixed compile error
 * 
 * 7     10/25/00 4:27p Marce
 * o Comments
 * 
 * 6     10/25/00 3:57p Marce
 * o Now handles fixing the log history string
 * 
 * 5     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 4     10/05/00 1:48p Marce
 * o Now gives the correct subsystem and file to CLogConfig and
 * destination objects
 * 
 * 3     10/05/00 10:25a Marce
 * o Fixed bug where calling log() from multiple threads caused a crash
 * 
 * 2     10/04/00 5:38p Marce
 * o Implemented log filtering
 */

#include "StdAfx.h"

#include "MsgBoxDestination.h"	// Me

#include "LogFileDestination.h"	// CLogFileDestination::getLogHistory()

#include "ErrorDialog.h"		// CErrorDialog

#include "HelperFunctions.h"	// ExtractFileAndSubsystem()

#include <process.h>			// _begin_thread

// Constants

static const TCHAR* const szSUB_TYPE			= _T("MsgBox");
static const TCHAR* const szMAX_ON_SCREEN_ATTR	= _T("MAX_ON_SCREEN");
static const TCHAR* const szMODAL_ATTR			= _T("MODAL");
static const wchar_t* const szUSE_DEFAULT		= L"default";

static const long nDEFAULT_MAX_ON_SCREEN		= 5;
static const bool nDEFAULT_IS_MODAL				= false;

// Helper functions
static void ConvertNewlines(tstring* pstr);
static void TrimWhitespace(tstring* pstr);

// Thread function
static void __cdecl ShowErrorDialog(void* pvParam);

// GLOBALS

static long	g_nCurrentOnScreen = 0;

////////////////////////////////////////
// LOG FILE DESTINATION IMPLEMENTATION
////////////////////////////////////////

CMsgBoxDestination::CMsgBoxDestination()
	: CLogDestination(szSUB_TYPE)
	, m_nMaxOnScreen(nDEFAULT_MAX_ON_SCREEN)
	, m_bIsModal(nDEFAULT_IS_MODAL)
{ }

CMsgBoxDestination::~CMsgBoxDestination()
{ }

CXmlElement* CMsgBoxDestination::clone() const
{
	return new CMsgBoxDestination(); // We create a new object not a clone
}

HRESULT CMsgBoxDestination::initialize()
{
	HRESULT hr = S_OK;

	bool bGotMaxOnScreen = false;

	// Invariants: Only call initialize() once
	_ASSERTE(!m_bIsInitialized);

	// Step 1:	Initialize member variables

	const _variant_t& varMaxOnScreen = m_mapAttributes[szMAX_ON_SCREEN_ATTR];
	if (varMaxOnScreen.vt != VT_BSTR || !*varMaxOnScreen.bstrVal)
	{
		m_nMaxOnScreen = nDEFAULT_MAX_ON_SCREEN; // Default
	}
	else if (wcsicmp(varMaxOnScreen.bstrVal, szUSE_DEFAULT) == 0)
	{
		// MAX_ON_SCREEN is "default" so use default value
		m_nMaxOnScreen = nDEFAULT_MAX_ON_SCREEN;
	}
	else
	{
		// Use the specified size
		m_nMaxOnScreen = varMaxOnScreen;
		bGotMaxOnScreen = true;
	}

	const _variant_t& varModal = m_mapAttributes[szMODAL_ATTR];
	if (varModal.vt != VT_BSTR || !*varModal.bstrVal)
	{
		m_bIsModal = nDEFAULT_IS_MODAL; // Default
	}
	else
	{
		// Use the specified size
		m_bIsModal = varModal;

		// Only one modal dialog can be shown at a time
		if (m_bIsModal && !bGotMaxOnScreen)
		{
			m_nMaxOnScreen = 1;
		}
	}

	// SANITY CHECK
	if (m_bIsModal && bGotMaxOnScreen && m_nMaxOnScreen != 1)
	{
		_ASSERTE("Can't specify both MODAL = True and MAX_ON_SCREEN > 1!");
	}

	_ASSERTE(m_nMaxOnScreen >= 0);

	m_bIsInitialized = true;

	return S_OK;
}

//*********************************************************************
// FUNCTION		:	CMsgBoxDestination::log
//					
// DESCRIPTION	:	
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	See header file for info.
//					
//*********************************************************************
HRESULT CMsgBoxDestination::log(	const TCHAR* szSubsystem, 
									const TCHAR* szPath,
									const TCHAR* szFilename, 
									long nLine, 
									long nThreadId, 
									long hrErrorCode,
									SIlxStackFrames* pvIlxStackFrames,
									long nNumStackFrames,
									const TCHAR* szMessage, 
									long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions
	if (!m_bIsInitialized)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	// Pre-conditions

	// This check is not exactly thread safe but its close enough
	// for government work. :) Just kidding.  Because this variable
	// is tied to the user clicking OK on the dialog, its not likely
	// that it will change quickly enough.
	if (g_nCurrentOnScreen >= m_nMaxOnScreen)
	{
		AtlTrace(	_T("ILXLOG: The maximum number of error dialogs (%d) are already being displayed. ")
					_T("Ignoring request to create another one."),
					m_nMaxOnScreen);
		return ERROR_TOO_MANY_TCBS; // Overloading the error code for
									// our own evil purposes
	}

	// Step 1:	Trim whitespace and change "\n" to "\r\n"

	tstring strMessage = szMessage;
	TrimWhitespace(&strMessage);
	ConvertNewlines(&strMessage);

	// Step 2:	Pop-up the modal error dialog

	CErrorDialog* pdlgError = new CErrorDialog; // Thread deallocates this
	_ASSERTE(pdlgError);
	if (!pdlgError) return E_OUTOFMEMORY;

	pdlgError->init(strMessage.c_str(), 
					//pvIlxStackFrames,
					//nNumStackFrames,
					szSubsystem,
					nSeverity);

	if (!m_bIsModal)
	{
		// Here is why I used _beginthread() instead of CreateThread():
		//
		// "A thread that uses functions from the C run-time libraries should
		// use the beginthread and endthread C run-time functions for thread
		// management rather than CreateThread and ExitThread. Failure to do
		// so results in small memory leaks when ExitThread is called."
		// - MSDN Library help documentation for CreateThread()

		long hThreadHandle = ::_beginthread(ShowErrorDialog,
											0,
											pdlgError); // Thread owns the memory
		if (hThreadHandle == -1)
		{
			hr = HRESULT_FROM_WIN32(errno);

			delete pdlgError;

			//ILXLOG(hr, _T("Failed to create event dispatcher thread."), 
			//		ILXLOG_ERROR);
			_ASSERTE(!"Failed to create event dispatcher thread.");
			return hr;
		}
	}
	else
	{
		// Instead of starting a thread to display the dialog we
		// simply call the thread function directly so that the
		// dialog is modal.
		ShowErrorDialog(pdlgError);
	}

	return S_OK;
}

void CMsgBoxDestination::setIsModal(bool bIsModal)
{
	m_bIsModal = bIsModal;

	// Only one modal dialog can be shown at a time
	if (m_bIsModal)
	{
		m_nMaxOnScreen = 1;
	}
}

bool CMsgBoxDestination::isModal() const
{
	return m_bIsModal;
}

// HELPER FUNCTIONS

//*********************************************************************
// FUNCTION		:	ConvertNewlines
//					
// DESCRIPTION	:	Replaces "\n" with "\r\n" in a string.
//					
// COMMENTS		:	Don't call this twice for the same string or else
//					you will get "\r\r\n"!
//					
//*********************************************************************
void ConvertNewlines(tstring* pstr)
{
	_ASSERTE(pstr);
	long nFind = pstr->find(_T('\n'));
	while (nFind != tstring::npos)
	{
		pstr->replace(nFind, 1, _T("\r\n"));
		nFind = pstr->find(_T('\n'), nFind + 2); // +2 to skip over "\r\n"
	}
}

//*********************************************************************
// FUNCTION		:	TrimWhitespace
//					
// DESCRIPTION	:	Removes leading whitespace from a string.
//					
//*********************************************************************
void TrimWhitespace(tstring* pstr)
{
	tstring::iterator it = pstr->begin();

	while (it == pstr->begin() && it != pstr->end() && _istspace(*it))
	{
		it = pstr->erase(it);
	}
}

void __cdecl ShowErrorDialog(void* pvParam)
{
	_ASSERTE(pvParam);
	if (!pvParam) return;

	CErrorDialog* pdlgError = reinterpret_cast<CErrorDialog*>(pvParam);
	_ASSERTE(pdlgError);
	if (!pdlgError) return;

	::InterlockedIncrement(&g_nCurrentOnScreen);

	int nRet = pdlgError->DoModal();

	long nNewVal = ::InterlockedDecrement(&g_nCurrentOnScreen);
	_ASSERTE(nNewVal >= 0);

	#ifdef _DEBUG

	// DoModal() returns -1 when we force ActiveILX to be closed.
	// You should return *immediately* if -1 is returned.
	if (nRet != IDCANCEL && nRet != -1 && pdlgError->debugBtnClicked())
	{
		// Welcome to the call stack!!!  Have at it!
		// (You will only see this in debug builds.)
	    _CrtDbgBreak();
	}

	#endif // _DEBUG

	delete pdlgError; // We own the memory so we must delete it
}
