///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/MsgBoxDestination.h,v 1.2 2004/09/07 06:06:05 me133 Exp $
//
//    File name:	MsgBoxDestination.h
//   Created By:	Marc Eaddy
//   Created On:	10/04/2000 1:55:53 PM
//
//      Purpose:	Displays the message in a modal error dialog box
//
//  Description:	Currently, this log destination is dangerous to use
//					because the error dialog box is modal and causes
//					other threads to block when they try to log a message.
//					This results in the application hanging.
//
//   Xml Format:	<DEST TYPE="MsgBox"/>
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: MsgBoxDestination.h,v $
 * Revision 1.2  2004/09/07 06:06:05  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 16    7/25/02 2:51p Marce
 * o Now that we use CrashFinder we no longer display the last 20 lines of
 * the log file in the Details section of the Error dialog box
 * 
 * 15    7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 12    3/27/01 2:06p Marce
 * Now supports MODAL attribute for forcing a modal dialog box
 * 
 * 11    12/19/00 3:31p Marce
 * o Added comment
 * 
 * 10    12/19/00 3:29p Marce
 * o Now supports modal and modeless operation
 * 
 * 9     11/10/00 5:37p Marce
 * o Now limits the number of error dialogs that will be displayed at one
 * time
 * 
 * 8     11/10/00 3:15p Marce
 * o Added isLogFileDestinationSet()
 * 
 * 6     10/25/00 4:27p Marce
 * o Comments
 * 
 * 5     10/25/00 4:13p Marce
 * o Comments
 * 
 * 4     10/10/00 6:46p Marce
 * o Now passes the full path as well as the file name
 * o Fixed bug where trace messages were not clickable
 * 
 * 3     10/05/00 10:25a Marce
 * o Fixed bug where calling log() from multiple threads caused a crash
 * 
 * 2     10/04/00 5:38p Marce
 * o Implemented log filtering
 */

#ifndef INCLUDED_MSG_BOX_DESTINATION_H
#define INCLUDED_MSG_BOX_DESTINATION_H

#include "LogDestination.h"	// Parent

// Forward declarations
class CLogFileDestination;

//*****************************
// CMsgBoxDestination
//*****************************

class CMsgBoxDestination : public CLogDestination
{
public:

	CMsgBoxDestination();
	virtual ~CMsgBoxDestination();

public:

	virtual HRESULT initialize();
		//
		// See LogDestination.h for a description
		//

	virtual HRESULT log(const TCHAR* szSubsystem, 
						const TCHAR* szPath,
						const TCHAR* szFilename, 
						long nLine, 
						long nThreadId, 
						long hrErrorCode,
						SIlxStackFrames* pvIlxStackFrames,
						long nNumStackFrames,
						const TCHAR* szMessage, 
						long nSeverity);
		//
		// See LogDestination.h for a description.
		//

	void setIsModal(bool bIsModal);
		// Whether the dialog box should be displayed modal or modeless.
		//
		// By default the dialog box is modeless.  This is fine for
		// non-fatal errors and warnings because the application can
		// still run while the dialog box is displayed.  For fatal
		// errors though the application usually exits after displaying
		// the dialog.  In this case we want the dialog to be modal
		// so that it does not disappear before the user can see it.

	bool isModal() const;

protected:
	// Overrides CXmlElement
	virtual CXmlElement* clone() const;
		//
		// Clones this element.
		//

private: // NOT IMPLEMENTED

	CMsgBoxDestination(const CMsgBoxDestination&);
	CMsgBoxDestination& operator=(const CMsgBoxDestination&);

private:	// DATA

	long					m_nMaxOnScreen;

	bool					m_bIsModal;
};

#endif // INCLUDED_MSG_BOX_DESTINATION_H