///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/StdOutDestination.cpp,v 1.4 2004/09/29 05:38:43 me133 Exp $
//
//    File name:	StdOutDestination.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/04/2000 2:29:21 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: StdOutDestination.cpp,v $
 * Revision 1.4  2004/09/29 05:38:43  me133
 * o Now outputs the subsystem for non-verbose messages. This makes
 * it easier to intermix messages from other processes in the same console
 * window.
 *
 * Revision 1.3  2004/09/23 05:51:13  me133
 * o Now prints INFO and SUCCESS messages
 * o Now supports verbose output
 *
 * Revision 1.2  2004/09/07 06:06:05  me133
 * o Now uses most recent ILXLog code with ILXCrashFinder
 * 
 * 
 * 5     7/17/02 3:10p Marce
 * o Added preliminary support for passing exception info via the
 * SIlxStackFrame structure used by IlxCrashFinder
 * 
 * 2     5/01/01 4:10p Marce
 * o Added StdOut destination
 */

#include "StdAfx.h"

#include "StdOutDestination.h"		// Me

#include "HelperFunctions.h"		// GetSeverity()

// Constants

static const TCHAR* const szSUB_TYPE = _T("StdOut");
static const TCHAR* const szSUB_TYPE_VERBOSE = _T("StdOutVerbose");

////////////////////////////////////////
// STD OUT DESTINATION IMPLEMENTATION
////////////////////////////////////////

CStdOutDestination::CStdOutDestination(bool bVerbose)
	: CLogDestination(bVerbose ? szSUB_TYPE_VERBOSE : szSUB_TYPE )
	, m_bVerbose(bVerbose)
{ }

CStdOutDestination::~CStdOutDestination()
{ }

CXmlElement* CStdOutDestination::clone() const
{
	return new CStdOutDestination(m_bVerbose); // We create a new object not a clone
}

//*********************************************************************
// FUNCTION		:	CStdOutDestination::log
//					
// DESCRIPTION	:	Sends message to the standard console output.
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	See header file.
//					
//*********************************************************************
HRESULT CStdOutDestination::log(	const TCHAR* szSubsystem, 
									const TCHAR* szPath,
									const TCHAR* szFilename, 
									long nLine, 
									long nThreadId, 
									long hrErrorCode,
									SIlxStackFrames* pvIlxStackFrames,
									long nNumStackFrames,
									const TCHAR* szMessage, 
									long nSeverity)
{
	HRESULT hr = S_OK;

	// Pre-conditions
	if (!m_bIsInitialized)
	{
		_ASSERTE(!"Must call initialize() first");
		return E_UNEXPECTED;
	}

	// Verify arguments

	if (!szMessage) return E_INVALIDARG;

	const TCHAR* szSeverity = ::GetSeverity(nSeverity);
	_ASSERTE(szSeverity && *szSeverity);

	SYSTEMTIME SystemTime;
	::GetLocalTime( &SystemTime ); 

	// Determine the source filename

	tstring strFile = szFilename;
	tstring strSubsystem = szSubsystem;
	tstring strFileAndLine;

	// Ex: "ActiveILX: "
	if (!strSubsystem.empty())
	{
		strSubsystem += _T(": ");
	}

	// Ex: " [File: ActiveILX.cpp, Line: 1234]"
	// Ex: " [File: ActiveILX.cpp]"
	if (!strFile.empty())
	{
		strFileAndLine += _T(" [File: ") + strFile;

		if (nLine >= 0)
		{
			TCHAR szLineBuf[50];
			_stprintf(szLineBuf, _T(", Line: %d]"), nLine);
			strFileAndLine += szLineBuf;
		}
		else
		{
			strFileAndLine += _T("]");
		}
	} 

	if (m_bVerbose)
	{
		_tprintf(_T("%02ld/%02ld %2ld:%02ld:%02ld %3d - %s - %s%s%s\n"), 
			SystemTime.wMonth,
			SystemTime.wDay,
			SystemTime.wHour,
			SystemTime.wMinute, 
			SystemTime.wSecond,
			nThreadId,
			szSeverity,
			strSubsystem.c_str(), 
			szMessage,
			strFileAndLine.c_str());
	}
	else
	{
		_tprintf(_T("%s%s - %s\n"), strSubsystem.c_str(), szSeverity, szMessage);
	}

	return S_OK;
}
