///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/ILXLog/XmlLogHelper.cpp,v 1.1 2004/09/03 16:19:09 me133 Exp $
//
//    File name:	XmlLogHelper.cpp
//   Created By:	Marc Eaddy
//   Created on:	05/01/2000 10:25:48 AM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: XmlLogHelper.cpp,v $
 * Revision 1.1  2004/09/03 16:19:09  me133
 * o Initial checkin to CVS
 * 
 * 
 * 13    4/09/01 11:34a Marce
 * o Cosmetic
 * 
 * 12    4/09/01 11:32a Marce
 * o Simplified and logs more errors
 * 
 * 11    11/10/00 4:54p Marce
 * o Made sure that the error buffer length was observed
 * 
 * 10    11/10/00 4:39p Marce
 * o Changed function name for clarity
 * 
 * 8     10/18/00 1:41p Marce
 * o Should not initialize a CComBSTR to 0 or NULL
 * 
 * 7     9/25/00 1:47p Marce
 * o Okay, okay.  Logs an info message if the file is not found.
 * 
 * 6     9/25/00 1:44p Marce
 * o Does not log anything at all if the file is not found.  Let's caller
 * handle this.
 * 
 * 5     9/25/00 1:32p Marce
 * o Trimmed whitespace from reason
 * 
 * 4     9/25/00 12:05p Marce
 * o Now does not log an error if a file is not found since this may be
 * expected
 * 
 * 3     5/13/00 12:04a Marce
 * o Removes the trailing newline on the reason message
 * o Fixed tab formatting problem
 * 
 * 2     5/03/00 12:43p Marce
 * o Removed assert since we are now popping up a message box
 * 
 * 1     5/01/00 10:41a Marce
 * 
 * 2     5/01/00 10:38a Marce
 * o Modified HandleError() so that it logs the error
 * 
 * 1     5/01/00 10:26a Marce
 */

#include "StdAfx.h"

#include "XmlLogHelper.h"	// Me

#include "ILXLogHelper.h"	// Logging singleton

#include <urlmon.h>			// INET_E_OBJECT_NOT_FOUND

// Imported type libraries
// Creates .tlh files in the build directory.  Because the compiler
// does not generate include guards for the type library header
// files it creates, we have to add our own include guards.  If
// you want to import these libraries inside of StdAfx.h, make sure
// that you define the symbol so that it does not get included
// twice.
#ifndef INCLUDED_MSXML_TYPE_LIBRARY
#define INCLUDED_MSXML_TYPE_LIBRARY
#import "msxml.dll" rename_namespace("MSXML"), named_guids
#endif // INCLUDED_MSXML_TYPE_LIBRARY

using namespace MSXML;

//*********************************************************************
// FUNCTION		:	LogXmlError
//					
// DESCRIPTION	:	
//					
// RETURNS		:	
//   HRESULT 		
//					
// PARAMETERS	:	
//   piXmlDocument	
//					
// COMMENTS		:	
//					
//*********************************************************************
HRESULT LogXmlError(MSXML::IXMLDOMDocument* piXmlDocument)
{
	// Pre-conditions
	if (!piXmlDocument) return E_INVALIDARG;

	HRESULT hr = S_OK;

	CComPtr<MSXML::IXMLDOMParseError> spiXmlError = 0;

	hr = piXmlDocument->get_parseError(&spiXmlError);
	VERIFY_HR_COND(hr, hr == S_OK && spiXmlError != 0, 
		_T("Failed to get the parsing error information."), ILXLOG_ERROR);

	long nLine			= 0;
	long nLinePos		= 0;
	long hrErrorCode		= E_FAIL;
	CComBSTR bstrUrl;
	CComBSTR bstrReason;

	hr = spiXmlError->get_line(&nLine);
	if (hr != S_OK)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		ILXLOG(hr, _T("Failed to get the line number for the XML error."), ILXLOG_WARNING);
	}

	hr = spiXmlError->get_linepos(&nLinePos);
	if (hr != S_OK)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		ILXLOG(hr, _T("Failed to get the line position for the XML error."), ILXLOG_WARNING);
	}

	hr = spiXmlError->get_errorCode(&hrErrorCode);
	if (hr != S_OK)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		ILXLOG(hr, _T("Failed to get the error code for the XML error."), ILXLOG_WARNING);
	}

	hr = spiXmlError->get_url(&bstrUrl);
	if (hr != S_OK)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		ILXLOG(hr, _T("Failed to get the url for the XML error."), ILXLOG_WARNING);
	}
	
	hr = spiXmlError->get_reason(&bstrReason);
	if (hr != S_OK)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		ILXLOG(hr, _T("Failed to get the reason for the XML error."), ILXLOG_WARNING);
	}

	// Trim whitespace from the reason

	long nReasonLen = bstrReason.Length();

	while (nReasonLen > 0 && _istspace(bstrReason[nReasonLen-1]))
	{
		bstrReason[--nReasonLen] = L'\0';
	}

	if (hrErrorCode == INET_E_RESOURCE_NOT_FOUND)
	{
		return ILXLOG1(hrErrorCode, ILXLOG_INFO, 
			_T("Failed to open XML file ('%ls')."), 
			bstrUrl ? bstrUrl : _T("<NO URL>"));
	}
	else
	{
		return ILXLOG5(hrErrorCode, ILXLOG_ERROR,
					_T("Error parsing XML file:\n")
					_T("	File  : %ls\n")
					_T("	Reason: %ls\n")
					_T("	Error : 0x%08X\n")
					_T("	Line  : %d\n")
					_T("	Column: %d\n"),
					bstrUrl ? bstrUrl : _T("<NO URL>"),
					bstrReason ? bstrReason : _T("<NO REASON>"),
					hrErrorCode,
					nLine,
					nLinePos);
	}
}
