///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/IlxPersistenceImpl.h,v 1.1 2004/09/07 06:10:52 me133 Exp $
//
//    File name:	IlxPersistenceImpl.h
//   Created By:	Marc Eaddy
//   Created On:	04/10/2002 3:46:00 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: IlxPersistenceImpl.h,v $
 * Revision 1.1  2004/09/07 06:10:52  me133
 * o Initial checkin to CVS
 * 
 * 
 * 10    8/02/02 6:08p Marce
 * o Fixed comments
 * 
 * 9     7/15/02 5:49p Marce
 * o Fixed Save As New Page access violation (QAB 18753)
 * 
 * 8     7/06/02 4:39a Marce
 * o Added comment
 * 
 * 7     7/06/02 4:34a Marce
 * o Added ability to get the InstanceFile based on the Common Data
 * Directory.  This can be used to support upgrading from an old
 * workspace.
 * 
 * 4     5/29/02 3:26p Jason
 * Added CommonDataDirectory property.
 * 
 * 3     4/30/02 9:17p Marce
 * o Added ILXCommon properties and macros for InstallDirectory
 * o CIlxPersistenceImpl now supports InstallDirectory
 * 
 * 2     4/11/02 5:59p Marce
 * o Decoupled IlxContextImpl and IlxPersistenceImpl to make life simpler
 * 
 * 1     4/11/02 12:24p Marce
 */

#ifndef INCLUDED_ILX_PERSISTENCE_IMPL_H
#define INCLUDED_ILX_PERSISTENCE_IMPL_H

#include "IPerPropertyBrowsingImpl.h"	// IPerPropertyBrowsingImpl_Fix
#include "ILXCommonPersistence.h"		// ILXCommonLoadFromStream(), ILXCommonSaveToStream()
#include "BugFixes.h"					// FixedAtlIPersistPropertyBag_Load
#include "IlxResult.h"					// IILXResultManager, IILXResult
#include "IlxLogHelper.h"				// Logging macros

//*****************************
// CIlxPersistenceImpl
//*****************************
//
// 1. Control must derive from CComControlBase<T>, CComControl<T>
//
// 2. Must have the following entries in the PROP_MAP
//
//	You should not store InstanceName, DataDirectory, or InstallDirectory
//  in the property map.  You container will always set these properties.
//
// 3. You must define these in your interface map (BEGIN_COM_MAP)
//      COM_INTERFACE_ENTRY(IPersistStreamInit)
//      COM_INTERFACE_ENTRY2(IPersist, IPersistStreamInit)
//      COM_INTERFACE_ENTRY(IPersistStorage)
//      COM_INTERFACE_ENTRY(IPersistPropertyBag)
//      COM_INTERFACE_ENTRY(ISpecifyPropertyPages)
//      COM_INTERFACE_ENTRY(IPerPropertyBrowsing)
//
// 4. Should have the following macros specified in the IDL file
//
//  GET_INSTANCE_NAME
//  PUT_INSTANCE_NAME
//  GET_DATA_DIRECTORY
//  PUT_DATA_DIRECTORY
//  GET_INSTALL_DIRECTORY
//  PUT_INSTALL_DIRECTORY
//
// 5. Should have the following methods specified in the IDL file
// 
//	(none)
//
// 6. Derive from: (substitute ILXBrowser for your class, iface, iid, and libid)
//
//    If you currently derive form CStockPropImpl<> use
//
// public CIlxPersistenceImpl<CILXBrowser, CStockPropImpl<CILXBrowser, IILXBrowser,  &IID_IILXBrowser, &LIBID_ILXBROWSERLib> >
//
//    or, if you currently derive from CDispatchImpl<> use
//
// public CIlxPersistenceImpl<CILXBrowser, CDispatchImpl<IILXBrowser,  &IID_IILXBrowser, &LIBID_ILXBROWSERLib> >
//
//    otherwise use
//
// public CIlxPersistenceImpl<CILXBrowser>
//

template <class T, class Base=T>
class ATL_NO_VTABLE CIlxPersistenceImpl
		: public Base
        , public IPersistStreamInitImpl<T>
        , public IPersistStorageImpl<T>
        , public IPersistPropertyBagImpl<T>
        , public ISpecifyPropertyPagesImpl<T>
        , public IPerPropertyBrowsingImpl_Fix<T>
{
public:

	typedef CIlxPersistenceImpl<T, Base> TypeMyPersistenceImpl;

	CIlxPersistenceImpl()
		: m_bstrCommonDataDirectory()
		, m_bstrDataDirectory()
		, m_bstrInstallDirectory()
		, m_bstrInstanceName()
		, m_bstrControlName()
		, m_setDispidsToIgnore()
		, m_spiResultManager(0)
	{ }

	virtual ~CIlxPersistenceImpl()
	{ 
		if (m_spiResultManager)
		{
			_ASSERTE(!"Must call FinalRelease()");
			FinalRelease();
		}
	}

	HRESULT FinalConstruct(	IILXResultManager* piResultManager, 
							const TCHAR* szControlName)
	{
		HRESULT hr = S_OK;

		_ASSERTE(piResultManager && szControlName && *szControlName);
		if (!piResultManager || !szControlName || !*szControlName) return E_POINTER;

		if (m_spiResultManager)
		{
			_ASSERTE(!"Must call FinalConstruct only once");
			return E_UNEXPECTED;
		}

		m_spiResultManager = piResultManager;
		m_bstrControlName = szControlName;

		return S_OK;
	}

	void FinalRelease()
	{
		m_spiResultManager = 0;
	}

	STDMETHODIMP InitNew()
	{
		m_bstrCommonDataDirectory.Empty();
		m_bstrDataDirectory.Empty();
		m_bstrInstallDirectory.Empty();
		m_bstrInstanceName.Empty();
		m_setDispidsToIgnore.clear();

		return OnInitNew();
	}

	STDMETHODIMP Load(LPSTREAM pStm)
	{
		HRESULT hr = S_OK;
		
		ATLTRACE2(atlTraceCOM, 0, _T("IPersistStreamInit::Load\n"));
		
		T* pT = static_cast<T*>(this);
		
		// Since control's store the InstanceName in their stream,
		// when it gets loaded it can trigger LoadInstanceData if
		// its different than before.  This is how we shared controls
		// are able to load their instance file just by calling this
		// Load() method.
		hr = ILXCommonLoadFromStream(	pStm,
										pT->GetPropertyMap(),
										this,
										pT->GetUnknown(),
										0,
										&m_setDispidsToIgnore);
		if (FAILED(hr))
		{
			ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to load stream for '%s' control."), 
				m_bstrControlName);
		}
		else
		{
			pT->m_bRequiresSave = FALSE;
		}
		
		return hr;
	}
	
	STDMETHODIMP Save(LPSTREAM pStm, BOOL fClearDirty)
	{
		HRESULT hr = S_OK;

		T* pT = static_cast<T*>(this);
		
		hr = ILXCommonSaveToStream(	pStm,
									pT->GetPropertyMap(),
									this,
									pT->GetUnknown(),
									&m_setDispidsToIgnore);

		if (FAILED(hr))
		{
			ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to save stream for '%s' control."), 
				m_bstrControlName);
		}

		if (IsInitialized())
		{
			hr = OnSaveInstanceData();
			if (FAILED(hr))
			{
				ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to save instance files for '%s' control."), 
					m_bstrControlName);
			}
		}

		return S_OK;
	}
	
	// IPersistPropertyBag::Load()
	#if _ATL_VER == 0x0300          // Fix for ATL 3.0 bug in AtlIPersistPropertyBag_Load
	STDMETHODIMP Load(LPPROPERTYBAG pPropBag, LPERRORLOG pErrorLog)
	{
		if (!pPropBag) return E_POINTER;
		
		HRESULT hr = S_OK;
		
		T* pT = static_cast<T*>(this);
		
		ATLTRACE2(atlTraceCOM, 0, _T("IPersistPropertyBag::Load\n"));
		
		ATL_PROPMAP_ENTRY* pMap = pT->GetPropertyMap();
		
		// Load the properties described in the PROP_MAP
		
		// Work around ATL 3.0 bug in AtlIPersistPropertyBag_Load
		hr = FixedAtlIPersistPropertyBag_Load(	pPropBag,
												pErrorLog,
												pMap,
												this,
												pT->GetUnknown() );
		if (FAILED(hr))
		{
			ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to load property bag for '%s' control."), 
				m_bstrControlName);
		}
		else
		{
			pT->m_bRequiresSave = FALSE;
		}
		
		return hr;
	}
	#endif // _ATL_VER == 0x0300
	
	//
	// IPerPropertyBrowsing implementation
	//
	STDMETHODIMP MapPropertyToPage(DISPID dispid, CLSID* pClsid)
	{
		HRESULT hr = S_OK;

		if (!pClsid) return E_POINTER;
		
		// Look up the property page in the PROP_MAP
		hr = IPerPropertyBrowsingImpl_Fix<T>::MapPropertyToPage(dispid,pClsid);
		VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to locate property page for dispid %d."),
			dispid);
		
		// When there is no property page for a listed property, fail the call
		if (CLSID_NULL == *pClsid)
			return E_INVALIDARG;

		return S_OK;
	}
	
    // ILXCommon methods
		
	STDMETHODIMP IgnorePersistedProperty(	/*[in]*/ long dispidToIgnore,
											/*[in]*/ VARIANT_BOOL vbIgnore)
	{
		if (vbIgnore == VARIANT_FALSE)
		{
			m_setDispidsToIgnore.erase(dispidToIgnore);
		}
		else
		{
			m_setDispidsToIgnore.insert(dispidToIgnore);
		}

		return S_OK;
	}

	STDMETHODIMP get_InstanceName(/*[retval,out]*/ BSTR* pbstrInstanceName)
	{
		if (!pbstrInstanceName) return E_POINTER;
		return m_bstrInstanceName.CopyTo(pbstrInstanceName);
	}
	
	STDMETHODIMP put_InstanceName(/*[in]*/ BSTR bstrInstanceName)
	{
		HRESULT hr = S_OK;

		// Verify arguments
		if (!bstrInstanceName) return E_POINTER;

		if (!m_spiResultManager) return E_UNEXPECTED;

		CComPtr<IILXResult> spiLastResult;

		hr = m_spiResultManager->GetLastResult(	_variant_t( long(ILX_CLIENT_INIT) ), 
												&spiLastResult);
		_ASSERTE(SUCCEEDED(hr));
		
		bool bInitialized = SUCCEEDED(hr) && spiLastResult == 0;

		// Don't do anything if the value is the same
		// Must check for self-assignment ourselves!
		if (bInitialized &&
			(m_bstrInstanceName.m_str == bstrInstanceName ||
			(m_bstrInstanceName && wcsicmp(m_bstrInstanceName, bstrInstanceName) == 0)))
		{
			return S_FALSE;
		}

		T* pT = static_cast<T*>(this);
		_ASSERTE(pT);

		if (0 == pT->m_nFreezeEvents )
			if (S_FALSE == pT->FireOnRequestEdit( DISPID_ILX_INSTANCE_NAME ) )
				return S_FALSE;

		// Set the property
		m_bstrInstanceName = bstrInstanceName;

		// Allow the control to be notified of the change
		hr = OnInstanceNameChanged();
		if (FAILED(hr)) return hr;

		if (IsInitialized())
		{
			hr = OnLoadInstanceData();
			if (FAILED(hr))
			{
				hr = m_spiResultManager->SetResult(E_ILX_INSTANCE_NAME, m_bstrControlName, 0, 0, 0, 0);
				if (FAILED(hr)) return hr;

				return E_ILX_DATA_DIR;
			}
			else
			{
				hr = m_spiResultManager->ClearResult(ILX_CLIENT_INIT);
				_ASSERTE(SUCCEEDED(hr));
			}
		}

		// Persistence is dirty
		pT->SetDirty(TRUE);

		// Tell the container that the property has changed
		if (0 == pT->m_nFreezeEvents)
		{
			hr = pT->FireOnChanged( DISPID_ILX_INSTANCE_NAME );
			_ASSERTE(SUCCEEDED(hr));
		}

		return S_OK;
	}

	STDMETHODIMP get_CommonDataDirectory(/*[retval,out]*/ BSTR* pbstrCommonDataDirectory)
	{
		if (!pbstrCommonDataDirectory) return E_POINTER;
		return m_bstrCommonDataDirectory.CopyTo(pbstrCommonDataDirectory);
	}

	STDMETHODIMP put_CommonDataDirectory(/*[in]*/ BSTR bstrCommonDataDirectory)
	{
		HRESULT hr = S_OK;

		// Verify arguments
		if (!bstrCommonDataDirectory) return E_POINTER;

		if (!m_spiResultManager) return E_UNEXPECTED;

		CComPtr<IILXResult> spiLastResult;

		hr = m_spiResultManager->GetLastResult(	_variant_t( long(ILX_CLIENT_INIT) ), 
												&spiLastResult);
		_ASSERTE(SUCCEEDED(hr));
		
		bool bInitialized = SUCCEEDED(hr) && spiLastResult == 0;

		// Don't do anything if the value is the same
		// Must check for self-assignment ourselves!
		if (bInitialized && 
			(m_bstrCommonDataDirectory.m_str == bstrCommonDataDirectory ||
			(m_bstrCommonDataDirectory && wcsicmp(m_bstrCommonDataDirectory, bstrCommonDataDirectory) == 0)))
		{
			return S_FALSE;
		}

		T* pT = static_cast<T*>(this);
		_ASSERTE(pT);

		if (0 == pT->m_nFreezeEvents )
			if (S_FALSE == pT->FireOnRequestEdit( DISPID_ILX_COMMON_DATA_DIRECTORY ) )
				return S_FALSE;

		// Set the property
		m_bstrCommonDataDirectory = bstrCommonDataDirectory;

		// Allow the control to be notified of the change
		hr = OnCommonDataDirectoryChanged();
		if (FAILED(hr)) return hr;

		// Persistence is dirty
		pT->SetDirty(TRUE);

		// Tell the container that the property has changed
		if (0 == pT->m_nFreezeEvents)
		{
			hr = pT->FireOnChanged( DISPID_ILX_COMMON_DATA_DIRECTORY );
			_ASSERTE(SUCCEEDED(hr));
		}

		return S_OK;
	}

	STDMETHODIMP get_DataDirectory(/*[retval,out]*/ BSTR* pbstrDataDirectory)
	{
		if (!pbstrDataDirectory) return E_POINTER;
		return m_bstrDataDirectory.CopyTo(pbstrDataDirectory);
	}
	
	STDMETHODIMP put_DataDirectory(/*[in]*/ BSTR bstrDataDirectory)
	{
		HRESULT hr = S_OK;

		// Verify arguments
		if (!bstrDataDirectory) return E_POINTER;

		if (!m_spiResultManager) return E_UNEXPECTED;

		CComPtr<IILXResult> spiLastResult;

		hr = m_spiResultManager->GetLastResult(	_variant_t( long(ILX_CLIENT_INIT) ), 
												&spiLastResult);
		_ASSERTE(SUCCEEDED(hr));
		
		bool bInitialized = SUCCEEDED(hr) && spiLastResult == 0;

		// Don't do anything if the value is the same
		// Must check for self-assignment ourselves!
		if (bInitialized && 
			(m_bstrDataDirectory.m_str == bstrDataDirectory ||
			(m_bstrDataDirectory && wcsicmp(m_bstrDataDirectory, bstrDataDirectory) == 0)))
		{
			return S_FALSE;
		}

		T* pT = static_cast<T*>(this);
		_ASSERTE(pT);

		if (0 == pT->m_nFreezeEvents )
			if (S_FALSE == pT->FireOnRequestEdit( DISPID_ILX_DATA_DIRECTORY ) )
				return S_FALSE;

		// Set the property
		m_bstrDataDirectory = bstrDataDirectory;

		// Allow the control to be notified of the change
		hr = OnDataDirectoryChanged();
		if (FAILED(hr)) return hr;

		if (IsInitialized())
		{
			hr = OnLoadInstanceData();
			if (FAILED(hr))
			{
				hr = m_spiResultManager->SetResult(E_ILX_DATA_DIR, m_bstrControlName, 0, 0, 0, 0);
				if (FAILED(hr)) return hr;

				return E_ILX_DATA_DIR;
			}
			else
			{
				hr = m_spiResultManager->ClearResult(ILX_CLIENT_INIT);
				_ASSERTE(SUCCEEDED(hr));
			}
		}

		// Persistence is dirty
		pT->SetDirty(TRUE);

		// Tell the container that the property has changed
		if (0 == pT->m_nFreezeEvents)
		{
			hr = pT->FireOnChanged( DISPID_ILX_DATA_DIRECTORY );
			_ASSERTE(SUCCEEDED(hr));
		}

		return S_OK;
	}

	STDMETHODIMP get_InstallDirectory(/*[retval,out]*/ BSTR* pbstrInstallDirectory)
	{
		if (!pbstrInstallDirectory) return E_POINTER;
		return m_bstrInstallDirectory.CopyTo(pbstrInstallDirectory);
	}
	
	STDMETHODIMP put_InstallDirectory(/*[in]*/ BSTR bstrInstallDirectory)
	{
		HRESULT hr = S_OK;

		// Verify arguments
		if (!bstrInstallDirectory) return E_POINTER;

		if (!m_spiResultManager) return E_UNEXPECTED;

		CComPtr<IILXResult> spiLastResult;

		hr = m_spiResultManager->GetLastResult(	_variant_t( long(ILX_CLIENT_INIT) ), 
												&spiLastResult);
		_ASSERTE(SUCCEEDED(hr));
		
		bool bInitialized = SUCCEEDED(hr) && spiLastResult == 0;

		// Don't do anything if the value is the same
		// Must check for self-assignment ourselves!
		if (bInitialized && 
			(m_bstrInstallDirectory.m_str == bstrInstallDirectory ||
			(m_bstrInstallDirectory && wcsicmp(m_bstrInstallDirectory, bstrInstallDirectory) == 0)))
		{
			return S_FALSE;
		}

		T* pT = static_cast<T*>(this);
		_ASSERTE(pT);

		if (0 == pT->m_nFreezeEvents )
			if (S_FALSE == pT->FireOnRequestEdit( DISPID_ILX_INSTALL_DIRECTORY ) )
				return S_FALSE;

		// Set the property
		m_bstrInstallDirectory = bstrInstallDirectory;

		// Allow the control to be notified of the change
		hr = OnInstallDirectoryChanged();
		if (FAILED(hr)) return hr;

		// Persistence is dirty
		pT->SetDirty(TRUE);

		// Tell the container that the property has changed
		if (0 == pT->m_nFreezeEvents)
		{
			hr = pT->FireOnChanged( DISPID_ILX_INSTALL_DIRECTORY );
			_ASSERTE(SUCCEEDED(hr));
		}

		return S_OK;
	}

	virtual bool GetInstanceFilePath(TCHAR* szPath, long nMaxLen) const
	{
		_ASSERTE(IsInitialized());
		if (!IsInitialized() || m_bstrControlName.Length() == 0)
			return false;

		long nRet = 0;
		
		long nDataDirLen = m_bstrDataDirectory.Length();

		// Ex: "News Retrieval.NRT201_1234.xml"
		if (m_bstrDataDirectory[nDataDirLen-1] == _T('\\'))
		{
			nRet = _sntprintf(szPath, nMaxLen, _T("%s%s.%s.xml"), 
				m_bstrDataDirectory, m_bstrControlName, m_bstrInstanceName);
		}
		else
		{
			nRet = _sntprintf(szPath, nMaxLen, _T("%s\\%s.%s.xml"),
				m_bstrDataDirectory, m_bstrControlName, m_bstrInstanceName);
		}

		szPath[nMaxLen] = _T('\0');

		if (nRet == -1 || nRet > nMaxLen)
		{
			ILXLOGMSG3(ILXLOG_ERROR, _T("File path is too long '%s\\%s'.  Max length is %d."),
				m_bstrDataDirectory,
				m_bstrInstanceName,
				nMaxLen);
			return false;
		}
		else
		{
			return true;
		}
	}

	virtual bool GetContextFilePath(TCHAR* szPath, 
									long nMaxLen, 
									bool bUseCommonDataDirectory = false) const
	{
		_ASSERTE(IsInitialized());
		if (!IsInitialized())
			return false;

		long nRet = 0;

		const CComBSTR& bstrDirToUse = 
			bUseCommonDataDirectory ? m_bstrCommonDataDirectory : m_bstrDataDirectory;
		
		long nDataDirLen = bstrDirToUse.Length();

		if (nDataDirLen <= 0)
		{
			return false;
		}

		// Ex: "NRT201_1234.Context.xml"
		if (bstrDirToUse[nDataDirLen-1] == _T('\\'))
		{
			nRet = _sntprintf(szPath, nMaxLen, _T("%s%s.Context.xml"), 
				bstrDirToUse, m_bstrInstanceName);
		}
		else
		{
			nRet = _sntprintf(szPath, nMaxLen, _T("%s\\%s.Context.xml"),
				bstrDirToUse, m_bstrInstanceName);
		}

		szPath[nMaxLen] = _T('\0');

		if (nRet == -1 || nRet > nMaxLen)
		{
			ILXLOGMSG3(ILXLOG_ERROR, _T("File path is too long '%s\\%s'.  Max length is %d."),
				bstrDirToUse,
				m_bstrInstanceName,
				nMaxLen);
			return false;
		}
		else
		{
			return true;
		}
	}

	virtual bool IsInitialized() const
	{
		if (m_bstrInstanceName.Length() == 0 || m_bstrDataDirectory.Length() == 0)
		{
			return false;
		}
		else
		{
			return true;
		}
	}

	virtual HRESULT OnInstanceNameChanged()
	{
		return S_OK;
	}

	virtual HRESULT OnCommonDataDirectoryChanged()
	{
		return S_OK;
	}

	virtual HRESULT OnDataDirectoryChanged()
	{
		return S_OK;
	}

	virtual HRESULT OnInstallDirectoryChanged()
	{
		return S_OK;
	}

	virtual HRESULT OnLoadInstanceData()
	{
		return S_OK;
	}

	virtual HRESULT OnSaveInstanceData()
	{
		return S_OK;
	}

	virtual HRESULT OnInitNew()
	{
		return S_OK;
	}

public: // DATA

	CComBSTR			m_bstrCommonDataDirectory;
	CComBSTR			m_bstrDataDirectory;
	CComBSTR			m_bstrInstallDirectory;
	CComBSTR			m_bstrInstanceName;
	CComBSTR			m_bstrControlName;
	TypeSetDispids		m_setDispidsToIgnore;
	IILXResultManager*	m_spiResultManager;
};

#endif // INCLUDED_ILX_PERSISTENCE_IMPL_H
