///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/StringParsing.cpp,v 1.3 2004/09/07 06:08:57 me133 Exp $
//
//    File name:	StringParsing.cpp
//   Created By:	Marc Eaddy
//   Created On:	10/03/2000 7:21:10 PM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: StringParsing.cpp,v $
 * Revision 1.3  2004/09/07 06:08:57  me133
 * o Second checkin to CVS
 * 
 * 
 * 9     7/15/02 5:42p Marce
 * o Added function for removing leading and trailing whitespace in place
 * (no extra memory allocations)
 * 
 * 6     2/11/02 1:07p Marce
 * o Merged changes made for optimmizations from the AppControls versions
 * of these files
 * 
 * 8     1/29/02 12:35a Marce
 * o Added CopyAndGrowString()
 * 
 * 7     10/29/01 11:15a Marce
 * o Added DelimitedString2StringListEx() for handling string delimiters
 * 
 * 4     11/10/00 4:35p Marce
 * o Comments
 * o Miniscule optimizations
 * 
 * 2     10/04/00 5:37p Marce
 * o Fixed compile errors and warnings
 */

#include "StringParsing.h"		// Me

#include <assert.h>				// assert()

static const size_t nZERO = 0;

void DelimitedString2StringListInPlace(	TCHAR* szBuffer, 
										TypeListStrings* pList, 
										const TCHAR cDelimiter)
{
	assert(szBuffer);
	assert(pList);

	if (!szBuffer || !pList) return;

	TypeListStrings& list = *pList;	// Shortcut

	TCHAR szDelimiter[2];
	szDelimiter[0] = cDelimiter;
	szDelimiter[1] = _T('\0');
	
	TCHAR* pszToken = _tcstok(szBuffer, szDelimiter);
	while (pszToken)
	{
		list.push_back(pszToken);
		pszToken = _tcstok(NULL, szDelimiter);
	}
}

void DelimitedString2StringList(	const TCHAR* szBuffer, 
									TypeListStrings* pList, 
									const TCHAR cDelimiter)
{
	assert(szBuffer);
	assert(pList);

	TCHAR szStaticBuffer[1000];

	long nLength = (long) _tcslen(szBuffer);

	if (nLength+1 < sizeof szStaticBuffer / sizeof *szStaticBuffer)
	{
		_tcscpy(szStaticBuffer, szBuffer);
		DelimitedString2StringListInPlace(szStaticBuffer, pList, cDelimiter);
	}
	else
	{
		TCHAR* szTempBuffer = new TCHAR[nLength + 1];
		_tcscpy(szTempBuffer, szBuffer);

		DelimitedString2StringListInPlace(szTempBuffer, pList, cDelimiter);

		delete[] szTempBuffer;
	}
}

void DelimitedString2StringListEx(	const TCHAR* szBuffer, 
									TypeListStrings* pList, 
									const TCHAR* szDelimiter)
{
	assert(szBuffer);
	assert(pList);
	assert(szDelimiter && *szDelimiter);

	if (!szBuffer || !pList || !szDelimiter || !*szDelimiter) return;

	TypeListStrings& list = *pList;	// Shortcut

	tstring strBuffer;

	TCHAR szOneChar[2]	= { 0 }; // Last char is always NULL
		// Used because you cannot call += with a TCHAR, you must
		// use a string.

	long nLength = (long) _tcslen(szBuffer);
	long nDelimiterLen = (long) _tcslen(szDelimiter);

	for(int i = 0; i < nLength; ++i)
	{
		if (_tcsncmp(&szBuffer[i], szDelimiter, nDelimiterLen) == 0)
		{
			list.push_back(strBuffer);
			strBuffer.assign(_T(""), nZERO); // Clear the string
			i += nDelimiterLen - 1;
		}
		else
		{
			szOneChar[0] = szBuffer[i];
			strBuffer += szOneChar;
		}
	}

	if (!strBuffer.empty())
	{
		list.push_back(strBuffer);
	}
}

void DelimitedString2NumList(	const TCHAR* szBuffer, 
								TypeListNums* pList, 
								const TCHAR cDelimiter)
{
	assert(szBuffer);
	assert(pList);

	TypeListNums& list = *pList;	// Shortcut

	tstring strBuffer	= _T("");
	TCHAR szOneChar[2]	= { 0 };
	TCHAR* pStopChar = 0;

	for(unsigned int i = 0; i < _tcslen(szBuffer); ++i)
	{
		if (szBuffer[i] == cDelimiter)
		{
			unsigned long nVal = _tcstoul(strBuffer.c_str(), &pStopChar, 0);
			assert(*pStopChar == _T('\0'));
			list.push_back(nVal);
			strBuffer = _T("");
		}
		else
		{
			szOneChar[0] = szBuffer[i];
			strBuffer += szOneChar;
		}
	}

	if (strBuffer.length() > 0)
	{
		unsigned long nVal = _tcstoul(strBuffer.c_str(), &pStopChar, 0);
		assert(*pStopChar == _T('\0'));
		list.push_back(nVal);
	}
}


//*********************************************************************
// FUNCTION		:	CopyAndGrowString
//					
// DESCRIPTION	:	Copy source string to destination string. If
//					destination string is null allocate memory for it
//					and then copy to it. If destination string capacity
//					is less than source string, free memory first, then
//					allocate memory and copy to it.
//
// PARAMETERS	:	
//   ppDest			- Destination buffer.  Can't be NULL.
//   pnDestCapacity	- Capacity of destination buffer.  Can be NULL.
//   szSrc			- Source buffer.  If NULL or empty causes ppDest to
//					  be set to empty.
//   nSrcLenHint	- Len of source buffer if known.  If -1 and szSrc is
//					  not NULL, the length will be found by calling
//					  _tcslen(szSrc).
//					
// COMMENTS		:	Taken from ILXGrid. Modifications by VipinK and MarcE.
//					
//*********************************************************************
void CopyAndGrowString(	TCHAR** ppDest, 
						long* pnDestCapacity, 
						const TCHAR* szSrc,
						long nSrcLenHint)
{
	static const long nGROW_FACTOR = 2;

	assert(!pnDestCapacity || *pnDestCapacity >= 0);
	assert(ppDest);

    if (!ppDest) return;

	long nSrcLenPlusNull = nSrcLenHint;

	if (nSrcLenPlusNull == -1)
	{
		nSrcLenPlusNull = szSrc ? (long) _tcslen(szSrc) : 0;
	}

	++nSrcLenPlusNull; // Account for terminating null

	long nCapacityToUse = pnDestCapacity ? *pnDestCapacity : 0;

    if (!*ppDest || nSrcLenPlusNull > nCapacityToUse)
	{
		nCapacityToUse = nSrcLenPlusNull;
		if (pnDestCapacity)
		{
			// Keep doubling capacity
			nCapacityToUse *= nGROW_FACTOR;
		}
		
		if (*ppDest) delete [] *ppDest;
		*ppDest = new TCHAR[nCapacityToUse];

		if (pnDestCapacity)
		{
			*pnDestCapacity = nCapacityToUse;
		}
    }

	if (szSrc)
		memcpy(*ppDest, szSrc, sizeof(TCHAR) * nSrcLenPlusNull);
	else
		**ppDest = _T('\0');
}

void TrimWhitespaceInPlace(wchar_t* psz)
{
	if (!psz || !*psz) return;

	// First count the leading whitespace
	wchar_t* szSkipLeadingWhitespace = psz;
	while (	szSkipLeadingWhitespace && 
			*szSkipLeadingWhitespace &&
			iswspace(*szSkipLeadingWhitespace))
	{
		++szSkipLeadingWhitespace;
	}

	long nLeadingWhitespaceToSkip = (long) (szSkipLeadingWhitespace - psz);

	// Now go to the end of the string

	while (	szSkipLeadingWhitespace &&
			*szSkipLeadingWhitespace)
	{
		++szSkipLeadingWhitespace;
	}

	// Backup off the null terminator
	--szSkipLeadingWhitespace;

	// Erase trailing whitespace
	while (szSkipLeadingWhitespace >= psz &&
			iswspace(*szSkipLeadingWhitespace))
	{
		*szSkipLeadingWhitespace-- = L'\0';
	}

	// Copy characters left on top of the leading whitespace
	if (nLeadingWhitespaceToSkip > 0)
	{
		// "    Joe" -> "Joe Joe"
		for(long i = 0; psz[i + nLeadingWhitespaceToSkip]; ++i)
		{
			psz[i] = psz[i + nLeadingWhitespaceToSkip];
		}

		// Copy null terminator left
		// "Joe Joe" -> "Joe"
		psz[i] = L'\0';
	}
}