///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/StringParsing.h,v 1.2 2004/09/07 06:08:57 me133 Exp $
//
//    File name:	StringParsing.h
//   Created By:	Marc Eaddy
//   Created On:	10/03/2000 7:19:59 PM
//
//      Purpose:	Convert a delimited string into a list
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: StringParsing.h,v $
 * Revision 1.2  2004/09/07 06:08:57  me133
 * o Second checkin to CVS
 * 
 * 
 * 9     7/15/02 5:42p Marce
 * o Added function for removing leading and trailing whitespace in place
 * (no extra memory allocations)
 * 
 * 6     2/11/02 1:07p Marce
 * o Merged changes made for optimmizations from the AppControls versions
 * of these files
 * 
 * 8     1/29/02 12:35a Marce
 * o Added CopyAndGrowString()
 * 
 * 7     10/29/01 11:15a Marce
 * o Added DelimitedString2StringListEx() for handling string delimiters
 * 
 * 4     11/10/00 12:18p Marce
 * o Added comments
 * o Reduced dependencies
 * 
 * 2     10/04/00 5:37p Marce
 * o Fixed compile errors and warnings
 */

#ifndef INCLUDE_STRING_PARSING_H
#define INCLUDE_STRING_PARSING_H

#pragma warning(disable:4786)	// identifier was truncated

#include <string>		// STL basic_string<>
#include <list>			// STL list<>
#include <tchar.h>		// TCHAR

// Macros

// Macros

//
// Usage: Use when the string can be NULL.  NULL and
// empty are considered equal.
//
//	bool operator<(const Foo& lhs, const Foo& rhs)
//	{
//		SAFE_TCSCMP_LESS(lhs.szName, rhs.szName)
//		return false;
//	}
//
#define SAFE_TCSCMP_LESS(x, y)									\
	if (x != y)													\
	{															\
		if (x && y)												\
		{														\
			int cmp = _tcscmp(x, y);							\
			if (cmp != 0) return cmp < 0;						\
		}														\
		else if (x && *x != _T('\0'))							\
		{														\
			/* y is null so return true if x is non-empty */	\
			return false;										\
		}														\
		else if (y && *y != _T('\0'))							\
		{														\
			/* x is null so return true if y is non-empty */	\
			return true;										\
		}														\
	}

//
// Usage: Use when the string can be NULL
//
//	bool operator==(const Foo& lhs, const Foo& rhs)
//	{
//		return SAFE_TCSCMP_EQUAL(lhs.szName, rhs.szName);		
//	}
//
#define SAFE_TCSCMP_EQUAL(x, y)							\
	( (x == y) || (x && y && _tcscmp(x, y) == 0) )

//
// Usage: Use when the string can be NULL
//
//	int nLen = SAFE_TCSLEN(szText);
//
#define SAFE_TCSLEN(x) (x ? _tcslen(x) : 0)

// Type defines

typedef std::basic_string<TCHAR>	tstring;
typedef std::list<tstring>			TypeListStrings;
typedef std::list<long>				TypeListNums;

// Helper functions

void DelimitedString2StringList(const TCHAR* szBuffer, 
								TypeListStrings* pList, 
								const TCHAR cDelimiter = _T(','));
	//
	// Converts a (comma) delimited string into a list of
	// tstrings.
	//
	// Note: Whitespace is not allowed between the delimiters.
	//
	// Ex: "eeeniee,meanie,minee,moe"
	//

void DelimitedString2StringListInPlace(	TCHAR* szBuffer, 
										TypeListStrings* pList, 
										const TCHAR cDelimiter = _T(','));


void DelimitedString2StringListEx(	const TCHAR* szBuffer, 
									TypeListStrings* pList, 
									const TCHAR* szDelimiter = _T(","));
	//
	// Converts a (comma) delimited string into a list of
	// tstrings.
	//
	// Note: Whitespace is not allowed between the delimiters.
	//
	// Ex: "eeeniee,meanie,minee,moe"
	//

void DelimitedString2NumList(	const TCHAR* szBuffer, 
								TypeListNums* pList, 
								const TCHAR cDelimiter = _T(','));
	//
	// Converts a (comma) delimited string of numbers into a list
	// of longs.
	//
	// See documentation on strtoul() for a description of the
	// number format allowed.  The base used is zero so the
	// initial characters of the string are used to determine
	// the base.
	//
	// Note: Whitespace is not allowed between the delimiters.
	//
	// Ex: "123,56,789" or "0x80004005,123"
	// 

void CopyAndGrowString(	TCHAR** ppDest, 
						long* pnDestCapacity, 
						const TCHAR* szSrc,
						long nSrcLenHint);
	//
	// Copies the source string to the destination string, allocating
	// memory if needed.  The caller should call delete[] on the
	// returned string.
	//
	// You can avoid allocations by keeping track of the current
	// buffer capacity and passing it as a parameter.  Furthermore,
	// if you know the length of the source string for some reason
	// you can pass that in.  Otherwise, pass in -1 and strlen() will
	// be called.
	//
	// PARAMETERS	:	
	//   ppDest			- Destination buffer
	//   pnDestCapacity	- Capacity of destination buffer.  Can be NULL.
	//   szSrc			- Source buffer
	//   nSrcLenHint	- Len of source buffer if known.  If -1, the length
	//					  will be found by calling _tcslen(szSrc).
	//					

void TrimWhitespaceInPlace(wchar_t* psz);

#endif // INCLUDE_STRING_PARSING_H