///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/XmlHelpers2.cpp,v 1.3 2004/09/07 06:08:58 me133 Exp $
//
//    File name:	XmlHelpers2.cpp
//   Created By:	Marc Eaddy
//   Created on:	03/07/2000 11:05:07 AM
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: XmlHelpers2.cpp,v $
 * Revision 1.3  2004/09/07 06:08:58  me133
 * o Second checkin to CVS
 * 
 * 
 * 23    4/10/02 8:19p Marce
 * o Can now specify NULL for the filename
 * 
 * 22    4/09/02 12:50p Marce
 * o No longer puts the latest year in the copyright.  We should update
 * the year manually (every Jan 1).
 * 
 * 21    3/22/02 12:51p Marce
 * o Removed space in "File :"
 * 
 * 20    3/18/02 9:15p Marce
 * o Now saves XML whitespace correctly
 * o Now includes the date in the XML comments
 * 
 * 18    1/04/02 10:32a Marce
 * Fixed compile errors
 * 
 * 17    1/03/02 1:32p Marce
 * o Minor change
 * 
 * 16    1/03/02 12:28p Marce
 * o Now the copyright comment has the current year
 * 
 * 15    11/02/01 3:09p Marce
 * o Now handles BYREF variants
 * 
 * 14    8/17/01 3:17p Marce
 * o Now saved XML looks much better
 * 
 * 12    7/17/01 6:59p Marce
 * o Now compiles in isolation
 * 
 * 11    7/03/01 6:24p Marce
 * o Added SaveAttributesToXML()
 * 
 * 10    4/09/01 12:20p Marce
 * o Now checks for S_OK instead of just FAILED(hr)
 * 
 * 9     1/04/01 12:22p Marce
 * o Changed copyright to 2001
 * 
 * 8     12/15/00 11:35a Marce
 * o Now always adds a beginning space
 * 
 * 7     12/07/00 7:38p Marce
 * o Added helper function for adding ILX copyright info and description
 * comments to an XML files
 * 
 * 5     3/09/00 3:55p Marce
 * o Fixed a bug which caused a null ptr read
 * o Changed ProgId to Iface
 * o Now checks iface in event table when looking for the best event table
 * entry
 * 
 * 3     3/07/00 12:06p Marce
 * o Fixed memory leak
 * 
 * 2     3/07/00 11:29a Marce
 * o Added some XML helpers
 * 
 * 1     3/07/00 11:06a Marce
 */

#include "StdAfx.h"

#include "XmlHelpers2.h"	// Me

#ifndef INCLUDED_MSXML_TYPE_LIBRARY
#define INCLUDED_MSXML_TYPE_LIBRARY
#import "msxml.dll" rename_namespace("MSXML"), named_guids, raw_interfaces_only, raw_native_types
#endif // INCLUDED_MSXML_TYPE_LIBRARY

#include <time.h>			// time(), localtime()
#include <limits.h>			// MAX_PATH

using namespace MSXML;

// Constants

static const TCHAR* const szDELIM_COMMENT = 
	_T("************************************************************** ");
static const TCHAR* const szCOPYRIGHT_COMMENT =
	_T("Copyright (C) 2001-2002 ILX Systems.  All Rights Reserved.     ");

static const TCHAR* const szBLANK = _T("");

static const long nFIXED_COMMENT_LEN = _tcslen(szDELIM_COMMENT) + 1;

// Helper functions
HRESULT AppendFormattedComment(	MSXML::IXMLDOMDocument* piDocument, 
								MSXML::IXMLDOMNode* piNode,
								const TCHAR* szComment,
								long nFixedLength = -1);

HRESULT LoadAttributesFromXML(	/*[in]*/ MSXML::IXMLDOMNode* piNodeWithAttributes, 
								/*[in,out]*/ TypeMapStr2Var* pmapAttributes)
{
	HRESULT hr = S_OK;

	USES_CONVERSION;

	// Pre-conditions
	if (piNodeWithAttributes == 0 || !pmapAttributes) return E_POINTER;

	MSXML::IXMLDOMNamedNodeMapPtr spiAttributeNodeMap = 0;

	hr = piNodeWithAttributes->get_attributes(&spiAttributeNodeMap);
	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;
	if (spiAttributeNodeMap == 0) return E_UNEXPECTED;

	long nTotalAttributes = 0;
	hr = spiAttributeNodeMap->get_length(&nTotalAttributes);
	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;

	for(int i = 0; i < nTotalAttributes; ++i)
	{
		MSXML::IXMLDOMNodePtr spiNamedAttribute = 0;
		
		hr = spiAttributeNodeMap->get_item(i, &spiNamedAttribute);
		if (hr != S_OK || spiNamedAttribute == 0)
		{
			_ASSERTE(spiNamedAttribute != 0);
			continue;
		}

		// Verify that the name node is a text attribute
		MSXML::DOMNodeType typeNode = (MSXML::DOMNodeType) MSXML::NODE_INVALID;
		hr = spiNamedAttribute->get_nodeType(&typeNode);
		
		if (hr != S_OK || typeNode != MSXML::NODE_ATTRIBUTE)
		{
			_ASSERTE(!_T("Failed to retrieve node type or node is not a text node."));
			continue;
		}

		// Get the node name

		CComBSTR bstrNodeName;

		hr = spiNamedAttribute->get_nodeName(&bstrNodeName);
	
		if (hr != S_OK || !bstrNodeName)
		{
			_ASSERTE(!_T("Failed to retrieve the node name"));
			continue;
		}

		// Get the node value

		_variant_t varNodeValue;

		hr = spiNamedAttribute->get_nodeValue(&varNodeValue);
		if (hr != S_OK) return hr;

		const TCHAR* szName = OLE2CT(bstrNodeName);

		(*pmapAttributes)[szName] = varNodeValue;

//		AtlTrace(_T("%s=%s\n"), szName,
//			static_cast<const TCHAR*>(static_cast<_bstr_t>(varNodeValue)));

	} // Iterate through all the attributes

	return S_OK;
}

HRESULT SaveAttributesToXML(	/*[in]*/ MSXML::IXMLDOMNode* piNode, 
								/*[in]*/ const TypeMapStr2Var& mapAttributes)
{
	HRESULT hr = S_OK;

	MSXML::IXMLDOMElementPtr spiElement(piNode);
	_ASSERTE(spiElement != 0);

	for(TypeMapStr2Var::const_iterator it = mapAttributes.begin();
		it != mapAttributes.end();
		++it)
	{
		const tstring& strName		= (*it).first;
		const _variant_t& varValue	= (*it).second;

		if (varValue.vt == (VT_VARIANT | VT_BYREF) && 
			varValue.pvarVal && 
			varValue.pvarVal->vt != VT_EMPTY)
		{
			hr = spiElement->setAttribute(_bstr_t(strName.c_str()), *varValue.pvarVal);
			if (FAILED(hr)) return hr;
		}
		else if (varValue.vt != VT_EMPTY)
		{
			hr = spiElement->setAttribute(_bstr_t(strName.c_str()), varValue);
			if (FAILED(hr)) return hr;
		}
	}
		
	return S_OK;
}

HRESULT AppendILXComment(	MSXML::IXMLDOMNode* piNode, 
							const TCHAR* szLine1,
							const TCHAR* szLine2,
							const TCHAR* szLine3,
							const TCHAR* szLine4)
{
	// Verify parameters (only szLine1 is necessary)
	_ASSERTE(szLine1 && *szLine1);
	_ASSERTE(piNode);

	HRESULT hr = S_OK;

	CComQIPtr<MSXML::IXMLDOMDocument> spiDocument(piNode);

	if (!spiDocument)
	{
		hr = piNode->get_ownerDocument(&spiDocument);
	}

	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;
	if (!spiDocument) return E_UNEXPECTED;

	// <!-- ********************************************************* -->
	hr = AppendFormattedComment(spiDocument, piNode, szDELIM_COMMENT, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	time_t timNow = time(0);
	struct tm* ptmNow = localtime(&timNow);
	_ASSERTE(ptmNow);

//	TCHAR szCopyrightComment[1000];
//	_tcsftime(	szCopyrightComment, 
//				(sizeof szCopyrightComment / sizeof *szCopyrightComment) - 1,
//				szCOPYRIGHT_COMMENT, 
//				ptmNow);

	// <!-- Copyright (C) 2000 ILX Systems.  All Rights Reserved.     -->
	hr = AppendFormattedComment(spiDocument, piNode, szCOPYRIGHT_COMMENT, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// <!--                                                           -->
	hr = AppendFormattedComment(spiDocument, piNode, szBLANK, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// Ex: <!-- ActiveILX Print Settings -->
	hr = AppendFormattedComment(spiDocument, piNode, szLine1, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	if (szLine2)
	{
		// Ex: <!-- ActiveILX Print Settings -->
		hr = AppendFormattedComment(spiDocument, piNode, szLine2, nFIXED_COMMENT_LEN);
		if (FAILED(hr)) return hr;
	}

	if (szLine3)
	{
		// Ex: <!-- ActiveILX Print Settings -->
		hr = AppendFormattedComment(spiDocument, piNode, szLine3, nFIXED_COMMENT_LEN);
		if (FAILED(hr)) return hr;
	}

	if (szLine4)
	{
		// Ex: <!-- ActiveILX Print Settings -->
		hr = AppendFormattedComment(spiDocument, piNode, szLine4, nFIXED_COMMENT_LEN);
		if (FAILED(hr)) return hr;
	}

	// <!--                                                           -->
	hr = AppendFormattedComment(spiDocument, piNode, szBLANK, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// <!-- ********************************************************* -->
	hr = AppendFormattedComment(spiDocument, piNode, szDELIM_COMMENT, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	return S_OK;
}

HRESULT AppendILXComment2(	MSXML::IXMLDOMNode* piNode, 
							const TCHAR* szFilename,
							bool bAlignColons,
							const TCHAR* szLine1,
							const TCHAR* szLine2,
							const TCHAR* szLine3,
							const TCHAR* szLine4)
{
	// Verify parameters
	_ASSERTE(piNode);

	HRESULT hr = S_OK;

	CComQIPtr<MSXML::IXMLDOMDocument> spiDocument(piNode);

	if (!spiDocument)
	{
		hr = piNode->get_ownerDocument(&spiDocument);
	}

	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;
	if (!spiDocument) return E_UNEXPECTED;

	// Step 1:	Determine the number of spaces between the field and the colon

	long nColonIndent = 0;

	const TCHAR* szLines[] = { _T("File:"), _T("Date:"), szLine1, szLine2, szLine3, szLine4 };

	if (bAlignColons)
	{
		for(int i = 0; i < sizeof szLines / sizeof *szLines; ++i)
		{
			const TCHAR* pszPos = szLines[i] ? _tcschr(szLines[i], _T(':')) : 0;
			if (pszPos && pszPos - szLines[i] > nColonIndent)
			{
				nColonIndent = pszPos - szLines[i];
			}
		}
	}

	// Sanity Check: Ensure indent is between 1 and 100
	_ASSERTE(nColonIndent >= 0 && nColonIndent <= 100);
	nColonIndent -= 4; // Length of "File" and "Date"
	nColonIndent = max(min(nColonIndent, 100), 0);

	// Step 2:	Build File string

	TCHAR szFilenameText[MAX_PATH+1000];
	if (szFilename)
	{
		_sntprintf(	szFilenameText, 
					(sizeof szFilenameText / sizeof *szFilenameText) - 1,
					_T("File%*s: %s"), nColonIndent, _T(""), szFilename);
		szFilenameText[(sizeof szFilenameText / sizeof *szFilenameText) - 1] = _T('\0');
	}

	// Step 3:	Build Copyright string

	time_t timNow = time(0);
	struct tm* ptmNow = localtime(&timNow);
	_ASSERTE(ptmNow);

	//TCHAR szCopyrightComment[1000];
	//_tcsftime(	szCopyrightComment, 
	//			(sizeof szCopyrightComment / sizeof *szCopyrightComment) - 1,
	//			szCOPYRIGHT_COMMENT, 
	//			ptmNow);

	// Step 4:	Build Date string

	TCHAR szDate[1000];
	TCHAR szDateFmt[1000];

	_sntprintf(szDateFmt, 
		(sizeof szDateFmt / sizeof *szDateFmt) - 1,
		_T("Date%*s: %%A, %%B %%#d, %%#Y, %%#I:%%M:%%S %%p"),
		nColonIndent, _T(""));
	szDateFmt[(sizeof szDateFmt / sizeof *szDateFmt) - 1] = _T('\0');

	_tcsftime(szDate, 
		(sizeof szDate / sizeof *szDate) - 1,
		szDateFmt,
		ptmNow);

	// Step 5:	Output the comment

	// <!-- ********************************************************* -->
	hr = AppendFormattedComment(spiDocument, piNode, szDELIM_COMMENT, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// <!-- Copyright (C) 2000 ILX Systems.  All Rights Reserved.     -->
	hr = AppendFormattedComment(spiDocument, piNode, szCOPYRIGHT_COMMENT, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// <!--                                                           -->
	hr = AppendFormattedComment(spiDocument, piNode, szBLANK, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// Ex: <!-- File : ILXPrintDef.xml                                 -->
	if (szFilename)
	{
		hr = AppendFormattedComment(spiDocument, piNode, szFilenameText, nFIXED_COMMENT_LEN);
		if (FAILED(hr)) return hr;
	}

	// Ex: <!-- Date : Wednesday, February 27, 2002, 9:33:54 PM        -->
	hr = AppendFormattedComment(spiDocument, piNode, szDate, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	for(int i = 2; i < sizeof szLines / sizeof *szLines; ++i)
	{
		if (szLines[i])
		{
			hr = AppendFormattedComment(spiDocument, piNode, szLines[i], nFIXED_COMMENT_LEN);
			if (FAILED(hr)) return hr;
		}
	}

	// <!--                                                           -->
	hr = AppendFormattedComment(spiDocument, piNode, szBLANK, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	// <!-- ********************************************************* -->
	hr = AppendFormattedComment(spiDocument, piNode, szDELIM_COMMENT, nFIXED_COMMENT_LEN);
	if (FAILED(hr)) return hr;

	return S_OK;
}

HRESULT AppendFormattedComment(	MSXML::IXMLDOMDocument* piDocument, 
								MSXML::IXMLDOMNode* piNode,
								const TCHAR* szComment,
								long nFixedLength)
{
	_ASSERTE(piDocument);
	_ASSERTE(szComment);

	static const TCHAR cBLANK = _T(' ');

	HRESULT hr = S_OK;

	tstring strComment = szComment;

	bool bAppendSpace = false;

	if (strComment.length() > 0)
	{
		// Start with a space
		strComment.insert(strComment.begin(), cBLANK);

		// Do we need to append a trailing space?
		if (strComment[strComment.length()-1] != cBLANK)
		{
			bAppendSpace = true;
		}
	}

	// Add spaces to pad the comment so its the same length as the
	// rest of the comments
	if ( (long)strComment.length() < nFixedLength)
	{
		if (bAppendSpace)
		{
			--nFixedLength;
		}
		strComment.resize(nFixedLength, cBLANK);
	}

	// Append a space
	if (bAppendSpace)
	{
		strComment.append(1, cBLANK);
	}

	CComPtr<MSXML::IXMLDOMComment> spiComment;
	hr = piDocument->createComment(CComBSTR(strComment.c_str()), &spiComment);
	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;
	if (!spiComment) return E_UNEXPECTED;

	CComPtr<MSXML::IXMLDOMNode> spiNewNode;
	hr = piNode->appendChild(spiComment, &spiNewNode);
	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;
	if (!spiNewNode) return E_UNEXPECTED;

	return S_OK;
}

HRESULT AppendWhitespace(	MSXML::IXMLDOMDocument* piDocument,
							MSXML::IXMLDOMNode* piNode,
							const TCHAR* szWhiteSpace)
{
	HRESULT hr = S_OK;

	if (!piDocument || !szWhiteSpace ||!*szWhiteSpace) return E_POINTER;

	CComPtr<MSXML::IXMLDOMText> spiWhitespace;
	hr = piDocument->createTextNode(CComBSTR(szWhiteSpace), &spiWhitespace);
	if (FAILED(hr) || !spiWhitespace)
	{
		return FAILED(hr) ? hr : E_FAIL;
	}

	CComPtr<MSXML::IXMLDOMNode> spiNewNode;
	hr = piNode->appendChild(spiWhitespace, &spiNewNode);
	if (FAILED(hr)) return hr;
	if (hr != S_OK) return E_FAIL;
	if (!spiNewNode) return E_UNEXPECTED;
	
	return S_OK;
}
