///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2000 ILX Systems.  All Rights Reserved.
//
//      $Header: /u/4/m/me133/cvs/Common/XmlHelpers2.h,v 1.2 2004/09/07 06:08:58 me133 Exp $
//
//    File name:	XmlHelpers2.h
//   Created By:	Marc Eaddy
//   Created on:	03/07/2000 10:49:43 AM
//
//      Purpose:	Copy attributes from an XML node into an STL map
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: XmlHelpers2.h,v $
 * Revision 1.2  2004/09/07 06:08:58  me133
 * o Second checkin to CVS
 * 
 * 
 * 16    3/18/02 9:15p Marce
 * o Now saves XML whitespace correctly
 * o Now includes the date in the XML comments
 * 
 * 14    10/19/01 7:18p Marce
 * o Fixed compile errors
 * 
 * 13    8/17/01 3:17p Marce
 * o Now saved XML looks much better
 * 
 * 11    7/03/01 6:24p Marce
 * o Added SaveAttributesToXML()
 * 
 * 10    12/07/00 7:46p Marce
 * o More and More comments
 * 
 * 9     12/07/00 7:42p Marce
 * o More comments
 * 
 * 8     12/07/00 7:41p Marce
 * o Comments
 * 
 * 7     12/07/00 7:38p Marce
 * o Added helper function for adding ILX copyright info and description
 * comments to an XML files
 * 
 * 6     11/10/00 1:17p Marce
 * o Added comments
 * 
 * 3     3/07/00 12:31p Marce
 * o Now uses case insensitive strings for the XML
 * 
 * 2     3/07/00 11:29a Marce
 * o Added some XML helpers
 */

#ifndef INCLUDED_XML_HELPERS_2_H
#define INCLUDED_XML_HELPERS_2_H

#include "CaseInsensitiveString.h"	// tstring

#include <map>						// STL map<>

#include <comdef.h>					// _variant_t

typedef std::map<tstring, _variant_t, TypeFuncLessTString >	TypeMapStr2Var;

// Forward delcare the IXMLDOMNode and IXMLDOMDocument interface
#ifndef XML_DOM_NODE_FWD_DECLARED
#define XML_DOM_NODE_FWD_DECLARED

// Forward delcare the IXMLDOMNode and IXMLDOMNodePtr classes
namespace MSXML
{
	struct __declspec(uuid("2933bf80-7b36-11d2-b20e-00c04f983e60"))
	/* dual interface */ IXMLDOMNode;
	struct __declspec(uuid("2933bf81-7b36-11d2-b20e-00c04f983e60"))
	/* dual interface */ IXMLDOMDocument;
};

#endif // XML_DOM_NODE_FWD_DECLARED

HRESULT LoadAttributesFromXML(	/*[in]*/ MSXML::IXMLDOMNode* piNode, 
								/*[in,out]*/ TypeMapStr2Var* pmapAttributes);
	//
	// Copies all the attributes for the node specified into the
	// map of string keys to variant values.
	//
	// The string keys are case-insensitive (via TypeFuncLesTString
	// above) so that myMap[_T("GRID_LAYOUT")] has the same effect as
	// myMap[_T("Grid_Layout")].
	//
	// The map passed in can be non-empty.  Attributes will simply be
	// added to it.  If the is attribute already in the map then it
	// will be overwritten.
	//

HRESULT SaveAttributesToXML(	/*[in]*/ MSXML::IXMLDOMNode* piNode, 
								/*[in]*/ const TypeMapStr2Var& mapAttributes);

HRESULT AppendILXComment(	MSXML::IXMLDOMNode* piNode, 
							const TCHAR* szDescriptionLine1,
							const TCHAR* szDescriptionLine2 = 0,
							const TCHAR* szDescriptionLine3 = 0,
							const TCHAR* szDescriptionLine4 = 0);
	//
	// Appends an ILX copyright and description comments to an XML
	// file.  The copyright and asterik headers are inserted
	// automatically.  You must specify at least one description line.
	// You can optionally specify more description lines.
	//
	// If you want the comment to appear at the top of the file, you
	// should call this function immediately before calling createElement/
	// appendChild.
	//
	// Here is an example: szDescriptionLine1 = "ActiveILX Print Settings"
	//
	// <!-- ********************************************************* -->
	// <!-- Copyright (C) 2000 ILX Systems.  All Rights Reserved.     -->
	// <!--                                                           -->
	// <!-- ActiveILX Print Settings                                  -->
	// <!--                                                           -->
	// <!-- ********************************************************* -->
	//

HRESULT AppendILXComment2(	MSXML::IXMLDOMNode* piNode, 
							const TCHAR* szFileName,
							bool bAlignColons,
							const TCHAR* szDescriptionLine1 = 0,
							const TCHAR* szDescriptionLine2 = 0,
							const TCHAR* szDescriptionLine3 = 0,
							const TCHAR* szDescriptionLine4 = 0);
	//
	// Appends an ILX copyright and description comments to an XML
	// file.  The copyright and asterik headers are inserted
	// automatically.  You must specify at least the file name.
	// You can optionally specify more description lines.
	//
	// If you want the comment to appear at the top of the file, you
	// should call this function immediately before calling createElement/
	// appendChild.
	//
	// Here is an example: szDescriptionLine1 = "ActiveILX Print Settings"
	//
	// <!-- ********************************************************* -->
	// <!-- Copyright (C) 2000 ILX Systems.  All Rights Reserved.     -->
	// <!--                                                           -->
	// <!-- File: ILXPrintDef.xml                                     -->
	// <!-- Date: Wednesday, February 27, 2002, 9:33:54 PM            -->
	// <!--                                                           -->
	// <!-- ActiveILX Print Settings                                  -->
	// <!--                                                           -->
	// <!-- This file was automatically generated.                    -->
	// <!--                                                           -->
	// <!-- ********************************************************* -->
	//

HRESULT AppendWhitespace(	MSXML::IXMLDOMDocument* piDocument, 
							MSXML::IXMLDOMNode* piNode,
							const TCHAR* szWhiteSpace);

#endif // INCLUDED_XML_HELPERS_2_H