///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/AssemblyDiff.cpp 23    12/16/05 8:40p Eaddy $
//
//    File name:	AssemblyDiff.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/24/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/AssemblyDiff.cpp $
 * 
 * 23    12/16/05 8:40p Eaddy
 * o Now only counts legal methods
 * 
 * 22    12/16/05 4:30p Eaddy
 * o Improved error msg
 * 
 * 21    12/15/05 2:35a Eaddy
 * o Added code to count methods updated
 * 
 * 20    12/14/05 2:16a Eaddy
 * o Even if we don't support emitting new types, methods, and fields, 
 * we still need to fix them up.  And we can't assume that they can
 * be iterated in lock step since they may have been reordered.
 * Actually, this only really happens when the build description file
 * doesn't match the executable because the exe was built using
 * VS.
 * 
 * 19    12/10/05 12:39a Eaddy
 * o Reduce log noise
 * 
 * 18    12/09/05 9:33p Eaddy
 * o Added support for introducing fields but of course this doesn't work
 * either because the CLR requires DefineField to be called in
 * ModuleLoadFinished
 * 
 * 17    12/09/05 9:01p Eaddy
 * o CMethodInfo is now stored in the methodDef iterator
 * 
 * 16    12/09/05 2:02p Eaddy
 * o Removed spurious comments
 * 
 * 15    12/09/05 1:34p Eaddy
 * o Improved log messages
 * 
 * 14    12/08/05 10:33p Eaddy
 * o Now really fixed problem with changing a method's local variables
 * 
 * 13    12/07/05 11:36p Eaddy
 * o Now emits new typeRefs
 * o Emitting assemblyRefs doesn't seem to work
 * 
 * 12    12/07/05 7:49p Eaddy
 * o Fixed bug where we were not updating tokens for typeRefs and
 * memberRefs
 * o Now emits new member refs
 * o Still need to check and emit other tokens (e.g. typeRefs)
 * o Still need to consolidate emitting and patching local vars
 * 
 * 11    10/07/05 8:02p Eaddy
 * o Minor change
 * 
 * 10    5/28/05 1:49p Eaddy
 * o Changed copyright
 *
 * Revision 1.4  2005/01/14 23:24:07  me133
 * o Improved logging
 *
 * Revision 1.3  2005/01/14 22:59:39  me133
 * o Trying to debug func ptr problems
 *
 * Revision 1.2  2004/10/26 04:33:04  me133
 * o Fixed bug where methods with exceptions weren't being decoded
 * properly
 *
 * Revision 1.1  2004/10/25 01:03:48  me133
 * o Method IL comparison is now complete and correct
 */

#include "stdafx.h"

#include "AssemblyDiff.h"		// Me

#include "MetadataIterators.h"	// typedef_iterator, etc.

#include "MethodInfo.h"			// CMethodInfo

#include "MetadataHelpers.h"	// Token2TypeName()

#include "ILHelper.h"			// opcode_iterator

#include <map>					// STL map<>

using namespace std;

// HELPER FUNCTIONS

HRESULT CreateTypePatches(	typeDef_iterator& itTypeDefOrig, 
							typeDef_iterator& itTypeDefNew,
							const WCHAR* szTypeName,
							ModuleID moduleIDOrig,
							ICorProfilerInfo* piProfInfo,
							IMetaDataEmit* piEmit,
							IMethodMalloc* piAlloc,
							const TypeListTokenPairs& listNewOrReorderedTokens,
							/*[in]*/ bool bFirstTime,
							TypeMapMethodInfo* pmapChangedFunctions);

HRESULT CreateMethodPatches(	methodDef_iterator& itMethodDefOrig,
								methodDef_iterator& itMethodDefNew,
								const WCHAR* szTypeName,
								ModuleID moduleIDOrig,
								ICorProfilerInfo* piProfInfo,
								IMetaDataEmit* piEmit,
								IMethodMalloc* piAlloc,
								const TypeListTokenPairs& listNewOrReorderedTokens,
								TypeMapMethodInfo* pmapChangedFunctions);

HRESULT PatchTokens(	const TypeMapMethodInfo& listChangedMethods,
						const TypeListTokenPairs& listNewOrReorderedTokens);

///////////////////
// IMPLEMENTATION
///////////////////

template<typename MetadataValue, typename MetadataIterType>
HRESULT MetadataDiff(	IMetaDataImport* piImportOrig, 
						LPCBYTE pbLoadAddressOrig,
						mdToken tokParentOrig,
						IMetaDataImport* piImportNew,
						LPCBYTE pbLoadAddressNew,
						mdToken tokParentNew,
						IMetaDataEmit* piMetaDataEmit, // For emitting the new method refs
						TypeListTokenPairs* plistNewOrReorderedTokens)
{
	HRESULT hr = S_OK;

	if (!piImportOrig || !pbLoadAddressOrig || !piImportNew ||!pbLoadAddressNew || !plistNewOrReorderedTokens) 
		return E_POINTER;

	typedef map<MetadataValue, mdToken> TypeMapStr2Tok;

	// Map of all the names of the metadata elements in the original metadata
	TypeMapStr2Tok mapMetadata2Tok;

	// Step 1:	Create a map of the original metadata elements
	//
	//			This has to be done each time since each time
	//			since we may have emitted new metadata last time.

	MetadataIterType itMetadataOrig(piImportOrig, pbLoadAddressOrig, false, tokParentOrig);

	for(; 
		!itMetadataOrig.end(); 
		++itMetadataOrig)
	{
		const MetadataValue& valOrig = itMetadataOrig.getValue();

		mdToken tokOrig = itMetadataOrig.getToken();

		const WCHAR* szDump = valOrig.toDump();

		//ILXLOGMSG5(ILXLOG_INFO, _T("ORIG %ls: %ls%ls%ls (0x%08X)."),
		//	Token2TypeName(tokOrig),
		//	valOrig.toString(),
		//	szDump ? L" " : L"",
		//	szDump ? szDump : L"",
		//	tokOrig);

		if (mapMetadata2Tok[valOrig] != 0)
		{
			ILXLOGMSG1(ILXLOG_ERROR, _T("Element is not unique (0x%08X). Comparison operator probably not discriminatory enough."),
				mapMetadata2Tok[valOrig]);
			_ASSERTE(!"Element is not unique. Comparison operator probably not discriminatory enough.");
		}
		mapMetadata2Tok[valOrig] = tokOrig;
	}

	// Step 2:	Diff the old and new metadata tables

	for(MetadataIterType itMetadataNew(piImportNew, pbLoadAddressNew, true, tokParentNew); 
		!itMetadataNew.end(); 
		++itMetadataNew)
	{
		const TypeMapStr2Tok::key_type& valNew = itMetadataNew.getValue();
		mdToken tokNew = itMetadataNew.getToken();

		const WCHAR* szDump = valNew.toDump();

		// Step 3:	First see if the NEW metadata element exists somewhere
		//			in the ORIG table

		TypeMapStr2Tok::const_iterator it_find =
			mapMetadata2Tok.find(valNew);

		if (it_find != mapMetadata2Tok.end())
		{
			// CASE 1: String is in both tables

			// Now see if they are at the same position (token)
			mdToken tokOrig = (*it_find).second;

			if (tokOrig != tokNew)
			{
				// CASE 1a: String is at different locations
				// Just need to fixup the location address
				ILXLOGMSG6(ILXLOG_INFO, _T("FIX %ls: %ls%ls%ls (0x%08X -> 0x%08X)."),
					Token2TypeName(tokNew), 
					valNew.toString(), 
					szDump ? L" " : L"", 
					szDump ? szDump : L"", 
					tokNew,
					tokOrig);

				plistNewOrReorderedTokens->push_back(
					TypePairTokens(tokNew, tokOrig));
			}
			else
			{
				//ILXLOGMSG5(ILXLOG_INFO, _T("SAME %ls: %ls%ls%ls (0x%08X)."),
				//	Token2TypeName(tokNew), 
				//	valNew.toString(),
				//	szDump ? L" " : L"", 
				//	szDump ? szDump : L"",
				//	tokNew);
			}
		}
		else
		{
			// CASE 2:	String is new

			mdString tokEmitted = mdTokenNil;

			hr = itMetadataNew.emit(piMetaDataEmit, tokParentOrig, &tokEmitted);
			VERIFY_HR_COND4(hr, tokEmitted, ILXLOG_ERROR, _T("Failed to emit %ls: %ls%ls%ls."), 
				Token2TypeName(tokNew),
				valNew.toString(), 
				szDump ? L" " : L"", 
				szDump ? szDump : L"");

			if (tokEmitted != tokNew)
			{
				// Its very likely that the emitted position will not
				// be the same as the new position
				plistNewOrReorderedTokens->push_back(
					TypePairTokens(tokNew, tokEmitted));

				// TODO: If a TypeRef is new, we need to emit all its member refs

				ILXLOGMSG7(ILXLOG_INFO, _T("EMIT-AND-FIX %ls: %ls%ls%ls (0x%08X->0x%08X) Scope: 0x%08X."), 
					Token2TypeName(tokNew), 
					valNew.toString(), 
					szDump ? L" " : L"", 
					szDump ? szDump : L"",
					tokNew, 
					tokEmitted,
					tokParentOrig);
			}
			else
			{
				ILXLOGMSG5(ILXLOG_INFO, _T("EMIT ONLY %ls: %ls%ls%ls (0x%08X)."), 
					Token2TypeName(tokNew), 
					valNew.toString(), 
					szDump ? L" " : L"", 
					szDump ? szDump : L"", 
					tokNew);
			}
		}
	}

	return S_OK;
}

HRESULT CreateUpdatePatches(	/*[in]*/ ICorProfilerInfo* piProfInfo,
								/*[in]*/ ModuleID moduleIDOrig,
								/*[in]*/ IMetaDataImport* piImportNew,
								/*[in]*/ LPCBYTE pbLoadAddressNewModule,
								/*[in]*/ bool bFirstTime,
								/*[out]*/ TypeMapMethodInfo* pmapChangedFunctions)
{
	HRESULT hr = S_OK;

	USES_CONVERSION;

	if (!piProfInfo || !piImportNew || !pbLoadAddressNewModule || !pmapChangedFunctions) 
		return E_POINTER;

	LPCBYTE pbLoadAddressOrigModule = 0;
	WCHAR szModuleName[MAX_PATH] = { 0 };
	ULONG nLen = MAX_PATH;
	hr = piProfInfo->GetModuleInfo(moduleIDOrig, &pbLoadAddressOrigModule, 
										nLen, &nLen, szModuleName, 0);
	VERIFY_HR_COND(hr, pbLoadAddressOrigModule, _T("GetModuleInfo() failed."),
		ILXLOG_ERROR);

	CComPtr<IMetaDataImport> spiImportOrigModule;
	hr = piProfInfo->GetModuleMetaData(	moduleIDOrig, 
										ofRead | ofWrite,
										IID_IMetaDataImport,
										(IUnknown**) &spiImportOrigModule.p);
	VERIFY_HR_COND(hr, spiImportOrigModule.p, 
		_T("GetModuleMetaData() failed."), ILXLOG_ERROR);

	CComPtr<IMetaDataEmit> spiMetaDataEmit;
	hr = spiImportOrigModule->QueryInterface(	IID_IMetaDataEmit, 
												(void**) &spiMetaDataEmit.p );
	VERIFY_HR_COND(hr, spiMetaDataEmit, _T("QI for IMetaDataEmit failed."), ILXLOG_ERROR);

	CComPtr<IMethodMalloc> spiMethodAlloc;
	hr = piProfInfo->GetILFunctionBodyAllocator(	moduleIDOrig, 
													&spiMethodAlloc.p );
	VERIFY_HR_COND(hr, spiMethodAlloc, _T("GetILFunctionBodyAllocator() failed."),
		ILXLOG_ERROR);

	// Step 1: Emit new strings

	TypeListTokenPairs listNewOrReorderedTokens;

	hr = MetadataDiff<SStrWLen, userString_iterator>(
		spiImportOrigModule, pbLoadAddressOrigModule, mdTokenNil,
		piImportNew, pbLoadAddressNewModule, mdTokenNil,
		spiMetaDataEmit,
		&listNewOrReorderedTokens);
	VERIFY_HR(hr, _T("Failed to compare user strings."), ILXLOG_ERROR);

	// Step 2:	Emit new signatures

	hr = MetadataDiff<SSigProps, signature_iterator>(
		spiImportOrigModule, pbLoadAddressOrigModule, mdTokenNil,
		piImportNew, pbLoadAddressNewModule, mdTokenNil,
		spiMetaDataEmit,
		&listNewOrReorderedTokens);
	VERIFY_HR(hr, _T("Failed to compare signatures."), ILXLOG_ERROR);

	// Step 3:	Compare CLR build version

	// Step 4:	Compare assembly permissions

	// Step 5:	Compare assembly custom attributes

	// Step 6:	Compare global functions

	// Step 7:	Compare global fields

	// Step 8:	Compare global member refs

	// Step 9:	Emit new assembly refs

	hr = MetadataDiff<SAssemblyProps, assemblyRef_iterator>(
		spiImportOrigModule, pbLoadAddressOrigModule, mdTokenNil,
		piImportNew, pbLoadAddressNewModule, mdTokenNil,
		spiMetaDataEmit,
		&listNewOrReorderedTokens);
	VERIFY_HR(hr, _T("Failed to compare assembly refs."), ILXLOG_ERROR);

	// Step 10:	Compare type refs

	assemblyRef_iterator itAssemblyRefNew(piImportNew, pbLoadAddressNewModule, true, mdTokenNil);
	for(; !itAssemblyRefNew.end(); ++itAssemblyRefNew)
	{
		mdAssemblyRef tokAssemblyRefEmit = itAssemblyRefNew.getToken();
	
		for(TypeListTokenPairs::const_iterator it = listNewOrReorderedTokens.begin();
			it != listNewOrReorderedTokens.end();
			++it)
		{
			const mdToken& tokFrom = (*it).first;
			if (tokFrom == tokAssemblyRefEmit)
			{
				tokAssemblyRefEmit = (*it).second;
				break;
			}
		}
	
		_ASSERTE(tokAssemblyRefEmit != mdTokenNil);
	
		hr = MetadataDiff<SStrProp, typeRef_iterator>(
			spiImportOrigModule, pbLoadAddressOrigModule, tokAssemblyRefEmit,
			piImportNew, pbLoadAddressNewModule, itAssemblyRefNew.getToken(),
			spiMetaDataEmit,
			&listNewOrReorderedTokens);
		VERIFY_HR(hr, _T("Failed to compare type refs."), ILXLOG_ERROR);
	}

	// Step 11:	Compare member refs

	typeRef_iterator itTypeRefOrig(spiImportOrigModule, pbLoadAddressOrigModule, false, mdTokenNil);
	typeRef_iterator itTypeRefNew(piImportNew, pbLoadAddressNewModule, true, mdTokenNil);

	for(; !itTypeRefNew.end(); ++itTypeRefNew)
	{
		mdToken tokTypeRefEmit = itTypeRefNew.getToken();

		for(TypeListTokenPairs::const_iterator it = listNewOrReorderedTokens.begin();
			it != listNewOrReorderedTokens.end();
			++it)
		{
			const mdToken& tokFrom = (*it).first;
			if (tokFrom == tokTypeRefEmit)
			{
				tokTypeRefEmit = (*it).second;
				break;
			}
		}

		_ASSERTE(tokTypeRefEmit != mdTokenNil);

		if (!itTypeRefOrig.find(tokTypeRefEmit))
		{
			_ASSERTE(false);
		}

		if(wcscmp(itTypeRefOrig.getName(), itTypeRefNew.getName()) != 0)
		{
			ILXLOGMSG2(ILXLOG_ERROR,
				_T("Type names are different: '%ls' != '%ls'."), 
				itTypeRefOrig.getName(),
				itTypeRefNew.getName());
			continue;
		}

		hr = MetadataDiff<SMemberProps, memberRef_iterator>(
			spiImportOrigModule, pbLoadAddressOrigModule, itTypeRefOrig.getToken(),
			piImportNew, pbLoadAddressNewModule, itTypeRefNew.getToken(),
			spiMetaDataEmit,
			&listNewOrReorderedTokens);

		VERIFY_HR1(hr, ILXLOG_ERROR,
			_T("Failed to compare member refs for type '%ls'."), itTypeRefOrig.getName());
	}

	// Step 12:	Compare type defs
	//
	//			Iterates over all the types and all the methods of those
	//			types.  If a method is different, a patch is created and
	//			added to pmapChangedFunctions.

	hr = MetadataDiff<SStrProp, typeDef_iterator>(
		spiImportOrigModule, pbLoadAddressOrigModule, mdTokenNil,
		piImportNew, pbLoadAddressNewModule, mdTokenNil,
		spiMetaDataEmit,
		&listNewOrReorderedTokens);
	VERIFY_HR(hr, _T("Failed to compare type refs."), ILXLOG_ERROR);

	int nMethodDefCount = 0;

	typeDef_iterator itTypeDefOrig(spiImportOrigModule, pbLoadAddressOrigModule, false, mdTokenNil);
	typeDef_iterator itTypeDefNew(piImportNew, pbLoadAddressNewModule, true, mdTokenNil);

	for(; !itTypeDefNew.end(); ++itTypeDefNew)
	{
		mdToken tokTypeDefEmit = itTypeDefNew.getToken();

		for(TypeListTokenPairs::const_iterator it = listNewOrReorderedTokens.begin();
			it != listNewOrReorderedTokens.end();
			++it)
		{
			const mdToken& tokFrom = (*it).first;
			if (tokFrom == tokTypeDefEmit)
			{
				tokTypeDefEmit = (*it).second;
				break;
			}
		}

		_ASSERTE(tokTypeDefEmit != mdTokenNil);

		if (!itTypeDefOrig.find(tokTypeDefEmit))
		{
			_ASSERTE(false);
		}

		if (itTypeDefOrig.isInterface() || itTypeDefOrig.isDelegate())
		{
			continue;
		}

		if (wcscmp(itTypeDefOrig.getName(), itTypeDefNew.getName()) != 0)
		{
			ILXLOGMSG2(ILXLOG_WARNING, _T("Type names are not the same: '%ls' != '%ls'."),
				itTypeDefOrig.getName(), itTypeDefNew.getName());
			continue;
		}

		// Count the number of changeable methods - REMOVE ME!
		for(methodDef_iterator itMethodDefs(itTypeDefNew);
			!itMethodDefs.end();
			++itMethodDefs)
		{
			const CMethodInfo& methodInfo = itMethodDefs.getValue();

			// A PInvoke or Abstract method has no managed bytecode method body
			// and therefore can't be modified by EnC
			if (methodInfo.isPInvoke() || 
				methodInfo.isAbstract() ||
				methodInfo.isCtor() ||
				methodInfo.isCCtor())
				continue;
		
			++nMethodDefCount;;
		}

		// Step 13:	Compare method defs

		// This will emit new methods and fields.  However, apparently 
		// you can only call DefineMethod and DefineField whilst inside
		// ModuleLoadFinished because the methods and fields are baked
		// in after that.  So introducing methods doesn't work... :(

		hr = MetadataDiff<CMethodInfo, methodDef_iterator>(
			spiImportOrigModule, pbLoadAddressOrigModule, itTypeDefOrig.getToken(),
			piImportNew, pbLoadAddressNewModule, itTypeDefNew.getToken(),
			spiMetaDataEmit,
			&listNewOrReorderedTokens);

		VERIFY_HR1(hr, ILXLOG_ERROR,
			_T("Failed to compare method defs for type '%ls'."), itTypeDefOrig.getName());

		hr = MetadataDiff<SFieldProps, fieldDef_iterator>(
			spiImportOrigModule, pbLoadAddressOrigModule, itTypeDefOrig.getToken(),
			piImportNew, pbLoadAddressNewModule, itTypeDefNew.getToken(),
			spiMetaDataEmit,
			&listNewOrReorderedTokens);

		VERIFY_HR1(hr, ILXLOG_ERROR,
			_T("Failed to compare field defs for type '%ls'."), itTypeDefOrig.getName());

		hr = CreateTypePatches(	itTypeDefOrig, 
								itTypeDefNew, 
								itTypeDefOrig.getName(),
								moduleIDOrig,
								piProfInfo,
								spiMetaDataEmit,
								spiMethodAlloc, 
								listNewOrReorderedTokens,
								bFirstTime,
								pmapChangedFunctions);
		if (FAILED(hr)) return hr;
	}

	// Counts the number of changed methods - REMOVE ME!
	
	int nNumChangedMethods = 0;
	for(TypeMapMethodInfo::iterator itChangedMethods = pmapChangedFunctions->begin();
		itChangedMethods != pmapChangedFunctions->end();
		++itChangedMethods)
	{
		CMethodInfo& patch = (*itChangedMethods).second;

		if (patch.isPInvoke() || 
			patch.isAbstract() ||
			patch.isCtor() ||
			patch.isCCtor())
			continue;

		if (!patch.m_bWasModified)
		{
			ILXLOGMSG2(ILXLOG_INFO, _T("Not modified: %ls.%ls."), 
				patch.m_szTypeName, patch.m_szMemberName);
			continue;
		}
		
		++nNumChangedMethods;
	}
	
	ILXLOGMSG1(ILXLOG_INFO, _T("Total number of changeable methods: '%d'."), nMethodDefCount);
	ILXLOGMSG2(ILXLOG_INFO, _T("Total number of changed methods: '%d' (%.1f%%)."), 
		nNumChangedMethods, (nNumChangedMethods / (float) nMethodDefCount) * 100.0f);

	// Step 14:	Compare type specs

	// Step 15:	Compare module refs

	// Step 16: Compare extra assembly info (assembly refs, files, exported
	//			types, manifest resources, ?)

	// Step 17: Patch token

	try
	{
		hr = PatchTokens(*pmapChangedFunctions, listNewOrReorderedTokens);
		if (FAILED(hr)) return hr;
	}
	catch(...)
	{
		ILXLOGMSG0(ILXLOG_ERROR, _T("PatchTokens() threw an exception."));
		return E_FAIL;
	}

	return S_OK;
}

HRESULT CreateTypePatches(	typeDef_iterator& itTypeDefOrig, 
							typeDef_iterator& itTypeDefNew,
							const WCHAR* szTypeName,
							ModuleID moduleIDOrig,
							ICorProfilerInfo* piProfInfo,
							IMetaDataEmit* piEmit,
							IMethodMalloc* piAlloc,
							const TypeListTokenPairs& listNewOrReorderedTokens,
							/*[in]*/ bool bFirstTime,
							TypeMapMethodInfo* pmapChangedFunctions)
{
	HRESULT hr = S_OK;

	if (!piAlloc || !pmapChangedFunctions || !piProfInfo) return E_POINTER;

	// Step 1:	Compare type def props

	// Step 2:	Compare class layout

	// Step 3:	Compare fields

	// Step 4:	Compare methods

	methodDef_iterator itMethodDefOrig(itTypeDefOrig);

	if (bFirstTime)
	{
		for(; !itMethodDefOrig.end(); ++itMethodDefOrig)
		{
			const CMethodInfo& methodInfoOrig = itMethodDefOrig.getValue();

			// Sanity check: MethodDef tokens are unique
			_ASSERTE(pmapChangedFunctions->find(methodInfoOrig.m_methodDef) ==
				pmapChangedFunctions->end());

			pmapChangedFunctions->insert(make_pair(methodInfoOrig.m_methodDef, methodInfoOrig));
		}

		itMethodDefOrig.reset();
	}

	methodDef_iterator itMethodDefNew(itTypeDefNew);

	for(; !itMethodDefOrig.end() && !itMethodDefNew.end(); ++itMethodDefOrig, ++itMethodDefNew)
	{
		hr = CreateMethodPatches(	itMethodDefOrig, 
									itMethodDefNew, 
									szTypeName,
									moduleIDOrig,
									piProfInfo,
									piEmit,
									piAlloc, 
									listNewOrReorderedTokens,
									pmapChangedFunctions);
		if (FAILED(hr)) 
			continue;
	}

	// Step 5:	Compare properties

	// Step 6:	Compare events

	// Step 7:	Compare method implementations

	// Step 8:	Compare permissions

	// Step 9:	Compare interface implementations

	// Step 10:	Compare custom attributes

	return S_OK;
}

HRESULT CreateMethodPatches(	methodDef_iterator& itMethodDefOrig,
								methodDef_iterator& itMethodDefNew,
								const WCHAR* szTypeName,
								ModuleID moduleIDOrig,
								ICorProfilerInfo* piProfInfo,
								IMetaDataEmit* piEmit,
								IMethodMalloc* piAlloc,
								const TypeListTokenPairs& listNewOrReorderedTokens,
								TypeMapMethodInfo* pmapChangedFunctions)
{
	HRESULT hr = S_OK;

	if (!piAlloc || !pmapChangedFunctions || !piProfInfo) return E_POINTER;

	// Step 1:	Get method info for original method
	//
	//			If the method was patched previously, we should already have
	//			the method info in our list of changed functions.

	CMethodInfo* pMethodInfoOrig = 0;
	bool bAlreadyInList = false;

	// See if we've already patched this method
	TypeMapMethodInfo::iterator it_find = pmapChangedFunctions->find(itMethodDefOrig.getToken());
	if (it_find != pmapChangedFunctions->end())
	{
		pMethodInfoOrig = &(*it_find).second;
		bAlreadyInList = true;
	}

	if (!pMethodInfoOrig)
	{
		// We haven't patched this method so lookup the metadata now
		pMethodInfoOrig = &itMethodDefOrig.getValue();
	}

	_ASSERTE(pMethodInfoOrig);

	// A PInvoke or Abstract method has no managed bytecode method body
	// and therefore can't be modified by EnC
	if (pMethodInfoOrig->isPInvoke() || pMethodInfoOrig->isAbstract())
		return S_OK;

	// Step 2:	Get method info for the new method

	const CMethodInfo& methodInfoNew = itMethodDefNew.getValue();

	// Step 3:	Compare the original and new methods to see if the new method has
	//			been modified
	hr = MethodsAreEquivalent(*pMethodInfoOrig, methodInfoNew, listNewOrReorderedTokens);
	if (FAILED(hr)) return hr;
	if (hr == S_OK) return hr;

	// Methods are different!
	_ASSERTE(hr == S_FALSE);

	//ILXLOGMSG2(ILXLOG_INFO, _T("'%ls.%ls' was modified."), szTypeName, methodInfoNew.m_szMemberName);
	
	// Step 6:	Get the FunctionId for the original method

	FunctionID functionID = 0;
	hr = piProfInfo->GetFunctionFromToken(	moduleIDOrig, 
											pMethodInfoOrig->m_methodDef,
											&functionID);
	VERIFY_HR4(hr, ILXLOG_ERROR,
		_T("'%ls.%ls' - GetFunctionFromToken(ModuleID:0x%08X,Def:0x%08X) failed."), 
		szTypeName,
		pMethodInfoOrig->m_szMemberName,
		moduleIDOrig, 
		pMethodInfoOrig->m_methodDef);

	// Step 7:	Create a patch to package the new method

	hr = pMethodInfoOrig->update(methodInfoNew, moduleIDOrig, functionID, piAlloc,
		piEmit, listNewOrReorderedTokens);
	if (FAILED(hr)) return hr;

	// Step 10:	Compare params
/*
	paramDef_iterator itParamsLhs(itMethodDefOrig);
	paramDef_iterator itParamsRhs(itMethodDefNew);

	for(; !itParamsLhs.end() && !itParamsRhs.end(); ++itParamsLhs, ++itParamsRhs)
	{
		CParamInfo paramInfoLhs(itParamsLhs);
		CParamInfo paramInfoRhs(itParamsRhs);

		if (paramInfoLhs != paramInfoRhs)
			return S_FALSE;
	}

	if (!itParamsLhs.end() || !itParamsRhs.end())
		return S_FALSE; // Different number of methods
*/
	// Step 11:	Compare custom attributes

	// Step 12:	Compare permissions

	// Step 13:	Compare member refs

	// Step 14:	Compare pinvoke info

	// Step 15:	Patch exception handlers (is this needed?)

	// Step 16:	Add to list of functions to patch
	if (!bAlreadyInList)
		pmapChangedFunctions->insert(make_pair(pMethodInfoOrig->m_methodDef, *pMethodInfoOrig));

	return S_OK;
}

HRESULT PatchTokens(	const TypeMapMethodInfo& listChangedMethods,
						const TypeListTokenPairs& listNewOrReorderedTokens)
{
	HRESULT hr = S_OK;

	// Fixup changed methods to use the correct token

	// Iterate over all the changed functions

	for(TypeMapMethodInfo::const_iterator itChangedMethods = listChangedMethods.begin();
		itChangedMethods != listChangedMethods.end();
		++itChangedMethods)
	{
		const CMethodInfo& patch = (*itChangedMethods).second;
		if (!patch.m_bWasModified)
			continue;
		
		// Step 4a:	Enumerate the opcodes for this method

		opcode_iterator itOpcode(	patch.m_pbMethodHeader + patch.m_ulMethodHeaderSize,
									patch.m_pbMethodHeader + patch.m_ulMethodHeaderSize + patch.m_ulCodeSize,
									CEE_COUNT,
									(OPCODE_FORMAT) (InlineString | InlineMethod));
		
		for(; !itOpcode.end(); ++itOpcode)
		{
			DWORD* pdwTok = (DWORD*) itOpcode.getOpCodeArg();
		
			// Step 4b:	Fix the token to point to the newly emitted one

			for(TypeListTokenPairs::const_iterator it = listNewOrReorderedTokens.begin();
				it != listNewOrReorderedTokens.end();
				++it)
			{
				const mdToken& tokFrom = (*it).first;

				if (tokFrom == *pdwTok)
				{
					const mdToken& tokTo = (*it).second;
					*pdwTok = tokTo;
					break;
				}
			} // for each reference to rewire

		} // for each instruction

	} // for each changed function

	return S_OK;
}
