///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/BugReportFile.cpp 9     5/28/05 1:20p Eaddy $
//
//    File name:	BugReportFile.cpp
//   Created By:	Marc Eaddy
//   Created on:	2/24/05
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/BugReportFile.cpp $
 * 
 * 9     5/28/05 1:20p Eaddy
 * Changed copyright
 */

#include "StdAfx.h"

#include "BugReportFile.h"		// Me

#include "Helpers.h"			// AppendXXX()

using namespace std;

static const char* szCOMPILER_OPTIONS_LINE		= "### Compiler command line";
static const size_t nCOMPILER_OPTIONS_LINE_LEN	= strlen(szCOMPILER_OPTIONS_LINE);

static const char* szSOURCE_FILE_LINE			= "### Source file: '";
static const size_t nSOURCE_FILE_LINE_LEN		= strlen(szSOURCE_FILE_LINE);

enum ECompilerOption
{
	coOUT,
	coTARGET,
	coDEFINE,
	coRECURSE,
	coRESOURCE,
	coREFERENCE,
	coDEBUG,
	coOPTIMIZE,
	coWARN_AS_ERROR,
	coCHECKED,
	coUNSAFE,
	coMAIN,
};

struct SCompilerOption
{
	const char* szName;
	size_t nNameLen;
	const char* szAltName;
	size_t nAltNameLen;
	bool bHasArg;
	bool bHasPlusMinus;
	ECompilerOption eOption;
} sCompilerOptions[] =
{
	// KLUDGE: Not all of these flags are actually checked!
	// NOTE: These have to be ordered since the alt form of "optimize" is "o"
	// which could match "out".
//	{ "out",			-1, NULL,	-1, true,	false,	coOUT			},
//	{ "target",			-1, "t",	-1, true,	false,	coTARGET		},
//	{ "define",			-1, "d",	-1, true,	false,	coDEFINE		},
//	{ "recurse",		-1, NULL,	-1, true,	fales,	coRECURSE		},
//	{ "resource",		-1, "res",	-1, true,	false,	coRESOURCE		},
	{ "reference",		-1, "r",	-1, true,	false,	coREFERENCE		},
	{ "debug",			-1, NULL,	-1, true,	true,	coDEBUG			},
	{ "optimize",		-1, "o",	-1, false,	true,	coOPTIMIZE		},
	{ "warnaserror",	-1, NULL,	-1, false,	true,	coWARN_AS_ERROR	},
	{ "checked",		-1, NULL,	-1, false,	true,	coCHECKED		},
	{ "unsafe",			-1, NULL,	-1, false,	true,	coUNSAFE		},
	{ "main",			-1, "m",	-1, true,	false,	coMAIN			},
};

static const int nNUM_COMPILER_OPTIONS = sizeof(sCompilerOptions) / sizeof(*sCompilerOptions);

///////////////////////////////////
// BUG REPORT FILE IMPLEMENTATION
///////////////////////////////////

CBugReportFile::CBugReportFile()
	: CCompileInfo()
{ 
	for(int i = 0; i < nNUM_COMPILER_OPTIONS; ++i)
	{
		if (sCompilerOptions[i].nNameLen == -1 &&
			sCompilerOptions[i].szName)
		{
			sCompilerOptions[i].nNameLen = strlen(sCompilerOptions[i].szName);
		}

		if (sCompilerOptions[i].nAltNameLen == -1 &&
			sCompilerOptions[i].szAltName)
		{
			sCompilerOptions[i].nAltNameLen = strlen(sCompilerOptions[i].szAltName);
		}
	}
}

CBugReportFile::~CBugReportFile()
{ }

// ICompileInfo implementation

const WCHAR* CBugReportFile::getFileExtension() const
{
	return L".bugreport";
}

HRESULT CBugReportFile::read(FILE* in, bool bIsUnicode, const WCHAR* szBugReportFilePath)
{
	USES_CONVERSION;

	HRESULT hr = S_OK;

	_ASSERTE(in);
	if (!in) return E_POINTER;

	static const nMAX_LINE_LEN = 10000;
    char buffer[nMAX_LINE_LEN];
	wchar_t wbuffer[nMAX_LINE_LEN];

	buffer[0] = '\0';
	wbuffer[0] = L'\0';

	bool bGotCompilerOptions = false;
	bool bGotSourceFiles = false;

	hr = ReadUnicodeOrASCIILineFromFile(in, bIsUnicode, buffer, nMAX_LINE_LEN);

	while(hr == S_OK)
	{
		if (strncmp(buffer, szCOMPILER_OPTIONS_LINE, nCOMPILER_OPTIONS_LINE_LEN) == 0)
		{
			_ASSERTE(!bGotCompilerOptions);
			readCompilerOptions(in, bIsUnicode);
			bGotCompilerOptions = true;
		}
		else if (strncmp(buffer, szSOURCE_FILE_LINE, nSOURCE_FILE_LINE_LEN) == 0)
		{
			char* pFilePath = buffer + nSOURCE_FILE_LINE_LEN;

			// Remove trailing '
			char* pTrailingQuote = strchr(pFilePath, '\'');
			_ASSERTE(pTrailingQuote);
			if (pTrailingQuote)
				*pTrailingQuote = '\0';

			m_vecSourceFiles.push_back(A2CW(pFilePath));

			bGotSourceFiles = true;
		}

		hr = ReadUnicodeOrASCIILineFromFile(in, bIsUnicode, buffer, nMAX_LINE_LEN);
	}

	if (!bGotCompilerOptions && !bGotSourceFiles)
	{
		ILXLOGMSG1(ILXLOG_ERROR, _T("Bugreport file '%ls' is invalid. No source files or compiler command-line found."), 
			szBugReportFilePath);
		return E_FAIL;
	}
	else if (!bGotCompilerOptions)
	{
		ILXLOGMSG1(ILXLOG_ERROR, _T("Bugreport file '%ls' is invalid. No compiler command-line options found."), 
			szBugReportFilePath);
		return E_FAIL;
	}
	else if (!bGotSourceFiles)
	{
		ILXLOGMSG1(ILXLOG_ERROR, _T("Bugreport file '%ls' is invalid. No source files found."), 
			szBugReportFilePath);
		return E_FAIL;
	}

	return S_OK;
}

HRESULT CBugReportFile::readCompilerOptions(FILE* in, bool bIsUnicode)
{
	USES_CONVERSION;

	HRESULT hr = S_OK;

	_ASSERTE(in);
	if (!in) return E_POINTER;

	static const nMAX_LINE_LEN = 10000;
    char buffer[nMAX_LINE_LEN];

	hr = ReadUnicodeOrASCIILineFromFile(in, bIsUnicode, buffer, nMAX_LINE_LEN);
	VERIFY_HR0(hr, ILXLOG_ERROR, _T("Expected compiler options line but got EOF."));

	char* pBufferStart = strchr(buffer, ' ');
	if (!pBufferStart)
	{
		ILXLOGMSG1(ILXLOG_ERROR, _T("Failed to parse compiler command-line args. Here's what we got: '%s'."),
			buffer);
		return E_UNEXPECTED; // Not the right format
	}

	size_t nLen = strlen(pBufferStart);
	if (nLen > 1 && pBufferStart[nLen-1] == '\n')
		pBufferStart[nLen-1] = '\0';

	// Set the compiler name
	m_strCompilerName.assign(A2CW(buffer), pBufferStart - buffer);
	m_strCompilerOptions.assign(A2CW(pBufferStart + 1));

	// Skip past compiler name
	char* pToken = strtok(pBufferStart, " /");
	while (pToken)
	{
		hr = parseCompilerOption(pToken);
		if (FAILED(hr))
		{
			ILXLOGMSG1(ILXLOG_ERROR, _T("Bug report file contained unrecognized compiler option '%s'."),
				pToken);
		}
		pToken = strtok(NULL, " /");
	}

	return S_OK;
}

HRESULT CBugReportFile::parseCompilerOption(const char* szOption)
{
	USES_CONVERSION;

	if (!szOption) return E_POINTER;

	for(int i = 0; i < nNUM_COMPILER_OPTIONS; ++i)
	{
		if (sCompilerOptions[i].szName &&
			sCompilerOptions[i].nNameLen > 0 &&
			strncmp(szOption, sCompilerOptions[i].szName, sCompilerOptions[i].nNameLen) == 0)
		{
			break;
		}
		else if (sCompilerOptions[i].szAltName &&
			sCompilerOptions[i].nAltNameLen > 0 &&
			strncmp(szOption, sCompilerOptions[i].szAltName, sCompilerOptions[i].nAltNameLen) == 0 &&
			(szOption[sCompilerOptions[i].nAltNameLen] == '\0' ||
			 szOption[sCompilerOptions[i].nAltNameLen] == ':' ||
			 szOption[sCompilerOptions[i].nAltNameLen] == '+' ||
			 szOption[sCompilerOptions[i].nAltNameLen] == '-') )
		{
			break;
		}
	}

	if (i == nNUM_COMPILER_OPTIONS)
	{
		// Unrecognized compiler option -- ignore
		return S_FALSE;
	}

	bool bEnabled = true;
	const char* pOptionArg = NULL;

	if (sCompilerOptions[i].bHasPlusMinus)
	{
		const char* pPlusMinus = strpbrk(szOption, "+-");
		if (pPlusMinus)
		{
			if (*pPlusMinus == '-')
				bEnabled = false;
			else if (*pPlusMinus == '+')
				bEnabled = true;
			else
			{
				_ASSERTE(!"Can't reach this");
			}
		}
		else
		{
			bEnabled = true; // Enabled by default if just the option name is specified
		}
	}

	if (sCompilerOptions[i].bHasArg)
	{
		pOptionArg = strchr(szOption, ':');
		if (pOptionArg)
		{
			++pOptionArg;
		}
	}

	switch(sCompilerOptions[i].eOption)
	{
		case coDEBUG: 
			m_bDebugFlag = bEnabled; 
			break;
		//case coOPTIMIZE: 
		//	m_bOptimizationFlag = bEnabled; 
		//	break;
		//case coWARN_AS_ERROR: 
		//	m_bWarnAsErrorsFlag = bEnabled; 
		//	break;
		//case coCHECKED: 
		//	m_bCheckedFlag = bEnabled; 
		//	break;
		//case coUNSAFE: 
		//	m_bUnsafeFlag = bEnabled; 
		//	break;
		//case coMAIN:
		//	VERIFY_COND1(pOptionArg && *pOptionArg, ILXLOG_ERROR, 
		//		_T("Bug report file contained invalid compiler option '%s'. Expected '/main:<type>'."),
		//		szOption);
		//	m_strMainType = A2CW(pOptionArg);
		//	break;
		//case coREFERENCE:
 		//	VERIFY_COND1(pOptionArg && *pOptionArg, ILXLOG_ERROR, 
		//		_T("Bug report file contained invalid compiler option '%s'. Expected '/ref:<file list>'."),
		//		szOption);
		//	m_vecReferenceFiles.push_back(A2CW(pOptionArg));
		//	break;
		// KLUDGE: We don't handle '/out:<file>'
		// KLUDGE: We don't handle '/target:[exe|winexe|library|module]'
		// KLUDGE: We don't handle '/define:<symbol list>'
		// KLUDGE: We don't handle '/doc:<file>'
		// IGNORE: '/recurse:<wildcard>'
		// KLUDGE: We don't handle '/addmodule:<file list>'
		// KLUDGE: We don't handle '/win32res:<file>'
		// KLUDGE: We don't handle '/win32icon:<file>'
		// KLUDGE: We don't handle '/resource:<resinfo>'
		// KLUDGE: We don't handle '/linkresource:<resinfo>'
		// KLUDGE: We don't handle '/debug:{full|pdbonly}'
		// KLUDGE: We don't handle '/warn:<n>'
		// KLUDGE: We don't handle '/nowarn:<warning list>'
		// KLUDGE: We don't handle '@<file>'
		// IGNORE: '/help'
		// IGNORE: '/nologo'
		// KLUDGE: We don't handle '/noconfig'
		// KLUDGE: We don't handle '/baseaddress:<address>'
		// IGNORE: We don't handle '/bugreport:<file>'
		// KLUDGE: We don't handle '/codepage:<n>'
		// KLUDGE: We don't handle '/utf8output'
		// KLUDGE: We don't handle '/fullpaths'
		// KLUDGE: We don't handle '/filealign:<n>'
		// KLUDGE: We don't handle '/nostdlib[+|-]'
		// KLUDGE: We don't handle '/lib:<file list>'
		default:
			return S_FALSE; // Unknown/ignored compiler option
	}

	return S_OK;
}
