///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/COMStuff.cpp 2     5/28/05 1:20p Eaddy $
//
//    File name:	COMStuff.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	Adapted from Matt Pietrek's DNProfiler
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/COMStuff.cpp $
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2004/10/24 19:33:10  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.3  2004/10/10 06:07:44  me133
 * o Added header comment
 *
 */

#include "COMStuff.h"					// Me

#include "PatcherProfilerCallback.h"

// MACROS

#define ARRAY_SIZE( s ) (sizeof( s ) / sizeof( s[0] ))

// CONSTANTS

#define MAX_LENGTH 256

#define PROFILER_GUID "{876F4C08-9270-4146-A37A-24F0DC1C9B50}"

extern const GUID __declspec( selectany ) CLSID_PROFILER = 
        { 0x876F4C08, 0x9270, 0x4146,
        { 0xA3, 0x7A, 0x24, 0xF0, 0xDC, 0x1C, 0x9B, 0x50 } };

static const char* g_szProgIDPrefix = "Patcher";

static const char* g_szDesc = "Patcher";

// GLOBAL VARIABLES

static CClassFactory	g_PatcherClassFactory;

// HELPER FUNCTIONS

HRESULT RegisterClassBase( REFCLSID rclsid,
                            const char *szDesc,                 
                            const char *szProgID,               
                            const char *szIndepProgID,          
                            char *szOutCLSID );

HRESULT UnregisterClassBase( REFCLSID rclsid,
                            const char *szProgID,
                            const char *szIndepProgID,
                            char *szOutCLSID );

BOOL SetRegValue( char *szKeyName, char *szKeyword, char *szValue );

BOOL DeleteKey( const char *szKey, const char *szSubkey );

BOOL SetKeyAndValue(const char *szKey,
                    const char *szSubkey,
                    const char *szValue );

/////////////////////////////////
// CLASS FACTORY IMPLEMENTATION
/////////////////////////////////

CClassFactory::CClassFactory()
	: m_refCount(1)
{ }

CClassFactory::~CClassFactory()
{ }

COM_METHOD( ULONG ) CClassFactory::AddRef()
{
	return InterlockedIncrement( &m_refCount );
}

COM_METHOD( ULONG ) CClassFactory::Release()
{
	return InterlockedDecrement( &m_refCount );
}

HRESULT CClassFactory::QueryInterface( REFIID riid, void **ppInterface )
{   
	if (!ppInterface) return E_POINTER;

    if ( riid == IID_IUnknown )
        *ppInterface = static_cast<IUnknown *>( this ); 
    else if ( riid == IID_IClassFactory )
        *ppInterface = static_cast<IClassFactory *>( this );
    else
    {
        *ppInterface = NULL;                                  
        return E_NOINTERFACE;
    }
    
    reinterpret_cast<IUnknown *>( *ppInterface )->AddRef();
    
    return S_OK;
}

COM_METHOD( HRESULT ) CClassFactory::LockServer( BOOL fLock ) 
{ 
	return S_OK; 
}

HRESULT CClassFactory::CreateInstance( IUnknown *pUnkOuter, REFIID riid,
                                        void **ppInstance )
{   
	if (!ppInstance) return E_POINTER;

    // aggregation is not supported by these objects
    if ( pUnkOuter )
        return CLASS_E_NOAGGREGATION;

    *ppInstance = (void*) new CPatcherProfilerCallback;

    return S_OK;
}

//=============================================================
// DllMain
//=============================================================
BOOL WINAPI DllMain( HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved )
{    
    // save off the instance handle for later use
    switch ( dwReason )
    {
        case DLL_PROCESS_ATTACH:
            DisableThreadLibraryCalls( hInstance );
            CPatcherProfilerCallback::s_hInst = hInstance;
            break;        
    } 
        
    return TRUE;
}

//=============================================================
// DllRegisterServer
//=============================================================
STDAPI DllRegisterServer()
{    
    HRESULT hr = S_OK;
    char  szModule[_MAX_PATH];  

    DllUnregisterServer();
    GetModuleFileNameA( CPatcherProfilerCallback::s_hInst, szModule, ARRAY_SIZE(szModule) );

    char rcCLSID[MAX_LENGTH];           // CLSID\\szID.
    char rcProgID[MAX_LENGTH];          // szProgIDPrefix.szClassProgID
    char rcIndProgID[MAX_LENGTH];       // rcProgID.iVersion
    char rcInproc[MAX_LENGTH + 2];      // CLSID\\InprocServer32

    // format the prog ID values.
    sprintf( rcIndProgID, "%s.%s", g_szProgIDPrefix, g_szProgIDPrefix ) ;
    sprintf( rcProgID, "%s.%d", rcIndProgID, 1 );

    // do the initial portion.
    hr = RegisterClassBase( CLSID_PROFILER, 
                            g_szDesc, rcProgID, rcIndProgID, rcCLSID );

    if ( SUCCEEDED( hr ) )
    {
        // set the server path.
        SetKeyAndValue( rcCLSID, "InprocServer32", szModule );

        // add the threading model information.
        sprintf( rcInproc, "%s\\%s", rcCLSID, "InprocServer32" );
        SetRegValue( rcInproc, "ThreadingModel", "Both" );
    }   
    else
        DllUnregisterServer();
    
    return hr;    
}

//=============================================================
// DllUnregisterServer
//=============================================================
STDAPI DllUnregisterServer()
{    
    char szID[64];         // the class ID to unregister.
    char szCLSID[64];      // CLSID\\szID.
    OLECHAR szWID[64];     // helper for the class ID to unregister.
    char rcProgID[128];    // szProgIDPrefix.szClassProgID
    char rcIndProgID[128]; // rcProgID.iVersion

    // format the prog ID values.
    sprintf( rcProgID, "%s.%s", g_szProgIDPrefix, PROFILER_GUID );
    sprintf( rcIndProgID, "%s.%d", rcProgID, 1 );

    UnregisterClassBase( CLSID_PROFILER, rcProgID, rcIndProgID, szCLSID );
    DeleteKey( szCLSID, "InprocServer32" );

    StringFromGUID2(CLSID_PROFILER, szWID, ARRAY_SIZE( szWID ) );
    WideCharToMultiByte(CP_ACP, 0, szWID, -1, szID, sizeof(szID), NULL,NULL);

    DeleteKey( "CLSID", szCLSID );

    return S_OK;   
}

//=============================================================
// DllGetClassObject
//=============================================================
STDAPI DllGetClassObject( REFCLSID rclsid, REFIID riid, LPVOID FAR *ppv )                  
{
    HRESULT hr = E_OUTOFMEMORY;

    if ( rclsid == CLSID_PROFILER )
        hr = g_PatcherClassFactory.QueryInterface( riid, ppv );

    return hr;   
}

// HELPER FUNCTIONS

//=============================================================
// RegisterClassBase
//=============================================================
HRESULT RegisterClassBase( REFCLSID rclsid,
                            const char *szDesc,                 
                            const char *szProgID,               
                            const char *szIndepProgID,          
                            char *szOutCLSID )              
{
    char szID[64];     // the class ID to register.
    OLECHAR szWID[64]; // helper for the class ID to register.

    StringFromGUID2( rclsid, szWID, ARRAY_SIZE( szWID ) );
    WideCharToMultiByte(CP_ACP, 0, szWID, -1, szID, sizeof(szID), NULL, NULL);

    strcpy( szOutCLSID, "CLSID\\" );
    strcat( szOutCLSID, szID );

    // create ProgID keys.
    SetKeyAndValue( szProgID, NULL, szDesc );
    SetKeyAndValue( szProgID, "CLSID", szID );

    // create VersionIndependentProgID keys.
    SetKeyAndValue( szIndepProgID, NULL, szDesc );
    SetKeyAndValue( szIndepProgID, "CurVer", szProgID );
    SetKeyAndValue( szIndepProgID, "CLSID", szID );

    // create entries under CLSID.
    SetKeyAndValue( szOutCLSID, NULL, szDesc );
    SetKeyAndValue( szOutCLSID, "ProgID", szProgID );
    SetKeyAndValue( szOutCLSID, "VersionIndependentProgID", szIndepProgID );
    SetKeyAndValue( szOutCLSID, "NotInsertable", NULL );
    
    return S_OK;
}

//=============================================================
// UnregisterClassBase
//=============================================================
HRESULT UnregisterClassBase( REFCLSID rclsid,
                            const char *szProgID,
                            const char *szIndepProgID,
                            char *szOutCLSID )
{
    char szID[64];     // the class ID to register.
    OLECHAR szWID[64]; // helper for the class ID to register.

    StringFromGUID2( rclsid, szWID, ARRAY_SIZE( szWID ) );
    WideCharToMultiByte(CP_ACP, 0, szWID, -1, szID, sizeof(szID), NULL, NULL);

    strcpy( szOutCLSID, "CLSID\\" );
    strcat( szOutCLSID, szID );

    // delete the version independant prog ID settings.
    DeleteKey( szIndepProgID, "CurVer" );
    DeleteKey( szIndepProgID, "CLSID" );
    RegDeleteKeyA( HKEY_CLASSES_ROOT, szIndepProgID );

    // delete the prog ID settings.
    DeleteKey( szProgID, "CLSID" );
    RegDeleteKeyA( HKEY_CLASSES_ROOT, szProgID );

    // delete the class ID settings.
    DeleteKey( szOutCLSID, "ProgID" );
    DeleteKey( szOutCLSID, "VersionIndependentProgID" );
    DeleteKey( szOutCLSID, "NotInsertable" );
    RegDeleteKeyA( HKEY_CLASSES_ROOT, szOutCLSID );
    
    return S_OK;
}

//=============================================================
// DeleteKey
//=============================================================
BOOL DeleteKey( const char *szKey, const char *szSubkey )
{
    char rcKey[MAX_LENGTH]; // buffer for the full key name.

    // init the key with the base key name.
    strcpy( rcKey, szKey );

    // append the subkey name (if there is one).
    if ( szSubkey )
    {
        strcat( rcKey, "\\" );
        strcat( rcKey, szSubkey );
    }

    // delete the registration key.
    RegDeleteKeyA( HKEY_CLASSES_ROOT, rcKey );
    
    return TRUE;
}

//=============================================================
// SetKeyAndValue
//=============================================================
BOOL SetKeyAndValue(	const char *szKey,
                        const char *szSubkey,
                        const char *szValue )
{
    HKEY hKey = NULL;       // handle to the new reg key.
    char rcKey[MAX_LENGTH]; // buffer for the full key name.

    // init the key with the base key name.
    strcpy( rcKey, szKey );

    // append the subkey name (if there is one).
    if ( szSubkey )
    {
        strcat( rcKey, "\\" );
        strcat( rcKey, szSubkey );
    }

    // create the registration key.
    if ( RegCreateKeyExA( HKEY_CLASSES_ROOT, 
                          rcKey, 
                          0, 
                          NULL,
                          REG_OPTION_NON_VOLATILE, 
                          KEY_ALL_ACCESS, 
                          NULL,
                          &hKey, 
                          NULL ) == ERROR_SUCCESS )
    {
        // set the value (if there is one).
        if ( szValue )
        {
            RegSetValueExA( hKey, NULL, 0, REG_SZ, (BYTE *)szValue,
                            (DWORD)(((strlen(szValue) + 1) * sizeof (char))));
        }
        
        RegCloseKey( hKey );

        return TRUE;
    }

    return FALSE;   
}

//=============================================================
// SetRegValue
//=============================================================
BOOL SetRegValue(char *szKeyName, char *szKeyword, char *szValue)
{
    HKEY hKey = NULL; // handle to the new reg key.

    // create the registration key.
    if ( RegCreateKeyExA( HKEY_CLASSES_ROOT, 
                          szKeyName, 0,  NULL,
                          REG_OPTION_NON_VOLATILE, 
                          KEY_ALL_ACCESS, 
                          NULL, &hKey, 
                          NULL) == ERROR_SUCCESS )
    {
        // set the value (if there is one).
        if ( szValue )
        {
            RegSetValueExA( hKey, szKeyword, 0, REG_SZ, 
                            (BYTE *)szValue, 
                            (DWORD)((strlen(szValue) + 1) * sizeof ( char )));
        }

        RegCloseKey( hKey );
        
        return TRUE;
    }
    
    return FALSE;
}
