///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/CompileInfo.cpp 7     5/28/05 1:20p Eaddy $
//
//    File name:	CompileInfo.cpp
//   Created By:	Marc Eaddy
//   Created on:	2/24/05
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/CompileInfo.cpp $
 * 
 * 7     5/28/05 1:20p Eaddy
 * Changed copyright
 */

#include "StdAfx.h"

#include "CompileInfo.h"		// Me

#include "Helpers.h"			// AppendXXX()

#include <io.h>					// _open_osfhandle()
#include <fcntl.h>				// _O_TEXT 

CCompileInfo::CCompileInfo()
	: m_strCompilerName()
	, m_strCompilerOptions()
	, m_bDebugFlag(false)
	, m_vecSourceFiles()
{ }

CCompileInfo::~CCompileInfo()
{ }

// CCompileInfo implementation

HRESULT CCompileInfo::findAndLoad(/*[in]*/ const WCHAR* szAssemblyFilePath)
{
	USES_CONVERSION;

	_ASSERTE(szAssemblyFilePath && *szAssemblyFilePath);
	_ASSERTE(wcsstr(szAssemblyFilePath, L".bugreport") == 0);

	HRESULT hr = S_OK;

	std::wstring strBugReportFilePath = szAssemblyFilePath;
	strBugReportFilePath += getFileExtension();

    FILE* in = NULL;

    // Open the file for reading
    HANDLE hFile = CreateFileW(	strBugReportFilePath.c_str(), 
								FILE_READ_DATA, 
								FILE_SHARE_READ, 
								NULL, 
								OPEN_EXISTING, 
								0, 
								NULL);
	if (hFile != INVALID_HANDLE_VALUE && hFile != 0)
	{
		in = _fdopen( _open_osfhandle((INT_PTR)hFile, _O_RDONLY | _O_TEXT), "rt" );
	}

	if (!in)
	{
		long nLastError = ::GetLastError();
		hr = HRESULT_FROM_WIN32(nLastError);
		if (nLastError == ERROR_FILE_NOT_FOUND)
		{
			// Could not open file -- probably isn't one. We're done.
			return hr;
		}
		else
		{
			return ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to open file '%ls'."), 
				strBugReportFilePath.c_str());
		}
	}

	// Step 1:	Read the first four bytes to see if the file is Unicode
	bool bIsUnicode = false;
	unsigned char bom[4] = { 0 }; // Get the byte-order mark, if there is one
	size_t nRead = fread(bom, 1, 4, in);
	long nBomSkip = 0;
	
	if (nRead == 4)
	{
		if (bom[0] == 0xef && bom[1] == 0xbb && bom[2] == 0xbf)
		{
			nBomSkip = 3;
			bIsUnicode = true;
		}
		else if ((bom[0] == 0xff && bom[1] == 0xfe) || // ucs-2le, ucs-4le, and ucs-16le
				 (bom[0] == 0xfe && bom[1] == 0xff))
		{
			nBomSkip = 2;
			bIsUnicode = true;
		}
		else if (bom[0] == 0 && bom[1] == 0 && bom[2] == 0xfe && bom[3] == 0xff) // ucs-4
		{
			nBomSkip = 4;
			bIsUnicode = true;
		}
	}

	// Seek back to the beginning
	fseek(in, nBomSkip, SEEK_SET);

	if (bIsUnicode)
	{
		// NOTE: Must switch to binary mode to read in the UNICODE file.
		// Otherwise, you'll get a null wchar_t after each normal wchar_t.
		_setmode(_fileno(in), _O_BINARY);
	}

	hr = read(in, bIsUnicode, strBugReportFilePath.c_str());

	fclose(in);

	return hr;
}

const WCHAR* CCompileInfo::getCompilerName() const
{
	return m_strCompilerName.c_str();
}

const WCHAR* CCompileInfo::getCompilerOptions() const
{
	return m_strCompilerOptions.c_str();
}

bool CCompileInfo::getDebugFlag() const
{
	return m_bDebugFlag;
}

size_t CCompileInfo::getSourceFileCount() const
{
	return m_vecSourceFiles.size();
}

const WCHAR* CCompileInfo::getSourceFileName(size_t nIndex) const
{
	if (nIndex < 0 || nIndex >= m_vecSourceFiles.size())
	{
		_ASSERTE(!"Out of range");
		return 0;
	}

	return m_vecSourceFiles[nIndex].c_str();
}

void CCompileInfo::debugPrint(int nIndentLevel) const
{
	/*
	USES_CONVERSION;

	char szMsg[1000];
	char szOutputFileInfo[1000];

	// Print header
	m_incfilehdr.debugPrint(nIndentLevel+1);

	// Print defines
	szMsg[0] = '\0';
	AppendIndent(szMsg, nIndentLevel+1);
	AppendString(szMsg, "PreprocessorDefines", W2CA(m_szDefines));
	strcat(szMsg, "\n");
	OutputDebugString(szMsg);

	// Print metadata files

	szOutputFileInfo[0] = '\0';
	m_outputFileMetadata.appendString(szOutputFileInfo);

	szMsg[0] = '\0';
	AppendIndent(szMsg, nIndentLevel+1);
	AppendString(szMsg, "Metadata Info", szOutputFileInfo);
	strcat(szMsg, "\n");
	OutputDebugString(szMsg);

	// Print source files

	szOutputFileInfo[0] = '\0';
	m_outputFileSource.appendString(szOutputFileInfo);

	szMsg[0] = '\0';
	AppendIndent(szMsg, nIndentLevel+1);
	AppendString(szMsg, "Source File Info", szOutputFileInfo);
	strcat(szMsg, "\n");
	OutputDebugString(szMsg);

	// Print input files

	for(TypeVecInputFiles::const_iterator it = m_vecInputFiles.begin();
		it != m_vecInputFiles.end();
		++it)
	{
		(*it).debugPrint(nIndentLevel+2);
	}

	// Print resource files

	for(TypeVecResourceFiles::const_iterator it = m_vecResourceFiles.begin();
		it != m_vecResourceFiles.end();
		++it)
	{
		(*it).debugPrint(nIndentLevel+2);
	}
	*/
}
