///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/DebugHelper.cpp 2     5/28/05 1:20p Eaddy $
//
//    File name:	DebugHelper.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/DebugHelper.cpp $
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.2  2004/10/10 06:07:44  me133
 * o Added header comment
 *
 */

#include "StdAfx.h"

#include "DebugHelper.h"	// Me

#include <dbghelp.h>		// Sym* functions

// HELPER FUNCTIONS

BOOL CALLBACK SymEnumSourceFilesCallback(	PSOURCEFILE pSourceFile,
											PVOID UserContext);
////////////////////////////////
// DEBUG HELPER IMPLEMENTATION
////////////////////////////////

// Singleton definition and initialization
SDebugHelper SDebugHelper::s_singleton;

// Static variables
HANDLE SDebugHelper::s_hProcess = 0;

SDebugHelper::SDebugHelper()
{
	_ASSERTE(s_hProcess == 0); // Singleton only gets initialized once!

	s_hProcess = ::GetCurrentProcess();
	VERIFY_COND_V(s_hProcess, _T("GetCurrentProcess() failed."), ILXLOG_ERROR);

	BOOL bRet = ::SymInitialize(	s_hProcess,
									NULL,		// SearchPath
									FALSE);		// InvadeProcess
	if (!bRet)
	{
		HRESULT hr = HRESULT_FROM_WIN32(::GetLastError());
		ILXLOG(hr, _T("SymInitialize() failed."), ILXLOG_ERROR);
		s_hProcess = 0;
	}
}

SDebugHelper::~SDebugHelper()
{
	if (s_hProcess)
	{
		BOOL bRet = ::SymCleanup(s_hProcess);
		if (!bRet)
		{
			HRESULT hr = HRESULT_FROM_WIN32(::GetLastError());
			ILXLOG(hr, _T("SymCleanup() failed."), ILXLOG_ERROR);
		}
	}
}

HRESULT SDebugHelper::GetSourceFilesForModule(	const WCHAR* szModulePath,
												LPCBYTE pbBaseLoadAddress,
												TypeListStrings* pListSourceFiles)
{
	USES_CONVERSION;

	HRESULT hr = S_OK;

	if (!szModulePath || !*szModulePath || !pListSourceFiles)
		return E_POINTER;

	if (!s_hProcess)
	{
		return E_FAIL;
	}

	DWORD64 dwRet = SymLoadModule64(	s_hProcess,
										NULL,							// hFile
										W2A(szModulePath),				// ImageName
										NULL,							// ModuleName
										(DWORD64) pbBaseLoadAddress,	// LoadAddress
										0);								// SizeOfDll
	if (dwRet == 0)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG(hr, _T("SymLoadModule64() failed."), ILXLOG_ERROR);
	}

	BOOL bRet = SymEnumSourceFiles(	s_hProcess,
									(ULONG64) pbBaseLoadAddress,
									NULL, // Mask
									SymEnumSourceFilesCallback,
									pListSourceFiles);
	if (!bRet)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());
		ILXLOG(hr, _T("SymEnumSourceFiles() failed."), ILXLOG_ERROR);
	}

	return S_OK;
}

BOOL CALLBACK SymEnumSourceFilesCallback(	PSOURCEFILE pSourceFile,
											PVOID UserContext)
{
	TypeListStrings* pListSourceFiles = reinterpret_cast<TypeListStrings*>(UserContext);
	if (!pListSourceFiles) return FALSE;

	if (pSourceFile && pSourceFile->FileName && *pSourceFile->FileName)
	{
		pListSourceFiles->push_back(pSourceFile->FileName);
	}

	return TRUE;
}
