///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/Helpers.cpp 5     12/08/05 8:28p Eaddy $
//
//    File name:	Helpers.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/Helpers.cpp $
 * 
 * 5     12/08/05 8:28p Eaddy
 * o Can now pass NULL separator
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.5  2005/01/14 23:25:19  me133
 * o Added include file (not sure why)
 *
 * Revision 1.4  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.3  2004/10/17 00:24:55  me133
 * o Now provides a class-based iface to get to the incremental file info
 *
 * Revision 1.2  2004/10/10 06:07:44  me133
 * o Added header comment
 */

#include "stdafx.h"

#include "Helpers.h"			// Me

// Make sure you have: #define _WIN32_IE 0x0500
#include <shlobj.h>				// SHGetFolderPath()
#include <stdio.h>				// sprintf()

// MACROS

#define NumItems(s) (sizeof(s) / sizeof(s[0]))
#define ISFLAG(p,x) if (Is##p##x(dwFlags)) strcat(sFlags, "["#x "] ");
#define ENUM_BUFFER_SIZE 10

// TYPE DEFINES

typedef BYTE* (*PfnGetPreStub)(void);

// GetClassObject declare
typedef HRESULT __stdcall DLLGETCLASSOBJECT(REFCLSID rclsid,
                                            REFIID   riid,
                                            void   **ppv);
// GLOBALS

static PfnGetPreStub g_pfnThePreStub = 0; //to avoid recalculating it every time.

static HINSTANCE g_gPalRtDll = NULL; // PalRt Dll handle

// CONSTANTS

static const TCHAR* szROTOR_ENV_VAR_NAME		= _T("ROTOR");
static const TCHAR* szROTOR_DEBUG_ENV_VAR_NAME	= _T("DEBUGGING_SUPPORTED_BUILD");
static const TCHAR* szROTOR_BIN_DIR				= _T("TARGETCOMPLUS");
static const char* szCLIX_FILENAME				= "\\clix.exe";

// Size of stub assembly code. From cgencpu.h (SSCLI).
// NOTE: This will only work for x86!!
// x86: CALL(E8) xx xx xx xx
static const DWORD dwMETHOD_CALL_PRESTUB_SIZE	= 5;

// sizeof(Stub). Stub is declared in stublink.h (SSCLI)
static const DWORD dwSTUB_STRUCT_SIZE_RELEASE	= 8;
static const DWORD dwSTUB_STRUCT_SIZE_DEBUG		= 20;

// HELPER FUNCTIONS

static HRESULT PAL_CoCreateInstance(REFCLSID rclsid, REFIID riid, void** ppv);

//*********************************************************************
// FUNCTION		:	WaitForDispatchThreadEvents
//					
// DESCRIPTION	:	
//					
// RETURNS		:	
//   EDispatchThreadWaitStatus 		
//					
// PARAMETERS	:	
//   hQueuedObjectAvailable	
//   hEventDispatchThreadPleaseTerminate	
//					
// COMMENTS		:	
//					
//*********************************************************************
EDispatchThreadWaitStatus WaitForDispatchThreadEvents(
							HANDLE hQueuedObjectAvailable, 
							HANDLE hEventDispatchThreadPleaseTerminate)
{
	_ASSERTE(hQueuedObjectAvailable);
	_ASSERTE(hEventDispatchThreadPleaseTerminate);

	HANDLE events[2] =
	{	
		hQueuedObjectAvailable,
		hEventDispatchThreadPleaseTerminate 
	};

	DWORD nEventCount = sizeof(events) / sizeof(*events);

	while (true) // Infinite loop
	{
		DWORD dwRet = MsgWaitForMultipleObjects(	nEventCount, 
													events, 
													FALSE, 
													INFINITE, 
													QS_ALLINPUT);
		if (dwRet == WAIT_OBJECT_0 + 0)
		{
			return DISPATCH_THREAD_EVENT_WAITING;
		}
		else if (dwRet == WAIT_OBJECT_0 + 1)
		{
			return DISPATCH_THREAD_SHUTDOWN_SIGNALED;
		}
		else if (dwRet == WAIT_OBJECT_0 + nEventCount)
		{
			MSG msg;

			// There is one or more window message available. Dispatch them
			while (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else
		{
			// Something else happened (it is not a message)
			return DISPATCH_THREAD_WAIT_ERROR;
		}
	}
}

// You should only call this from a MethodDesc that hasn't been JIT compiled,
// because that's when the prestub will be valid
DWORD ExtractPreStubFuncAddr(FunctionID functionID)
{
	DWORD* dwPreStubMemoryAddr = (DWORD*)(functionID-4);
	DWORD dwPreStubRelativeFuncAddr = *dwPreStubMemoryAddr;
	DWORD dwPreStubAbsoluteFuncAddr = dwPreStubRelativeFuncAddr + (DWORD) functionID;
	return dwPreStubAbsoluteFuncAddr;
}

//
// This code is adapted from Patrick Grawehr's PhD Thesis (written in
// German!):
//
// http://www.iks.inf.ethz.ch/publications/files/dotproseda03.pdf
//
// Patrick's code didn't work because 1) LoadLibrary() was
// failing, 2) his hardcoded relative address for the ThePreStub()
// function was wrong (he probably had a different version of the CLR),
// and 3) he added 8 to skip to the next Stub but for debug builds you
// needed to add 20.
//
// I was able to fix #1 by using GetModuleHandle() instead of LoadLibrary()
// and #3 by checking for the DEBUGGING_SUPPORTED_BUILD env variable.  #2
// I could only fix for Rotor because I could step into the ThePreStub()
// function call and put &ThePreStub in the Watch window to find the relative
// address.  I didn't know how to obtain the address of a private function
// inside the CLR mscorwks.dll that doesn't have debug info.  Rather than figure
// that out, I found a more portable and less-kludgy kludge.  At some point when I
// get a JITCompilationStarted call, I call my ExtractPreStubFuncAddr()
// helper function to extract out the prestub address.  I store it in a
// global variable and pass it to SetFunctionReJITBugFix().  Patrick's
// code was a little neater since the kludge was self-contained in one
// function.
//
HRESULT SetFunctionReJITBugFix(ICorProfilerInfo* piProfInfo, 
							   FunctionID functionID,
							   DWORD dwPfnThePreStubFunc)
{
	HRESULT hr = S_OK;
	
	hr = piProfInfo->SetFunctionReJIT(functionID);
	if (FAILED(hr)) return hr;

	BYTE *patchloc = (BYTE*)(functionID-dwMETHOD_CALL_PRESTUB_SIZE);
	DWORD *addrloc = (DWORD*)(functionID-(dwMETHOD_CALL_PRESTUB_SIZE-1));

	// functionID points to the end of the instruction, just
	// from where we need to calculate the offset for the call
	// warning C4311: 'type cast' : pointer truncation from 'HMODULE' to 'DWORD'
	#pragma warning(disable: 4311)
	DWORD relprestub = (DWORD)(dwPfnThePreStubFunc - functionID);
	#pragma warning(default: 4311)

	// This code is not 100% thread-safe, but quite secure
	// We would need to use something like InterlockedExchange8b()
	*patchloc = 0xe8;		// CALL(E8)
	*addrloc = relprestub;

	return S_OK;
};

void GetRotorAndCLRInfo(	bool* pbIsRotor, 
							bool* pbIsDebugBuild,
							char* szClixPath,
							int nMaxClixPathLen)
{
	_ASSERTE(pbIsRotor || pbIsDebugBuild || szClixPath);
	_ASSERTE(!szClixPath || nMaxClixPathLen > (int) strlen(szCLIX_FILENAME));

	static bool s_bGotEnv		= false;

	// Cached environment variable values
	static bool s_bIsRotor		= false;
	static bool s_bIsDebugBuild = false;
	static char s_szClixPath[MAX_PATH] = { 0 };;

	if (!s_bGotEnv)
	{
		// See if we are running in Rotor (clix.exe)
		TCHAR szIsRotor[_MAX_PATH] = { 0 };

		DWORD dwRet = ::GetEnvironmentVariable(	szROTOR_ENV_VAR_NAME, 
												szIsRotor,
												_MAX_PATH );

		if (dwRet && szIsRotor && *szIsRotor)
		{
			s_bIsRotor = _variant_t(szIsRotor);
		}
		
		// See if we are running a debug build of Rotor
		TCHAR szIsDebugBuild[_MAX_PATH] = { 0 };

		dwRet = ::GetEnvironmentVariable(	szROTOR_DEBUG_ENV_VAR_NAME, 
											szIsDebugBuild,
											_MAX_PATH );

		if (dwRet && szIsDebugBuild && *szIsDebugBuild)
		{
			s_bIsDebugBuild = _variant_t(szIsDebugBuild);
		}

		dwRet = ::GetEnvironmentVariable(	szROTOR_BIN_DIR, 
											s_szClixPath,
											_MAX_PATH );
		if (dwRet)
		{
			strcat(s_szClixPath, szCLIX_FILENAME);
			_ASSERTE(::PathFileExists(s_szClixPath));
		}

		// So we only do this once
		s_bGotEnv = true;
	}

	if (pbIsRotor) *pbIsRotor = s_bIsRotor;
	if (pbIsDebugBuild) *pbIsDebugBuild = s_bIsDebugBuild;
	if (szClixPath)
	{
		if (s_szClixPath && *s_szClixPath && ::PathFileExists(s_szClixPath))
			strcpy(szClixPath, s_szClixPath);
		else
			szClixPath[0] = '\0';
	}
}

STDAPI MyCoCreateInstance(	REFCLSID rclsid,
							LPUNKNOWN pUnkOuter,
							DWORD dwClsContext,
							REFIID riid,
							LPVOID* ppv)
{
	HRESULT hr = S_OK;

	hr = CoCreateInstance(rclsid, NULL, dwClsContext, riid, ppv);
	if (FAILED(hr))
	{
		bool bIsRotor = false;
		GetRotorAndCLRInfo(&bIsRotor, NULL, NULL, 0);

		if (bIsRotor)
		{
			// NOTE: This only works for creating CLR COM objects
			return PAL_CoCreateInstance(rclsid, riid, ppv);
		}
		else
		{
			return hr;
		}
	}

	return S_OK;
}

// Call into the palrt to CoCreate the clsid
HRESULT PAL_CoCreateInstance(REFCLSID   rclsid,
                             REFIID     riid,
                             void     **ppv)
{
	HRESULT hr = S_OK;

	g_gPalRtDll = LoadLibraryW(L"rotor_palrt.dll");
    if (!g_gPalRtDll)
	{
        hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG(hr, _T("Failed to load rotor_palrt.dll."), ILXLOG_ERROR);
	}

    DLLGETCLASSOBJECT* pfnCoCreateInstance = 
		(DLLGETCLASSOBJECT*) GetProcAddress(	g_gPalRtDll, 
												"PAL_CoCreateInstance");
	if (!pfnCoCreateInstance)
	{
        hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG(hr, _T("Failed to get proc address of PAL_CoCreateInstance."), ILXLOG_ERROR);
	}

    hr = (*pfnCoCreateInstance)(rclsid, riid, (void**)ppv);
	VERIFY_HR(hr, _T("Rotor's CoCreateInstance() failed."), ILXLOG_ERROR);

	return S_OK;
}


void AppendBitSet(char* szMsg, const char* szName, 
				  const BYTE* pBitset, unsigned int nNumBytes, 
				  const char* szSeparator)
{
	AppendMsg(szMsg, szName, szSeparator);
	strcat(szMsg, ": ");

    const unsigned int nMASK = 1 << ((sizeof(nMASK)*8)-1);

	char szByte[100];
	for(unsigned int j = 0; j < nNumBytes; ++j)
	{
		char cByte = pBitset[j];

		for(unsigned int i = 0; i < sizeof(cByte) * 8; ++i)
		{
			unsigned int bit = cByte & nMASK;
			szByte[(j*8) + i] = bit == nMASK ? '1' : '0';
			cByte <<= 1;
		}
	}

	szByte[j*8] = '\0';

	strcat(szMsg, szByte);
}

void AppendString(char* szMsg, const char* szName, const char* szValue,
				  const char* szSeparator)
{
	char szNameValue[1000];
	sprintf(szNameValue, "%s: %s", szName, szValue);
	AppendMsg(szMsg, szNameValue, szSeparator);
}

void AppendFlag(char* szMsg, const char* szName, bool bValue,
				const char* szSeparator)
{
	AppendString(szMsg, szName, bValue ? "T" : "F", szSeparator);
}

void AppendNum(char* szMsg, const char* szName, DWORD nValue,
			   const char* szSeparator)
{
	char szNum[100];
	sprintf(szNum, "%s: %d", szName, nValue);
	AppendMsg(szMsg, szNum, szSeparator);
}

void AppendTime(char* szMsg, const char* szName, FILETIME fileTime,
				const char* szSeparator)
{
	char szTime[100];
	strcpy(szTime, szName);

	if (fileTime.dwHighDateTime == 0 && fileTime.dwLowDateTime == 0)
	{
		strcat(szTime, ": NONE");
	}
	else
	{
		SYSTEMTIME systemTime;
		BOOL bSuccess = FileTimeToSystemTime(&fileTime, &systemTime);
		if (!bSuccess)
		{
			strcat(szTime, ": ERROR");
		}
		else
		{
			sprintf(szTime, "%s: %02d:%02d:%02d:%03d %02d/%02d/%02d", 
				szName,
				systemTime.wHour,
				systemTime.wMinute,
				systemTime.wSecond,
				systemTime.wMilliseconds,
				systemTime.wMonth,
				systemTime.wDay,
				systemTime.wYear);
		}
	}

	AppendMsg(szMsg, szTime, szSeparator);
}

void AppendNumHex(char* szMsg,	const char* szName, DWORD nValue,
				  const char* szSeparator)
{
	char szNum[100];
	sprintf(szNum, "%s: 0x%08X", szName, nValue);
	AppendMsg(szMsg, szNum, szSeparator);
}

void AppendMsg(char* szMsg, const char* szAppend,
			   const char* szSeparator)
{
	_ASSERTE(szMsg);
	if (!szMsg) return;

	if (*szMsg && szSeparator && *szSeparator)
	{
		size_t len = strlen(szMsg);
		if (len > 1 && !isspace(szMsg[len-1]))
			strcat(szMsg, szSeparator);
	}

	strcat(szMsg, szAppend);
}

void AppendMsg(std::wstring str, const WCHAR* szAppend,
			   const WCHAR* szSeparator)
{
	if (!str.empty())
	{
		str.append(szSeparator);
	}

	str.append(szAppend);
}

void AppendMsgWithIndent(char* szMsg, const char* szAppend, int nIndentLevel,
						 const char* szSeparator)
{
	AppendIndent(szMsg, nIndentLevel);
	AppendMsg(szMsg, szAppend, szSeparator);
}

void AppendIndent(char* szMsg, int nIndentLevel)
{
	_ASSERTE(szMsg);
	if (!szMsg) return;

	if (nIndentLevel == 0)
		return;

	char* p = szMsg;
	while(*p) { ++p; } // Find end of string

	for(int i = 0; i < nIndentLevel; ++i, ++p)
	{
		*p = '\t';
	}

	*p = '\0';
}

HRESULT ReadUnicodeOrASCIILineFromFile(FILE* in, bool bIsUnicode, char* buffer, int nMaxLen)
{
	if (!in || !buffer || nMaxLen <= 0)
		return E_POINTER;

	if (nMaxLen <= 0)
		return E_INVALIDARG;

	buffer[0] = '\0';

	void* pSuccess = 0;

	if (bIsUnicode)
	{
		// NOTE: You must call _setmode(_fileno(in), _O_BINARY) once
		// before calling this function to switch to binary mode to read
		// in the UNICODE file. Otherwise, you'll get a null wchar_t after
		// each normal wchar_t.

		static const nMAX_LINE_LEN = 10000;
		_ASSERTE(nMAX_LINE_LEN >= nMaxLen);

	    wchar_t wbuffer[nMAX_LINE_LEN];
		wbuffer[0] = '\0';
		pSuccess = fgetws(wbuffer, nMAX_LINE_LEN-1, in);
		if (pSuccess)
		{
			int nNumBytes = WideCharToMultiByte(CP_ACP, 0, wbuffer, -1, buffer, nMaxLen-1, NULL, NULL);
			if (nNumBytes == 0)
			{
				return ILXLOG(HRESULT_FROM_WIN32(::GetLastError()),
					_T("WideCharToMultiByte() failed."), ILXLOG_ERROR);
			}
		}
	}
	else
	{
		pSuccess = fgets(buffer, nMaxLen-1, in);
	}

	if (!pSuccess)
	{
		if (feof(in))
		{
			return S_FALSE;
		}
		else if (ferror(in) != 0)
		{
			return E_FAIL;
		}
		else
		{
			return E_UNEXPECTED;
		}
	}

	// Force a null terminator
	buffer[nMaxLen-1] = '\0';

	size_t nLen = strlen(buffer);

	// Trim off newlines/carriage return
	while (nLen-1 >= 0 && 
		(buffer[nLen-1] == '\r' || buffer[nLen-1] == '\n'))
	{
		--nLen;
	}

	if (nLen >= 0)
		buffer[nLen] = '\0';

	return S_OK;
}
