///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/ILHelper.cpp 6     12/08/05 10:34p Eaddy $
//
//    File name:	ILHelper.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/ILHelper.cpp $
 * 
 * 6     12/08/05 10:34p Eaddy
 * o Removed incorrect assert
 * 
 * 5     12/08/05 7:54p Eaddy
 * o Fixed signature comparison
 * 
 * 4     12/07/05 7:50p Eaddy
 * o Can now filter on operand type as well as opcode
 * o Added getNextOpCode and sigCmp
 * 
 * 3     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.6  2005/01/14 23:00:16  me133
 * o Added invariant check
 *
 * Revision 1.5  2004/10/26 04:35:58  me133
 * o Now displays opcode name if possible
 *
 * Revision 1.4  2004/10/25 01:03:48  me133
 * o Method IL comparison is now complete and correct
 *
 * Revision 1.3  2004/10/11 23:21:52  me133
 * o Log message now contains opcode value
 *
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "stdafx.h"

#include "ILHelper.h"			// Me

#include "MetadataHelpers.h"	// DisplaySignature()

#include "MethodInfo.h"			// CMethodInfo

// MACROS

// this macro borrowed from Rotor's code >>
#define GET_VAL( ip, val, type )								\
        {                                                       \
            _ASSERTE( ( ip + sizeof(type) ) <= pILCodeEnd );	\
            val = ( *((type *&)(ip))++ );					    \
        }
// this macro borrowed from Rotor's code <<

// HELPER FUNCTIONS

static HRESULT FindOpCode(	LPBYTE pILCodeFindStart,
							LPCBYTE pILCodeEnd,
							OPCODE opcodeFilter,
							OPCODE_FORMAT opcodeFormatFilter,
							/*[out]*/ LPBYTE* ppIP,
							/*[out]*/ LPBYTE* ppNextIP);

static HRESULT GetNextOpCode(	/*[in]*/ LPCBYTE pILCodeEnd,
								/*[in,out]*/ LPBYTE* ppIP,
								/*[out]*/ OPCODE* pOpCode);

///////////////////////////////////
// OPCODE ITERATOR IMPLEMENTATION
///////////////////////////////////

opcode_iterator::opcode_iterator(	LPBYTE pIPStart, 
									LPBYTE pIPEnd, 
									OPCODE opcodeFilter,
									OPCODE_FORMAT opcodeFormatFilter)
	: m_pIPStart(pIPStart)
	, m_pIPEnd(pIPEnd)
	, m_pIPCur(0)
	, m_opcodeFilter(opcodeFilter)
	, m_opcodeFormatFilter(opcodeFormatFilter)
	, m_hrLast(S_OK)
{ 
	this->operator++();
}

opcode_iterator::opcode_iterator(	const CMethodInfo& methodInfo,
									OPCODE opcodeFilter,
									OPCODE_FORMAT opcodeFormatFilter)
	: m_pIPStart(methodInfo.getCodeStart())
	, m_pIPEnd(methodInfo.getCodeEnd())
	, m_pIPCur(0)
	, m_opcodeFilter(opcodeFilter)
	, m_opcodeFormatFilter(opcodeFormatFilter)
	, m_hrLast(S_OK)
{ 
	this->operator++();
}

opcode_iterator& opcode_iterator::operator++()
{
	m_hrLast = FindOpCode(m_pIPStart, m_pIPEnd, m_opcodeFilter, m_opcodeFormatFilter, 
		&m_pIPCur, &m_pIPStart);
	return *this;
}

LPBYTE opcode_iterator::operator*() const
{
	return m_pIPCur;
}

OPCODE opcode_iterator::getOpCode() const
{
	return (OPCODE) *m_pIPCur;
}

OPCODE opcode_iterator::getNextOpCode() const
{
	_ASSERTE(getOpCode() == CEE_PREFIX1);
	return OPCODE((*(m_pIPCur + 1)) + 256); // Decode next instr
}

LPBYTE opcode_iterator::getOpCodeArg() const
{
	_ASSERTE(m_pIPCur);
	if (!m_pIPCur) return 0;

	OPCODE opcode = getOpCode();

	if (opcode == CEE_PREFIX1)
	{
		opcode = getNextOpCode(); // Decode next instr
	}

	return m_pIPCur + g_ILcodes[opcode].Len;
}

bool opcode_iterator::end() const
{
	return m_hrLast != S_OK;
}

bool opcode_iterator::isError() const
{
	return FAILED(m_hrLast);
}

bool opcode_iterator::isOpCodeEqual(OPCODE opcodeFirstRhs, OPCODE opcodeSecondRhs) const
{
	_ASSERTE(opcodeFirstRhs != CEE_PREFIX1 || opcodeSecondRhs != CEE_COUNT);

	OPCODE opcodeFirstLhs = getOpCode();

	if (opcodeFirstLhs != opcodeFirstRhs)
	{
		return false;
	}
	else if (opcodeFirstLhs == CEE_PREFIX1)
	{
		// Handle two-byte instuctions
		_ASSERTE(opcodeFirstRhs == CEE_PREFIX1);
		_ASSERTE(opcodeSecondRhs == CEE_COUNT);

		OPCODE opcodeSecondLhs = getNextOpCode(); // Decode next instr

		return opcodeSecondLhs == opcodeSecondRhs;
	}
	else
	{
		return true;
	}
}

////////////////////
// IMPLEMENTATIONS
////////////////////

HRESULT FindOpCode(	LPBYTE pILCodeFindStart,
					LPCBYTE pILCodeEnd,
					OPCODE opcodeFilter,
					OPCODE_FORMAT opcodeFormatFilter,
					/*[out]*/ LPBYTE* ppIP,
					/*[out]*/ LPBYTE* ppNextIP)
{
	HRESULT hr = S_OK;

	// Verify args
	if (!pILCodeFindStart || !pILCodeEnd) return E_POINTER;
	if (!ppIP) return E_POINTER;

	// Only allowed to set one filter
	_ASSERTE(opcodeFilter == CEE_COUNT || opcodeFormatFilter == Uninitialized);

	// Initialize out params
	*ppIP = 0;

	LPBYTE pIPNext = pILCodeFindStart;
	while (pIPNext < pILCodeEnd)
	{
		OPCODE opcodeCur = CEE_COUNT;

		LPBYTE pIPCur = pIPNext; // Save the current IP

		hr = GetNextOpCode(pILCodeEnd, &pIPNext, &opcodeCur);
		if (FAILED(hr)) return hr;
		_ASSERTE(hr != S_FALSE);
		_ASSERTE(pIPNext != 0);

		if (opcodeFormatFilter != Uninitialized)
		{
			// OPERAND_FORMAT is a bit flag
			if (g_ILcodes[opcodeCur].operandFormat & opcodeFormatFilter)
			{
				_ASSERTE(opcodeCur != CEE_COUNT);
				*ppIP = pIPCur;
				*ppNextIP = pIPNext;
				return S_OK;
			}
		}
		else if (opcodeFilter != CEE_COUNT)
		{
			if (opcodeCur == opcodeFilter)
			{
				_ASSERTE(opcodeCur != CEE_COUNT);
				*ppIP = pIPCur;
				*ppNextIP = pIPNext;
				return S_OK;
			}
		}
		else
		{
			_ASSERTE(opcodeCur != CEE_COUNT);
			*ppIP = pIPCur;
			*ppNextIP = pIPNext;
			return S_OK;
		}
	}

	return S_FALSE;
}

HRESULT GetNextOpCode(	/*[in]*/ LPCBYTE pILCodeEnd,
						/*[in,out]*/ LPBYTE* ppIP,
						/*[out]*/ OPCODE* pOpCode)
{
	HRESULT hr = S_OK;

	if (!pILCodeEnd) return E_POINTER;
	if (!ppIP) return E_POINTER;
	if (!*ppIP) return E_INVALIDARG;
	if (!pOpCode) return E_POINTER;

	// Initialize out params
	*pOpCode = CEE_COUNT;

	//	Step 1:	See if we have reached the last instruction

	if (*ppIP >= pILCodeEnd)
	{
		return S_FALSE;
	}

	// Step 2:	Parse and decode the current opcode, advancing
	//			the instruction pointer (IP) as we do
	
	BYTE  Val = 0;

	GET_VAL( *ppIP, Val, BYTE );
    *pOpCode = (OPCODE) Val;

	// helper vars
	unsigned int	nOffset = 0; 
	signed int		i4 = 0;
	signed			branch = 0;
	unsigned int	token = 0;

DECODE_OPCODE:

    switch ( *pOpCode )
    {
    case CEE_PREFIX1:
        GET_VAL( *ppIP, Val, BYTE );
        *pOpCode = OPCODE( Val + 256 );
        goto DECODE_OPCODE;

    case CEE_LDARG_0: // ldarg.0 : 0x02
    case CEE_LDARG_1: // ldarg.1 : 0x03
    case CEE_LDARG_2: // ldarg.2 : 0x04
    case CEE_LDARG_3: // ldarg.3 : 0x05
        nOffset = ( *pOpCode - CEE_LDARG_0 );
        // make sure that opcode is valid, can have only 0..3 "super short" forms of ldarg
        _ASSERTE( nOffset < 4 ); 
        break;

    case CEE_LDARG_S: //ldarg.s <unsigned int8> : 0x0E <unsigned int8>
        GET_VAL( *ppIP, nOffset, BYTE );// skip one more byte (=<unsigned int8>) 
        break;

    case CEE_LDARG: //ldarg <unsigned int16> : 0xFE 0x09 <unsigned int16>
        GET_VAL( *ppIP, nOffset, unsigned short ); // skip unsigned short
        break;

    case CEE_LDLOC_0: // ldloc.0 : 0x06
    case CEE_LDLOC_1: // ldloc.0 : 0x07
    case CEE_LDLOC_2: // ldloc.0 : 0x08
    case CEE_LDLOC_3: // ldloc.0 : 0x09
        nOffset = ( *pOpCode - CEE_LDLOC_0 );
        // make sure that opcode is valid, can have only 0..3 "super short" forms of ldloc
        _ASSERTE( nOffset < 4 );
        break;

    case CEE_LDLOC_S: // ldloc.s <unsigned int8> : 0x11 <unsigned int8>
        GET_VAL( *ppIP, nOffset, BYTE ); // skip one more byte (=<unsigned int8>)
        break;

    case CEE_LDLOC: //ldloc <unsigned int16> : 0xFE 0x0C <unsigned int16>
        GET_VAL( *ppIP, nOffset, unsigned short ); // skip unsigned short (= <unsigned int16>)
        break;

    case CEE_STARG_S: // starg.s <unsigned int8> : 0x10 <unsigned int8>
        GET_VAL( *ppIP, nOffset, BYTE ); // skip byte (=<unsigned int8>)
        break;

    case CEE_STARG: // starg <unsigned int16> : 0xFE 0x0B <unsigned int16>
        GET_VAL( *ppIP, nOffset, unsigned short ); // skip unsigned short (= <unsigned int16>)
        break;

    case CEE_STLOC_0: // stloc.0 : 0x0A
    case CEE_STLOC_1: // stloc.1 : 0x0B
    case CEE_STLOC_2: // stloc.2 : 0x0C
    case CEE_STLOC_3: // stloc.3 : 0x0D
        nOffset = ( *pOpCode - CEE_STLOC_0 );
        // make sure that opcode is valid, can have only 0..3 "super short" forms of stloc
        _ASSERTE( nOffset < 4 );
        break;

    case CEE_STLOC_S: // stloc.s <unsigned int8> : 0x13 <unsigned int8>
        GET_VAL( *ppIP, nOffset, BYTE ); // skip <unsigned int8>
        break;

    case CEE_STLOC: // stloc <unsigned int16> : 0xFE 0x0E <unsigned int16>
        GET_VAL( *ppIP, nOffset, unsigned short ); // skip <unsigned int16>
        break;

    case CEE_ADD: // add : 0x58
        break;

    case CEE_ADD_OVF: // add.ovf: 0xD6
        break;

    case CEE_ADD_OVF_UN: // add.ovf.un: 0xD7
        break;

    case CEE_SUB: // sub : 0x59
        break;

    case CEE_SUB_OVF: // sub.ovf : 0xDA
        break;

    case CEE_SUB_OVF_UN: // sub.ovf.un : 0xDB 
        break;

    case CEE_MUL: // mul : 0x5A
        break;

    case CEE_MUL_OVF: // mul.ovf : 0xD8
        break;

    case CEE_MUL_OVF_UN: // mul.ovf.un : 0xD9
        break;

    case CEE_DIV:  // div : 0x5B
        break;

    case CEE_DIV_UN: // div.un : 0x5C
        break;

    case CEE_REM: // rem : 0x5D
        break;

    case CEE_REM_UN: // rem.un : 0x5E
        break;

    case CEE_LOCALLOC: // localloc : 0xFE 0x0F //localloc is unverifiable!!!
        break;

    case CEE_NEG: // neg : 0x65
        break;

    case CEE_LDIND_U1: // ldind.u1 : 0x47
        break;
    case CEE_LDIND_U2: // ldind.u2 : 0x49
        break;
    case CEE_LDIND_U4: // ldind.u4 : 0x4B
        break;
    case CEE_LDIND_I1: // ldind.i1 : 0x46
        break;
    case CEE_LDIND_I2: // ldind.i2 : 0x48
        break;
    case CEE_LDIND_I4: // ldind.i4 : 0x4A
        break;
    case CEE_LDIND_I8: // ldind.i8 : 0x4C
        break;
    case CEE_LDIND_R4: // ldind.r4 : 0x4E
        break;
    case CEE_LDIND_R8: // ldind.r8 : 0x4F
        break;
    case CEE_LDIND_I: // ldind.i : 0x4D
        break;

    case CEE_LDIND_REF: // ldind.ref : 0x50
        break;

    case CEE_STIND_I1: // stind.i1 : 0x52
        break;

    case CEE_STIND_I2: // stind.i2 : 0x53
        break;

    case CEE_STIND_I4: // stind.i4 : 0x54
        break;

    case CEE_STIND_I8: // stind.i8 : 0x55
        break;

    case CEE_STIND_I: // stind.i : 0xDF
        break;

    case CEE_STIND_R4: // stind.r4 : 0x56
        break;

    case CEE_STIND_R8: // stind.r8 : 0x57
        break;

    case CEE_STIND_REF: // stind.ref : 0x51
        break;

    case CEE_LDC_I4_M1:  // ldc.i4.m1 : 0x15, load -1 on the stack
    case CEE_LDC_I4_0 :  // ldc.i4.0  : 0x16, load 0
    case CEE_LDC_I4_1 :  // ldc.i4.1  : 0x17, load 1
    case CEE_LDC_I4_2 :  // ldc.i4.2  : 0x18, load 2
    case CEE_LDC_I4_3 :  // ldc.i4.3  : 0x19, load 3
    case CEE_LDC_I4_4 :  // ldc.i4.4  : 0x1A, load 4
    case CEE_LDC_I4_5 :  // ldc.i4.5  : 0x1B, load 5
    case CEE_LDC_I4_6 :  // ldc.i4.6  : 0x1C, load 6
    case CEE_LDC_I4_7 :  // ldc.i4.7  : 0x1D, load 7
    case CEE_LDC_I4_8 :  // ldc.i4.8  : 0x1E, load 8
        i4 = ( *pOpCode - CEE_LDC_I4_0 );
        // make sure that opcode is valid, we can only load -1 .. 8
        _ASSERTE( (-1 <= i4) && (i4 <= 8) );
        break;

    case CEE_LDC_I4_S: // ldc.i4.s <int8> : 0x1F <int8>
        GET_VAL( *ppIP, i4, signed char ); // get <int8> as param
        break;

    case CEE_LDC_I4: // ldc.i4 <int32> : 0x20 <int32>
        GET_VAL( *ppIP, i4, signed int ); // get <int32> as param
        break;

    case CEE_LDC_I8: // ldc.i8 <int64> : 0x21 <int64>
		signed __int64  i8;
		GET_VAL( *ppIP, i8, signed __int64 ); // get <int64> as param
        break;

    case CEE_LDC_R4: // ldc.r4 <float32> : 0x22 <float32>
		float _float; // 4 bytes
		GET_VAL( *ppIP, _float, float ); // get <float32> as param
        break;
    case CEE_LDC_R8:  // ldc.r8 <float64>: 0x23 <float64>
		double _double; // 8 bytes
		GET_VAL( *ppIP, _double, double ); // get <float64> as param
        break;

    case CEE_LDNULL: // ldnull : 0x14
        break;

    case CEE_LDLOCA_S: // ldloca.s <unsigned int8> : 0x12 <unsigned int8>
        GET_VAL( *ppIP, nOffset, BYTE ); // skip <unsigned int8>
        break;

    case CEE_LDLOCA: // ldloca <unsigned int16> : 0xFE 0x0D <unsigned int16>
        GET_VAL( *ppIP, nOffset, unsigned short ); // skip <unsigned int16>
        break;

    case CEE_LDSTR: // ldstr <token> : 0x72 <token>
		GET_VAL( *ppIP, token, unsigned int); // get str token
        break;

    case CEE_CPBLK: // cpblk : 0xFE 0x17 // unverifiable !
        break;

    case CEE_INITBLK: //initblk : 0xFE 0x18 // unverifiable !
        break;

        case CEE_INITOBJ: // initobj <token>: 0xFE 0x15 <token>
		GET_VAL( *ppIP, token, unsigned int );  // get token
        break;

    case CEE_CPOBJ: // cpobj <token> : 0x70 <token>
		GET_VAL( *ppIP, token, unsigned int );  // get token
        break;

    case CEE_LDOBJ: // ldobj <token> : 0x71 <token>
		GET_VAL( *ppIP, token, unsigned int );  // get token
        break;

    case CEE_STOBJ: // stobj <token> : 0x81 <token>
		GET_VAL( *ppIP, token, unsigned int );  // get token
        break;

    case CEE_MKREFANY:  // mkrefany <token> : 0xC6 <token>
		GET_VAL( *ppIP, token, unsigned int );  // get token
        break;

    case CEE_SIZEOF: // sizeof <token> : 0xFE 0x1C <token>
		GET_VAL( *ppIP, token, unsigned int );  // get token
        break;

    case CEE_LEAVE_S: // leave.s <int8> : 0xDE <int8>
        GET_VAL( *ppIP, branch, signed char ); // get <int8> branch
        break;

    case CEE_LEAVE: // leave <int32> : 0xDD <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32> branch
        break;

    case CEE_BR: // br <int32> : 0x38 <int32> -- branch <int32> bytes from the cur. position
        GET_VAL( *ppIP, branch, int ); // get <int32> branch
        break;

    case CEE_BR_S: // br.s <int8> : 0x2B <int8>
        GET_VAL( *ppIP, branch, signed char ); // get <int8> branch
        break;

    case CEE_BRTRUE: // brtrue <int32> : 0x3A <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32> branch
        break;

    case CEE_BRTRUE_S: // brtrue.s <int8> : 0x2D <int8>
        GET_VAL( *ppIP, branch, signed char ); // get <int8> branch
        break;

    case CEE_BRFALSE: // brfalse <int32> : 0x39 <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32> branch
        break;

    case CEE_BRFALSE_S: // brfalse.s <int8> : 0x2C <int8> 
        GET_VAL( *ppIP, branch, signed char ); // get <int8> branch
        break;

    case CEE_CEQ: // ceq : 0xFE 0x01
        break;

    case CEE_CGT: // cgt : 0xFE 0x02
        break;

    case CEE_CGT_UN: // cgt.un : 0xFE 0x03
        break;

    case CEE_CLT: // clt : 0xFE 0x04
        break;

    case CEE_CLT_UN: // clt.un : 0xFE 0x05
        break;

    case CEE_BEQ_S: // beq.s <int8> : 0x2E <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BEQ: // beq <int32> : 0x3B <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32> branch
        break;

    case CEE_BNE_UN_S: // bne.un.s <int8> : 0x33 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BNE_UN: // bne.un <int32> : 0x40 <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BGT_S: // bgt.s <int8> : 0x30 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BGT: // bgt <int32> : 0x3D <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BGT_UN_S: // bgt.un.s <int8> : 0x35 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BGT_UN: // bgt.un <int32> : 0x42 <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BGE_S: // bge.s <int8> : 0x2F <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BGE: // bge <int32> : 0x3C <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BGE_UN_S: // bge.un.s <int8> : 0x34 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BGE_UN: // bge.un <int32> : 0x41 <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BLT_S: // blt.s <int8> : 0x32 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BLT: // blt <int32> : 0x3F <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BLT_UN_S: // blt.un.s <int8> : 0x37 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BLT_UN: // blt.un <int32> : 0x44 <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BLE_S: // ble.s <int8> : 0x31 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BLE: // ble <int32> : 0x3E <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BLE_UN_S: // ble.un.s <int8> : 0x36 <int8>
        GET_VAL( *ppIP, branch, char ); // get <int8>
        break;

    case CEE_BLE_UN: // ble.un <int32> : 0x43 <int32>
        GET_VAL( *ppIP, branch, int ); // get <int32>
        break;

    case CEE_BREAK: // break : 0x01
        break;

    case CEE_AND: 
        break;

    case CEE_OR:
        break;

    case CEE_XOR:
        break;

    case CEE_NOT:
        break;

    case CEE_SHR:
        break;

    case CEE_SHR_UN:
        break;

    case CEE_SHL:
        break;

    case CEE_DUP:
        break;

    case CEE_POP:
        break;

    case CEE_NOP: // nop : 0x00
        break;

    case CEE_LDARGA_S: // ldarga.s <unsigned int8> : 0x0F <unsigned int8>
        GET_VAL( *ppIP, nOffset, signed char ); // skip <unsigned int8>
        break;

    case CEE_LDARGA: // ldarga <unsigned int16> : OxFE 0x0A <unsigned int16>
        GET_VAL( *ppIP, nOffset, unsigned short ); // skip <unsigned int16>
        break;

    case CEE_REFANYVAL: // refanyval <token> : 0xC2 <token> ???
		_ASSERTE( !"CEE_REFANYVAL" );
// #pragma TODO( "check if it's a correct way to handle CEE_REFANYVAL" )
		GET_VAL( *ppIP, token, unsigned __int32 ); // get <token>
        break;

    case CEE_REFANYTYPE: // refanytype:  0xFE 0x1D
        break;

    case CEE_ARGLIST: // arglist : 0xFE 0x00
        break;

    case CEE_ILLEGAL:
        _ASSERTE( false );
        break;

    case CEE_CALLI: // calli <token> : 0x29 <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token> for sig of function
        break;

    case CEE_CALL: // call <token> : 0x28 <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token> for the target method
        break;

    case CEE_CALLVIRT: // callvirt <token> : 0x6F <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token> for the virtual method
        break;

    case CEE_CASTCLASS: // castclass <token> : 0x74 <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_CONV_I1: // conv.i1 : 0x67
        break;

    case CEE_CONV_I2: // conv.i2 : 0x68
        break;

    case CEE_CONV_I: // conv.i : 0xD3
    case CEE_CONV_I4: // conv.i4 : 0x69
        break;

    case CEE_CONV_U1: // conv.u1 : 0xD2
        break;

    case CEE_CONV_U2: // conv.u2 : 0xD1
        break;

    case CEE_CONV_U: // conv.u : 0xE0
    case CEE_CONV_U4: // conv.u4 : 0x6D 
        break;

    // case CEE_CONV_I: purposely!!!
    case CEE_CONV_I8: // conv.i8 : 0x6A 
        break;

    // case CEE_CONV_U: purposely!!!
    case CEE_CONV_U8: // conv.u8 : 0x6E 
        break;

    case CEE_CONV_R4: // conv.r4 : 0x6B 
        break;

    case CEE_CONV_R8: // conv.r8 : 0x6C 
        break;

    case CEE_CONV_R_UN: // conv.r.un : 0x76 
        break;

    case CEE_CONV_OVF_I1: // conv.ovf.i1 : 0xB3 
        break;

    case CEE_CONV_OVF_U1: // conv.ovf.u1 : 0xB4 
        break;

    case CEE_CONV_OVF_I2: // conv.ovf.i2 : 0xB5 
        break;

    case CEE_CONV_OVF_U2: // conv.ovf.u2 : 0xB6 
        break;

    case CEE_CONV_OVF_I: // conv.ovf.i : 0xD4 
    case CEE_CONV_OVF_I4: // conv.ovf.i4 : 0xB7 
        break;

    case CEE_CONV_OVF_U: // conv.ovf.u : 0xD5
    case CEE_CONV_OVF_U4: // conv.ovf.u4 : 0xB8
        break;

    // case CEE_CONV_OVF_I: purposely!!!
    case CEE_CONV_OVF_I8: // conv.ovf.i8 : 0xB9
        break;

    // case CEE_CONV_OVF_U: purposely!!!
    case CEE_CONV_OVF_U8: // conv.ovf.u8 : 0xBA
        break;

    case CEE_CONV_OVF_I1_UN: // conv.ovf.i1.un : 0x82
        break;

    case CEE_CONV_OVF_U1_UN: // conv.ovf.u1.un : 0x86
        break;

    case CEE_CONV_OVF_I2_UN: // conv.ovf.i2.un : 0x83
        break;

    case CEE_CONV_OVF_U2_UN: // conv.ovf.u2.un : 0x87
        break;

    case CEE_CONV_OVF_I_UN: // conv.ovf.i.un : 0x8A
    case CEE_CONV_OVF_I4_UN: // conv.ovf.i4.un : 0x84
        break;

    case CEE_CONV_OVF_U_UN: // conv.ovf.u.un : 0x8B
    case CEE_CONV_OVF_U4_UN: // conv.ovf.u4.un : 0x88
        break;

    // case CEE_CONV_OVF_I_UN: purposely!!!
    case CEE_CONV_OVF_I8_UN: // conv.ovf.i8.un : 0x85
        break;

    // case CEE_CONV_OVF_U_UN: purposely!!!
    case CEE_CONV_OVF_U8_UN: // conv.ovf.u8.un : 0x89
        break;

    case CEE_LDTOKEN: // ldtoken <token> : 0xD0 <token>
		unsigned int token;
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_BOX: // box <token> : 0x8C <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_UNBOX: // unbox <token> : 0x79 <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_ISINST: // isinst <token> : 0x75 <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_JMP: // jmp <token> : 0x27 <token> // "jmp" instruction is unverifiable
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_LDELEM_U1: // ldelem.u1
        break;

    case CEE_LDELEM_U2: // ldelem.u2
        break;

    case CEE_LDELEM_U4: // ldelem.u4
        break;

    case CEE_LDELEM_I: // ldelem.i
        break;

    case CEE_LDELEM_I1: // ldelem.i1
        break;

    case CEE_LDELEM_I2: // ldelem.i2
        break;

    case CEE_LDELEM_I4: // ldelem.i4
        break;

    case CEE_LDELEM_I8: // ldelem.i8
        break;

    case CEE_LDELEM_R4: // ldelem.r4
        break;

    case CEE_LDELEM_R8: // ldelem.r8
        break;

    case CEE_LDELEM_REF: // ldelem.ref
        break;

	case CEE_LDELEMA: // ldelema <token> : 0x8F <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_LDSFLD:
    case CEE_LDFLD:
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_LDFLDA:
    case CEE_LDSFLDA: 
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_STSFLD:
    case CEE_STFLD: 
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_LDFTN: // ldftn <token> : 0xFE 0x06 <token>
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_LDLEN: // ldlen : 0x8E
        break;

    case CEE_LDVIRTFTN: // ldvirtftn <token> : 0xFE 0x07 <token> 
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_NEWARR:
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_NEWOBJ:
		GET_VAL( *ppIP, token, unsigned int);   // get <token>
        break;

    case CEE_ENDFILTER:
        break;

    case CEE_ENDFINALLY:
        break;

    case CEE_RET:
        break;

    case CEE_STELEM_I1:
        break;

    case CEE_STELEM_I2:
        break;

    case CEE_STELEM_I4:
        break;

    case CEE_STELEM_I8:
        break;

    case CEE_STELEM_I:
        break;

    case CEE_STELEM_R4:
        break;

    case CEE_STELEM_R8:
        break;

    case CEE_STELEM_REF:
        break;
    
    case CEE_CKFINITE:
        break;

    case CEE_SWITCH:
		// switch <unsigned int32> <int32>...<int32> : 0x45,
		// <unsigned int32>  is # of cases
		// branch to one of the <int32> offsets ( 1..# of cases )
		unsigned int nOfCases;
		unsigned int i;
		GET_VAL( *ppIP, nOfCases, unsigned int ); // get <unsigned int32> # of cases
		_ASSERTE( nOfCases > 0 );
		for ( i = 0; i < nOfCases; ++i )
		{
			GET_VAL( *ppIP, i4, signed int ); // get <int32> as param
		}
        break;             

    case CEE_THROW:
        break;

    case CEE_RETHROW: 
        break;

    case CEE_TAILCALL:  // tail : 0xFE 0x14
		_ASSERTE( !"CEE_TAILCALL" );

		// a tail call looks like this:
		// tail.
		// callvirt int32 SomeMethodThatReturnsInt32()
		// ret // is "RET" necessary for verification?

		// 1. get the next instruction // e.g. callvirt
		unsigned char _val;
		GET_VAL( *ppIP, _val, unsigned char );
		*pOpCode = OPCODE( _val );
		
		// the next opcode should be one of them: call or calli or callvirt
		_ASSERTE( (*pOpCode == CEE_CALL) || (*pOpCode == CEE_CALLI) || (*pOpCode == CEE_CALLVIRT) );
		// there should be room for method's token (4 bytes) + "RET" (1 byte)
		_ASSERTE( ( (pILCodeEnd - *ppIP) > 4 ) );
		// is "RET"?
		_ASSERTE( *ppIP[ 4 ] == CEE_RET );
		
		// 2. Get method's token // int32 SomeMethodThatReturnsInt32()
		GET_VAL( *ppIP, token, unsigned int );
        
		break;        

    case CEE_UNALIGNED: 
        // ignore the alignment
        unsigned __int8 alignment;
        GET_VAL( *ppIP, alignment, unsigned __int8 );
        break;

    case CEE_VOLATILE:
        break;      // do nothing

    default:
		ILXLOGMSG2(ILXLOG_WARNING, _T("Cannot verify IL opcode '%s' (0x%02X)."), 
			*pOpCode < CEE_COUNT ? g_ILcodes[*pOpCode].szName : "<UNKNOWN>", 
			*pOpCode );
		return E_FAIL;

    } /*switch*/

	return S_OK;
}

// Performs a binary comparison of the current opcode
bool OpcodesAreEqual(	const opcode_iterator& itLhs,
						const opcode_iterator& itRhs)
{
	OPCODE opcodeLhs = itLhs.getOpCode();
	OPCODE opcodeRhs = itRhs.getOpCode();

	if (opcodeLhs != opcodeRhs)
	{
		return false;
	}
	else if (opcodeLhs == CEE_PREFIX1)
	{
		// Handle two-byte instuctions
		_ASSERTE(opcodeRhs == CEE_PREFIX1);
		opcodeLhs = itLhs.getNextOpCode(); // Decode next instr
		opcodeRhs = itRhs.getNextOpCode(); // Decode next instr

		return opcodeLhs == opcodeRhs;
	}
	else
	{
		return true;
	}
}

int SigCmp(	const PCCOR_SIGNATURE lhs, 
			ULONG lhsLen,
			IMetaDataImport* piImportLhs,
			const PCCOR_SIGNATURE rhs,
			ULONG rhsLen,
			IMetaDataImport* piImportRhs,
			const WCHAR* szSigType,
			const WCHAR* szName)
{
	if (lhs == rhs)
	{
		_ASSERTE(lhsLen == rhsLen);
		return 0;
	}
	else if (lhsLen != rhsLen && !(szSigType && szName))
	{
		return lhsLen - rhsLen;
	}

	_ASSERTE(lhsLen >= 0 && rhsLen >= 0);
	_ASSERTE(piImportLhs && piImportRhs);
 
	WCHAR szBufLhs[2048];
	szBufLhs[0] = L'\0';
	Sig2String(piImportLhs, lhs, lhsLen, szBufLhs, (sizeof(szBufLhs)/sizeof(*szBufLhs))-1);

	WCHAR szBufRhs[2048];
	szBufRhs[0] = L'\0';
	Sig2String(piImportRhs, rhs, rhsLen, szBufRhs, (sizeof(szBufRhs)/sizeof(*szBufRhs))-1);

	int cmp = wcscmp(szBufLhs, szBufRhs);
	if (cmp != 0 && szSigType && szName)
	{
		ILXLOGMSG2(ILXLOG_INFO, _T("%ls signature size has changed for method '%ls':"), 
			szSigType, szName);
		ILXLOGMSG2(ILXLOG_INFO, _T("  Old: %ls (%d bytes)"), szBufLhs, lhsLen);
		ILXLOGMSG2(ILXLOG_INFO, _T("  New: %ls (%d bytes)"), szBufRhs, rhsLen);
	}

	return cmp;
}

bool InstructionsAreEqual(	const opcode_iterator& itLhs,
							const opcode_iterator& itRhs)
{
	return memcmp(itLhs.m_pIPCur, itRhs.m_pIPCur, 
		itLhs.m_pIPStart - itLhs.m_pIPCur) == 0;
}
