///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/ILHelper.h 4     12/08/05 7:54p Eaddy $
//
//    File name:	ILHelper.h
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/ILHelper.h $
 * 
 * 4     12/08/05 7:54p Eaddy
 * o Fixed signature comparison
 * 
 * 3     12/07/05 7:51p Eaddy
 * o Can now filter on operand type as well as opcode
 * o Added getNextOpCode and sigCmp
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2004/10/26 04:36:12  me133
 * no message
 *
 * Revision 1.3  2004/10/25 01:03:48  me133
 * o Method IL comparison is now complete and correct
 *
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#ifndef INCLUDED_IL_HELPER_H
#define INCLUDED_IL_HELPER_H

#include <cor.h>				// Needed to compile corprof.h!

#include <iterator>		// STL iterator<>

// Forward declarations
class CMethodInfo;

// openum.h from Rotor
typedef enum opcode_format_t 
{
	Uninitialized	= -1, // Used to indicate 'uninitialized'
	InlineNone		= 0,	// no inline args       
	InlineVar		= 1,	// local variable       (U2 (U1 if Short on))
	InlineI			= 2,	// an signed integer    (I4 (I1 if Short on))
	InlineR			= 3,	// a real number        (R8 (R4 if Short on))
	InlineBrTarget	= 4,    // branch target        (I4 (I1 if Short on))
	InlineI8		= 5,
	InlineMethod	= 6,   // method token (U4)
	InlineField		= 7,   // field token  (U4)
	InlineType		= 8,   // type token   (U4)
	InlineString	= 9,   // string TOKEN (U4)
	InlineSig		= 10,  // signature tok (U4)
	InlineRVA		= 11,  // ldptr token  (U4)
	InlineTok		= 12,  // a meta-data token of unknown type (U4)
	InlineSwitch	= 13,  // count (U4), pcrel1 (U4) .... pcrelN (U4)
	InlinePhi		= 14,  // count (U1), var1 (U2) ... varN (U2) 

	// WATCH OUT we are close to the limit here, if you add
	// more enumerations you need to change ShortIline definition below

	// The extended enumeration also encodes the size in the IL stream
	ShortInline 	= 16,						// if this bit is set, the format is the 'short' format
	PrimaryMask   	= (ShortInline-1),			// mask these off to get primary enumeration above
	ShortInlineVar 	= (ShortInline + InlineVar),
	ShortInlineI	= (ShortInline + InlineI),
	ShortInlineR	= (ShortInline + InlineR),
	ShortInlineBrTarget = (ShortInline + InlineBrTarget),
	InlineOpcode	= (ShortInline + InlineNone),    // This is only used internally.  It means the 'opcode' is two byte instead of 1
} OPCODE_FORMAT;

// Uses #include file magic to create array of opcodes

struct OPCODE_ENCODING
{
    BYTE OpCode1;
    BYTE OpCode2;
	ULONG Len;
	opcode_format_t operandFormat;
	const char* szName;
};

#define OPDEF( id, name, pop, push, operandFormat, type, len, OpCode1, OpCode2, cf ) { OpCode1, OpCode2, len, operandFormat, name },

// array of IL opcodes
const OPCODE_ENCODING g_ILcodes [] = 
{
#include "opcode.def"
};
#undef OPDEF

// Use #include file magic to create enumeration of opcodes

#define OPDEF( id, s, pop, push, args, type, l, OpCode1, OpCode2, ctrl ) id,
typedef enum enumOpcode
{
#include "opcode.def"
  CEE_COUNT,  /* number of instructions and macros pre-defined */
} OPCODE;
#undef OPDEF

// opcode_iterator

class opcode_iterator : public std::iterator<std::input_iterator_tag,
                                            LPBYTE>
{
public:

	// Set opcodeFilter to CEE_COUNT to indicate no opcode filter
	// Se opcodeFormatFilter to Uninitialized to indicate no format filter

	opcode_iterator(LPBYTE pIPStart, 
					LPBYTE pIPEnd, 
					OPCODE opcodeFilter=CEE_COUNT,
					OPCODE_FORMAT opcodeFormatFilter=Uninitialized);

	opcode_iterator(const CMethodInfo& methodInfo,
					OPCODE opcodeFilter=CEE_COUNT,
					OPCODE_FORMAT opcodeFormatFilter=Uninitialized);

	opcode_iterator& operator++();
    LPBYTE operator*() const;
	bool end() const;

	bool isOpCodeEqual(OPCODE opcodeFirst, OPCODE opcodeSecond=CEE_COUNT) const;

	OPCODE getOpCode() const;
	OPCODE getNextOpCode() const;
	LPBYTE getOpCodeArg() const;

	// Returns true if there was an IL parse error, ie we encountered
	// an unknown opcode
	bool isError() const;

	// Performs a binary comparison of the current opcode and
	// its parameters
	friend bool InstructionsAreEqual(	const opcode_iterator& itLhs,
										const opcode_iterator& itRhs);

private: // NOT IMPLEMENTED
	opcode_iterator& operator=(const opcode_iterator& that);
	opcode_iterator(const opcode_iterator&);

private: // DATA
	LPBYTE m_pIPStart;
	LPBYTE m_pIPEnd;
	LPBYTE m_pIPCur;
	OPCODE m_opcodeFilter;
	OPCODE_FORMAT m_opcodeFormatFilter;
	HRESULT m_hrLast;
};

int SigCmp(	const PCCOR_SIGNATURE lhs, 
			ULONG lhsLen,
			IMetaDataImport* piImportLhs,
			const PCCOR_SIGNATURE rhs,
			ULONG rhsLen,
			IMetaDataImport* piImportRhs,
			const WCHAR* szSigType,
			const WCHAR* szName);

// Performs a binary comparison of the current opcode
bool OpcodesAreEqual(	const opcode_iterator& itLhs,
						const opcode_iterator& itRhs);

#endif // INCLUDED_IL_HELPER_H
