///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/IncrFile.cpp 8     5/28/05 1:20p Eaddy $
//
//    File name:	IncrFile.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/16/04
//
//      Purpose:	
//
//  Description:	Structure declarations, code and some comments
//					were taken from incbuild.cpp/.h from Rotor.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/IncrFile.cpp $
 * 
 * 8     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2004/10/26 04:38:01  me133
 * o Added kludges to try to handle problems with parsing incr file
 *
 * Revision 1.3  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.2  2004/10/17 00:24:55  me133
 * o Now provides a class-based iface to get to the incremental file info
 *
 * Revision 1.1  2004/10/16 11:17:28  me133
 * no message
 */

#include "StdAfx.h"

#include "IncrFile.h"			// Me

#include "Helpers.h"			// AppendXXX()

#include <io.h>					// _open_osfhandle()

using namespace std;

// CONSTANTS

// From incbuild.cpp/.h
#define INCBUILD_HEADERSTRING "Incremental Build Data for C# Compiler -- this file stores data to enable fast rebuilds.\n"
#define INCBUILD_ID  0xC380A349     // Identify this as a incremental build file.
#define INCBUILD_VERSION 7          // Version number of this format.

#define NUM_GLOBAL_FIELD_TOKENS		8	// number of elements in INFILESYM.globalFields.pTokens (from symbol.h)

// HELPER FUNCTIONS

static bool ReadString(FILE* file, LPWSTR s, unsigned cbMax);	// (from incbuild.cpp)
static void SkipAllocMapRead(FILE *file);

//////////////////////////////////
// INC BUILD FILE IMPLEMENTATION
//////////////////////////////////

CIncrBuildFile::CIncrBuildFile()
	: CCompileInfo()
	, m_incfilehdr()
	, m_szDefines()
	, m_outputFileMetadata()
	, m_outputFileSource()
	, m_nSourceFilesStartIndex(0)
	, m_vecInputFiles()
	, m_vecResourceFiles()
{ 
	m_szDefines[0] = '\0';

	// Only the C# compiler produces the incremental build files we
	// can parse
	m_strCompilerName = L"csc";
}

CIncrBuildFile::~CIncrBuildFile()
{ }

const WCHAR* CIncrBuildFile::getFileExtension() const
{
	return L".incr";
}

HRESULT CIncrBuildFile::read(FILE* in, bool bIsUnicode, const WCHAR* szFilePath)
{
	HRESULT hr = S_OK;

	_ASSERTE(in);
	if (!in) return E_POINTER;

    WCHAR buffer[MAX_PATH];
	buffer[0] = '\0';

    // Read and check signature/version.
    _ASSERTE(sizeof(INCBUILD_HEADERSTRING) < MAX_PATH);
    fread(buffer, sizeof(INCBUILD_HEADERSTRING), 1, in);
    if (_getw(in) != (int) INCBUILD_ID || _getw(in) != (int) INCBUILD_VERSION)
	{
		fclose(in);
		ILXLOGMSG1(ILXLOG_ERROR, _T("Incremental compilation file is invalid or incompatible: '%ls'."),
			szFilePath);
        return E_FAIL;
	}

    // Get fixed size header
    fread(&m_incfilehdr, sizeof(m_incfilehdr), 1, in);

	if (m_incfilehdr.emitChecked)
	{
		AppendMsg(m_strCompilerOptions, L"/checked", L" ");
	}

	if (m_incfilehdr.emitUnsafe)
	{
		AppendMsg(m_strCompilerOptions, L"/unsafe", L" ");
	}

	m_bDebugFlag = m_incfilehdr.emitDebugInfo;

	if (m_incfilehdr.optUnusedVars)
	{
		AppendMsg(m_strCompilerOptions, L"/o", L" ");
	}

	if (m_incfilehdr.warnAsErrors)
	{
		AppendMsg(m_strCompilerOptions, L"/warnaserror", L" ");
	}

	buffer[0] = '\0';
    // read in preprocessor definitions(one bug ';' delimited string)
    if (!ReadString(in, m_szDefines, sizeof(m_szDefines)))
	{
		fclose(in);
		ILXLOGMSG1(ILXLOG_ERROR, _T("Failed to read preprocessor definitions from incr file: '%ls'."),
			szFilePath);
        return E_FAIL;
	}

    // Go through each output file in the database.
	hr = m_outputFileMetadata.read(in, &m_vecInputFiles, &m_vecResourceFiles);
	if (FAILED(hr)) return hr;
	
	for(TypeVecInputFiles::const_iterator it = m_vecInputFiles.begin();
		it != m_vecInputFiles.end();
		++it)
	{
		AppendMsg(m_strCompilerOptions, L"/r:", L" ");
		m_strCompilerOptions.append((*it).getFileName());
	}

	if (hr == S_OK) // S_OK means read the next output file
	{
		m_nSourceFilesStartIndex = m_vecInputFiles.size();

		hr = m_outputFileSource.read(in, &m_vecInputFiles, &m_vecResourceFiles);
		if (FAILED(hr)) return hr;
		_ASSERTE(hr == S_FALSE); // S_FALSE means stop reading output files

		for(size_t i = m_nSourceFilesStartIndex; i < m_vecInputFiles.size(); ++i)
		{
			m_vecSourceFiles.push_back(m_vecInputFiles[i].getFileName());
		}
	}

	if (m_outputFileSource.getMainType())
	{
		AppendMsg(m_strCompilerOptions, L"/main:", L" ");
		m_strCompilerOptions.append(m_outputFileSource.getMainType());
	}

	return S_OK;
}
/*
void CIncrBuildFile::debugPrint(int nIndentLevel) const
{
	USES_CONVERSION;

	char szMsg[1000];
	char szOutputFileInfo[1000];

	// Print header
	m_incfilehdr.debugPrint(nIndentLevel+1);

	// Print defines
	szMsg[0] = '\0';
	AppendIndent(szMsg, nIndentLevel+1);
	AppendString(szMsg, "PreprocessorDefines", W2CA(m_szDefines));
	strcat(szMsg, "\n");
	OutputDebugString(szMsg);

	// Print metadata files

	szOutputFileInfo[0] = '\0';
	m_outputFileMetadata.appendString(szOutputFileInfo);

	szMsg[0] = '\0';
	AppendIndent(szMsg, nIndentLevel+1);
	AppendString(szMsg, "Metadata Info", szOutputFileInfo);
	strcat(szMsg, "\n");
	OutputDebugString(szMsg);

	// Print source files

	szOutputFileInfo[0] = '\0';
	m_outputFileSource.appendString(szOutputFileInfo);

	szMsg[0] = '\0';
	AppendIndent(szMsg, nIndentLevel+1);
	AppendString(szMsg, "Source File Info", szOutputFileInfo);
	strcat(szMsg, "\n");
	OutputDebugString(szMsg);

	// Print input files

	for(TypeVecInputFiles::const_iterator it = m_vecInputFiles.begin();
		it != m_vecInputFiles.end();
		++it)
	{
		(*it).debugPrint(nIndentLevel+2);
	}

	// Print resource files

	for(TypeVecResourceFiles::const_iterator it = m_vecResourceFiles.begin();
		it != m_vecResourceFiles.end();
		++it)
	{
		(*it).debugPrint(nIndentLevel+2);
	}
}
*/

/////////////////////////////////////////
// INC BUILD OUTPUT FILE IMPLEMENTATION
/////////////////////////////////////////

CIncrBuildOutputFile::CIncrBuildOutputFile()
	: m_outfilehdr()
	, m_szFileName()
	, m_szMainType()
{ 
	m_szFileName[0] = '\0';
	m_szMainType[0] = '\0';
}

CIncrBuildOutputFile::~CIncrBuildOutputFile()
{ }

/*
 * Read the incremental build info for a single outfile.
 * May be the metadata outfile, or a regular source outfile
 */
HRESULT CIncrBuildOutputFile::read(	FILE* in, 
									TypeVecInputFiles* pvecInputFiles,
									TypeVecResourceFiles* pvecResourceFiles)
{
	HRESULT hr = S_OK;

	_ASSERTE(in && pvecInputFiles && pvecResourceFiles);
	if (!in || !pvecInputFiles || !pvecResourceFiles) return E_POINTER;

    // read the output file header
	fread(&m_outfilehdr, sizeof(m_outfilehdr), 1, in);

    // read in the output file name
    if (!ReadString(in, m_szFileName, sizeof(m_szFileName)))
	{
		return ILXLOG0(E_FAIL, ILXLOG_ERROR, _T("Failed to read output filename from incr file."));
	}

    // check the /main switch
    if (m_outfilehdr.bExplicitMain) 
	{
        // read in the old /main option
        if (!ReadString(in, m_szMainType, sizeof(m_szMainType)))
		{
			return ILXLOG0(E_FAIL, ILXLOG_ERROR, _T("Failed to read main class from incr file."));
		}
	}

	if (*m_szFileName)
	{
		SkipAllocMapRead(in);
		SkipAllocMapRead(in);
	}

	size_t cCurrentInputFiles = pvecInputFiles->size();
	pvecInputFiles->resize(cCurrentInputFiles + m_outfilehdr.cInputFiles);

    // Go through each input file
    for (DWORD i = 0; i < m_outfilehdr.cInputFiles; ++i) 
	{
		// Read the fixed size part, and the filename
		hr = (*pvecInputFiles)[cCurrentInputFiles+i].read(	in, 
															m_outfilehdr.cInputFiles);
		if (FAILED(hr)) return hr;
    }

	size_t cCurrentResourceFiles = pvecResourceFiles->size();

	pvecResourceFiles->resize(cCurrentResourceFiles + m_outfilehdr.cResFiles);

    for (DWORD i = 0; i < m_outfilehdr.cResFiles; ++i)
	{
        // Read the fixed size part, and the filename
		hr = (*pvecResourceFiles)[cCurrentResourceFiles+i].read(in);
		if (FAILED(hr)) return hr;
	}

	if (*m_szFileName)
	{
		return S_FALSE;
	}
	else
	{
		return S_OK;
	}
}

const WCHAR* CIncrBuildOutputFile::getMainType() const
{
	return m_szMainType;
}

void CIncrBuildOutputFile::appendString(char* szMsg) const
{
	USES_CONVERSION;

	m_outfilehdr.appendString(szMsg);

	if (*m_szFileName)
	{
		AppendString(szMsg, "Output file", W2CA(m_szFileName));
	}

	if (*m_szMainType)
	{
		AppendString(szMsg, "Main Type", W2CA(m_szMainType));
	}
}

void CIncrBuildOutputFile::debugPrint(int nIndentLevel) const
{
	char szMsg[1000];
	szMsg[0] = '\0';

	AppendIndent(szMsg, nIndentLevel);

	appendString(szMsg);

	OutputDebugString(szMsg);
}

/////////////////////////////////////////
// INCR BUILD INPUT FILE IMPLEMENTATION
/////////////////////////////////////////

CIncrBuildInputFile::CIncrBuildInputFile()
	: m_infilehdr()
	, m_vecTypes()
	, m_szFileName()
{ 
	m_szFileName[0] = '\0';
}

CIncrBuildInputFile::~CIncrBuildInputFile()
{ }

HRESULT CIncrBuildInputFile::read(FILE* in, unsigned int cInputFiles)
{
	HRESULT hr = S_OK;

	_ASSERTE(*m_szFileName == '\0'); // Only call read() once

	_ASSERTE(in);
	if (!in) return E_POINTER;

	// Read the header
	fread(&m_infilehdr, sizeof(m_infilehdr), 1, in);

	// Read the filename
	ReadString(in, m_szFileName, sizeof(m_szFileName));

	_ASSERTE(*m_szFileName);

	if (!m_infilehdr.isSource)
	{
		return S_OK;
	}

	SkipAllocMapRead(in);

	m_vecTypes.resize(m_infilehdr.cTypeDefs);

    for (ULONG k = 0; k < m_infilehdr.cTypeDefs; ++k ) 
	{
        hr = m_vecTypes[k].read(in, cInputFiles);
		if (FAILED(hr)) return hr;
	}

    // read the Global FieldDefs
	if (m_infilehdr.cGlobalFields > 0)
	{
		mdToken* pTokens = 
			(mdToken*) malloc(sizeof(mdToken) * m_infilehdr.cGlobalFields);

		// We should do fseek() instead to avoid malloc!!!
		fread(pTokens, sizeof(mdToken), m_infilehdr.cGlobalFields, in);

		// We aren't doing anything with them
		free(pTokens);
		pTokens = 0;
	}

	return S_OK;
}

bool CIncrBuildInputFile::isSource() const
{
	return m_infilehdr.isSource;
}

const WCHAR* CIncrBuildInputFile::getFileName() const
{
	return m_szFileName;
}

void CIncrBuildInputFile::debugPrint(int nIndentLevel) const
{
	USES_CONVERSION;

	char szIndent[10];
	*szIndent = '\0';
	AppendIndent(szIndent, nIndentLevel);

	char szMsg[1000];
	sprintf(szMsg, 
		"%sInput file %d: %s\n",
		szIndent, m_infilehdr.id+1, W2CA(m_szFileName));

	AppendIndent(szMsg, nIndentLevel+1);
	m_infilehdr.appendString(szMsg);
	strcat(szMsg, "\n");

	OutputDebugString(szMsg);

	for(TypeVecTypes::const_iterator it = m_vecTypes.begin();
		it != m_vecTypes.end();
		++it)
	{
		(*it).debugPrint(nIndentLevel+1);
	}
}

////////////////////////////////////////////
// INCR BUILD RESOURCE FILE IMPLEMENTATION
////////////////////////////////////////////

CIncrBuildResourceFile::CIncrBuildResourceFile()
	: m_resfilehdr()
	, m_szResourceName()
	, m_szFileName()
{ 
	m_szResourceName[0] = '\0';
	m_szFileName[0] = '\0';
}

CIncrBuildResourceFile::~CIncrBuildResourceFile()
{ }

HRESULT CIncrBuildResourceFile::read(FILE* in)
{
	_ASSERTE(in);
	if (!in) return E_POINTER;

	// Read the header
	fread(&m_resfilehdr, sizeof(m_resfilehdr), 1, in);

	// Read the resource name
	ReadString(in, m_szResourceName, sizeof(m_szResourceName));

	// Read the filename
	ReadString(in, m_szFileName, sizeof(m_szFileName));

	return S_OK;
}

void CIncrBuildResourceFile::debugPrint(int nIndentLevel) const
{
	USES_CONVERSION;

	char szHeader[1000];
	*szHeader = '\0';

	AppendIndent(szHeader, nIndentLevel);
	m_resfilehdr.appendString(szHeader);

	char szIndent[10];
	*szIndent = '\0';
	AppendIndent(szIndent, nIndentLevel);

	char szMsg[1000];
	sprintf(szMsg, 
		"%sResource name: %s\n"
		"%sResource file: %s\n"
		"%s\n", 
		szIndent, W2CA(m_szResourceName), 
		szIndent, W2CA(m_szFileName), 
		szHeader);

	OutputDebugString(szMsg);
}

// HELPER FUNCTIONS

/*
 * Read a length-prefix Unicode string.
 * Returns false if too long for the buffer.
 */
bool ReadString(FILE* file, LPWSTR s, unsigned cbMax)
{
    unsigned cch = _getw(file);

    if ((cch + 1) * sizeof(WCHAR) > cbMax || cch * sizeof(WCHAR) > cbMax) {
        // Can't read it, just seek past it.
        fseek(file, cch * sizeof(WCHAR), SEEK_CUR);
        return false;
    }
    else {
        // Read filename and zero-terminate.
        fread(s, sizeof(WCHAR), cch, file);
        s[cch] = L'\0';
        return true;
    }
}

void SkipAllocMapRead(FILE *file)
{
    ULONG nUsed = 0;
    fread(&nUsed, sizeof(nUsed), 1, file);

	RVAALLOCATION* pRvaAllocation = (RVAALLOCATION*) malloc(sizeof(RVAALLOCATION) * nUsed);

	// We should do fseek() instead to avoid malloc!!!
    fread(pRvaAllocation, sizeof(RVAALLOCATION), nUsed, file);

	// Since we aren't using it
	free(pRvaAllocation);
	pRvaAllocation = 0;
}
