///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/IncrFile.h 6     5/28/05 1:20p Eaddy $
//
//    File name:	IncrFile.h
//   Created By:	Marc Eaddy
//   Created on:	10/16/04
//
//      Purpose:	
//
//  Description:	Structure declarations taken from incbuild.cpp/.h
//					from Rotor
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/IncrFile.h $
 * 
 * 6     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2004/10/26 04:38:01  me133
 * o Added kludges to try to handle problems with parsing incr file
 *
 * Revision 1.3  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.2  2004/10/17 00:24:55  me133
 * o Now provides a class-based iface to get to the incremental file info
 *
 * Revision 1.1  2004/10/16 11:17:28  me133
 * no message
 */

#ifndef INCLUDED_INCR_FILE_H
#define INCLUDED_INCR_FILE_H

#include "CompileInfo.h"		// CCompileInfo

#include "IncrFileStructs.h"	// INCBUILD_HDR, etc.

#include <vector>				// STL vector

// Type defines

typedef std::vector<INCREMENTALTYPE>		TypeVecTypes;

class CIncrBuildInputFile
{
public:
	CIncrBuildInputFile();
	~CIncrBuildInputFile();

	bool isSource() const;

	const WCHAR* getFileName() const;

	HRESULT read(FILE* in, unsigned int cInputFiles);
	void debugPrint(int nIndentLevel=0) const;

private: // DATA
	INCBUILD_INFILE m_infilehdr;
    WCHAR m_szFileName[MAX_PATH];
	TypeVecTypes m_vecTypes;
};

class CIncrBuildResourceFile
{
public:
	CIncrBuildResourceFile();
	~CIncrBuildResourceFile();

	HRESULT read(FILE* in);
	void debugPrint(int nIndentLevel=0) const;

private: // DATA
	INCBUILD_RESFILE m_resfilehdr;
    WCHAR m_szResourceName[MAX_PATH];
    WCHAR m_szFileName[MAX_PATH];
};

// Type defines
typedef std::vector<CIncrBuildInputFile>	TypeVecInputFiles;
typedef TypeVecInputFiles::const_iterator	TypeVecInputFilesCIter;
typedef std::vector<CIncrBuildResourceFile>	TypeVecResourceFiles;

class CIncrBuildOutputFile
{
public:
	CIncrBuildOutputFile();
	~CIncrBuildOutputFile();

	HRESULT read(	FILE* in, 
					TypeVecInputFiles* pvecInputFiles,
					TypeVecResourceFiles* pvecResourceFiles);

	const WCHAR* getMainType() const;

	void appendString(char* szMsg) const;
	void debugPrint(int nIndentLevel=0) const;

private: // DATA
    INCBUILD_OUTFILE m_outfilehdr;
    WCHAR m_szFileName[MAX_PATH];
	WCHAR m_szMainType[MAX_PATH];
};

class CIncrBuildFile : public CCompileInfo
{
public:
	CIncrBuildFile();
	virtual ~CIncrBuildFile();

	// ICompileInfo implementation
	HRESULT read(FILE* in, bool bIsUnicode, const WCHAR* szFilePath);
	const WCHAR* getFileExtension() const;

private: // HELPER FUNCTIONS
	HRESULT ReadOutputFileInfo(FILE* in, bool bFirstCall);

private: // NOT IMPLEMENTED
	CIncrBuildFile(const CIncrBuildFile&);
	CIncrBuildFile& operator=(const CIncrBuildFile&);

private: // DATA
    INCBUILD_HDR m_incfilehdr;
    WCHAR m_szDefines[MAX_PATH];

	CIncrBuildOutputFile m_outputFileMetadata;
	CIncrBuildOutputFile m_outputFileSource;

	size_t m_nSourceFilesStartIndex;

	TypeVecInputFiles m_vecInputFiles;
	TypeVecResourceFiles m_vecResourceFiles; 
};

#endif // INCLUDED_INCR_FILE_H
