///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/IncrFileStructs.cpp 3     5/28/05 1:20p Eaddy $
//
//    File name:	IncrFileStructs.h
//   Created By:	Marc Eaddy
//   Created on:	10/16/04
//
//      Purpose:	
//
//  Description:	Structure declarations taken from incbuild.cpp/.h
//					from Rotor
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/IncrFileStructs.cpp $
 * 
 * 3     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2005/01/14 23:03:06  me133
 * o Trying to fix incr file parsing problem with Text3D vs hello
 *
 * Revision 1.3  2004/10/26 04:38:01  me133
 * o Added kludges to try to handle problems with parsing incr file
 *
 * Revision 1.2  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.1  2004/10/17 00:24:55  me133
 * o Now provides a class-based iface to get to the incremental file info
 */

#include "StdAfx.h"

#include "IncrFileStructs.h"	// Me

#include "Helpers.h"			// AppendXXX()

// HELPER FUNCTIONS

static unsigned int RawSizeInBytes(unsigned int numberOfBits);	// (from fncpbind.cpp)	

// CONSTANTS

#define BIT_SHIFT					5   // Shift by this much (from fncpbind.cpp)
#define MAX_SIZE					32  // This is the size of a 'slot' (from fncpbind.cpp)

/*

******* IMPORTANT: The INCBUILD_VERSION constant MUST be changed every
******* IMPORTANT: time a file format change is made to the incremental
******* IMPORTANT: build file.

The file format for the incremental build file:

ascii string "Incremental Build Data for C# Compiler -- this file stores data to enable fast rebuilds."
DWORD INCBUILD_ID
DWORD INCBUILD_VERSION
struct INCBUILD_HDR
length-prefixed unicode strings: preprocessor definitions(';' seperated list).

for each output file:
  struct INCBUILD_OUTFILE
  length-prefixed unicode string: output file name
  ALLOCMAP for outfile
  list of input files

for each input file:
  struct INCBUILD_INFILE
  length-prefixed unicode string: input file name.
  ALLOCMAP for inputfile
  list of types
  list of Global FieldDef tokens (from <PrivateImplementationDetails>)

for each type
  typedef token of the type
  bitfield of input files dependant on the type

*/


////////////////////////////////
// INCBUILD_HDR IMPLEMENTATION
////////////////////////////////

void INCBUILD_HDR::appendString(char* szMsg) const
{
	AppendFlag(szMsg,	"DebugInfo",	emitDebugInfo);
	AppendFlag(szMsg,	"Optimized",	optUnusedVars);
	AppendFlag(szMsg,	"WarnAsErrors",	warnAsErrors);
	AppendFlag(szMsg,	"Assemble",		assemble);
	AppendFlag(szMsg,	"Checked",		emitChecked);
	AppendFlag(szMsg,	"Unsafe",		emitUnsafe);
	AppendNum(szMsg,	"BuildCount",	cBuildCount);
	AppendNum(szMsg,	"OutputFiles",	cOutputFiles);
	AppendNum(szMsg,	"InputFiles",	cInputFiles);
}

void INCBUILD_HDR::debugPrint(int nIndentLevel) const
{
	char szMsg[1000];
	*szMsg = '\0';

	AppendIndent(szMsg, nIndentLevel);
	
	appendString(szMsg);
	
	strcat(szMsg, "\n");

	OutputDebugString(szMsg);
}

////////////////////////////////////
// INCBUILD_OUTFILE IMPLEMENTATION
////////////////////////////////////

void INCBUILD_OUTFILE::appendString(char* szMsg) const
{
	AppendTime(szMsg,		"OutputWritten",	timeOutputWritten);
	AppendTime(szMsg,		"DebugWritten",		timeDebugWritten);
	AppendNum(szMsg,		"InputFiles",		cInputFiles);
	AppendNum(szMsg,		"ResFiles",			cResFiles);
	AppendNumHex(szMsg,		"EntryClassToken",	entryClassToken);
	AppendFlag(szMsg,		"Dll",				bDll);
	AppendFlag(szMsg,		"Console",			bConsole);
	AppendFlag(szMsg,		"ExplicitMain",		bExplicitMain);
}

void INCBUILD_OUTFILE::debugPrint(int nIndentLevel) const
{
	char szMsg[1000];
	*szMsg = '\0';

	AppendIndent(szMsg, nIndentLevel);

	appendString(szMsg);

	strcat(szMsg, "\n");
	OutputDebugString(szMsg);
}

///////////////////////////////////
// INCBUILD_INFILE IMPLEMENTATION
///////////////////////////////////

void INCBUILD_INFILE::appendString(char* szMsg) const
{
	AppendFlag(szMsg,	"Source",		isSource);
	AppendFlag(szMsg,	"GlobalAttr",	hasGlobalAttr);
	AppendNum(szMsg,	"ID",			id);
	AppendTime(szMsg,	"LastChange",	timeLastChange);
	AppendNum(szMsg,	"TypeDefs",		cTypeDefs);
	AppendNum(szMsg,	"GlobalFields",	cGlobalFields);
}

void INCBUILD_INFILE::debugPrint(int nIndentLevel) const
{
	char szMsg[1000];
	*szMsg = '\0';

	AppendIndent(szMsg, nIndentLevel);

	appendString(szMsg);

	strcat(szMsg, "\n");
	OutputDebugString(szMsg);
}

////////////////////////////////////
// INCBUILD_RESFILE IMPLEMENTATION
////////////////////////////////////

void INCBUILD_RESFILE::appendString(char* szMsg) const
{
	AppendFlag(szMsg,	"Linked",		isLinked);
	AppendTime(szMsg,	"LastChange",	timeLastChange);
}

void INCBUILD_RESFILE::debugPrint(int nIndentLevel) const
{
	char szMsg[1000];
	*szMsg = '\0';

	AppendIndent(szMsg, nIndentLevel);

	appendString(szMsg);

	strcat(szMsg, "\n");
	OutputDebugString(szMsg);
}

///////////////////////////////////
// INCREMENTALTYPE IMPLEMENTATION
///////////////////////////////////

INCREMENTALTYPE::INCREMENTALTYPE()
	: token(0)
	, pBITSET_dependantFiles(0)
	, pAGGSYM_cls(0)
	, m_nNumBytesInBitSet(0)
{ }

INCREMENTALTYPE::~INCREMENTALTYPE()
{
	free(pBITSET_dependantFiles);
	free(pAGGSYM_cls);
}

HRESULT INCREMENTALTYPE::read(FILE* in, unsigned int cInputFiles)
{
	fread(&token, sizeof(token), 1, in);
    if (TypeFromToken(token) != mdtTypeDef)
	{
		return ILXLOG2(E_FAIL, ILXLOG_ERROR, 
				_T("Invalid token type in incr file: %d, expected %d."), 
				token, mdtTypeDef);
	}

	// NOTE: I am not interpretting the bitset currently (not sure
	// how). I need to do this in order to determine type/file
	// dependencies.

	_ASSERTE(!pBITSET_dependantFiles);

	m_nNumBytesInBitSet = RawSizeInBytes(cInputFiles);

	pBITSET_dependantFiles = (BYTE*) malloc(m_nNumBytesInBitSet);

	// We should do fseek() instead to avoid malloc!!!
	fread(pBITSET_dependantFiles, m_nNumBytesInBitSet, 1, in);

	//The problem seems to be here with Text3D. Need to read an extra DWORD here.
	return S_OK;
}

void INCREMENTALTYPE::appendString(char* szMsg) const
{
	AppendNumHex(szMsg,	"Token",			token);
	AppendBitSet(szMsg,	"DependantFiles",	pBITSET_dependantFiles, m_nNumBytesInBitSet);
	//AppendNumHex(szMsg,	"Cls",				(DWORD) pAGGSYM_cls);
}

void INCREMENTALTYPE::debugPrint(int nIndentLevel) const
{
	char szMsg[1000];
	*szMsg = '\0';

	AppendIndent(szMsg, nIndentLevel);

	appendString(szMsg);

	strcat(szMsg, "\n");
	OutputDebugString(szMsg);
}

// HELPER FUNCTIONS

unsigned int RawSizeInBytes(unsigned int numberOfBits)
{
    return ((numberOfBits + MAX_SIZE) >> BIT_SHIFT) * sizeof(DWORD_PTR);
}
