///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/IncrFileStructs.h 2     5/28/05 1:20p Eaddy $
//
//    File name:	IncrFileStructs.h
//   Created By:	Marc Eaddy
//   Created on:	10/16/04
//
//      Purpose:	
//
//  Description:	Structure declarations taken from incbuild.cpp/.h
//					from Rotor
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/IncrFileStructs.h $
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.2  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.1  2004/10/17 00:24:55  me133
 * o Now provides a class-based iface to get to the incremental file info
 */

#ifndef INCLUDED_INCR_FILE_STRUCTS_H
#define INCLUDED_INCR_FILE_STRUCTS_H

#include <cor.h>		// mdToken

#include <stdio.h>		// FILE

// Fixed length information about the output file.
struct INCBUILD_HDR
{
    // These variables indicate compiler options that must
    // match to do an incremental build.
    bool emitDebugInfo: 1;      // Debug info on?
    bool optUnusedVars: 1;      // Optimization on?
    bool warnAsErrors: 1;       // Warn as errors on?
    bool assemble: 1;           // emit an assembly
    bool emitChecked: 1;        // chcked or unchecked casts
    bool emitUnsafe: 1;         // allow unsafe constructs

    DWORD cBuildCount;          // Number of incremental builds without a full-rebuild
    DWORD cOutputFiles;         // Number of output files to follow.
    DWORD cInputFiles;          // Number of input files

	void appendString(char* szMsg) const;
	void debugPrint(int nIndentLevel) const;
};

struct INCBUILD_OUTFILE
{
    FILETIME timeOutputWritten; // Time output file was written.
    FILETIME timeDebugWritten;  // Time pdb file was written. (or all 0 to indicate no PDB info)
    DWORD cInputFiles;          // Number of input files
    DWORD cResFiles;            // Number of resource files
    mdToken entryClassToken;    // TypeDef of class containing entrypoint
    bool bDll : 1;              // Is this a DLL (or module)
    bool bConsole : 1;          // Is this a Console or WinApp
    bool bExplicitMain: 1;      // have a /main switch on the command line

	void appendString(char* szMsg) const;
	void debugPrint(int nIndentLevel) const;
};

// Fixed length information about an input file
struct INCBUILD_INFILE
{
    bool isSource: 1;           // true = source, false = metadata
    bool hasGlobalAttr: 1;
    DWORD id;                   // Input file ID number
    FILETIME timeLastChange;    // Time input file was last changed
    DWORD cTypeDefs;            // Number of typedefs owned by this input file.
    DWORD cGlobalFields;        // Number of global FieldDef tokens owned by this file.

	void appendString(char* szMsg) const;
	void debugPrint(int nIndentLevel) const;
};

// Fixed length information about a resource file
struct INCBUILD_RESFILE
{
    bool isLinked:1;
    FILETIME timeLastChange;    // Time input file was last changed

	void appendString(char* szMsg) const;
	void debugPrint(int nIndentLevel) const;
};

struct INCREMENTALTYPE
{
public:
	INCREMENTALTYPE();
	~INCREMENTALTYPE();
	
	HRESULT read(FILE* in, unsigned int cInputFiles);

	void appendString(char* szMsg) const;
	void debugPrint(int nIndentLevel) const;

private: // DATA
    mdTypeDef   token;
    BYTE*		pBITSET_dependantFiles; // set of files that depend on this type
    BYTE*		pAGGSYM_cls;

	unsigned int m_nNumBytesInBitSet;	// careful! This messes up the size!
};

// From incbuild.h
struct RVAALLOCATION
{
    ULONG offset;
    ULONG size;

    ULONG endOffset() const   { return offset + size; }
};

#endif // INCLUDED_INCR_FILE_STRUCTS_H
