///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/MetadataHelpers.cpp 8     12/14/05 12:58a Eaddy $
//
//    File name:	MetadataHelpers.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/MetadataHelpers.cpp $
 * 
 * 8     12/14/05 12:58a Eaddy
 * o Fixed release build compile error
 * 
 * 7     12/10/05 12:40a Eaddy
 * o Fixed bug where diff sigs were considered equal (szarray)
 * 
 * 6     12/09/05 9:01p Eaddy
 * o Now longer include "default" in the sig
 * 
 * 5     12/09/05 4:29p Eaddy
 * o Small sig formatting chgs
 * 
 * 4     12/08/05 8:28p Eaddy
 * o Added Sig2String
 * 
 * 3     12/07/05 7:52p Eaddy
 * o Added Token2TypeName
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.7  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 *
 * Revision 1.6  2004/10/24 19:33:10  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.5  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.4  2004/10/11 23:22:52  me133
 * o Patch now contains method name
 * o We now skip over methods we couldn't lookup (instead of crashing)
 * o Uses IsBadReadPtr() to avoid access violations when we fail to lookup
 * a method
 *
 * Revision 1.3  2004/10/10 07:40:51  me133
 * o Converted print statements to logging
 *
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "stdafx.h"

#include "MetadataHelpers.h"	// Me

#include <cor.h>				// IMAGE_DIRECTORY_ENTRY_COMHEADER

#include <dbghelp.h>			// ImageNtHeader()

#include <corhlpr.h>			// CQuickBytes

#include <wchar.h>				// swprintf

// MACROS

#define NumItems(s) (sizeof(s) / sizeof(s[0]))

// CONSTANTS

// Tables for mapping element type to text
static WCHAR* g_szMapElementType[] = 
{
	L"End",          // 0x0
	L"Void",         // 0x1
	L"Boolean",
	L"Char", 
	L"I1",
	L"UI1",
	L"I2",           // 0x6
	L"UI2",
	L"I4",
	L"UI4",
	L"I8",
	L"UI8",
	L"R4",
	L"R8",
	L"String",
	L"Ptr",          // 0xf
	L"ByRef",        // 0x10
	L"ValueClass",
	L"Class",
	L"CopyCtor",
	L"MDArray",      // 0x14
	L"GENArray",
	L"TypedByRef",
	L"VALUEARRAY",
	L"I",
	L"U",
	L"R",            // 0x1a
	L"FNPTR",
	L"Object",
	L"SZArray",
	L"GENERICArray",
	L"CMOD_REQD",
	L"CMOD_OPT",
	L"INTERNAL",
};

static const WCHAR* g_szCalling[] = 
{   
	L"",
	L"[C]",
	L"[STDCALL]",
	L"[THISCALL]",
	L"[FASTCALL]",
	L"[VARARG]",
	L"Field",
	L"Local Signature",
	L"Property",
	L"[UNMANAGED]",
};

///////////////////
// HELPER FUNCTIONS
///////////////////

static const WCHAR* TypeDefName(IMetaDataImport* piImport, mdTypeDef tok, WCHAR* buf, ULONG len);
static const WCHAR* TypeRefName(IMetaDataImport* piImport, mdTypeRef tok, WCHAR* buf, ULONG len);

///////////////////
// IMPLEMENTATION
///////////////////

// This function returns the address to the MapView of file and file size.
// Adapted from MetaInfo .NET sample.
HRESULT MyGetMapViewOfFile(	const char* szFile, 
							PBYTE *ppbMap, 
							DWORD *pdwFileSize)
{
	HRESULT hr = S_OK;

	if (!szFile || !*szFile) return E_POINTER;
	
	// Must specify one or the other
	if (!ppbMap && !pdwFileSize) return E_POINTER;

    HANDLE hFile = CreateFileA(szFile,
                               GENERIC_READ,
                               FILE_SHARE_READ,
                               NULL,
                               OPEN_EXISTING,
                               FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL); 
	VERIFY_COND1(hFile != INVALID_HANDLE_VALUE, ILXLOG_ERROR, _T("CreateFileA() failed: '%s'."), 
		szFile);

	if (pdwFileSize)
	{
	    DWORD dwHighSize = 0;
		*pdwFileSize = ::GetFileSize(hFile, &dwHighSize);
		DWORD dwLastError = ::GetLastError();
		if (*pdwFileSize == 0xFFFFFFFF && dwLastError != NO_ERROR)
		{
			::CloseHandle(hFile);
			return ILXLOG1(HRESULT_FROM_WIN32(dwLastError), ILXLOG_ERROR,
				_T("GetFileSize() failed: '%s'."), szFile);
		}

		_ASSERTE(dwHighSize == 0); // We don't support very big files
	}

	if (!ppbMap)
	{
		::CloseHandle(hFile);
		return S_OK;
	}

	HANDLE hMapFile = ::CreateFileMappingA(	hFile, 
											NULL, 
											PAGE_READONLY, 
											0, 
											0, 
											NULL);
	::CloseHandle(hFile);
	hFile = 0; // So we don't accidentally use it

	VERIFY_COND1(hMapFile, ILXLOG_ERROR, _T("CreateFileMappingA() failed: '%s'."), szFile);

    *ppbMap = (PBYTE) MapViewOfFile(hMapFile, FILE_MAP_READ, 0, 0, 0);
	::CloseHandle(hMapFile);
	hMapFile = 0;

	VERIFY_COND1(*ppbMap, ILXLOG_ERROR, _T("MapViewOfFile() failed: '%s'."), szFile);

	return S_OK;

} // void GetMapViewOfFile()

// Adapted from:
//		Roshan James, rosh_spark@gmx.net
//		May 2003
//		http://www27.brinkster.com/sparksite/work/trashbin/trashbin.htm
//
// ME: I had to modify his code to call ImageRvaToVa().  For some
// reason I couldn't get the FindObjMetaData in mdobj.cpp
// (in MetaInfo sample) to work (access violations).  I think it
// only works for .LIB files that have OBJ sections or something.
// I need it to work for EXEs and DLLs that have metadata.
//
HRESULT MyFindMetaData(	PVOID pImage, 
						DWORD dwFileLength,
						PVOID* ppMetaData, 
						ULONG* pcbMetaData)
{
	// Verify arguments
	if (!pImage || !ppMetaData || !pcbMetaData) return E_POINTER;

    // Get a pointer to the header and the first section.

    // Initialize out params
    *ppMetaData = NULL;
    *pcbMetaData = 0;

	_IMAGE_DOS_HEADER* pDosH = (_IMAGE_DOS_HEADER*) pImage;
	char* pPESign = (char*)pImage + pDosH->e_lfanew;	//e_lfanew contains the address of the rest of the header

	if(!(	pPESign[0] == 'P' &&  
			pPESign[1] == 'E' &&  
			pPESign[2] == 0 &&  
			pPESign[3] == 0	))
	{
		ILXLOGMSG(_T("This is not a PE file according to the file signature."),
			ILXLOG_ERROR);;
		return E_FAIL;
	}

	_IMAGE_FILE_HEADER* pCoffH = (_IMAGE_FILE_HEADER*)(pPESign + 4) ; 	//4 byte PE signature
	_IMAGE_OPTIONAL_HEADER* pOptH = (_IMAGE_OPTIONAL_HEADER*)((char *)pCoffH + sizeof(_IMAGE_FILE_HEADER)); //these lie one after the other

	PIMAGE_NT_HEADERS pNtHeader = ::ImageNtHeader(pImage);
	if (!pNtHeader)
	{
		HRESULT hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG(hr, _T("ImageNtHeader failed."), ILXLOG_ERROR);
	}

	//these are content specific to a managed file
	_IMAGE_DATA_DIRECTORY* pDDCorHdr = &(pOptH->DataDirectory[IMAGE_DIRECTORY_ENTRY_COMHEADER]);
	if (!pDDCorHdr) return E_FAIL;

	//if this file is a managed file then this entry should exist
	IMAGE_COR20_HEADER* pCorHdr = 
		(IMAGE_COR20_HEADER*) ImageRvaToVa(	pNtHeader, 
											pImage, 
											pDDCorHdr->VirtualAddress, 
											0);
	if (!pCorHdr) return E_FAIL;

	// KLUDGE: Sometimes the CLR treates the RVA as a virtual address (VA) and
	// sometimes as an offset.  For example, if the CLR is loading
	// assemblies it seems to treat rva as an offset. If we are mapping
	// PE files directly, the rva appears to be a VA.  If ImageRvaToVa()
	// doesn't appear to work we try treating it as an offset.  This is
	// very kludgy but I don't know another way to do this.
	// See CorMap::ReadHeaders() in cormap.cpp in Rotor to see an example.
	if (pCorHdr->MajorRuntimeVersion == 0)
	{
		IMAGE_COR20_HEADER* pCorHdr2 = 
			(IMAGE_COR20_HEADER*) pImage + pDDCorHdr->VirtualAddress;
		if (pCorHdr->MajorRuntimeVersion != 0)
		{
			pCorHdr = pCorHdr2;
		}
	}

	*ppMetaData = (PVOID) ImageRvaToVa(		pNtHeader, 
											pImage, 
											pCorHdr->MetaData.VirtualAddress, 
											0);
	*pcbMetaData = pCorHdr->MetaData.Size;

	return S_OK;
}

#define CASE_TYPE(type) case mdt##type: return L#type;

const WCHAR* Token2TypeName(mdToken tok)
{
	switch(TypeFromToken(tok))
	{
		CASE_TYPE(Module);
		CASE_TYPE(TypeRef);
		CASE_TYPE(TypeDef);
		CASE_TYPE(FieldDef);
		CASE_TYPE(MethodDef);
		CASE_TYPE(ParamDef);
		CASE_TYPE(InterfaceImpl);
		CASE_TYPE(MemberRef);
		CASE_TYPE(CustomAttribute);
		CASE_TYPE(Permission);
		CASE_TYPE(Signature);
		CASE_TYPE(Event);
		CASE_TYPE(Property);
		CASE_TYPE(ModuleRef);
		CASE_TYPE(TypeSpec);
		CASE_TYPE(Assembly);
		CASE_TYPE(AssemblyRef);
		CASE_TYPE(File);
		CASE_TYPE(ExportedType);
		CASE_TYPE(ManifestResource);
		CASE_TYPE(String);
		CASE_TYPE(Name);
		CASE_TYPE(BaseType);
	default: 
		return L"Unknown token type";
	}
}

// Stolen from Rotor (mdinfo.cpp) - ME
const WCHAR* TypeDefName(IMetaDataImport* piImport, mdTypeDef tok, WCHAR* buf, ULONG len)
{
	// Verify params
	_ASSERTE(piImport && buf);

	// Initialize out params
	buf[0] = 'L\0';

	HRESULT hr = piImport->GetTypeDefProps(tok, buf, len, NULL, NULL, NULL);                  // [OUT] Put base class TypeDef/TypeRef here.
	VERIFY_HR_(L"", hr, _T("GetTypeDefProps failed."), ILXLOG_WARNING);
	return buf;
}

// Stolen from Rotor (mdinfo.cpp) - ME
const WCHAR* TypeRefName(IMetaDataImport* piImport, mdTypeRef tok, WCHAR* buf, ULONG len)
{
	// Verify params
	_ASSERTE(piImport && buf);

	// Initialize out params
	buf[0] = 'L\0';

	HRESULT hr = piImport->GetTypeRefProps(tok, NULL, buf, len, NULL);
	VERIFY_HR_(L"", hr, _T("GetTypeRefProps failed."), ILXLOG_WARNING);
	return buf;
} 

/////////////////////////////////////////////////////////////////////////
// Adds description of element type to the end of buffer -- caller must ensure
// buffer is large enough. - Stolen from Rotor (mdinfo.cpp) - ME
/////////////////////////////////////////////////////////////////////////
HRESULT GetOneElementType(IMetaDataImport* piImport,
						  PCCOR_SIGNATURE pbSigBlob, ULONG ulSigBlob, ULONG* pcb,
						  WCHAR* szBuf, int nMaxLen)
{
	HRESULT hr = S_OK;

	szBuf[0] = L'\0';

	WCHAR szTempBuf[2048];
	WCHAR szTempBuf2[2048];

	ULONG ulData = 0;
	ULONG cb = CorSigUncompressData(pbSigBlob, &ulData);

	ULONG& cbCur = *pcb;
	cbCur = cb;

	// Handle the modifiers.
	if (ulData & ELEMENT_TYPE_MODIFIER)
	{
		if (ulData == ELEMENT_TYPE_SENTINEL)
			wcscat(szBuf, L"<ELEMENT_TYPE_SENTINEL>");
		else if (ulData == ELEMENT_TYPE_PINNED)
			wcscat(szBuf, L"PINNED");
		else
		{
			hr = E_FAIL;
			goto ErrExit;
		}

		hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb, szBuf, nMaxLen);
		if (FAILED(hr))
			goto ErrExit;

		cbCur += cb;
		goto ErrExit;
	}

	// Handle the underlying element types.
	if (ulData >= ELEMENT_TYPE_MAX) 
	{
		hr = E_FAIL;
		goto ErrExit;
	}

	bool bFirst = true;
	while (ulData == ELEMENT_TYPE_PTR || ulData == ELEMENT_TYPE_BYREF)
	{
		if (!bFirst)
			wcscat(szBuf, L" ");

		wcscat(szBuf, g_szMapElementType[ulData]);

		cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
		cbCur += cb;
		bFirst = false;
	}

	if (!bFirst)
		wcscat(szBuf, L" ");

	wcscat(szBuf, g_szMapElementType[ulData]);

	if (CorIsPrimitiveType((CorElementType)ulData) || 
		ulData == ELEMENT_TYPE_TYPEDBYREF ||
		ulData == ELEMENT_TYPE_OBJECT ||
		ulData == ELEMENT_TYPE_I ||
		ulData == ELEMENT_TYPE_U /*||
		ulData == ELEMENT_TYPE_R*/) // ELEMENT_TYPE_R is a private CLR type - ME
	{
		// If this is a primitive type, we are done
		goto ErrExit;
	}
	else if (	ulData == ELEMENT_TYPE_VALUETYPE || 
				ulData == ELEMENT_TYPE_CLASS || 
				ulData == ELEMENT_TYPE_CMOD_REQD ||
				ulData == ELEMENT_TYPE_CMOD_OPT)
	{
		mdToken tk = mdTokenNil;
		cb = CorSigUncompressToken(&pbSigBlob[cbCur], &tk);
		cbCur += cb;

		// get the name of type ref. Don't care if truncated
		if (TypeFromToken(tk) == mdtTypeDef)
		{
			swprintf(szTempBuf, L" %ls", 
				TypeDefName(piImport, tk, szTempBuf2, NumItems(szTempBuf2)));
		}
		else if (TypeFromToken(tk) == mdtTypeRef)
		{
			swprintf(szTempBuf, L" %ls", 
				TypeRefName(piImport, tk, szTempBuf2, NumItems(szTempBuf2)));
		}
		else
		{
			_ASSERTE(TypeFromToken(tk) == mdtTypeSpec);
			swprintf(szTempBuf, L" %8x", tk);
		}

		wcscat(szBuf, szTempBuf);

		if (ulData == ELEMENT_TYPE_CMOD_REQD ||
			ulData == ELEMENT_TYPE_CMOD_OPT)
		{
			hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb,
				szBuf, nMaxLen);
			if (FAILED(hr))
				goto ErrExit;

			cbCur += cb;
		}

		goto ErrExit;
	}

	// TODO - ELEMENT_TYPE_VALUEARRAY is a private CLR type - ME
	//if (ulData == ELEMENT_TYPE_VALUEARRAY)
	//{
	//	// display the base type of SDARRAY
	//	if (FAILED(GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb,
	//		szBuf, nMaxLen)))
	//		goto ErrExit;
	//	cbCur += cb;
	//
	//	// display the size of SDARRAY
	//	cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
	//	cbCur += cb;
	//	swprintf(szTempBuf, L" %d", ulData);
	//	wcscat(szBuf, szTempBuf));
	//	goto ErrExit;
	//}
	if (ulData == ELEMENT_TYPE_SZARRAY)
	{
		szTempBuf[0] = L'\0';
		
		// display the base type of SZARRAY or GENERICARRAY
		hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb,
			szTempBuf, nMaxLen);

		wcscat(szBuf, L" ");
		wcscat(szBuf, szTempBuf);

		if (FAILED(hr))
			goto ErrExit;

		cbCur += cb;
		goto ErrExit;
	}
	else if (ulData == ELEMENT_TYPE_FNPTR) 
	{
		cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
		cbCur += cb;

		if (ulData & IMAGE_CEE_CS_CALLCONV_EXPLICITTHIS)
			wcscat(szBuf, L" explicit");
		if (ulData & IMAGE_CEE_CS_CALLCONV_HASTHIS)
			wcscat(szBuf, L" hasThis");

		wcscat(szBuf, L" ");
		wcscat(szBuf, g_szCalling[ulData & IMAGE_CEE_CS_CALLCONV_MASK]);
		wcscat(szBuf, L":");

		// Get number of args
		ULONG numArgs = 0;
		cb = CorSigUncompressData(&pbSigBlob[cbCur], &numArgs);
		cbCur += cb;

		// do return type
		hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb,
			szBuf, nMaxLen);
		if (FAILED(hr))
			goto ErrExit;

		cbCur += cb;

		wcscat(szBuf, L"(");
		while (numArgs > 0) 
		{
			if (cbCur > ulSigBlob)
				goto ErrExit;
			hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb,
				szBuf, nMaxLen);
			if (FAILED(hr))
				goto ErrExit;

			cbCur += cb;
			--numArgs;
			if (numArgs > 0) 
				wcscat(szBuf, L",");
		}

		wcscat(szBuf, L" )");
		goto ErrExit;
	}

	if(ulData != ELEMENT_TYPE_ARRAY) 
		return E_FAIL;

	// display the base type of SDARRAY
	hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob-cbCur, &cb,
		szBuf, nMaxLen);
	if (FAILED(hr))
		goto ErrExit;

	cbCur += cb;

	// display the rank of MDARRAY
	cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
	cbCur += cb;
	swprintf(szTempBuf, L" %d", ulData);
	wcscat(szBuf, szTempBuf);
	if (ulData == 0)
		// we are done if no rank specified
		goto ErrExit;

	// how many dimensions have size specified?
	cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
	cbCur += cb;
	swprintf(szTempBuf, L" %d", ulData);
	wcscat(szBuf, szTempBuf);

	while (ulData)
	{
		ULONG ulTemp = 0;
		cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulTemp);
		swprintf(szTempBuf, L" %d", ulTemp);
		wcscat(szBuf, szTempBuf);
		cbCur += cb;
		ulData--;
	}
	// how many dimensions have lower bounds specified?
	cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
	cbCur += cb;
	swprintf(szTempBuf, L" %d", ulData);
	wcscat(szBuf, szTempBuf);

	while (ulData)
	{
		int iTemp = 0;
		cb = CorSigUncompressSignedInt(&pbSigBlob[cbCur], &iTemp);
		swprintf(szTempBuf, L" %d", iTemp);
		wcscat(szBuf, szTempBuf);
		cbCur += cb;
		ulData--;
	}

ErrExit:
	if (cbCur > ulSigBlob)
		hr = E_FAIL;

	return hr;
} // HRESULT MDInfo::GetOneElementType()

/////////////////////////////////////////////////////////////////////////
// Display COM+ signature -- taken from cordump.cpp's DumpSignature
/////////////////////////////////////////////////////////////////////////
void Sig2String(	IMetaDataImport* piImport, 
					PCCOR_SIGNATURE pbSigBlob, 
					ULONG ulSigBlob, 
					WCHAR* szBuf,
					int nMaxLen)
{
	HRESULT hr = S_OK;

	// Verify arguments
	_ASSERTE(piImport);
	_ASSERTE(szBuf);
	_ASSERTE(nMaxLen >= 1);

	// Initialize out params
	szBuf[0] = L'\0';

	if (!pbSigBlob)
	{
		wcscpy(szBuf, L"Null signature");
		return;
	}

	WCHAR szTempBuf[2048];

	//ULONG ulSigBlobStart = ulSigBlob;

	ULONG cbCur = 0;
	ULONG ulData = 0;

	ULONG cb = CorSigUncompressData(pbSigBlob, &ulData);

	// CallCnvntn
	if (*g_szCalling[ulData & IMAGE_CEE_CS_CALLCONV_MASK])
	{
		wcscat(szBuf, g_szCalling[ulData & IMAGE_CEE_CS_CALLCONV_MASK]);
		wcscat(szBuf, L": ");
	}

	if (cb > ulSigBlob) 
		goto ErrExit;

	cbCur += cb;
	ulSigBlob -= cb;

	if (ulData & IMAGE_CEE_CS_CALLCONV_HASTHIS)
		wcscat(szBuf, L"hasThis ");
	if (ulData & IMAGE_CEE_CS_CALLCONV_EXPLICITTHIS)
		wcscat(szBuf, L"explicit ");

	if (isCallConv(ulData, IMAGE_CEE_CS_CALLCONV_FIELD))
	{
		// display field type
		if (FAILED(hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob, &cb, 
			szTempBuf, NumItems(szTempBuf)-1)))
			goto ErrExit;

		wcscat(szBuf, szTempBuf);
		wcscat(szBuf, L" ");

		if (cb>ulSigBlob) 
			goto ErrExit;

		cbCur += cb;
		ulSigBlob -= cb;
	}
	else 
	{
		ULONG ulArgs = 0;
		cb = CorSigUncompressData(&pbSigBlob[cbCur], &ulArgs);
		if (cb>ulSigBlob) 
			goto ErrExit;

		cbCur += cb;
		ulSigBlob -= cb;

		if (ulData != IMAGE_CEE_CS_CALLCONV_LOCAL_SIG)
		{
			// display return type when it is not a local varsig
			if (FAILED(hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob, &cb,
				szTempBuf, NumItems(szTempBuf)-1)))
				goto ErrExit;

			wcscat(szBuf, L"ReturnType: ");
			wcscat(szBuf, szTempBuf);
			wcscat(szBuf, L" ");

			if (cb>ulSigBlob) 
				goto ErrExit;

			cbCur += cb;
			ulSigBlob -= cb;
		}

		wcscat(szBuf, L"(");

		ULONG       i = 0;
		while (i < ulArgs && ulSigBlob > 0)
		{
			ULONG       ulData;

			// Handle the sentinal for varargs because it isn't counted in the args.
			CorSigUncompressData(&pbSigBlob[cbCur], &ulData);
			++i;

			if (FAILED(hr = GetOneElementType(piImport, &pbSigBlob[cbCur], ulSigBlob, &cb,
				szTempBuf, NumItems(szTempBuf)-1)))
				goto ErrExit;

			if (i > 1)
				wcscat(szBuf, L", ");

			wcscat(szBuf, szTempBuf);

			if (cb>ulSigBlob) 
				goto ErrExit;

			cbCur += cb;
			ulSigBlob -= cb;
		}

		wcscat(szBuf, L")");
	}

	// Nothing consumed but not yet counted.
	cb = 0;

ErrExit:
	// We should have consumed all signature blob.  If not, dump the sig in hex.
	//  Also dump in hex if so requested.

	if (/*m_DumpFilter & dumpMoreHex ||*/ ulSigBlob != 0)
	{
		// Did we not consume enough, or try to consume too much?
		if (cb > ulSigBlob)
		{
			ILXLOGMSG(_T("Error in signature:  Signature should be larger."), ILXLOG_ERROR);
		}
		else if (cb < ulSigBlob)
		{
			ILXLOGMSG1(ILXLOG_ERROR, 
				_T("Error in signature:  Not all of signature blob was consumed.  %d byte(s) remain."), 
				ulSigBlob);

			// If it is short, just append it to the end.
			//if (ulSigBlob < 4)
			//{
			//	Write(": ");
			//	for (; ulSigBlob; ++cbCur, --ulSigBlob)
			//		VWrite("%02x ", pbSigBlob[cbCur]);
			//	WriteLine("");
			//	goto ErrExit2;
			//}
			//WriteLine("");
		}

		// Any appropriate error message has been issued.  Dump sig in hex, as determined
		//  by error or command line switch.
		//cbCur = 0;
		//ulSigBlob = ulSigBlobStart;
		//char rcNewPrefix[80];
		//swprintf(rcNewPrefix, L"%s\t\tSignature ", preFix);
		//DumpHex(rcNewPrefix, pbSigBlob, ulSigBlob, false, 24);
	}

	CHECK_HR(hr, _T("Bad signature blob value."), ILXLOG_ERROR);
} // Sig2String()
