///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/ModuleInfo.h 5     5/28/05 1:20p Eaddy $
//
//    File name:	ModuleInfo.h
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/ModuleInfo.h $
 * 
 * 5     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2004/10/24 19:33:11  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.3  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.2  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#ifndef INCLUDED_MODULE_INFO_H
#define INCLUDED_MODULE_INFO_H

#include "IncrFile.h"		// CIncrBuildFile
#include "BugReportFile.h"	// CBugReportFile

#define _WINSOCKAPI_   /* Prevent inclusion of winsock.h in windows.h */

// Modify the following defines if you have to target a platform prior to the ones specified below.
// Refer to MSDN for the latest info on corresponding values for different platforms.
#ifndef WINVER				// Allow use of features specific to Windows 95 and Windows NT 4 or later.
#define WINVER 0x0400		// Change this to the appropriate value to target Windows 98 and Windows 2000 or later.
#endif

#include <cor.h>		// Needed to compile corprof.h!
#include <corprof.h>	// ModuleID

#include <list>			// STL list
#include <queue>		// STL deque
#include <string>		// STL wstring

typedef std::list<std::string>	TypeListStrings;

//*****************************
// CModuleInfo
//*****************************

class CModuleInfo
{
public:
	CModuleInfo();
	~CModuleInfo();

	HRESULT init(	const ModuleID& moduleID,
					const WCHAR* szModulePath,
					LPCBYTE pbBaseLoadAddress);

	ModuleID getModuleID() const;
	const WCHAR* getModulePath() const;
	LPCBYTE getLoadAddr() const;

	HRESULT findAndLoadCompileInfo();

private: // NOT IMPLEMENTED
	CModuleInfo(const CModuleInfo&);
	CModuleInfo& operator=(const CModuleInfo&);

private:	// DATA
	ModuleID				m_moduleID;
	std::wstring			m_strModulePath;
	LPCBYTE					m_pbBaseLoadAddress;
	
	mutable std::wstring	m_strTmpPath;
	mutable bool			m_bGotTmpPath;

public:
	CCompileInfo*			m_pCompileInfo;

private:
	CIncrBuildFile			m_incrBuildFile;
	CBugReportFile			m_bugReportFile;
};

//*****************************
// CModuleQueue
//*****************************

typedef std::queue<CModuleInfo*> TypeQueueModules;

class CModuleQueue
{
public:
	CModuleQueue(const char* szEventName);
	~CModuleQueue();

	void push(	const ModuleID& moduleID,
				const WCHAR* szModulePath,
				LPCBYTE pbBaseLoadAddress);

	bool pop(CModuleInfo** ppModuleInfo);

	HANDLE getEvent() const;

private:
	void push(CModuleInfo* pModuleInfo);

private: // DATA
	CRITICAL_SECTION			m_cs;
	HANDLE						m_hQueuedObjectAvailable;
	TypeQueueModules			m_queModules;
	TCHAR						m_szEventName[MAX_PATH];
};

#endif // INCLUDED_MODULE_INFO_H
