///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/ModuleWatcherThread.cpp 5     11/11/05 5:29p Eaddy $
//
//    File name:	ModuleWatcherThread.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/ModuleWatcherThread.cpp $
 * 
 * 5     11/11/05 5:29p Eaddy
 * o Asserts
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.8  2005/01/14 23:30:10  me133
 * o No longer logs error for mscorlib
 * o Disabled verification that debug flag was used
 *
 * Revision 1.7  2004/10/26 04:44:57  me133
 * o Removed unused member var
 * o No longer sends shutdown message since WatcherKeeper does this
 *
 * Revision 1.6  2004/10/24 19:33:11  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.5  2004/10/17 08:46:52  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.4  2004/10/10 07:40:51  me133
 * o Converted print statements to logging
 *
 * Revision 1.3  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "StdAfx.h"

#include "ModuleWatcherThread.h"	// Me

#include "ModuleInfo.h"				// CModuleQueue

#include "WatcherSocket.h"			// CWatcherSocket

#include "Helpers.h"				// WaitForDispatchThreadEvents()

CModuleWatcherArg::CModuleWatcherArg(	CModuleQueue* pModuleQueue,
										int nPortToUse,
										HANDLE hWatcherReadyEvent,
										HANDLE hThreadPleaseTerminate,
										HANDLE hHeartbeatTimer)
	: m_pModuleQueue(pModuleQueue)
	, m_nPortToUse(nPortToUse)
	, m_hWatcherReadyEvent(hWatcherReadyEvent)
	, m_hThreadPleaseTerminate(hThreadPleaseTerminate)
	, m_hHeartbeatTimer(hHeartbeatTimer)
{ }

CModuleWatcherArg::~CModuleWatcherArg()
{ }

//////////////////////////////////////////////
// MODULE WATCHER THREAD PROC IMPLEMENTATION
//////////////////////////////////////////////

void __cdecl ModuleWatcherThreadProc(void* pvCModuleWatcherArg)
{
	ILXLOGMSG(_T("MODULE WATCHER: Starting."), ILXLOG_INFO);

	USES_CONVERSION;

	HRESULT hr = S_OK;

	// All local variables must be declared here because of 
	// the goto statement

	CWatcherSocket watcherSocket;

	CModuleWatcherArg* pModuleWatcherArg		= 0;

	HANDLE hQueuedObjectAvailable				= 0;
		// Signaled by ModuleQueue when an ModuleID has been queued

	HANDLE hThreadPleaseTerminate				= NULL;
		// Signaled by CProfilerCallback when the thread should exit

	BOOL bRet									= FALSE;

	CModuleQueue* pModuleQueue					= 0;
	int nPortToUse								= 0;
	HANDLE hWatcherReadyEvent					= NULL;
	HANDLE hHeartbeatTimer						= NULL;

	// Verify input parameters

	_ASSERTE(pvCModuleWatcherArg);
	if (!pvCModuleWatcherArg)
	{
		goto LABEL_EXIT_THREAD;
	}

	// Step 1:	Unpack thread argument

	pModuleWatcherArg = reinterpret_cast<CModuleWatcherArg*>(pvCModuleWatcherArg);
	if (!pModuleWatcherArg)
	{
		goto LABEL_EXIT_THREAD;
	}

	// Get the module queue
	nPortToUse			= pModuleWatcherArg->m_nPortToUse;
	pModuleQueue		= pModuleWatcherArg->m_pModuleQueue;
	hWatcherReadyEvent	= pModuleWatcherArg->m_hWatcherReadyEvent;
	hThreadPleaseTerminate = pModuleWatcherArg->m_hThreadPleaseTerminate;
	hHeartbeatTimer		= pModuleWatcherArg->m_hHeartbeatTimer;

	if (nPortToUse <= 0 || !pModuleQueue || !hWatcherReadyEvent || !hThreadPleaseTerminate || !hHeartbeatTimer)
	{
		ILXLOG4(hr, ILXLOG_ERROR, _T("Invalid thread parameters - %p - %p - %p - %p."), 
			pModuleQueue, hWatcherReadyEvent, hThreadPleaseTerminate, hHeartbeatTimer);
		goto LABEL_EXIT_THREAD;
	}

	// Free the memory for the arg object
	delete pModuleWatcherArg;
	pModuleWatcherArg = 0;
	
	// Step 2:	Wait for the Watcher process to start

	// Here we are treating a WSAEvent just like a standard Win32 event.
	// This seems to work even though WSAEvents have their own wait function.
	EDispatchThreadWaitStatus eWaitStatus = 
		WaitForDispatchThreadEvents(hWatcherReadyEvent,
									hThreadPleaseTerminate);
	
	if (eWaitStatus == DISPATCH_THREAD_WAIT_ERROR)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());

		ILXLOG(hr, _T("Event queue wait failed."), ILXLOG_ERROR);
		_ASSERTE(!"Event queue wait failed!");

		goto LABEL_EXIT_THREAD;
	}
	else if (eWaitStatus == DISPATCH_THREAD_SHUTDOWN_SIGNALED)
	{
		// We are being told to close down
		ILXLOGMSG(_T("Watcher process never got started."), ILXLOG_INFO);
		goto LABEL_EXIT_THREAD;
	}
	
	_ASSERTE(eWaitStatus == DISPATCH_THREAD_EVENT_WAITING);

	// Step 3:	Connect to Watcher

	static const int nMAX_RETRIES = 10;

	ILXLOGMSG(_T("MODULE WATCHER: Attempting to connect to Watcher."), ILXLOG_INFO);

	hr = watcherSocket.Connect(nPortToUse, nMAX_RETRIES, true);
	if (FAILED(hr))
	{
		ILXLOG(hr, _T("Failed to connect to Watcher server."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	// Step 4:	Create events

	hQueuedObjectAvailable = pModuleQueue->getEvent();
	if (!hQueuedObjectAvailable)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());
		ILXLOG(hr, _T("CreateEvent() failed."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	// Step 4:	Handle events forever!

	while (true)
	{
		// Step 5:	Wait for either a new queued event or until
		//			we are told to exit

		EDispatchThreadWaitStatus eWaitStatus = 
			WaitForDispatchThreadEvents(hQueuedObjectAvailable,
										hThreadPleaseTerminate);
		
		if (eWaitStatus == DISPATCH_THREAD_WAIT_ERROR)
		{
			hr = HRESULT_FROM_WIN32(::GetLastError());

			ILXLOG(hr, _T("Event queue wait failed."), ILXLOG_ERROR);
			_ASSERTE(!"Event queue wait failed!");

			goto LABEL_EXIT_THREAD;
		}
		else if (eWaitStatus == DISPATCH_THREAD_SHUTDOWN_SIGNALED)
		{
			// We are being told to close down
			goto LABEL_EXIT_THREAD;
		}
		
		_ASSERTE(eWaitStatus == DISPATCH_THREAD_EVENT_WAITING);

		CModuleInfo* pModuleInfo = 0;
		bool bSuccess = pModuleQueue->pop(&pModuleInfo);
		_ASSERTE(bSuccess);
		_ASSERTE(pModuleInfo);
		if (!pModuleInfo)
		{
			continue;
		}

		_ASSERTE(pModuleInfo->getModuleID() != 0);

		// Load the incremental build file (.incr) which contains all the
		// info needed to compile the module
		hr = pModuleInfo->findAndLoadCompileInfo();
		
		size_t nSourceFileCount = 0;
		if (SUCCEEDED(hr) && pModuleInfo->m_pCompileInfo)
		{
			nSourceFileCount = pModuleInfo->m_pCompileInfo->getSourceFileCount();
		}
		
		if (FAILED(hr) || nSourceFileCount == 0)
		{
			const WCHAR* szModuleName = pModuleInfo->getModulePath();
			_ASSERTE(szModuleName && *szModuleName);
			if (szModuleName && wcsstr(szModuleName, L"mscorlib") == 0)
			{
				ILXLOGMSG2(ILXLOG_INFO, _T("Incremental build file '%ls.incr' and bug report file '%ls.bugreport' not found. Did you forget to specify '/incremental' or '/bugreport'?"), 
					szModuleName,
					szModuleName);
			}

			delete pModuleInfo;
			pModuleInfo = 0;
			continue;
		}

		pModuleInfo->m_pCompileInfo->debugPrint();

		ILXLOGMSG1(ILXLOG_INFO, _T("Module added to watch queue: '%ls'."), 
			pModuleInfo->getModulePath());

		for(size_t i = 0; i < nSourceFileCount; ++i)
		{
			ILXLOGMSG1(ILXLOG_INFO, _T("WATCHING SOURCE FILE: %ls."), 
				pModuleInfo->m_pCompileInfo->getSourceFileName(i));
		}

		static int c = 0;

		//if (c++ == 1)
		if (true)
		{
		// Step 5:	Signal the Watcher
		watcherSocket.SendModuleInfo(	(DWORD) pModuleInfo->getModuleID(),
										W2CA(pModuleInfo->getModulePath()),
										pModuleInfo->m_pCompileInfo);
		}
		else
		{
			ILXLOGMSG1(ILXLOG_INFO, _T("SKIPPING MODULE: %ls!"), 
				pModuleInfo->getModulePath());
		}

		delete pModuleInfo;
		pModuleInfo = 0;
	}

LABEL_EXIT_THREAD:

	ILXLOGMSG(_T("Stopping watcher thread."), ILXLOG_INFO);

	_endthread();
}
