///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/ParamInfo.cpp 2     5/28/05 1:20p Eaddy $
//
//    File name:	ParamInfo.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/24/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/ParamInfo.cpp $
 * 
 * 2     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.1  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 *
 */

#include "stdafx.h"

#include "ParamInfo.h"			// Me

// HELPER FUNCTIONS

static HRESULT _FillVariant2(	BYTE        bCPlusTypeFlag, 
								const void*	pValue,
								VARIANT*	pvar);

//////////////////////////////
// PARAM INFO IMPLEMENTATION
//////////////////////////////

CParamInfo::CParamInfo(IMetaDataImport* piImport, mdParamDef paramDef)
	: m_methodDef(0)
	, m_ulNum(0)
	, m_szParamName()
	, m_ulNameLen(0)
	, m_dwFlags(0)
	, m_dwCPlusFlags(0)
	, m_varDefaultValue()
	, m_pValue(0)
	, m_cbValue(0)
{
	init(piImport, paramDef);
}

CParamInfo::CParamInfo(paramDef_iterator& itParamDef)
	: m_methodDef(0)
	, m_ulNum(0)
	, m_szParamName()
	, m_ulNameLen(0)
	, m_dwFlags(0)
	, m_dwCPlusFlags(0)
	, m_varDefaultValue()
	, m_pValue(0)
	, m_cbValue(0)
{
	init(itParamDef.getImport(), *itParamDef);
}

HRESULT CParamInfo::init(IMetaDataImport* piImport, mdParamDef paramDef)
{
	HRESULT hr = S_OK;

	// No AddRef/Release necessary
	_ASSERTE(piImport);
	if (!piImport) return E_POINTER;

	hr = piImport->GetParamProps(	paramDef, 
									&m_methodDef, 
									&m_ulNum, 
									m_szParamName, 
									NumItems(m_szParamName),
									&m_ulNameLen, 
									&m_dwFlags, 
									&m_dwCPlusFlags, 
									&m_pValue, 
									&m_cbValue);
	VERIFY_HR(hr, _T("GetMethodProps() failed."), ILXLOG_ERROR);

	_FillVariant2((BYTE)m_dwCPlusFlags, m_pValue, &m_varDefaultValue);

	return S_OK;
}

bool operator==(const CParamInfo& lhs, const CParamInfo& rhs)
{
	// Ignore m_methodDef differences due to different assemblies
	// Ignore m_ulNum differences due to different assemblies
	// Ignore m_pValue differences due to different assemblies

	if (lhs.m_ulNameLen != rhs.m_ulNameLen)
		return false;

	if (lhs.m_dwFlags != rhs.m_dwFlags)
		return false;

	if (lhs.m_dwCPlusFlags != rhs.m_dwCPlusFlags)
		return false;

	if (lhs.m_varDefaultValue != rhs.m_varDefaultValue)
		return false;

	if (lhs.m_cbValue != rhs.m_cbValue)
		return false;

	if (lhs.m_szParamName != rhs.m_szParamName)
	{
		if (!lhs.m_szParamName || !rhs.m_szParamName)
			return false; // One is null but not the other

		if (wcscmp(lhs.m_szParamName, rhs.m_szParamName) != 0)
			return false; // Strings are different
	}

	return true;
}

bool operator!=(const CParamInfo& lhs, const CParamInfo& rhs)
{
	return !(operator==(lhs, rhs));
}

// From MetaInfo sample
HRESULT _FillVariant2(	BYTE        bCPlusTypeFlag, 
						const void*	pValue,
						VARIANT*	pvar) 
{
    HRESULT hr = NOERROR;

    switch (bCPlusTypeFlag)
    {
    case ELEMENT_TYPE_BOOLEAN:
        pvar->vt = VT_BOOL;
        pvar->boolVal = *((BYTE*)pValue); //*((VARIANT_BOOL *)pValue);
        break;
    case ELEMENT_TYPE_I1:
        pvar->vt = VT_I1;
        pvar->cVal = *((CHAR*)pValue);
        break;  
    case ELEMENT_TYPE_U1:
        pvar->vt = VT_UI1;
        pvar->bVal = *((BYTE*)pValue);
        break;  
    case ELEMENT_TYPE_I2:
        pvar->vt = VT_I2;
        pvar->iVal = *((SHORT*)pValue);
        break;  
    case ELEMENT_TYPE_U2:
    case ELEMENT_TYPE_CHAR:
        pvar->vt = VT_UI2;
        pvar->uiVal = *((USHORT*)pValue);
        break;  
    case ELEMENT_TYPE_I4:
        pvar->vt = VT_I4;
        pvar->lVal = *((LONG*)pValue);
        break;  
    case ELEMENT_TYPE_U4:
        pvar->vt = VT_UI4;
        pvar->ulVal = *((ULONG*)pValue);
        break;  
    case ELEMENT_TYPE_R4:
        pvar->vt = VT_R4;
        pvar->fltVal = *((FLOAT*)pValue);
        break;  
    case ELEMENT_TYPE_R8:
        pvar->vt = VT_R8;
        pvar->dblVal = *((DOUBLE*)pValue);
        break;  
    case ELEMENT_TYPE_STRING:
        pvar->vt = VT_BSTR;

        // allocated bstr here
        pvar->bstrVal = ::SysAllocString((LPWSTR)pValue);
        if (pvar->bstrVal == NULL)
            hr = E_OUTOFMEMORY;
        break;  
    case ELEMENT_TYPE_CLASS:
        pvar->punkVal = NULL;
        pvar->vt = VT_UNKNOWN;
        _ASSERTE( *((IUnknown **)pValue) == NULL );
        break;  
    case ELEMENT_TYPE_I8:
        pvar->vt = VT_I8;
        pvar->cyVal.int64 = *((LONGLONG*)pValue);
        break;
    case ELEMENT_TYPE_U8:
        pvar->vt = VT_UI8;
        pvar->cyVal.int64 = *((LONGLONG*)pValue);
        break;
    case ELEMENT_TYPE_VOID:
        pvar->vt = VT_EMPTY;
        break;
    default:
        _ASSERTE(!"bad constant value type!");
    }

    return hr;
}
