///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/PatcherProfilerCallback.cpp 11    12/15/05 2:37a Eaddy $
//
//    File name:	PatcherProfilerCallback.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	Some function header comments came from corprof.idl
//					- Marc Eaddy
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/PatcherProfilerCallback.cpp $
 * 
 * 11    12/15/05 2:37a Eaddy
 * o Improved statistics gathering and display
 * 
 * 10    12/10/05 12:41a Eaddy
 * o Now displays elapsed times
 * 
 * 9     11/11/05 5:30p Eaddy
 * o Commented out the IL-native mapping dump
 * 
 * 8     10/07/05 8:04p Eaddy
 * o Now the JITer keeps track of the IL-to-native mapping which is
 * critical for source-level debugging
 * 
 * 7     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.11  2005/01/14 23:22:05  me133
 * o Improved logging
 *
 * Revision 1.10  2004/10/26 04:45:18  me133
 * o Removed unused member var
 *
 * Revision 1.9  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 *
 * Revision 1.8  2004/10/24 19:33:11  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.7  2004/10/17 08:46:53  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.6  2004/10/17 00:24:55  me133
 * o Now provides a class-based iface to get to the incremental file info
 *
 * Revision 1.5  2004/10/11 23:24:09  me133
 * o Added wishful comment
 *
 * Revision 1.4  2004/10/11 05:59:37  me133
 * no message
 *
 * Revision 1.3  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#include "StdAfx.h"

#include "PatcherProfilerCallback.h"	// Me

#include "WatcherKeeperThread.h"		// WatcherKeeperThreadProc()
#include "ModuleWatcherThread.h"		// ModuleWatcherThreadProc()
#include "UpdaterThread.h"				// UpdateThreadProc()

#include "Helpers.h"					// ExtractPreStubFuncAddr()

#include "MethodInfo.h"					// GetMethodNameFromFunctionId()

#include "XmlLogHelper.h"				// LogXmlError()

#include <corerror.h>					// CORPROF_E_DATAINCOMPLETE

// CONSTANTS

static const TCHAR* szBREAK_IMMEDIATELY_ENV_VAR_NAME	= _T("PATCHER_BREAK");
static const TCHAR* szPROFILING_LOG_FILE				= _T("Patcher.log");
static const TCHAR* szPROFILING_MODULE					= _T("Patcher");

static const int nPORT_TO_USE = 3456;

//////////////////////////////////////////////////////
// PATCHER PROFILER CALLBACK IMPLEMENTATION (Part 1)
//////////////////////////////////////////////////////

HINSTANCE CPatcherProfilerCallback::s_hInst = 0;

CPatcherProfilerCallback::CPatcherProfilerCallback()
	: m_spiProfInfo()
	, m_bBreakImmediately(false)
	, m_hWatcherStartedEvent(0)
	, m_hThreadsPleaseTerminate(0)
	, m_hHeartbeatTimer(0)
	, m_modulesToWatchQueue("m_modulesToWatchQueue")
	, m_mapChangedFunctions()
	, m_csListChangedFunctions()
	, m_dwPfnThePreStubFunc(0)
{ 
	HRESULT hr = S_OK;

	TCHAR szBreakImmediately[_MAX_PATH] = { 0 };

	DWORD dwRet = ::GetEnvironmentVariable(	szBREAK_IMMEDIATELY_ENV_VAR_NAME, 
											szBreakImmediately,
											_MAX_PATH );
	if (dwRet)
	{
		_variant_t varBreakImmediately(szBreakImmediately);
		m_bBreakImmediately = varBreakImmediately;
	}
	else
	{
		DWORD dwLastError = ::GetLastError();
		_ASSERTE(dwLastError == ERROR_ENVVAR_NOT_FOUND);
	}

	if (m_bBreakImmediately)
		DebugBreak();

	// Step 1:	Initializes logging

	if (GetLog().isInitialized())
	{
		return;
	}

	hr = GetLog().initialize( szPROFILING_MODULE );
	//_ASSERTE(SUCCEEDED(hr));
	if (FAILED(hr)) return;

	CComPtr<MSXML::IXMLDOMNode> spiLogSettingsNode;
	hr = GetLogSettingsNode(&spiLogSettingsNode);
	if (hr == S_OK && spiLogSettingsNode)
	{
		hr = ::GetLog().loadConfig(spiLogSettingsNode);
		_ASSERTE(SUCCEEDED(hr));
	}
	else
	{
		hr = ::GetLog().useLogFile(szPROFILING_LOG_FILE);
		ILXLOG(hr, _T("Failed to load log settings file."), ILXLOG_ERROR);
		_ASSERTE(SUCCEEDED(hr));
	}
}

CPatcherProfilerCallback::~CPatcherProfilerCallback()
{ 
	_ASSERTE(m_spiProfInfo == NULL);
	GetLog().shutdown();
}

ULONG CPatcherProfilerCallback::AddRef()
{
    return S_OK;
}
ULONG CPatcherProfilerCallback::Release()
{
    return S_OK;
}
HRESULT CPatcherProfilerCallback::QueryInterface( REFIID riid, void **ppInterface )
{
    return E_NOTIMPL;
}

/*
 *
 * STARTUP/SHUTDOWN EVENTS
 *
 */

/*
 * The CLR calls Initialize to setup the code profiler
 * whenever a new CLR application is started. The call provides
 * an IUnknown interface pointer that should be QI'd for an ICorProfilerInfo
 * interface pointer.
 *
 * NOTE: this is the only opportunity to enable callbacks that are a part
 * of COR_PRF_MONITOR_IMMUTABLE, since they can no longer be changed after
 * returning from this function.  This is done through SetEventMask on the
 * ICorProfilerInfo object.
 *
 * NOTE: Inprocess debugging not supported from this callback.
 */
STDMETHODIMP CPatcherProfilerCallback::Initialize(/*[in]*/ IUnknown* piunkCorProfilerInfo)
{
    HRESULT hr = S_OK;
    
	if (!piunkCorProfilerInfo) return E_POINTER;

	_ASSERTE(m_spiProfInfo == NULL);
	m_spiProfInfo = 0;

	ILXLOGMSG0(ILXLOG_INFO, _T("Initialize() called."));

	// Step 1:	Obtain the ICorProfilerInfo iface

    hr = piunkCorProfilerInfo->QueryInterface( __uuidof(ICorProfilerInfo), (void **)&m_spiProfInfo );
	VERIFY_HR_COND(hr, m_spiProfInfo, _T("Failed to obtain ICorProfilerInfo iface."), 
		ILXLOG_ERROR);

	// Step 2:	Specify which events we are interested in

	DWORD dwEventMask = (DWORD)(COR_PRF_MONITOR_MODULE_LOADS |
								COR_PRF_MONITOR_JIT_COMPILATION |
								COR_PRF_DISABLE_INLINING |
								COR_PRF_ENABLE_REJIT |
								COR_PRF_ENABLE_JIT_MAPS);

    hr = m_spiProfInfo->SetEventMask( dwEventMask );
	VERIFY_HR(hr, _T("Failed to set the event mask."), ILXLOG_ERROR);

	// Step 3:	Create Win32 event that will signal the threads
	//			to terminate

	m_hThreadsPleaseTerminate = ::CreateEvent(	0, FALSE, FALSE, NULL);
	if (!m_hThreadsPleaseTerminate)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		return ILXLOG(hr, _T("ERROR: CreateEvent() failed."), ILXLOG_ERROR);
	}

	m_hWatcherStartedEvent = ::CreateEvent(0, TRUE, FALSE, NULL);
	if (!m_hWatcherStartedEvent)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		return ILXLOG(hr, _T("ERROR: CreateEvent() failed."), ILXLOG_ERROR);
	}

	m_hHeartbeatTimer = ::CreateWaitableTimer(NULL, TRUE, NULL);
	if (!m_hHeartbeatTimer)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		return ILXLOG(hr, _T("Failed to initialize the watcher keeper thread."), ILXLOG_ERROR);
	}

	// Step 4:	***** Create and run the threads *****
	//

	// Here is why I use _beginthread() instead of CreateThread():
	//
	// "A thread that uses functions from the C run-time libraries should
	// use the beginthread and endthread C run-time functions for thread
	// management rather than CreateThread and ExitThread. Failure to do
	// so results in small memory leaks when ExitThread is called."
	// - MSDN Library help documentation for CreateThread()

	// Note: Don't need to marshal ICorProfilerInfo because it aggregates the
	// free-threaded marshaller (FTM).

	// Must be static in order for the thread to be able to access the object
	// when this function exits
	CWatcherKeeperArg* pWatcherKeeperArg = new CWatcherKeeperArg(	nPORT_TO_USE,
																	m_hWatcherStartedEvent,
																	m_hThreadsPleaseTerminate,
																	m_hHeartbeatTimer);

	ILXLOGMSG(_T("Spawning watcher keeper thread."), ILXLOG_INFO);

	uintptr_t hThreadHandle = ::_beginthread(	WatcherKeeperThreadProc,
												0,
												pWatcherKeeperArg);
	if (hThreadHandle == -1)
	{
		hr = HRESULT_FROM_WIN32(errno);
		ILXLOG(hr, _T("Failed to create watcher keeper thread."), ILXLOG_ERROR);
		_ASSERTE(!"Failed to create watcher keeper thread.");
		return hr;
	}

	// Must be static in order for the thread to be able to access the object
	// when this function exits
	CModuleWatcherArg* pModuleWatcherArg = new CModuleWatcherArg(	&m_modulesToWatchQueue,
																	nPORT_TO_USE,
																	m_hWatcherStartedEvent,
																	m_hThreadsPleaseTerminate,
																	m_hHeartbeatTimer);

	ILXLOGMSG(_T("Spawning module watcher thread."), ILXLOG_INFO);

	uintptr_t hThreadHandle2 = ::_beginthread(	ModuleWatcherThreadProc,
												0,
												pModuleWatcherArg);
	if (hThreadHandle2 == -1)
	{
		hr = HRESULT_FROM_WIN32(errno);
		ILXLOG(hr, _T("Failed to create module watcher thread."), ILXLOG_ERROR);
		_ASSERTE(!"Failed to create module watcher thread.");
		return hr;
	}

	// We store a ptr instead of the DWORD because initially the
	// value is zero and it gets updated in JITCompilationStarted().
	// It is possible although highly unlikely for the Updater thread
	// to need this value before its been set.  In this case, the
	// Updater thread determines this and fails gracefully.
	CUpdaterArg* pUpdaterArg = new CUpdaterArg(	&m_mapChangedFunctions,
												&m_csListChangedFunctions,
												&m_dwPfnThePreStubFunc,
												m_spiProfInfo,
												nPORT_TO_USE,
												m_hWatcherStartedEvent,
												m_hThreadsPleaseTerminate,
												m_hHeartbeatTimer);

	ILXLOGMSG(_T("Spawning updater thread."), ILXLOG_INFO);

	uintptr_t hThreadHandle3 = ::_beginthread(	UpdaterThreadProc,
												0,
												pUpdaterArg); // Thread responsible for deleting
	if (hThreadHandle3 == -1)
	{
		hr = HRESULT_FROM_WIN32(errno);
		ILXLOG(hr, _T("Failed to create updater thread."), ILXLOG_ERROR);
		_ASSERTE(!"Failed to create updater thread.");
		return hr;
	}

	return S_OK;
}

/*
 * The CLR calls Shutdown to notify the code profiler that
 * the application is exiting.  This is the profiler's last opportunity to
 * safely call functions on the ICorProfilerInfo interface.  After returning
 * from this function the runtime will proceed to unravel its internal data
 * structures and any calls to ICorProfilerInfo are undefined in their
 * behaviour.
 *
 * NOTE: to maintain overall API speed, the profiler omits many argument
 * checks.  This means that the profiler is responsible for making sure
 * that the arguments it provides to any ICorProfilerInfo API are valid.
 *
 * NOTE: Inprocess debugging not supported from this callback.
 */
STDMETHODIMP CPatcherProfilerCallback::Shutdown()
{
	ILXLOGMSG0(ILXLOG_INFO, _T("Shutdown() called."));

	// Step x:	Tell the threads to shutdown

	if (m_hThreadsPleaseTerminate)
	{
		::SetEvent(m_hThreadsPleaseTerminate);
		::CloseHandle(m_hThreadsPleaseTerminate);
		m_hThreadsPleaseTerminate = 0;
	}

	if (m_hWatcherStartedEvent)
	{
		::CloseHandle(m_hWatcherStartedEvent);
		m_hWatcherStartedEvent = 0;
	}

	if (m_hHeartbeatTimer)
	{
		::CloseHandle(m_hHeartbeatTimer);
		m_hHeartbeatTimer = 0;
	}

	m_spiProfInfo = 0;

	return S_OK;
}

/*
 *
 * JIT EVENTS
 *
 */

/*
 * The CLR calls ModuleLoadFinished to notify the code
 * profiler that a module has been loaded. The hrStatus provides the
 * success or failure of the operation
 */
STDMETHODIMP CPatcherProfilerCallback::ModuleLoadFinished(
            /*[in]*/ ModuleID moduleID,
            /*[in]*/ HRESULT  hrStatus)
{
	//ILXLOGMSG1(ILXLOG_INFO, _T("ModuleLoadFinished() called for ModuleID %d."), 
	//	moduleID);

	HRESULT hr = S_OK;

	// Verify args
	_ASSERTE(moduleID != 0);

	// Verify that Initialize() succeeded
	_ASSERTE(m_spiProfInfo != NULL);
	if (m_spiProfInfo == NULL) return E_UNEXPECTED;

	// Step 1:	Verify that module was loaded
	if (hrStatus != S_OK)
	{
		ILXLOG1(hrStatus, ILXLOG_WARNING, _T("Module %d failed to load. Skipping."), 
			moduleID);
		return S_OK;
	}

	// Step 2:	Obtain module information

	LPCBYTE pbBaseLoadAddress = 0;
	WCHAR szModulePath[MAX_PATH];
	ULONG nLen = MAX_PATH-1;
	AssemblyID assemblyID;

	*szModulePath = L'\0'; // Just in case

	hr = m_spiProfInfo->GetModuleInfo(moduleID, &pbBaseLoadAddress, nLen, &nLen,
		szModulePath, &assemblyID);
	
	if (hr != CORPROF_E_DATAINCOMPLETE) // Not sure why this is a special case
	{
		VERIFY_HR(hr, _T("GetModuleInfo() failed."), ILXLOG_ERROR);
	}

	// Step 3:	Verify returned values

	VERIFY_COND2(nLen < MAX_PATH, ILXLOG_ERROR, _T("Module name is too long (%d >= %d)."),
		nLen, MAX_PATH);
	VERIFY_COND(moduleID != 0, _T("Invalid ModuleID (0)."), ILXLOG_ERROR);
	VERIFY_COND1(pbBaseLoadAddress, ILXLOG_ERROR, _T("Invalid BaseLoadAddress (0) for ModuleID: %d."),
			moduleID);
	VERIFY_COND1(szModulePath && *szModulePath, ILXLOG_WARNING,  
		_T("Invalid ModulePath (\"\") for ModuleID: %d."), moduleID);

	ILXLOGMSG2(ILXLOG_INFO, _T("ModuleLoadFinished() called for '%ls' (ModuleID %d)."), 
		szModulePath, moduleID);

	// Step 4:	Add the new module to the queue watched by the ModuleWatcher thread
	m_modulesToWatchQueue.push(moduleID, szModulePath, pbBaseLoadAddress);

	return S_OK;
}

/*
 * The CLR calls JITCompilationStarted to notify the code
 * profiler that the JIT compiler is starting to compile a function.
 *
 * The fIsSafeToBlock argument tells the profiler whether or not blocking
 * will affect the operation of the runtime.  If true, blocking may cause
 * the runtime to wait for the calling thread to return from this callback.
 * Although this will not harm the runtime, it will skew the profiling
 * results.
 *
 * NOTE: It is possible to receive more than one JITCompilationStarted/
 * JITCompilationFinished pair for each method.  This is because of how
 * the runtime handles class constructors: method A starts to be JIT'd,
 * then realizes that the class ctor for class B needs to be run, so
 * JIT's it and runs it, and while it's running makes a call to original
 * method A, which causes it to be JIT'd again, and causes the original
 * (incomplete) JIT'ing of A to be aborted.  However, both attempts to
 * JIT A are reported with JIT compilation events.  If the profiler is
 * going to replace IL code for this method with SetILFunctionBody, then
 * it must do so for both JITCompilationStarted events, but may use the
 * same IL block for both.  Note that the second IL body provided through
 * SetILFunctionBody will be the IL that is used.
 */
STDMETHODIMP CPatcherProfilerCallback::JITCompilationStarted(
            /*[in]*/ FunctionID functionID,
            /*[in]*/ BOOL       fIsSafeToBlock)
{
	HRESULT hr = S_OK;

	DWORD dwUpdateStartTime = ::GetTickCount();

	// Verify args
	_ASSERTE(functionID != 0);
	if (functionID == 0) return E_INVALIDARG;

	// Verify that Initialize() succeeded
	_ASSERTE(m_spiProfInfo);
	if (!m_spiProfInfo) return E_UNEXPECTED;

	// Step 1:	Obtain the metadata for this method

	CMethodInfo methodInfo;
	hr = methodInfo.init(m_spiProfInfo, functionID, 0, 0, false);
	if (FAILED(hr)) return hr;
	
	//ILXLOGMSG3(ILXLOG_INFO,
	//	_T("JITCompilationStarted() called - %ls.%ls (FunctionID:%p)."),
	//	methodInfo.m_szTypeName, methodInfo.m_szMemberName, functionID);

	// Step 2:	Obtain the PreStub function address if we haven't already

	if (m_dwPfnThePreStubFunc == 0)
	{
		m_dwPfnThePreStubFunc = ExtractPreStubFuncAddr(functionID);
	}

	// Step 3:	Replace the method body in the case where JITCompilationStarted()
	//			is being called as a result of SetFunctionReJIT()

	// Acquire critical section that safely holds the lock until we return
	CCritSecLock autoCritSec(m_csListChangedFunctions);

	CMethodInfo* pPatch = 0;

	TypeMapMethodInfo::iterator it_find = m_mapChangedFunctions.find(methodInfo.m_methodDef);
	if (it_find != m_mapChangedFunctions.end())
	{
		pPatch = &(*it_find).second;
	}
	else
	{
		// This function doesn't need to be patched
		return S_OK;
	}

	// Don't re-jit the same function if it hasn't changed since the last JIT
	if (!pPatch->m_bWasModified)
	{
		//ILXLOGMSG2(ILXLOG_INFO, 
		//	_T("JITCompilationStarted() was already called for function patch '%ls.%ls'."),
		//	pPatch->m_szTypeName, pPatch->m_szMemberName);
		return S_OK;
	}
/*
	// Dump out the IL-to-native mapping

	ULONG32 cMap = 50;
	ULONG32 cActual = 0;

	COR_DEBUG_IL_TO_NATIVE_MAP map[50];

	hr = m_spiProfInfo->GetILToNativeMapping( functionID,
		cMap,
		&cActual,
		map);
	
	if (SUCCEEDED(hr))
	{
		for(unsigned int j = 0; j < cActual; ++j)
		{
			if ((LONG32) map[j].ilOffset >= 0)
			{
				ILXLOGMSG3(ILXLOG_INFO, _T( "0x%08X-0x%08X => 0x%08X"), 
					map[j].nativeStartOffset, map[j].nativeEndOffset, map[j].ilOffset);
			}
			else
			{
				const TCHAR* pMappingType = 0;

				switch(map[j].ilOffset)
				{
					case NO_MAPPING: pMappingType = _T("NO MAPPING"); break;
					case PROLOG:     pMappingType = _T("--PROLOG--"); break;
					case EPILOG:     pMappingType = _T("--EPILOG--"); break;
					default:         pMappingType = _T("--UNKNOWN-"); break;
				}

				ILXLOGMSG3(ILXLOG_INFO, _T( "0x%08X-0x%08X => %s"), 
					map[j].nativeStartOffset, map[j].nativeEndOffset, pMappingType);
			}
		}
	}
	else
	{
		ILXLOG1(hr, ILXLOG_ERROR, _T( "GetILToNativeMapping() failed for function ID:0x%08X."), 
			functionID);
	}
*/
	// Step 4:	We found a patch, so apply it

	hr = m_spiProfInfo->SetILFunctionBody(	pPatch->m_moduleID,			// ModuleID of the given module
											pPatch->m_methodDef,			// Metadata token for method
											pPatch->m_pbMethodHeader);	// Pointer to the IL method body
	if (FAILED(hr))
	{
		ILXLOG(hr, _T( "SetILFunctionBody() failed."), ILXLOG_ERROR);
		// No use returning an error code here since the EE won't know
		// how to handle it.
	}

	// Indicate that the patch has been applied
	pPatch->m_bWasModified = false;

	DWORD dwCurrentTime = ::GetTickCount();

	WCHAR szTimes[1000];
	
	swprintf(szTimes, L"Weave:%dms Compile:%dms Patch:%dms Update:%dms Total:%dms", 
		pPatch->m_dwWeaveTimeEnd - pPatch->m_dwTicksStart,		// Elapsed weave time
		pPatch->m_dwCompileTimeEnd - pPatch->m_dwWeaveTimeEnd,	// Elapsed compile time
		pPatch->m_dwPatchTimeEnd - pPatch->m_dwCompileTimeEnd,	// Elapsed patch creation time
		dwCurrentTime - pPatch->m_dwPatchTimeEnd,				// Elapsed update time
		dwCurrentTime - pPatch->m_dwTicksStart);				// Total elapsed time

	ILXLOGMSG5(ILXLOG_SUCCESS, _T("Re-JIT'd '%ls.%ls' (ID:0x%08X, Hdr:%p) (%ls)."), 
		pPatch->m_szTypeName, 
		pPatch->m_szMemberName, 
		functionID, 
		pPatch->m_pbMethodHeader, 
		szTimes);

	return S_OK;
}

HRESULT CPatcherProfilerCallback::GetLogSettingsNode(MSXML::IXMLDOMNode** ppiXmlLogSettings)
{
	HRESULT hr = S_OK;

	_ASSERTE(ppiXmlLogSettings);
	_ASSERTE(!*ppiXmlLogSettings);

	// Step 1:	Find out the path to ourselves (Patcher.dll)

	wchar_t szModulePath[MAX_PATH+1];
	DWORD dwRet = GetModuleFileNameW(	CPatcherProfilerCallback::s_hInst,
										szModulePath,
										MAX_PATH);
	if (!dwRet)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		return ILXLOG(hr, _T("Failed to obtain module file name."), ILXLOG_ERROR);
	}

	// Step 2:	Extract our module directory

	wchar_t szDrive[_MAX_DRIVE];
	wchar_t szDir[_MAX_DIR];
	_wsplitpath(szModulePath, szDrive, szDir, 0, 0);

	// Step 3:	Create the path to ILXLogSettings.xml

	wchar_t szProfilerLogFilePath[MAX_PATH];
	_wmakepath(szProfilerLogFilePath, szDrive, szDir, L"ILXLogSettings", L"xml");

	// We may need to remove "Debug" from the directory name
	if (!::PathFileExistsW(szProfilerLogFilePath) && *szDir)
	{
		wchar_t* pLastSlash = wcsrchr(szDir, L'\\');
		if (pLastSlash)
		{
			*pLastSlash = L'\0';
			pLastSlash = wcsrchr(szDir, L'\\');
			if (pLastSlash)
			{
				*pLastSlash = L'\0';
				_wmakepath(szProfilerLogFilePath, szDrive, szDir, L"ILXLogSettings", L"xml");
			}
		}
	}

	// Step 4:	Load the log file

	_variant_t varUrlString = szProfilerLogFilePath;

	VARIANT_BOOL bIsSuccessful = VARIANT_FALSE;
	CComPtr<MSXML::IXMLDOMDocument> spiXmlDocument;

	hr = spiXmlDocument.CoCreateInstance(MSXML::CLSID_DOMDocument);
	VERIFY_HR(hr, _T("Failed to create XML parser."), ILXLOG_ERROR);

	hr = spiXmlDocument->load(varUrlString, &bIsSuccessful);
	VERIFY_HR(hr, _T("Failed to load log settings file."), ILXLOG_ERROR);
	
	if (!bIsSuccessful)
	{
		hr = ::LogXmlError(spiXmlDocument);
		return ILXLOG(hr, _T("Failed to load log settings file."), ILXLOG_ERROR);
	}

	CComPtr<MSXML::IXMLDOMElement> spiDocumentElement;
	hr = spiXmlDocument->get_documentElement(&spiDocumentElement);
	VERIFY_HR_COND(hr, spiDocumentElement, 
		_T("Failed to retrieve the document element for the log settings file."), ILXLOG_ERROR);

	hr = spiDocumentElement->QueryInterface(ppiXmlLogSettings);
	VERIFY_HR_COND(hr, *ppiXmlLogSettings, 
		_T("Failed to QI for IXMLDOMNode."), ILXLOG_ERROR);

	return S_OK;
}
