///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/PatcherProfilerCallback.h 4     5/28/05 1:20p Eaddy $
//
//    File name:	PatcherProfilerCallback.h
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/PatcherProfilerCallback.h $
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.6  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 *
 * Revision 1.5  2004/10/24 19:33:11  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.4  2004/10/17 08:46:53  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.3  2004/10/10 06:07:45  me133
 * o Added header comment
 */

#ifndef INCLUDED_PATCHER_PROFILER_CALLBACK_H
#define INCLUDED_PATCHER_PROFILER_CALLBACK_H

#pragma once

#include "ModuleInfo.h"				// CModuleQueue
#include "AssemblyDiff.h"			// TypeListFuncIdAndPtr

#include <cor.h>					// Needed to compile corprof.h!!
#include <corprof.h>				// ICorProfilerCallback
#include <atlsync.h>				// CCriticalSection

// Forward delcare the IXMLDOMNode class
namespace MSXML
{
	struct __declspec(uuid("2933bf80-7b36-11d2-b20e-00c04f983e60"))
	/* dual interface */ IXMLDOMNode;
	struct __declspec(uuid("2933bf81-7b36-11d2-b20e-00c04f983e60"))
	/* dual interface */ IXMLDOMDocument;
};

// CPatcherProfilerCallback

class CPatcherProfilerCallback : public ICorProfilerCallback
{
public:
	CPatcherProfilerCallback();
	virtual ~CPatcherProfilerCallback();

	// IUnknown implementation
    STDMETHOD_(ULONG, AddRef)();
    STDMETHOD_(ULONG, Release)();
    STDMETHOD(QueryInterface)( REFIID riid, void **ppInterface );           

	// ICorProfilerCallback Methods

    /*
     *
     * STARTUP/SHUTDOWN EVENTS
     *
     */

    /*
     * The CLR calls Initialize to setup the code profiler
     * whenever a new CLR application is started. The call provides
     * an IUnknown interface pointer that should be QI'd for an ICorProfilerInfo
     * interface pointer.
     *
     * NOTE: this is the only opportunity to enable callbacks that are a part
     * of COR_PRF_MONITOR_IMMUTABLE, since they can no longer be changed after
     * returning from this function.  This is done through SetEventMask on the
     * ICorProfilerInfo object.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(Initialize)(
                /*[in]*/ IUnknown* piCorProfilerInfoUnk);

    /*
     * The CLR calls Shutdown to notify the code profiler that
     * the application is exiting.  This is the profiler's last opportunity to
     * safely call functions on the ICorProfilerInfo interface.  After returning
     * from this function the runtime will proceed to unravel its internal data
     * structures and any calls to ICorProfilerInfo are undefined in their
     * behaviour.
     *
     * NOTE: to maintain overall API speed, the profiler omits many argument
     * checks.  This means that the profiler is responsible for making sure
     * that the arguments it provides to any ICorProfilerInfo API are valid.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(Shutdown)();

    /*
     *
     * APPLICATION DOMAIN EVENTS
     *
     */

    /*
     * Called when an application domain creation has begun and ended.
     * The id is not valid for any information request until after the app
     * domain has been fully created.  One may only cache the id provided in
     * AppDomainCreationStarted for later use.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(AppDomainCreationStarted)(
                /*[in]*/ AppDomainID appDomainId);

    STDMETHOD(AppDomainCreationFinished)(
                /*[in]*/ AppDomainID appDomainId,
                /*[in]*/ HRESULT     hrStatus);

    /*
     * Called before and after an app domain is unloaded from a process.
     * Once an AppDomain unload has started, its ID is no longer valid in the
     * system and cannot be used in calls to any ICorProfilerInfo APIs.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(AppDomainShutdownStarted)(
                /*[in]*/ AppDomainID appDomainId);

    STDMETHOD(AppDomainShutdownFinished)(
                /*[in]*/ AppDomainID appDomainId,
                /*[in]*/ HRESULT     hrStatus);

    /*
     *
     * ASSEMBLY EVENTS
     *
     */

    /*
     * Called when an Assembly load has begun and ended. The id is not valid
     * for any information request until after the assembly has been fully
     * loaded.  One may only cache the id provided in AssemblyLoadStarted for
     * later use.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(AssemblyLoadStarted)(
                /*[in]*/ AssemblyID assemblyId);

    STDMETHOD(AssemblyLoadFinished)(
                /*[in]*/ AssemblyID assemblyId,
                /*[in]*/ HRESULT    hrStatus);

    /*
     * Called before and after an assembly is unloaded.  AssemblyUnloadStarted
	 * is the last point at which the AssemblyID is valid for calls to the
	 * ICorProfilerInfo interface.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(AssemblyUnloadStarted)(
                /*[in]*/ AssemblyID assemblyId);

    STDMETHOD(AssemblyUnloadFinished)(
                /*[in]*/ AssemblyID assemblyId,
                /*[in]*/ HRESULT    hrStatus);


    /*
     *
     * MODULE EVENTS
     *
     */

    /*
     * The CLR calls ModuleLoadStarted to notify the code
     * profiler that a module is about to be loaded.  Because this module has not
     * yet been loaded, other module information methods are not valid to
     * call in this callback.  Wait until ModuleLoadFinished() is
     * received to do things like acquire the metadata.
     */
    STDMETHOD(ModuleLoadStarted)(
                /*[in]*/ ModuleID moduleId);

    /*
     * The CLR calls ModuleLoadFinished to notify the code
     * profiler that a module has been loaded. The hrStatus provides the
     * success or failure of the operation
     */
    STDMETHOD(ModuleLoadFinished)(
                /*[in]*/ ModuleID moduleId,
                /*[in]*/ HRESULT  hrStatus);

    /*
     * Called before a module is being unloaded.  Use this events to collect
     * final statics that require the ModuleID to be valid.  After returning
     * from ModuleUnloadStarted, the ModuleID is no longer valid.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(ModuleUnloadStarted)(
                /*[in]*/ ModuleID moduleId);

    STDMETHOD(ModuleUnloadFinished)(
                /*[in]*/ ModuleID moduleId,
                /*[in]*/ HRESULT  hrStatus);

    /*
     * A module can get loaded through legacy means (ie: IAT or LoadLibrary) or
     * through a metadata reference.  The CLR loader therefore has many code
     * paths for determining what assembly a module lives in.  It is therefore
     * possible that after a ModuleLoadFinished event, the module does not
     * know what assembly it is in and getting the parent AssemblyID is not possible.
     * This event is fired when the module is officially attached to its parent
     * assembly.  Calling GetModuleInfo after this point will return the proper parent
     * assembly.
     */
    STDMETHOD(ModuleAttachedToAssembly)(
                /*[in]*/ ModuleID   moduleId,
                /*[in]*/ AssemblyID AssemblyId);


    /*
     *
     *  CLASS EVENTS
     *
     */

    /*
     * The CLR calls ClassLoadStarted to notify the code profiler
     * that a class is being loaded.  The ClassID is not valid for calls to
	 * the ICorProfilerInfo interface until the profiler receives a
	 * ClassLoadFinished event for the same ClassID.
     */
    STDMETHOD(ClassLoadStarted)(
                /*[in]*/ ClassID classId);

    /*
     * The CLR calls ClassLoadFinished to notify the code
     * profiler that a class has been loaded.  The ClassID is now valid for
     * calls to the ICorProfilerInfo interface.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(ClassLoadFinished)(
                /*[in]*/ ClassID classId,
                /*[in]*/ HRESULT hrStatus);

    /*
     * The given class is about to be unloaded.  Use this event to gather final
     * status and clean up anything that requires the ClassID to be valid.
     * After returning from this callback the ClassID is no longer valid.
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(ClassUnloadStarted)(
                /*[in]*/ ClassID classId);

    STDMETHOD(ClassUnloadFinished)(
                /*[in]*/ ClassID classId,
                /*[in]*/ HRESULT hrStatus);

    /*
     *
     * JIT EVENTS
     *
     */

    /*
     * The CLR calls FunctionUnloadStarted to notify the code
     * profiler that a function is being unloaded.  After returning from this
     * call, the FunctionID is no longer valid.
     *
     * NOTE: In the current version of the runtime, this callback never occurs.
     */
    STDMETHOD(FunctionUnloadStarted)(
                /*[in]*/ FunctionID functionId);


    /*
     * The CLR calls JITCompilationStarted to notify the code
     * profiler that the JIT compiler is starting to compile a function.
     *
     * The fIsSafeToBlock argument tells the profiler whether or not blocking
     * will affect the operation of the runtime.  If true, blocking may cause
     * the runtime to wait for the calling thread to return from this callback.
     * Although this will not harm the runtime, it will skew the profiling
     * results.
     *
     * NOTE: It is possible to receive more than one JITCompilationStarted/
     * JITCompilationFinished pair for each method.  This is because of how
     * the runtime handles class constructors: method A starts to be JIT'd,
     * then realizes that the class ctor for class B needs to be run, so
     * JIT's it and runs it, and while it's running makes a call to original
     * method A, which causes it to be JIT'd again, and causes the original
     * (incomplete) JIT'ing of A to be aborted.  However, both attempts to
     * JIT A are reported with JIT compilation events.  If the profiler is
     * going to replace IL code for this method with SetILFunctionBody, then
     * it must do so for both JITCompilationStarted events, but may use the
     * same IL block for both.  Note that the second IL body provided through
     * SetILFunctionBody will be the IL that is used.
     */
    STDMETHOD(JITCompilationStarted)(
                /*[in]*/ FunctionID functionId,
                /*[in]*/ BOOL       fIsSafeToBlock);

    /*
     * The CLR calls JITCompilationFinished to notify the code
     * profiler that the JIT compiler has finished compiling a function.
     *
     * The fIsSafeToBlock argument tells the profiler whether or not blocking
     * will affect the operation of the runtime.  If true, blocking may cause
     * the runtime to wait for the calling thread to return from this callback.
     * Although this will not harm the runtime, it will skew the profiling
     * results.
     *
     * The FunctionID is now valid in ICorProfilerInfo APIs.
     *
     * The hrStatus provides the success or failure of the operation
     */
    STDMETHOD(JITCompilationFinished)(
                /*[in]*/ FunctionID functionId,
                /*[in]*/ HRESULT    hrStatus,
                /*[in]*/ BOOL       fIsSafeToBlock);

    /*
     * This notifies the profiler when a search for a prejitted function is
     * starting.
     *
     *    functionId: the function for which the search is being performed.
     *    bUseCachedFunction: if true, the EE uses the cached function (if applicable)
     *                        if false, the EE jits the function instead of
	 *                        using a pre-jitted version.
     */
    STDMETHOD(JITCachedFunctionSearchStarted)(
                /*[in]*/ FunctionID functionId,
                /*[out]*/ BOOL* pbUseCachedFunction);

    /*
     * This notifies the profiler when a search for a cached function has been
     * performed.
     *
     *    functionId: the function for which the search has been performed.
     *    result: the result of the search.  There are two possible results:
     *        COR_PRF_CACHED_FUNCTION_FOUND
     *        COR_PRF_CACHED_FUNCTION_NOT_FOUND
     *
     */
    STDMETHOD(JITCachedFunctionSearchFinished)(
                /*[in]*/ FunctionID        functionId,
                /*[in]*/ COR_PRF_JIT_CACHE result);

    /*
     * The CLR calls JITFunctionPitched to notify the profiler
     * that a jitted function was removed from memory.  If the pitched
     * function is called in the future, the profiler will receive new
     * JIT compilation events as it is re-jitted.
     *
     * NOTE: the FunctionID is not valid until it is re-jitted.  When it is
     * re-jitted, it will use the same FunctionID value.
     */
    STDMETHOD(JITFunctionPitched)(
                /*[in]*/ FunctionID functionId);

    /*
     * The CLR calls JITInlining to notify the profiler that the jitter
     * is about to inline calleeId into callerId.  Set pfShouldInline to FALSE
     * to prevent the callee from being inlined into the caller, and set to
     * TRUE to allow the inline to occur.
     *
     * NOTE: Inlined functions do not provide Enter/Leave events, so if you desire
     *       an accurate callgraph, you should set FALSE.  Be aware that
     *       setting FALSE will affect performance, since inlining typically
     *       increases speed and reduces separate jitting events for the inlined
     *       method.
     */
    STDMETHOD(JITInlining)(
                /*[in]*/ FunctionID callerId,
                /*[in]*/ FunctionID calleeId,
                /*[out]*/ BOOL* pfShouldInline);

    /*
     *
     * THREAD EVENTS
     *
     */

    /*
     * The CLR calls ThreadCreated to notify the code profiler
     * that a thread has been created.  The ThreadID is valid immediately.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(ThreadCreated)(
                /*[in]*/ ThreadID threadId);

    /*
     * The CLR calls ThreadDestroyed to notify the code profiler
     * that a thread has been destroyed.  The ThreadID is no longer valid.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(ThreadDestroyed)(
                /*[in]*/ ThreadID threadId);

    /*
     * The CLR calls ThreadAssignedToOSThread to tell the profiler
     * that a managed thread is being implemented via a particualr OS thread.
     * This callback exists so that the profiler can maintain an accurate
     * OS to Managed thread mapping across fibres.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(ThreadAssignedToOSThread)(
                /*[in]*/ ThreadID managedThreadId,
                /*[in]*/ DWORD    osThreadId);

    /*
     *
     * REMOTING EVENTS
     *
     *
     * NOTE: Inprocess debugging not supported from these callbacks.
     */

    //
    // Client-side events
    //

    /*
     * NOTE: each of the following pairs of callbacks will occur on the same
     *       thread
     *   RemotingClientInvocationStarted  & RemotingClientSendingMessage
     *   RemotingClientReceivingReply     & RemotingClientInvocationFinished
     *   RemotingServerInvocationReturned & RemotingServerSendingReply

    /*
     * The CLR calls RemotingClientInvocationStarted to notify the profiler that
     * a remoting call has begun.  This event is the same for synchronous and
     * asynchronous calls.
     */
    STDMETHOD(RemotingClientInvocationStarted)();

    /*
     * The CLR calls RemotingClientSendingMessage to notify the profiler that
     * a remoting call is requiring the the caller to send an invocation request through
     * a remoting channel.
     *
     * pCookie  - if remoting GUID cookies are active, this value will correspond with the
     *            the value provided in RemotingServerReceivingMessage, if the channel
     *            succeeds in transmitting the message, and if GUID cookies are active on
     *            the server-side process.  This allows easy pairing of remoting calls,
     *            and the creation of a logical call stack.
     * fIsAsync - is true if the call is asynchronous.
     */
    STDMETHOD(RemotingClientSendingMessage)(
                /*[in]*/ GUID* pCookie,
                /*[in]*/ BOOL fIsAsync);

    /*
     * The CLR calls RemotingClientReceivingReply to notify the profiler that
     * the server-side portion of a remoting call has completed and that the client is
     * now receiving and about to process the reply.
     *
     * pCookie  - if remoting GUID cookies are active, this value will correspond with the
     *            the value provided in RemotingServerSendingReply, if the channel
     *            succeeds in transmitting the message, and if GUID cookies are active on
     *            the server-side process.  This allows easy pairing of remoting calls.
     * fIsAsync - is true if the call is asynchronous.
     */
    STDMETHOD(RemotingClientReceivingReply)(
                /*[in]*/ GUID* pCookie,
                /*[in]*/ BOOL fIsAsync);

    /*
     * The CLR calls RemotingClientInvocationFinished to notify the profiler that
     * a remoting invocation has run to completion on the client side.  If the call was
     * synchronous, this means that it has also run to completion on the server side.  If
     * the call was asynchronous, a reply may still be expected when the call is handled.
     * If the call is asynchronous, and a reply is expected, then the reply will occur in
     * the form of a call to RemotingClientReceivingReply and an additional call to
     * RemotingClientInvocationFinished to indicate the required secondary processing of
     * an asynchronous call.
     */
    STDMETHOD(RemotingClientInvocationFinished)();

    //
    // Server-side events
    //

    /*
     * The CLR calls RemotingServerReceivingMessage to notify the profiler that
     * the process has received a remote method invocation (or activation) request.  If
     * the message request is asynchronous, then the request may be serviced by any
     * arbitrary thread.
     *
     * pCookie  - if remoting GUID cookies are active, this value will correspond with the
     *            the value provided in RemotingClientSendingMessage, if the channel
     *            succeeds in transmitting the message, and if GUID cookies are active on
     *            the client-side process.  This allows easy pairing of remoting calls.
     * fIsAsync - is true if the call is asynchronous.
     */
    STDMETHOD(RemotingServerReceivingMessage)(
                /*[in]*/ GUID* pCookie,
                /*[in]*/ BOOL fIsAsync);

    /*
     * The CLR calls RemotingServerInvocationStarted to notify the profiler that
     * the process is invoking a method due to a remote method invocation request.
     */
    STDMETHOD(RemotingServerInvocationStarted)();

    /*
     * The CLR calls RemotingServerInvocationReturned to notify the profiler that
     * the process has finished invoking a method due to a remote method invocation request.
     */
    STDMETHOD(RemotingServerInvocationReturned)();

    /*
     * The CLR calls RemotingServerSendingReply to notify the profiler that
     * the process has finished processing a remote method invocation request and is
     * about to transmit the reply through a channel.
     *
     * pCookie  - if remoting GUID cookies are active, this value will correspond with the
     *            the value provided in RemotingClientReceivingReply, if the channel
     *            succeeds in transmitting the message, and if GUID cookies are active on
     *            the client-side process.  This allows easy pairing of remoting calls.
     * fIsAsync - is true if the call is asynchronous.
     */
    STDMETHOD(RemotingServerSendingReply)(
                /*[in]*/ GUID* pCookie,
                /*[in]*/ BOOL fIsAsync);

    /*
     *
     * TRANSITION EVENTS
     *
     */

    /*
     * The CLR calls UnmanagedToManagedTransition to notify the
     * code profiler that a transition from unmanaged code to managed code has
     * occurred. functionId is always the ID of the callee, and reason
     * indicates whether the transition was due to a call into managed code from
     * unmanaged, or a return from an unmanaged function called by a managed one.
     *
     * Note that if the reason is COR_PRF_TRANSITION_RETURN, then the functionId
     * is that of the unmanaged function, and will never have been jitted.
     * Unmanaged functions still have some basic information associated with
     * them, such as a name, and some metadata.
	 *
	 * Note that if the reason is COR_PRF_TRANSITION_RETURN and the callee was
	 * a PInvoke call indirect, then the runtime does not know the destination
	 * of the call and functionId will be NULL.
	 *
	 * Note that if the reason is COR_PRF_TRANSITION_CALL then it may be possible
	 * that the callee has not yet been JIT-compiled.
     */
    STDMETHOD(UnmanagedToManagedTransition)(
                /*[in]*/ FunctionID                functionId,
                /*[in]*/ COR_PRF_TRANSITION_REASON reason);


    /*
     * The CLR calls ManagedToUnmanagedTransition to notify the
     * code profiler that a transition from managed code to unmanaged code has
     * occurred. functionId is always the ID of the callee, and reason
     * indicates whether the transition was due to a call into unmanaged code from
     * managed, or a return from an managed function called by an unmanaged one.
     *
     * Note that if the reason is COR_PRF_TRANSITION_CALL, then the functionId
     * is that of the unmanaged function, and will never have been jitted.
     * Unmanaged functions still have some basic information associated with
     * them, such as a name, and some metadata.
	 *
	 * Note that if the reason is COR_PRF_TRANSITION_CALL and the callee is
	 * a PInvoke call indirect, then the runtime does not know the destination
	 * of the call and functionId will be NULL.
     */
    STDMETHOD(ManagedToUnmanagedTransition)(
                /*[in]*/ FunctionID                functionId,
                /*[in]*/ COR_PRF_TRANSITION_REASON reason);


    /*
     *
     * RUNTIME SUSPENSION EVENTS
     *
     */

    /*
     * NOTE: None of the suspension callbacks are supported by in-process
     *       debugging *except* ObjectReferences.
     */

    /*
     * The CLR calls RuntimeSuspendStarted to notify the code profiler
     * that the runtime is about to suspend all of the runtime threads.
     * All runtime threads that are in unmanaged code are permitted to continue
     * running until they try to re-enter the runtime, at which point they will
     * also suspend until the runtime resumes.  This also applies to new threads
     * that enter the runtime.  All threads within the runtime are either
     * suspended immediately if they are in interruptible code, or asked to
     * suspend when they do reach interruptible code.
     *
     * suspendReason make be any of the following values:
     *  COR_PRF_SUSPEND_FOR_GC
     *      the runtime is suspending to service a GC request.  The GC-related
     *      callbacks will occur between the RuntimeSuspendFinished and
     *      RuntimeResumeStarted events.
     *  COR_PRF_SUSPEND_FOR_CODE_PITCHING
     *      the runtime is suspending so that code pitching may occur.  This
     *      only occurs when the EJit is active with code pitching enabled.
     *      Code pitching callbacks will occur between the
     *      RuntimeSuspendFinished and RuntimeResumeStarted events.
     *  COR_PRF_SUSPEND_FOR_APPDOMAIN_SHUTDOWN
     *      the runtime is suspending so that an AppDomain can be shut down.
     *      While the runtime is suspended, the runtime will determine which
     *      threads are in the AppDomain that is being shut down, set them to
     *      abort when they resume, and then resumes the runtime.  There are
     *      no AppDomain-specific callbacks during this suspension.
     *  COR_PRF_SUSPEND_FOR_SHUTDOWN
     *      the runtime is shutting down, and it must suspend all threads to
     *      complete the operation.
     *  COR_PRF_SUSPEND_FOR_GC_PREP
     *      the runtime is preparing for a GC.
     *  COR_PRF_SUSPEND_FOR_INPROC_DEBUGGER
     *      the runtime is suspending for in-process debugging.
     *  COR_PRF_SUSPEND_OTHER
     *      the runtime is suspending for a reason other than those above.
     */
    STDMETHOD(RuntimeSuspendStarted)(
            /*[in]*/ COR_PRF_SUSPEND_REASON suspendReason);

    /*
     * The CLR calls SyncForSuspendFinished to notify the code profiler
     * that the runtime has suspended all threads needed for a runtime
     * suspension.  Note that not all runtime threads are required to be
     * suspended, as described in the comment for SyncForSuspendStarted.
     *
     * NOTE: It is guaranteed that this event will occur on the same ThreadID
     * as RuntimeSuspendStarted occurred on.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RuntimeSuspendFinished)();

    /*
     * The CLR calls RuntimeSuspendAborted to notify the code profiler
     * that the runtime is aborting the runtime suspension that was occurring.
     * This may occur if two threads simultaneously attempt to suspend the
     * runtime.
     *
     * NOTE: It is guaranteed that this event will occur on the same ThreadID
     * as the RuntimeSuspendStarted occurred on, and that only one of
     * RuntimeSuspendFinished and RuntimeSuspendAborted may occur on a single
     * thread following a RuntimeSuspendStarted event.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RuntimeSuspendAborted)();

    /*
     * The CLR calls RuntimeResumeStarted to notify the code profiler
     * that the runtime is about to resume all of the runtime threads.
     *
     * NOTE: If a thread successfully suspended the runtime, then it is *NOT*
     * guaranteed that the call to RuntimeResumeStarted will be called on the
     * same ThreadID.  In this case, it is also not guaranteed that the thread
     * calling RuntimeResumeStarted will have a non-zero ThreadID.
     * However, if a thread unsuccessfully suspends the runtime
     * (i.e., there was a conflict with another suspension), then it *IS*
     * guaranteed that the calls to RuntimeSuspspendStarted,
     * RuntimeSuspendAborted and RuntimeResumeStarted will be called from the
     * same ThreadID.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RuntimeResumeStarted)();

    /*
     * The CLR calls RuntimeResumeFinished to notify the code profiler
     * that the runtime has finished resuming all of it's threads and is now
     * back in normal operation.
     *
     * NOTE: It is *NOT* guaranteed that this event will occur on the same
     * ThreadID as the RuntimeSuspendStarted occurred on, but is guaranteed
     * to occur on the same ThreadID as the RuntimeResumeStarted occurred on.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RuntimeResumeFinished)();

    /*
     * The CLR calls ThreadSuspended to notify the code profiler
     * that a particular thread has been suspended.  All threads within managed
     * code must be suspended.  If a thread is in unmanaged code, it will be
     * allowed to continue, but will suspend upon re-entering the runtime
     * and will fire this event.  Thus, this notification could occur after
     * a suspension has completed, but before the runtime resumes.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RuntimeThreadSuspended)(
                    /*[in]*/ ThreadID threadId);

    /*
     * The CLR calls ThreadResumed to notify the code profiler
     * that a particular thread has been resumed after being suspended due to
     * a runtime suspension.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RuntimeThreadResumed)(
                    /*[in]*/ ThreadID threadId);

    /*
     *
     * GC EVENTS
     *
     */

    /*
     * The CLR calls MovedReferences with information about
     * object references that moved as a result of garbage collection.
     *
     * cMovedObjectIDRanges is a count of the number of ObjectID ranges that
     *      were moved.
     * oldObjectIDRangeStart is an array of elements, each of which is the start
     *      value of a range of ObjectID values before being moved.
     * newObjectIDRangeStart is an array of elements, each of which is the start
     *      value of a range of ObjectID values after being moved.
     * cObjectIDRangeLength is an array of elements, each of which states the
     *      size of the moved ObjectID value range.
     *
     * The last three arguments of this function are parallel arrays.
     *
     * In other words, if an ObjectID value lies within the range
     *      oldObjectIDRangeStart[i] <= ObjectID < oldObjectIDRangeStart[i] + cObjectIDRangeLength[i] 
     * for 0 <= i < cMovedObjectIDRanges, then the ObjectID value has changed to
     *      ObjectID - oldObjectIDRangeStart[i] + newObjectIDRangeStart[i] 
     *
     * NOTE: All of these callbacks are made while the runtime is suspended, so
     *       none of the ObjectID values can change until the runtime resumes
     *       and another GC occurs.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     *
     * NOTE: The profiler will receive GC-related events (except ObjectAllocated)
     *       when the profiler has been suspended for COR_PRF_SUSPEND_FOR_GC *except*
     *       for one special case.  When the runtime is shutting down, there is a
     *       stage where it is suspended for COR_PRF_SUSPEND_FOR_SHUTDOWN reason and
     *       is never resumed.  But after this suspension a garbage collection may
     *       occur without a COR_PRF_SUSPEND_FOR_GC suspension notification, and
     *       the profiler will thus receive GC-related callbacks.
     */
    STDMETHOD(MovedReferences)(
                /*[in]*/                                ULONG    cMovedObjectIDRanges,
                /*[in, size_is(cMovedObjectIDRanges)]*/ ObjectID oldObjectIDRangeStart[] ,
                /*[in, size_is(cMovedObjectIDRanges)]*/ ObjectID newObjectIDRangeStart[] ,
                /*[in, size_is(cMovedObjectIDRanges)]*/ ULONG    cObjectIDRangeLength[] );

    /*
     * The CLR calls ObjectAllocated to notify the code profiler
     * an object was allocated on the heap.
     */
    STDMETHOD(ObjectAllocated)(
                /*[in]*/ ObjectID objectId,
                /*[in]*/ ClassID classId);

    /*
     * The CLR calls ObjectsAllocatedByClass to notify the code
     * profiler about the number of objects of a particular class that were
     * allocated since the previous garbage collection. The classes and the
     * counts are passed in parallel arrays.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     *
     * NOTE: This callback will not report objects allocated in the large
     *       object heap.
     */
    STDMETHOD(ObjectsAllocatedByClass)(
                /*[in]*/                       ULONG   cClassCount,
                /*[in, size_is(cClassCount)]*/ ClassID classIds[] ,
                /*[in, size_is(cClassCount)]*/ ULONG   cObjects[] );

    /*
     * The CLR calls ObjectReferences to provide information
     * about objects in memory referenced by a given object.  This function
     * is called for each object remaining in the GC heap after a collection
     * has completed.  If the profiler returns an error from this callback,
     * the profiling services will discontinue invoking this callback until the
     * next GC.  This callback can be used in conjunction with the
     * RootReferences callback to create a complete object reference graph for
     * the runtime.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(ObjectReferences)(
                /*[in]*/                       ObjectID objectId,
                /*[in]*/                       ClassID  classId,
                /*[in]*/                       ULONG    cObjectRefs,
                /*[in, size_is(cObjectRefs)]*/ ObjectID objectRefIds[] );

    /*
     * The CLR calls RootReferences with information about root
     * references after a garbage collection has occurred. Static object
     * references and references to objects on a stack are co-mingled in the
     * arrays.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(RootReferences)(
                /*[in]*/                     ULONG    cRootRefs,
                /*[in, size_is(cRootRefs)]*/ ObjectID rootRefIds[] );


    /*
     *
     * EXCEPTION EVENTS
     *
     *
     * NOTE: Inprocess debugging not supported from any exception callbacks
     *       except ExceptionThrown.
     */

    //
    // Exception creation
    //

    /*
     * The CLR calls ExceptionThrown to notify the code
     * profiler that an exception has been thrown.
     *
     * NOTE: This function is only called if the CLR exception handler
     *       is called to process an exception.
     *
     * NOTE: For in-process debugging, thrownObjectId is always NULL.  The
     *       actual thrown object can be found through
     *       ICorDebugThread::GetCurrentException
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionThrown)(
                /*[in]*/ ObjectID thrownObjectId);

    //
    // Search phase
    //

    /*
     * The CLR calls ExceptionSearchFunctionEnter to notify the profiler
     * that the search phase of exception handling has entered a function.
     */
    STDMETHOD(ExceptionSearchFunctionEnter)(
                /*[in]*/ FunctionID functionId);

    /*
     * The CLR calls ExceptionSearchFunctionLeave to notify the profiler
     * that the search phase of exception handling has left a function.
     */
    STDMETHOD(ExceptionSearchFunctionLeave)();

    /*
     * The CLR will call ExceptionSearchFilterEnter just before excecuting
     * a user filter.  The functionID is that of the function containing the filter.
     */
    STDMETHOD(ExceptionSearchFilterEnter)(
                /*[in]*/ FunctionID functionId);

    /*
     * The CLR will call ExceptionSearchFilterLeave immediately after
     * executing a user filter.
     */
    STDMETHOD(ExceptionSearchFilterLeave)();

    /*
     * The CLR will call ExceptionSearchCatcherFound when the search
     * phase of exception handling has located a handler for the exception that
     * was thrown.
     */
    STDMETHOD(ExceptionSearchCatcherFound)(
                /*[in]*/ FunctionID functionId);

    /*
     * The CLR calls ExceptionHandlerEnter when the runtime's
     * [Win32 SEH] exception handler is entered.  There may or may not be
     * managed functions guarded by this handler, the profiler may or may not
     * receive exception-related callbacks between ExceptionOSHandlerEnter and
     * ExceptionOSHandlerLeave callbacks.
     *
     * This notification is provided to allow profilers to detect unmanaged-to-
     * managed transitions in stack searches and unwinds.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     *
     * NOTE: NOT YET IMPLEMENTED.
     */
    STDMETHOD(ExceptionOSHandlerEnter)(
                /*[in]*/ UINT_PTR __unused);

    /*
     * This function is similar to ExceptionHandlerEnter, except that
     * it is called just before the runtime's exception handler returns.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     *
     * NOTE: NOT YET IMPLEMENTED.
     */
    STDMETHOD(ExceptionOSHandlerLeave)(
                /*[in]*/ UINT_PTR __unused);

    //
    // Unwind phase
    //

    /*
     * The CLR calls ExceptionUnwindFunctionEnter to notify the profiler
     * that the unwind phase of exception handling has entered a function.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionUnwindFunctionEnter)(
                /*[in]*/ FunctionID functionId);

    /*
     * The CLR calls ExceptionUnwindFunctionLeave to notify the profiler
     * that the unwind phase of exception handling has left a function.  The
     * function instance and it's stack data has now been removed from the stack.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionUnwindFunctionLeave)();

    /*
     * The CLR calls ExceptionUnwindFinallyEnter to notify the profiler
     * that the unwind phase of exception is entering a finally clause contained
     * in the specified function.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionUnwindFinallyEnter)(
                /*[in]*/ FunctionID functionId);

    /*
     * The CLR calls ExceptionUnwindFinallyLeave to notify the profiler
     * that the unwind phase of exception is leaving a finally clause.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionUnwindFinallyLeave)();

    /*
     * The CLR calls this function just before passing control to 
     * the appropriate catch block.  Note that this is called only if the 
     * catch point is in JIT'ed code.  An exception that is caught in 
     * unmanaged code, or in the internal code of the CLR will
     * not generate this notification.  The ObjectID is passed again since
     * a GC could have moved the object since the ExceptionThrown
     * notification.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionCatcherEnter)(
                /*[in]*/ FunctionID functionId,
                /*[in]*/ ObjectID   objectId);

    /*
     * The CLR calls ExceptionCatcherLeave when the runtime leaves
     * the catcher's code.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     */
    STDMETHOD(ExceptionCatcherLeave)();

    /*
     *  CLR<->COM interop vtable creation/destruction.
     */

    /*
     * The CLR calls this function when an CLR<->COM interop vtable
     * for a particular IID and for a particular class has been created.
     * This provides the ClassID of the class for which this vtable has been
     * created, the IID it implements, the start of the vtable and how many
     * slots are in it.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     *
     * NOTE: It is possible to receive a NULL GUID if the interface is
     *       internal only.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(COMClassicVTableCreated)(
               /*[in]*/ ClassID wrappedClassId,
               /*[in]*/ REFGUID implementedIID,
               /*[in]*/ void* pVTable,
               /*[in]*/ ULONG   cSlots);

    /*
     * The CLR calls this function when a CLR<->COM interop vtable is
     * being destroyed.  Provided are the ClassID, IID and vtable pointer for
     * the destroyed vtable.
     *
     * NOTE: The profiler should not block here, since the stack may not be in a
     *       GC-friendly state and so preemptive GC cannot be enabled.  If the
     *       profiler blocks here and a GC is attempted, the runtime will block
     *       until this callback returns.  Also, the profiler may NOT call into
     *       managed code or in any way cause a managed memory allocation.
     *
     * NOTE: It is possible to receive a NULL GUID if the interface is
     *       internal only.
     *
     * NOTE: Inprocess debugging not supported from this callback.
     */
    STDMETHOD(COMClassicVTableDestroyed)(
               /*[in]*/ ClassID wrappedClassId,
               /*[in]*/ REFGUID implementedIID,
               /*[in]*/ void* pVTable);

    /*
     * When the CLR itself catches an exception, these functions are called
     * to notify the profiler.  Between the Found and Execute callbacks the
     * CLR may also unwind managed frames and so normal ExceptionUnwind*
     * functions can be called in between.
     *
     * NOTE: for backward compatibility the profiler must also set the
     *       COR_PRF_MONITOR_CLR_EXCEPTIONS event flag to receive these events.
     *
     * NOTE: Cannot use inprocess debugging from these callbacks.
     */
    STDMETHOD(ExceptionCLRCatcherFound)();

    STDMETHOD(ExceptionCLRCatcherExecute)();

private: // HELPER FUNCTIONS

	HRESULT GetLogSettingsNode(MSXML::IXMLDOMNode** ppiXmlLogSettings);

public:	// DATA
	static HINSTANCE			s_hInst;        // instance handle to this piece of code

private: // DATA
	CComPtr<ICorProfilerInfo>	m_spiProfInfo;
	bool						m_bBreakImmediately;
	HANDLE						m_hWatcherStartedEvent;
	HANDLE						m_hThreadsPleaseTerminate;
	HANDLE						m_hHeartbeatTimer;
	CModuleQueue				m_modulesToWatchQueue;
	TypeMapMethodInfo			m_mapChangedFunctions;
	CCriticalSection			m_csListChangedFunctions;
	DWORD						m_dwPfnThePreStubFunc;
};

#endif // INCLUDED_PATCHER_PROFILER_CALLBACK_H
