///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/UpdaterThread.cpp 10    12/15/05 2:37a Eaddy $
//
//    File name:	UpdaterThread.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/UpdaterThread.cpp $
 * 
 * 10    12/15/05 2:37a Eaddy
 * o Improved statistics gathering and display
 * 
 * 9     12/10/05 12:42a Eaddy
 * o Now keeps track of elapsed times
 * 
 * 8     12/09/05 2:03p Eaddy
 * o Fixed bug where we were requesting rejit of a function that hadn't
 * been jit'd (called) yet.  Apparently, we now get a non-zero functionID
 * even if the function hasn't been jit'd.
 * 
 * 7     5/28/05 1:49p Eaddy
 * o Changed copyright
 *
 * Revision 1.5  2004/10/24 19:33:11  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.4  2004/10/17 08:46:53  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.3  2004/10/11 23:23:36  me133
 * o Now logs the method name not just the function id
 *
 * Revision 1.2  2004/10/10 06:07:46  me133
 * o Added header comment
 */

#include "StdAfx.h"

#include "UpdaterThread.h"			// Me

#include "MetadataHelpers.h"		// GetMethodNameFromFunctionId()

#include "MethodInfo.h"				// CMethodInfo

#include "WatcherSocket.h"			// CWatcherSocket

#include "Helpers.h"				// WaitForDispatchThreadEvents()

#include <corerror.h>				// CLDB_E_BADUPDATEMODE


CUpdaterArg::CUpdaterArg(	TypeMapMethodInfo* pmapChangedFunctions,
							CRITICAL_SECTION* pListCS,
							DWORD* pdwPfnThePreStubFunc,
							ICorProfilerInfo* piProfInfo,
							int nPortToUse,
							HANDLE hWatcherReadyEvent,
							HANDLE hThreadPleaseTerminate,
							HANDLE hHeartbeatTimer)
	: m_pmapChangedFunctions(pmapChangedFunctions)
	, m_pListCS(pListCS)
	, m_pdwPfnThePreStubFunc(pdwPfnThePreStubFunc)
	, m_piProfInfo(piProfInfo)
	, m_nPortToUse(nPortToUse)
	, m_hWatcherReadyEvent(hWatcherReadyEvent)
	, m_hThreadPleaseTerminate(hThreadPleaseTerminate)
	, m_hHeartbeatTimer(hHeartbeatTimer)
{ 
	_ASSERTE(m_piProfInfo);
	if (m_piProfInfo)
	{
		m_piProfInfo->AddRef();
	}
}

CUpdaterArg::~CUpdaterArg()
{
	_ASSERTE(m_piProfInfo);
	if (m_piProfInfo)
	{
		m_piProfInfo->Release();
	}
}

//*********************************************************************
// FUNCTION		:	UpdaterThreadProc
//					
// DESCRIPTION	:	
//					
// PARAMETERS	:	
//   pvParam	
//					
// COMMENTS		:	
//					
//*********************************************************************
void __cdecl UpdaterThreadProc(void* pvCUpdaterArg)
{
	ILXLOGMSG(_T("UPDATER: Starting."), ILXLOG_INFO);

	HRESULT hr = S_OK;

	// Step 1:	Declare local variables
	//
	//			All local variables must be declared here because of 
	//			the goto statement

	CComPtr<ICorProfilerInfo> spiProfInfo;
	CComPtr<IMetaDataDispenserEx> spiMetaDataDispenser;
	CComPtr<IMetaDataImport> spiImportNew;

	CWatcherSocket watcherSocket;

	CUpdaterArg* pUpdaterArg					= 0;

	HANDLE hQueuedObjectAvailable				= 0;
		// Signaled by ModuleQueue when an ModuleID has been queued

	HANDLE hThreadPleaseTerminate				= 0;
		// Signaled by CProfilerCallback when the thread should exit

	TypeMapMethodInfo*	pmapChangedFunctions	= 0;
	CRITICAL_SECTION* pListCS					= 0;
	DWORD* pdwPfnThePreStubFunc					= 0;
	int nPortToUse								= 0;
	HANDLE hWatcherReadyEvent					= NULL;
	bool bFirstTime								= true;

	// Verify input parameters

	_ASSERTE(pvCUpdaterArg);
	if (!pvCUpdaterArg)
	{
		goto LABEL_EXIT_THREAD;
	}

	// Step 2:	Unpack thread argument

	pUpdaterArg = reinterpret_cast<CUpdaterArg*>(pvCUpdaterArg);
	if (!pUpdaterArg)
	{
		goto LABEL_EXIT_THREAD;
	}

	// Grab the params
	pmapChangedFunctions			= pUpdaterArg->m_pmapChangedFunctions;
	pListCS							= pUpdaterArg->m_pListCS;
	pdwPfnThePreStubFunc			= pUpdaterArg->m_pdwPfnThePreStubFunc;
	nPortToUse						= pUpdaterArg->m_nPortToUse;
	hWatcherReadyEvent				= pUpdaterArg->m_hWatcherReadyEvent;
	hThreadPleaseTerminate			= pUpdaterArg->m_hThreadPleaseTerminate;

	if (!pmapChangedFunctions || !pListCS || !pdwPfnThePreStubFunc || 
		nPortToUse <= 0 || !hWatcherReadyEvent || !hThreadPleaseTerminate)
	{
		ILXLOG4(hr, ILXLOG_ERROR, _T("Invalid thread parameters - %p - %p - %p - %d."), 
			pmapChangedFunctions, pListCS, pdwPfnThePreStubFunc, nPortToUse);
		goto LABEL_EXIT_THREAD;
	}

	// Get the ICorProfilerInfo ptr
	spiProfInfo = pUpdaterArg->m_piProfInfo; // This had better do an AddRef()
	if (spiProfInfo == 0)
	{
		hr = E_UNEXPECTED;
		ILXLOG(hr, _T("Failed to initialize event dispatcher thread."), ILXLOG_ERROR);
		_ASSERTE(!"Failed to initialize event dispatcher thread.");
		goto LABEL_EXIT_THREAD;
	}

	// Free the memory for the arg object
	delete pUpdaterArg;
	pUpdaterArg = 0;

	// Step 3:	Create events

	// Step 4:	Initialize COM and enter a new STA

	hr = ::CoInitializeEx(0, COINIT_APARTMENTTHREADED);
	if (FAILED(hr))
	{
		ILXLOG(hr, _T("Failed to initialize COM."), ILXLOG_ERROR);
		_ASSERTE(!"Failed to initialize COM");
		return;
	}

	// Step 5:	Create the metadata dispenser

	_ASSERTE(!spiMetaDataDispenser.p);

	hr = MyCoCreateInstance(	CLSID_CorMetaDataDispenser, 
								NULL, 
								CLSCTX_SERVER,
								IID_IMetaDataDispenserEx,
								(void**)(&spiMetaDataDispenser.p));

    if (FAILED(hr) || !spiMetaDataDispenser)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		ILXLOG(hr, "Unable to create metadata dispenser.", ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	// Step 6:	Wait for the Watcher process to start

	ILXLOGMSG(_T("UPDATER: Waiting for Watcher to start."), ILXLOG_INFO);

	// Here we are treating a WSAEvent just like a standard Win32 event.
	// This seems to work even though WSAEvents have their own wait function.
	EDispatchThreadWaitStatus eWaitStatus = 
		WaitForDispatchThreadEvents(hWatcherReadyEvent,
									hThreadPleaseTerminate);
	
	if (eWaitStatus == DISPATCH_THREAD_WAIT_ERROR)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());

		ILXLOG(hr, _T("Event queue wait failed."), ILXLOG_ERROR);
		_ASSERTE(!"Event queue wait failed!");

		goto LABEL_EXIT_THREAD;
	}
	else if (eWaitStatus == DISPATCH_THREAD_SHUTDOWN_SIGNALED)
	{
		// We are being told to close down
		ILXLOG(hr, _T("Watcher process never got started."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}
	
	_ASSERTE(eWaitStatus == DISPATCH_THREAD_EVENT_WAITING);

	// Step 7:	Connect to Watcher

	static const int nMAX_RETRIES = 10;

	ILXLOGMSG(_T("UPDATER: Attempting to connect to Watcher."), ILXLOG_INFO);

	hr = watcherSocket.Connect(nPortToUse, nMAX_RETRIES, true);
	if (FAILED(hr))
	{
		ILXLOG(hr, _T("Failed to connect to Watcher server."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	hQueuedObjectAvailable = watcherSocket.GetReadEventHandle();
	if (!hQueuedObjectAvailable)
	{
		ILXLOGMSG(_T("GetReadEventHandle() failed."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	hr = watcherSocket.RequestChangedModuleInfo();
	if (FAILED(hr))
	{
		ILXLOG(hr, _T("RequestChangedModuleInfo() failed."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	// Step 8:	Handle events forever!

	while (true)
	{
		// Step 9:	Wait for either a new queued event or until
		//			we are told to exit

		// Here we are treating a WSAEvent just like a standard Win32 event.
		// This seems to work even though WSAEvents have their own wait function.
		EDispatchThreadWaitStatus eWaitStatus = 
			WaitForDispatchThreadEvents(hQueuedObjectAvailable,
										hThreadPleaseTerminate);
		
		if (eWaitStatus == DISPATCH_THREAD_WAIT_ERROR)
		{
			hr = HRESULT_FROM_WIN32(::GetLastError());

			ILXLOG(hr, _T("Event queue wait failed."), ILXLOG_ERROR);
			_ASSERTE(!"Event queue wait failed!");

			goto LABEL_EXIT_THREAD;
		}
		else if (eWaitStatus == DISPATCH_THREAD_SHUTDOWN_SIGNALED)
		{
			// We are being told to close down
			goto LABEL_EXIT_THREAD;
		}

		_ASSERTE(eWaitStatus == DISPATCH_THREAD_EVENT_WAITING);

		DWORD dwModuleID = 0;
		char szModulePath[MAX_PATH+1] = { 0 };

		DWORD dwTicksStart = 0;
		DWORD dwWeaveTimeEnd = 0;
		DWORD dwCompileTimeEnd = 0;

		hr = watcherSocket.GetChangedModuleInfo(&dwModuleID, szModulePath, MAX_PATH, 
			&dwTicksStart, &dwWeaveTimeEnd, &dwCompileTimeEnd);
		if (FAILED(hr))
		{
			ILXLOGMSG(_T("GetChangedModuleInfo() failed."), ILXLOG_ERROR);
			continue;
		}

		if (dwModuleID == 0 || !szModulePath || !*szModulePath)
		{
			ILXLOG(E_FAIL, _T("Module ID is 0 or module path is empty. Skipping."), ILXLOG_ERROR);
			continue;
		}

		//ILXLOGMSG2(ILXLOG_INFO, _T("UPDATER THREAD: %p %s."), dwModuleID, szModulePath);

		// Step 10:	Open the "new" assembly and map it into memory

		PBYTE pbMapViewOfFileNew = 0;
		DWORD dwFileSize = 0;
		hr = MyGetMapViewOfFile(szModulePath, &pbMapViewOfFileNew, &dwFileSize);
		if (FAILED(hr))
		{
			ILXLOG1(hr, ILXLOG_ERROR, _T("MyGetMapViewOfFile() failed: '%s'."), szModulePath);
			continue;
		}

		// Step 11:	Obtain the metadata from the assembly

		PVOID pvMetaDataNew = 0;
		ULONG nMetaDataNewSize = 0;
		hr = MyFindMetaData(pbMapViewOfFileNew, dwFileSize, &pvMetaDataNew, &nMetaDataNewSize);
		if (FAILED(hr))
		{
			ILXLOG1(hr, ILXLOG_ERROR, _T("MyFindObjMetaData() failed: '%s'."), szModulePath);
			continue;
		}

		if (pvMetaDataNew == 0 || nMetaDataNewSize == 0)
		{
			ILXLOGMSG3(ILXLOG_ERROR, _T("Metadata is invalid for file '%s' (Metadata=%p Size=%d)."),
				szModulePath, pvMetaDataNew, nMetaDataNewSize);
			continue;
		}

		// Step 12:	Open a scope directly on the in-memory metadata

		spiImportNew = 0; // Don't forget to release first

		hr = spiMetaDataDispenser->OpenScopeOnMemory(pvMetaDataNew, 
												nMetaDataNewSize, 
												0, // Flags
												IID_IMetaDataImport, 
												(IUnknown**)(&spiImportNew.p));
		if (hr == CLDB_E_BADUPDATEMODE)
		{
			VARIANT value;
			V_VT(&value) = VT_UI4;
			V_UI4(&value) = MDUpdateIncremental;

			hr = spiMetaDataDispenser->SetOption(MetaDataSetUpdate, &value);
			if (FAILED(hr))
			{
				ILXLOG(hr, "SetOption() failed.", ILXLOG_ERROR);
			}

			hr = spiMetaDataDispenser->OpenScopeOnMemory(pvMetaDataNew, 
													nMetaDataNewSize, 
													0, // Flags
													IID_IMetaDataImport, 
													(IUnknown**)(&spiImportNew.p));
		}

		if (FAILED(hr) || !spiImportNew.p)
		{
			hr = FAILED(hr) ? hr : E_FAIL;
			ILXLOG(hr, "OpenScope() failed.", ILXLOG_ERROR);
			continue;
		}

		// We should probably check here to see if hThreadPleaseTerminate
		// has been signaled while we've been away processing
		if (WaitForSingleObject(hThreadPleaseTerminate, 0) ==
			WAIT_OBJECT_0)
		{
			// Unmap the file
			::UnmapViewOfFile(pbMapViewOfFileNew);
			pbMapViewOfFileNew = 0;

			// We are being told to close down
			goto LABEL_EXIT_THREAD;
		}

		// Step 13:	Create the update patches

		hr = CreateUpdatePatches(	spiProfInfo,
									dwModuleID,
									spiImportNew, 
									pbMapViewOfFileNew,
									bFirstTime,
									pmapChangedFunctions);
		bFirstTime = false;

		// Unmap the file
		::UnmapViewOfFile(pbMapViewOfFileNew);
		pbMapViewOfFileNew = 0;

		// Quick check to see if we've been told to shutdown
		if (WaitForSingleObject(hThreadPleaseTerminate, 0) == WAIT_OBJECT_0)
		{
			goto LABEL_EXIT_THREAD;
		}

		// Quick sanity check
		if (!*pdwPfnThePreStubFunc)
		{
			ILXLOG(E_UNEXPECTED, 
				_T("We cannot ReJIT any functions because pdwPfnThePreStubFunc hasn't been set."),
				ILXLOG_ERROR);
			continue;
		}

		// Step 15:	Request the updated functions to be ReJIT'd

		for(TypeMapMethodInfo::iterator itChangedMethods = pmapChangedFunctions->begin();
			itChangedMethods != pmapChangedFunctions->end();
			++itChangedMethods)
		{
			CMethodInfo& patch = (*itChangedMethods).second;

			// Don't Re-JIT a function that hasn't changed since the last time we Re-JIT'd
			if (!patch.m_bWasModified)
			{
				continue;
			}

			patch.m_dwTicksStart = dwTicksStart;
			patch.m_dwWeaveTimeEnd = dwWeaveTimeEnd;
			patch.m_dwCompileTimeEnd = dwCompileTimeEnd;
			patch.m_dwPatchTimeEnd = ::GetTickCount();

			// Can't ReJit a function that hasn't been called (JIT'd) yet. If we try,
			// SetFunctionReJIT() will return CORPROF_E_CANNOT_UPDATE_METHOD.

			// First do a trivial check
			if (patch.m_functionID == 0)
			{
				ILXLOGMSG2(ILXLOG_INFO, _T("Not requesting re-JIT for '%ls.%ls' since it hasn't been called (JIT'd) yet."), 
					patch.m_szTypeName,
					patch.m_szMemberName);
				continue;
			}
			else
			{
				// Now do a definitive check
				const BYTE* pbIgnore = 0;
				ULONG nIgnore = 0;
				hr = spiProfInfo->GetCodeInfo(patch.m_functionID, &pbIgnore, &nIgnore);
				if (hr == CORPROF_E_FUNCTION_NOT_COMPILED)
				{
					ILXLOGMSG2(ILXLOG_INFO, _T("Not requesting re-JIT for '%ls.%ls' since it hasn't been called (JIT'd) yet."), 
						patch.m_szTypeName,
						patch.m_szMemberName);
					continue;
				}
			}

			CHECK_HR(hr, _T("GetCodeInfo() failed."), ILXLOG_ERROR);

			ILXLOGMSG4(ILXLOG_INFO, _T("Request Re-JIT for '%ls.%ls' (FunctionID: 0x%08X, MethodDef: 0x%08X)."), 
				patch.m_szTypeName,
				patch.m_szMemberName,
				patch.m_functionID,
				patch.m_methodDef);

			hr = SetFunctionReJITBugFix(	spiProfInfo, 
											patch.m_functionID, 
											*pdwPfnThePreStubFunc);
			CHECK_HR4(hr, ILXLOG_ERROR,
				_T("SetFunctionReJIT() failed for function '%ls.%ls' (FunctionID: 0x%08X, MethodDef: 0x%08X)."), 
				patch.m_szTypeName,
				patch.m_szMemberName,
				patch.m_functionID,
				patch.m_methodDef);
		}
	}

LABEL_EXIT_THREAD:

	ILXLOGMSG(_T("Stopping updater thread."), ILXLOG_INFO);

	// Release these before CoUninitialize()!
	spiProfInfo.Release();

	::CoUninitialize();

	_endthread();
}
