///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/WatcherKeeperThread.cpp 3     5/28/05 1:20p Eaddy $
//
//    File name:	WatcherKeeperThread.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/22/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/WatcherKeeperThread.cpp $
 * 
 * 3     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.3  2005/01/14 23:09:55  me133
 * o Now uses create no window
 * o Fixed err msg
 *
 * Revision 1.2  2004/10/26 04:43:53  me133
 * o Stops heartbeat if it doesn't work so we don't keep printing an error
 * message every second
 *
 * Revision 1.1  2004/10/24 21:07:21  me133
 * no message
 *
 */

#include "StdAfx.h"

#include "WatcherKeeperThread.h"	// Me

#include "WatcherSocket.h"			// CWatcherSocket

#include "Helpers.h"				// WaitForDispatchThreadEvents()

// CONSTANTS

static const long nHEARTBEAT_PERIOD					= 1000; // 1 second
static const TCHAR* szENABLE_PROFILING_ENV_VAR_NAME = _T("Cor_Enable_Profiling");
static const TCHAR* szWATCHER_PATH_ENV_VAR_NAME		= _T("WATCHER_PATH");
static const TCHAR* szDEFAULT_WATCHER_PATH			= _T("C:\\PROJECTS\\Goblin\\Watcher\\Watcher.exe");

// Helper functions

static HRESULT StartWatcher(int nPortToUse, PROCESS_INFORMATION* pWatcherProcessInfo);

//////////////////////////////////////
// WATCHER KEEPER ARG IMPLEMENTATION
//////////////////////////////////////

CWatcherKeeperArg::CWatcherKeeperArg(	int nPortToUse,
										HANDLE hWatcherReadyEvent,
										HANDLE hThreadPleaseTerminate,
										HANDLE hHeartbeatTimer)
	: m_nPortToUse(nPortToUse)
	, m_hWatcherReadyEvent(hWatcherReadyEvent)
	, m_hThreadPleaseTerminate(hThreadPleaseTerminate)
	, m_hHeartbeatTimer(hHeartbeatTimer)
{ }

CWatcherKeeperArg::~CWatcherKeeperArg()
{ }

//////////////////////////////////////////////
// MODULE WATCHER THREAD PROC IMPLEMENTATION
//////////////////////////////////////////////

void __cdecl WatcherKeeperThreadProc(void* pvCWatcherKeeperArg)
{
	//ILXLOGMSG(_T("WATCHER KEEPER: Starting."), ILXLOG_INFO);

	HRESULT hr = S_OK;

	// Step 1:	Declare all variables
	//
	//			All local variables must be declared here because
	//			of the goto statement

	CWatcherSocket watcherSocket;

	CWatcherKeeperArg* pWatcherKeeperArg		= 0;

	HANDLE hHeartbeatTimer						= 0;

	HANDLE hThreadPleaseTerminate				= 0;
		// Signaled by CProfilerCallback when the thread should exit

    LARGE_INTEGER liDueTime;

	BOOL bRet									= FALSE;

	PROCESS_INFORMATION	watcherProcessInfo;
	::ZeroMemory(&watcherProcessInfo, sizeof(watcherProcessInfo));

	int nPortToUse								= 0;
	HANDLE hWatcherReadyEvent					= NULL;

	bool bKeepSendingHeartbeat					= true;

	// Verify input parameters

	_ASSERTE(pvCWatcherKeeperArg);
	if (!pvCWatcherKeeperArg)
	{
		goto LABEL_EXIT_THREAD;
	}

	// Step 2:	Unpack thread arguments

	pWatcherKeeperArg = reinterpret_cast<CWatcherKeeperArg*>(pvCWatcherKeeperArg);
	if (!pWatcherKeeperArg)
	{
		goto LABEL_EXIT_THREAD;
	}

	nPortToUse			= pWatcherKeeperArg->m_nPortToUse;
	hWatcherReadyEvent	= pWatcherKeeperArg->m_hWatcherReadyEvent;
	hThreadPleaseTerminate = pWatcherKeeperArg->m_hThreadPleaseTerminate;
	hHeartbeatTimer		= pWatcherKeeperArg->m_hHeartbeatTimer;

	if (nPortToUse <= 0 || !hWatcherReadyEvent || !hThreadPleaseTerminate || !hHeartbeatTimer)
	{
		ILXLOG3(hr, ILXLOG_ERROR, _T("Invalid thread parameters - %d - %p - %p."), 
			nPortToUse, hWatcherReadyEvent, hHeartbeatTimer);
		goto LABEL_EXIT_THREAD;
	}

	// Free the memory for the arg object
	delete pWatcherKeeperArg;
	pWatcherKeeperArg = 0;
	
	// Step 3:	Start Watcher and connect to it

	static const int nMAX_RETRIES = 10;

	ILXLOGMSG(_T("WATCHER KEEPER: Attempting to connect to Watcher."), ILXLOG_INFO);

	hr = watcherSocket.Connect(nPortToUse, 1, false);
	if (FAILED(hr))
	{
		ILXLOGMSG(_T("WATCHER KEEPER: Watcher not started, attempting to start."), ILXLOG_INFO);

		hr = StartWatcher(nPortToUse, &watcherProcessInfo);
		if (FAILED(hr))
		{
			ILXLOG(hr, _T("Failed to start Watcher server."), ILXLOG_ERROR);
			goto LABEL_EXIT_THREAD;
		}

		hr = watcherSocket.Connect(nPortToUse, nMAX_RETRIES, true);
	}

	if (FAILED(hr))
	{
		ILXLOG(hr, _T("Failed to connect to Watcher server."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	// Signal the other threads that the Watcher is ready
	bRet = ::SetEvent(hWatcherReadyEvent);
	if (!bRet)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());
		ILXLOG(hr, _T("Failed to set the Watcher started event."), ILXLOG_ERROR);
		goto LABEL_EXIT_THREAD;
	}

	liDueTime.QuadPart = nHEARTBEAT_PERIOD * -10000; // 1 seconds (negative means relative)

	// Step 4:	Handle events forever!

	while (true)
	{
		// Reset the timer
		if (bKeepSendingHeartbeat)
		{
			bRet = ::SetWaitableTimer(hHeartbeatTimer, &liDueTime, 0, 0, 0, FALSE);
			if (!bRet)
			{
				hr = HRESULT_FROM_WIN32(GetLastError());
				ILXLOG(hr, _T("Failed to restart the watcher keeper thread timer."), ILXLOG_ERROR);
				goto LABEL_EXIT_THREAD;
			}
		}

		// Step 5:	Wait for either a timer event or until we are told to exit

		EDispatchThreadWaitStatus eWaitStatus = 
			WaitForDispatchThreadEvents(hHeartbeatTimer,
										hThreadPleaseTerminate);
		
		if (eWaitStatus == DISPATCH_THREAD_WAIT_ERROR)
		{
			hr = HRESULT_FROM_WIN32(::GetLastError());
			ILXLOG(hr, _T("Event queue wait failed."), ILXLOG_ERROR);
			_ASSERTE(!"Event queue wait failed!");

			goto LABEL_EXIT_THREAD;
		}
		else if (eWaitStatus == DISPATCH_THREAD_SHUTDOWN_SIGNALED)
		{
			// We are being told to close down
			goto LABEL_EXIT_THREAD;
		}
		
		_ASSERTE(eWaitStatus == DISPATCH_THREAD_EVENT_WAITING);

		// SEND HEARTBEAT MESSAGE
		//ILXLOGMSG(_T("Thump thump."), ILXLOG_INFO);
		hr = watcherSocket.SendHeartbeat();
		if (FAILED(hr))
		{
			static int nMaxMsgs = 1;
			if (nMaxMsgs > 0)
			{
				ILXLOG(hr, _T("Failed to send heartbeat message to Watcher."), ILXLOG_WARNING);
				--nMaxMsgs;
			}
			else
			{
				ILXLOG(hr, _T("Failed to send heartbeat message to Watcher. Stopping hearbeat."), ILXLOG_WARNING);

				bRet = ::SetWaitableTimer(hHeartbeatTimer, &liDueTime, 0, 0, 0, FALSE);
				if (bRet)
				{
					bRet = ::CancelWaitableTimer(hHeartbeatTimer);
					if (!bRet)
					{
						hr = HRESULT_FROM_WIN32(GetLastError());
						ILXLOG(hr, _T("Failed to stop watcher keeper thread timer."), ILXLOG_ERROR);
					}
				}
				bKeepSendingHeartbeat = false;
			}
		}
	}

LABEL_EXIT_THREAD:

	// Step 6:	Send shutdown message to Watcher
	if (watcherSocket.IsConnected())
	{
		hr = watcherSocket.SendShutdown();
		CHECK_HR(hr, _T("Failed to shutdown socket."), ILXLOG_WARNING);
	}

	// Step 7:	Shutdown and cleanup the Watcher process

	if (watcherProcessInfo.hProcess)
	{
		static const int nPROCESS_CLOSE_TIMEOUT = 100; // ms

		ILXLOGMSG(_T("Closing Watcher process."), ILXLOG_INFO);

		// Wait until child process exits.
		DWORD dwRet = WaitForSingleObject( watcherProcessInfo.hProcess, nPROCESS_CLOSE_TIMEOUT);
		if (dwRet == WAIT_OBJECT_0)
		{
			ILXLOGMSG(_T("Watcher process exited."), ILXLOG_SUCCESS);
		}
		else
		{
			hr = HRESULT_FROM_WIN32(::GetLastError());
			ILXLOG(hr, _T("Failed to close Watcher process."), ILXLOG_ERROR);
		}

		CloseHandle(watcherProcessInfo.hProcess);
		CloseHandle(watcherProcessInfo.hThread);
	}

	ILXLOGMSG(_T("Stopping watcher keeper thread."), ILXLOG_INFO);

	_endthread();
}

// Helper function implementations

HRESULT StartWatcher(int nPortToUse, PROCESS_INFORMATION* pWatcherProcessInfo)
{
	HRESULT hr = S_OK;

	if (!pWatcherProcessInfo) return E_POINTER;

	// Disable profiling of Watcher. If we don't do this then Watcher itself
	// will be profiled and the log file will have a sharing violation.
	BOOL bRet = SetEnvironmentVariable(szENABLE_PROFILING_ENV_VAR_NAME, "0");
	if (!bRet)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG(hr, _T("Failed to set Cor_Enable_Profiling env variable."), ILXLOG_ERROR);
	}

	STARTUPINFO si;
	ZeroMemory( &si, sizeof(si) );
	si.cb = sizeof(si);

	char szCLIX[MAX_PATH];
	bool bIsRotor = false;
	GetRotorAndCLRInfo(&bIsRotor, 0, szCLIX, MAX_PATH-1);

	char szPort[100];
	sprintf(szPort, "%d", nPortToUse);

	char szCmdLine[MAX_PATH*2];
	szCmdLine[0] = '\0';

	if (bIsRotor)
	{
		strcpy(szCmdLine, szCLIX);
		strcat(szCmdLine, " ");
	}

	TCHAR szWatcherPath[_MAX_PATH] = { 0 };

	DWORD dwRet = ::GetEnvironmentVariable(	szWATCHER_PATH_ENV_VAR_NAME, 
											szWatcherPath,
											_MAX_PATH );
	const TCHAR* pszWatcherPathToUse = 0;

	if (dwRet && *szWatcherPath)
	{
		pszWatcherPathToUse = szWatcherPath;
	}
	else
	{
		pszWatcherPathToUse = szDEFAULT_WATCHER_PATH;
	}

	strcat(szCmdLine, "\"");
	strcat(szCmdLine, pszWatcherPathToUse);
	strcat(szCmdLine, "\" ");
	strcat(szCmdLine, szPort);
	
	// FIX ME (USE CREATE_NO_WINDOW)
	bRet = CreateProcess(	0,
							szCmdLine,
							0,					// Process attr
							0,					// Thread attr
							TRUE,				// Inherits parent's handles
							0/*CREATE_NO_WINDOW*/,	// Avoid creating a console window when Patching
												//	 non-Console apps
							0,					// Use parent's environment
							0,					// Use parent's current dir
							&si,				// Default STARTUPINFO
							pWatcherProcessInfo);
	if (!bRet)
	{
		hr = HRESULT_FROM_WIN32(::GetLastError());
		return ILXLOG1(hr, ILXLOG_ERROR, _T("Failed to start the Watcher process %s."), 
			szCmdLine);
	}

	return S_OK;
}