///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Patcher/WatcherSocket.h 6     12/15/05 2:37a Eaddy $
//
//    File name:	WatcherSocket.h
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Patcher/WatcherSocket.h $
 * 
 * 6     12/15/05 2:37a Eaddy
 * o Improved statistics gathering and display
 * 
 * 5     12/10/05 1:07p Eaddy
 * o Now provides elapsed times
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2004/10/24 19:33:12  me133
 * o Implemented heartbeat message
 * o Created WatcherKeeper thread for starting and stopping the Watcher
 * and for sending the heartbeat message
 * o Now reads Watcher path from env variable
 * o Fixed bug where meerly changing a string didn't cause a change to be
 * detected
 *
 * Revision 1.3  2004/10/17 08:46:53  me133
 * o Now uses the incremental file info to pass the correct compile settings
 * to Watcher. Still need to pass resource files.  We aren't able to parse
 * the incr file in all cases (Text3D.exe.incr built with /debug).
 *
 * Revision 1.2  2004/10/10 06:07:46  me133
 * o Added header comment
 */

#ifndef INCLUDED_WATCHER_SOCKET_H
#define INCLUDED_WATCHER_SOCKET_H

#include "CompileInfo.h"	// ICompileInfo

#include <winsock2.h>		// WSA*

#include <list>				// STL list<>
#include <string>			// STL string<>
#include <atlsync.h>		// CCriticalSection

// Forward declarations
class CIncrBuildFile;

class CWatcherSocket
{
public:
	CWatcherSocket();
	~CWatcherSocket();

	HRESULT Connect(int nPortToUse, int nRetries, bool bLogConnectFailure);

	HRESULT SendModuleInfo(	DWORD dwModuleID,
							const char* szModulePath,
							const CCompileInfo* pCompileInfo);

	HRESULT RequestChangedModuleInfo();

	HRESULT GetChangedModuleInfo(	DWORD* pdwModuleID,
									char* szModulePath,
									int nMaxLen,
									DWORD* pdwStartTicks,
									DWORD* pdwWeaveTimeEnd,
									DWORD* pdwCompileTimeEnd);

	HRESULT SendHeartbeat();

	HRESULT SendShutdown();

	WSAEVENT GetReadEventHandle();

	bool IsConnected() const;

private: // HELPER FUNCTIONS

	HRESULT SendUINT32(UINT32 ui);
	HRESULT SendString(const char* sz, size_t nLen = -1);

	HRESULT ReadUINT32(UINT32* pui);
	HRESULT ReadString(char* sz, size_t nMaxLen);

private: // DATA
	SOCKET				m_socket;
	WSAEVENT			m_wsaEvent;
};

#endif // INCLUDED_WATCHER_SOCKET_H
