///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/AppConfig.cs 7     10/07/05 7:59p Eaddy $
//
//    File name:	AppConfig.cs
//   Created By:	Marc Eaddy
//   Created on:	3/22/05
//
//      Purpose:	Loads supported languages for App.config
//
//  Description:	Some of the App.config loader code was borrowed from
//					John Conwell's DotNetScript code.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/AppConfig.cs $
 * 
 * 7     10/07/05 7:59p Eaddy
 * o Fixed exception that occurs when adding a key that already exists
 * 
 * 6     5/28/05 1:24p Eaddy
 * o Changed copyright
 * 
 * 5     5/25/05 6:33p Eaddy
 * o Now uses EnC.Watcher namespace
 * 
 * 4     3/22/05 6:54p Eaddy
 * o Improved log message
 * 
 * 3     3/22/05 6:05p Eaddy
 * 
 * 2     3/22/05 6:04p Eaddy
 * o Added shout out
 * 
 * 1     3/22/05 5:58p Eaddy
 * o Initial checkin
 */

using System;
using System.Xml;
using System.Collections;
using System.Diagnostics;
using System.Configuration;

namespace EnC.Watcher
{
	/// <summary>
	/// Summary description for AppConfig.
	/// </summary>
	internal class AppConfigLanguageSectionHandler : IConfigurationSectionHandler
	{
		internal static Hashtable codeProviders = new Hashtable();

		public AppConfigLanguageSectionHandler()
		{ }

		public static CodeDomProviderDesc getProviderForCompiler(string compilerName)
		{
			return (AppConfigLanguageSectionHandler.CodeDomProviderDesc) codeProviders[compilerName];
		}

		public static CodeDomProviderDesc getProviderForLanguage(string languageName)
		{
			foreach(DictionaryEntry descPair in codeProviders)
			{
				CodeDomProviderDesc desc = (CodeDomProviderDesc) descPair.Value;

				if (desc.languageName == languageName)
				{
					return desc;
				}
			}
			
			return null;
		}

		/// <summary>
		/// IConfigurationSectionHandler implementation.  The XmlNode
		/// param passes in one section at a time.  So if you have 4 referenced assemblies,
		/// Create() will be called 4 times, once for each referenced assembly
		/// </summary>
		public object Create(object parent, object configContext, XmlNode section)
		{
			string languageName = section.Attributes["name"].Value;
			string assembly = section.Attributes["assembly"].Value;
			string codeProviderName = section.Attributes["codeProviderName"].Value;
			string compilerName = section.Attributes["compilerName"].Value;
		
			MyTrace.Log(TraceLevel.Info, "Loaded language info: {0} ({1}.exe), {2} ({3})",
				languageName, 
				compilerName,
				codeProviderName, 
				assembly);

			if (!codeProviders.Contains(compilerName))
			{
				codeProviders.Add(compilerName, 
					new CodeDomProviderDesc(languageName, assembly, codeProviderName, compilerName));
			}
		
			return codeProviders;
		}

		/// <summary>
		/// </summary>
		internal class CodeDomProviderDesc
		{
			internal readonly string languageName = string.Empty;
			internal readonly string assembly = string.Empty;
			internal readonly string codeProviderName = string.Empty;
			internal readonly string compilerName = string.Empty;

			internal CodeDomProviderDesc(	string languageName, 
				string assembly, 
				string codeProviderName,
				string compilerName)
			{
				this.languageName = languageName;
				this.assembly = assembly;
				this.codeProviderName = codeProviderName;
				this.compilerName = compilerName;
			}
		}
	}
}
