///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/DynModule.cs 33    12/16/05 8:43p Eaddy $
//
//    File name:	DynModule.cs
//   Created By:	Marc Eaddy
//   Created on:	10/17/04
//
//      Purpose:	
//
//  Description:	
//
// Currently, the C++/CLI and J# compilers do not work.  The C++/CLI CodeDomProvider
// returns null when CreateCompiler() is called.  The J# compiler compiles but there
// are problems when the function is patched that cause an InvalidProgramException.
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/DynModule.cs $
 * 
 * 33    12/16/05 8:43p Eaddy
 * o More source exclusions :(
 * 
 * 32    12/15/05 2:41a Eaddy
 * o Fixed stats
 * o Excludes some problematic Goblin files
 * 
 * 31    12/10/05 12:43a Eaddy
 * o Now keeps track of elapsed times
 * 
 * 30    12/08/05 3:57p Eaddy
 * o Minor change to log message
 * 
 * 29    12/07/05 11:38p Eaddy
 * o Now logs the output module
 * 
 * 28    12/07/05 12:41a Eaddy
 * o Now displays language name
 * o Aspects can now add import assemblies
 * 
 * 27    11/11/05 5:37p Eaddy
 * o Removes the /recurse option.  You must specify the exact file list in
 * the Makefile or else we have problems getting the file order exactly
 * the same.
 * 
 * 26    10/07/05 8:01p Eaddy
 * o Added ability to turn off load-time weving
 * 
 * 25    5/28/05 1:23p Eaddy
 * o Changed copyright
 * 
 * 24    5/25/05 6:33p Eaddy
 * o Now uses EnC.Watcher namespace
 * 
 * 23    4/22/05 7:36p Eaddy
 * o Can now turn off weaving using an attribute in the XML file
 * 
 * 22    4/04/05 2:27p Eaddy
 * o Abstracted watching multiple files into FileWatcherCollection
 * o Abstracted handling aspects into AspectRulesFile
 * o Added comments
 * o Made OnFileChanged thread-safe since it will be called from separate
 * threads via MyFileWatcher
 * 
 * 21    4/01/05 6:56p Eaddy
 * o Now handles parsing exceptions
 * 
 * 20    4/01/05 3:33p Eaddy
 * o Fixed exception when aspect xml file is not found
 * 
 * 19    4/01/05 1:01p Eaddy
 * o Fixed null ref exception
 * 
 * 18    4/01/05 12:09p Eaddy
 * o Fixed compile error
 * 
 * 17    4/01/05 11:15a Eaddy
 * o Now supports rudimentary weaving
 * 
 * 15    3/28/05 12:03p Eaddy
 * o Fixed bug where we were deleting all the compiler options flags
 * 
 * 14    3/23/05 5:44p Eaddy
 * o Added comment
 * 
 * 13    3/23/05 4:10p Eaddy
 * o Now removes source files from compiler options by default (unless
 * /recurse is specified)
 * 
 * 12    3/22/05 6:47p Eaddy
 * o Now removes double source files for VB.NET compiler only
 * o Fixed parsing bug
 * 
 * 11    3/22/05 6:25p Eaddy
 * o Now supports languages that are configured in app.config
 * o Now extracts references
 * 
 * 10    3/21/05 11:48a Eaddy
 * o Added Managed C++ support, however, CreateCompiler() always returns
 * null
 * 
 * 9     3/15/05 5:01p Eaddy
 * o Now passes through most compiler options
 * o Now creates a new temp exe each time
 * 
 * 8     3/10/05 5:00p Eaddy
 * o DEBUG: Now compiles right away
 * 
 * 7     3/10/05 1:28p Eaddy
 * 
 * 6     3/10/05 1:24p Eaddy
 * o Seems to work
 * 
 * 5     3/10/05 12:43p Eaddy
 * o Started to fix compiler options
 * 
 * 4     3/09/05 11:55p Eaddy
 * o Now receives compiler name
 * o Now uses compile options string exclusively
 * 
 * 3     2/27/05 1:56a Eaddy
 * o Made code for finding compiler provider more generic
 * 
 * 2     2/26/05 4:02a Eaddy
 * o Now supports Edit-and-Continue for VB.NET!!
 * o Can theorectically support Edit-and-Continue for JScript.NET (still
 * need to test)
 * Revision 1.2  2004/10/24 19:31:03  me133
 * o Implemented heartbeat message
 * o Now handles more network error conditions
 *
 * Revision 1.1  2004/10/17 08:45:25  me133
 * o Now uses DynModule which keeps track of all the compiler settings
 * and references (it gets these from Patcher)
 *
 * Revision 1.6  2004/10/15 06:57:24  me133
 * o Switched for WINNT for the laptop
 *
 * Revision 1.5  2004/10/11 23:21:25  me133
 * o Added a bunch more references to catch as many as I can
 *
 * Revision 1.4  2004/10/11 05:57:53  me133
 * no message
 *
 * Revision 1.3  2004/10/10 07:40:06  me133
 * o Now outputs trace messages to log file usnig System.Diagnostics.Trace
 *
 * Revision 1.2  2004/10/10 06:07:23  me133
 * o Added header comment
 */

using System;
using System.Diagnostics;
using System.Collections;
using System.Text;
using System.Reflection;
using System.IO;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Threading;
using System.Net.Sockets;

using DDW.CSharp.Parse;
using DDW.CSharp.Dom;
using DDW.CSharp.Gen;

namespace EnC.Watcher
{
	public delegate void OnModuleUpdatedDelegate(uint moduleID, string tmpModulePath, 
		int nTicksStart, int nTicksWeaveEnd, int nTicksCompileEnd);
	public delegate string OnCompilerOptionFound(string options, string optionName, int start, int length);

	/// <summary>
	/// Watches a module's source files and recompiles the module if they change.
	/// </summary>
	/// <remarks>
	/// Aspects described in AspectRules.xml are weaved into the source files
	/// according to the aspect rules.  Changes made to any of the aspect source
	/// files or AspectRules.xml will also trigger reloading the aspects, reweaving,
	/// and recompiling.
	/// </remarks>
	public class DynModule : FileWatcherCollection
	{
		/// <summary>
		/// Delegate that is called when the module has been recompiled successfully.
		/// </summary>
		OnModuleUpdatedDelegate	_onModuleUpdatedDelegate = null;

		/// <summary>
		/// Compiler used to build this module.
		/// </summary>
		ICodeCompiler			_compiler = null;

		/// <summary>
		/// Compiler name ("csc")
		/// </summary>
		string					_compilerName = null;

		/// <summary>
		/// Language name ("C#")
		/// </summary>
		string					_languageName = null;

		/// <summary>
		/// Compiler parameters used to build the module ("/debug").
		/// </summary>
		CompilerParameters		_compilerParams = new CompilerParameters();

		/// <summary>
		/// Opaque ModuleID associated with this module. Sent from Patcher.
		/// </summary>
		uint					_moduleID = 0;

		/// <summary>
		/// Source files for this module.
		/// </summary>
		string[]				_arySourceFiles = null;

		/// <summary>
		/// Weaved source files for this module. If no weaving has occured, this
		/// array is the same as _arySourceFiles.
		/// </summary>
		string[]				_aryWeavedSourceFiles = null;

		/// <summary>
		/// Path to the original module ("C:/blah/Hello.exe").
		/// </summary>
		string					_targetPathOrig = null;

		/// <summary>
		/// Name of the module without the extension ("Hello"). We cache this for efficiency.
		/// </summary>
		string					_moduleName = null;

		/// <summary>
		/// Module extension (".exe"). We cache this for efficiency.
		/// </summary>
		string					_moduleExt = null;

		/// <summary>
		/// Count used to create a unique temporary file name.
		/// </summary>
		int						_tmpCount = 0;

		/// <summary>
		/// Temporary directory for recompiled modules ("C:/blah/EnC/")
		/// </summary>
		string					_tmpDir = null;

		/// <summary>
		/// Manages the aspects.
		/// </summary>
		static AspectsFile	_aspectsFile = new AspectsFile();

		static System.Configuration.AppSettingsReader appSettingsReader 
			= new System.Configuration.AppSettingsReader();

		/// <summary>
		/// Creates a dynamic module. Must call init() or readFromSocket() to initialize.
		/// </summary>
		public DynModule()
		{
			// Initialize compiler options
			_compilerParams.CompilerOptions = "";
			_compilerParams.GenerateInMemory = false;
			_compilerParams.GenerateExecutable = true;
			_compilerParams.OutputAssembly = "";

			_aspectsFile.setFileChangedDelegate(new MyFileWatcher.OnChangedDelegate(OnFileChanged));
		}

		/// <summary>
		/// Initializes the dynamic module.
		/// </summary>
		/// <param name="moduleID">Opaque ModuleID of the module. Sent from Patcher.</param>
		/// <param name="pathToOrigModule">Path to original module ("C:\blah\Hello.exe")</param>
		/// <param name="compilerName">Name of the command-line compiler ("csc.exe")</param>
		/// <param name="compilerOptions">Compiler options ("/debug Hello.cs")</param>
		/// <param name="debugBuild">Is this a debug build?</param>
		/// <param name="sources">Module's source files ("Hello.cs", "HelloHelper.cs")</param>
		public void init(	uint moduleID, 
							string pathToOrigModule, 
							string compilerName, 
							string compilerOptions, 
							bool debugBuild,
							string[] sources)
		{
			_moduleID = moduleID;
			Debug.Assert(_moduleID >= 0);
			_targetPathOrig = pathToOrigModule;
			_compilerName = Path.GetFileNameWithoutExtension(compilerName);
			_compilerParams.IncludeDebugInformation = debugBuild;
			_compilerParams.CompilerOptions = compilerOptions;
			_arySourceFiles = sources;
			_aryWeavedSourceFiles = new string[_arySourceFiles.Length];

			// Prune some options

			MyTrace.Log(TraceLevel.Info, "Compiler options before removing some flags: '{0}'", 
				_compilerParams.CompilerOptions);
			
			findCompilerOption("/out:",			new OnCompilerOptionFound(removeCompilerOption));
			findCompilerOption("/bugreport:",	new OnCompilerOptionFound(removeCompilerOption));
			findCompilerOption("/r:",			new OnCompilerOptionFound(extractReference));
			findCompilerOption("/reference:",	new OnCompilerOptionFound(extractReference));

			// KLUDGE: Ideally, we would parse out all the source files from the compiler options.
			// However, the /recurse flags makes it difficult because it requires us to perform
			// the recurse algorithm to locate the source files.  The list of source files found
			// must be exactly the same (same contents and order) as the one produced by the compiler
			// or else the assemblies will be different. For now, if /recurse is used we don't try
			// to parse the source files.
			findCompilerOption("/recurse:",		new OnCompilerOptionFound(removeCompilerOption));
			findSources(						new OnCompilerOptionFound(removeCompilerOption));
			//if (_compilerParams.CompilerOptions.IndexOf("/recurse") == -1)
			//{
			//	findSources(						new OnCompilerOptionFound(removeCompilerOption));
			//}

			// Remove the /debug flag since this is specified using
			// _compilerParams.IncludeDebugInformation.
			//removeCompilerOption("/debug");

			// Remove extra whitespace
			_compilerParams.CompilerOptions = _compilerParams.CompilerOptions.Trim();

			MyTrace.Log(TraceLevel.Info, "Compiler options after removing some flags: '{0}'", 
				_compilerParams.CompilerOptions);

			// Register for changes to the source files

			for(int i = 0; i < _arySourceFiles.Length; ++i)
			{
				registerForChanges(_arySourceFiles[i]);
			}
		}

		/// <summary>
		/// Reads initialization data from TCP/IP message sent from Patcher.
		/// </summary>
		/// <param name="socket"></param>
		/// <returns>true if initialization data was read successfully</returns>
		public bool readFromSocket(Socket socket)
		{
			// Step 1:	Wait for module ID
			uint moduleID = Globals.ReadUINT32(socket);
			if (moduleID < 0)
			{
				return false;
			}

			MyTrace.Log(TraceLevel.Info, "GOT MODULE INFO!");

			// See if Patcher wants us to shutdown
			if (moduleID == (int) Globals.NetworkMessage.nSHUTDOWN) // Shutdown
			{
				MyTrace.Log(TraceLevel.Info, "GOT SHUTDOWN MESSAGE!");
				return false;
			}

			MyTrace.Log(TraceLevel.Info, "GOT MODULE ID: {0}", moduleID);

			// Step 2:	Extract target path

			NetworkStream networkStream = new NetworkStream(socket);
			BinaryReader binaryReader = new BinaryReader(networkStream);

			string pathToOrigModule = Globals.ReadLengthEncodedASCIIString(binaryReader);
			MyTrace.Log(TraceLevel.Info, "GOT MODULE PATH: {0}", pathToOrigModule);

			// Step 3:	Extract compiler name

			string compilerName = Globals.ReadLengthEncodedASCIIString(binaryReader);
			MyTrace.Log(TraceLevel.Info, "GOT COMPILER NAME: {0}", compilerName);

			// Step 4:	Get debug flag
			//
			// NOTE: The debug flag has to be set explicitly, you cannot just append
			// " /debug" to _compilerParams.CompilerOptions. Otherwise, all the RVAs
			// will be off in the resulting compiled assembly because the
			// DebuggableAttribute Custom Attribute will not be defined.

			bool debugBuild = Globals.ReadUINT32(socket) != 0;

			// Step 5:	Get compiler options

			string compilerOptions = 
				Globals.ReadLengthEncodedASCIIString(binaryReader);

			// Step 6:	Get source files

			uint nNumSourcePaths = Globals.ReadUINT32(socket);
			if (nNumSourcePaths == 0)
			{
				return false; // We expect at least one source file
			}
						
			MyTrace.Log(TraceLevel.Info, "GOT SOURCE PATH COUNT: {0}", nNumSourcePaths);

			string[] sources = new string[nNumSourcePaths];
			for(int i = 0; i < nNumSourcePaths; ++i)
			{
				sources[i] = Globals.ReadLengthEncodedASCIIString(binaryReader);
				MyTrace.Log(TraceLevel.Info, "GOT SOURCE PATH: {0}", sources[i]);
			}

			MyTrace.Log(TraceLevel.Info, "GOT EVERYTHING FOR THE MODULE");

			// Step 7: Initialize

			init(moduleID, pathToOrigModule, compilerName, compilerOptions, debugBuild, sources);
			return true;
		}

		/// <summary>
		/// Searches for a compiler option.
		/// </summary>
		/// <param name="strOptionName">Option name ("/r:")</param>
		/// <param name="onCompilerOptionFound">Delegate called each time the option is found.</param>
		private void findCompilerOption(	string strOptionName, 
											OnCompilerOptionFound onCompilerOptionFound)
		{
			Debug.Assert(onCompilerOptionFound != null);

			int nFind = _compilerParams.CompilerOptions.IndexOf(strOptionName);
			while (nFind > -1 && nFind+1 < _compilerParams.CompilerOptions.Length)
			{
				int nEnd = _compilerParams.CompilerOptions.IndexOf(" ", nFind+1);
				if (nEnd == -1)
				{
					nEnd = _compilerParams.CompilerOptions.Length;
				}

				Debug.Assert(nEnd > nFind && nEnd <= _compilerParams.CompilerOptions.Length);

				_compilerParams.CompilerOptions = 
					onCompilerOptionFound(_compilerParams.CompilerOptions, 
											strOptionName, nFind, nEnd-nFind);

				nFind = _compilerParams.CompilerOptions.IndexOf(strOptionName, nFind);
			}
		}

		/// <summary>
		/// Looks for source files in the compiler option string.
		/// </summary>
		/// <param name="onCompilerOptionFound">Delegate called for each found source file.</param>
		private void findSources(OnCompilerOptionFound onCompilerOptionFound)
		{
			Debug.Assert(onCompilerOptionFound != null);

			int nFind = 0;
			while(nFind > -1 && nFind+1 < _compilerParams.CompilerOptions.Length)
			{
				int nEnd = _compilerParams.CompilerOptions.IndexOf(" ", nFind+1);
				if (nEnd == -1)
				{
					nEnd = _compilerParams.CompilerOptions.Length;
				}

				Debug.Assert(nEnd > nFind && nEnd <= _compilerParams.CompilerOptions.Length);

				// Only compiler flags start with '/'. The rest must be source files.
				if (_compilerParams.CompilerOptions[nFind+1] != '/')
				{
					_compilerParams.CompilerOptions = 
						onCompilerOptionFound(_compilerParams.CompilerOptions, 
						"", nFind, nEnd-nFind);
				}
				else
				{
					++nFind; // Skip over the space
				}

				nFind = _compilerParams.CompilerOptions.IndexOf(" ", nFind);
			}
		}

		/// <summary>
		/// Implements the OnCompilerOptionFound delegate. Removes the found compiler option.
		/// </summary>
		/// <param name="options"></param>
		/// <param name="optionName"></param>
		/// <param name="start"></param>
		/// <param name="length"></param>
		/// <returns></returns>
		private string removeCompilerOption(string options, string optionName, int start, int length)
		{
			if (start-1 >= 0 &&
				options[start-1] == ' ')
			{
				--start;
				++length;
			}

			//MyTrace.Log(TraceLevel.Info, "Removing: '{0}'", options.Substring(start, length));

			return options.Remove(start, length);
		}

		/// <summary>
		/// Implements the OnCompilerOptionFound delegate. Extracts the assembly reference option
		/// and adds it to the list of ReferencedAssemblies.
		/// </summary>
		/// <param name="options"></param>
		/// <param name="optionName"></param>
		/// <param name="start"></param>
		/// <param name="length"></param>
		/// <returns></returns>
		private string extractReference(string options, string optionName, int start, int length)
		{
			string assembly = options.Substring(	start + optionName.Length, 
													length-optionName.Length);
			_compilerParams.ReferencedAssemblies.Add(assembly);

			return removeCompilerOption(options, optionName, start, length);
		}

		/// <summary>
		/// Creates a temporary filename for the module in the EnC subdirectory.
		/// </summary>
		/// <returns>Path to temporary file.</returns>
		private string createOutputFilePath()
		{
			if (_tmpDir == null)
			{
				// Crack the path
				_moduleName = Path.GetFileNameWithoutExtension(_targetPathOrig);
				_moduleExt = Path.GetExtension(_targetPathOrig);

				// Create the tmp directory and tmp output file
				_tmpDir = Path.Combine(Path.GetDirectoryName(_targetPathOrig), "EnC");
			}
			
			string tmpModulePath;

			// Keep incrementing file name until we find one that doesn't exist
			do
			{
				tmpModulePath = Path.Combine(_tmpDir, _moduleName + "-" + _tmpCount.ToString() + _moduleExt);
				++_tmpCount;

			} while (File.Exists(tmpModulePath));

			MyTrace.Log(TraceLevel.Info, "Tmp Module Path: " + tmpModulePath);

			return tmpModulePath;
		}

		/// <summary>
		/// Sets the callback used to notify when a module has been recompiled.
		/// </summary>
		/// <param name="onModuleUpdatedDelegate"></param>
		public void setModuleUpdatedCallback(OnModuleUpdatedDelegate onModuleUpdatedDelegate)	
		{
			_onModuleUpdatedDelegate = onModuleUpdatedDelegate;

			bool bCompileOnLoad = false;

			try
			{
				bCompileOnLoad = (bool) appSettingsReader.GetValue("CompileOnLoad", 
					typeof(bool));
			}
			catch(InvalidOperationException ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to retrieve value for CompileOnLoad. "
					+ ex.Message);
			}

			if (bCompileOnLoad)
			{
				// Kicks off the initial weave step at "load-time"

				int nTicksStart = Environment.TickCount;

				MyTrace.Log(TraceLevel.Info, "Load-time weaving pass. Recompiling.");

				lock(this)
				{
					weaveAndCompile(nTicksStart);
				}
			}
		}

		/// <summary>
		/// Internal callback called by MyFileWatcher when a file is changed.
		/// </summary>
		/// <param name="path">Path of changed file.</param>
		protected override void OnFileChanged(string path)
		{
			int nTicksStart = Environment.TickCount;

			MyTrace.Log(TraceLevel.Info, "Change detected: " + path + ". Recompiling.");

			// Always reweave and recompile when a file changes
			lock(this)
			{
				weaveAndCompile(nTicksStart);
			}
		}

		/// <summary>
		/// Weaves aspects into the module source code and then compiles the weaved module.
		/// </summary>
		public void weaveAndCompile(int nTicksStart)
		{
			if (_arySourceFiles == null || _arySourceFiles.Length == 0)
			{
				MyTrace.Log(TraceLevel.Error, "No source files found.");
				return;
			}

			// Step 1: Create temporary module filename ("EnC/Hello_CS-0.exe")
			_compilerParams.OutputAssembly = createOutputFilePath();

			// Step 2:	Create the temporary output directory ("EnC/")

			try
			{
				Directory.CreateDirectory(_tmpDir);
			}
			catch(Exception ex)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to create directory '{0}': {1}",
					_tmpDir, ex.Message);
				return;
			}

			// Step 3:	Create the compiler if it hasn't been created
			if (!createCompiler())
			{
				MyTrace.Log(TraceLevel.Error, "Failed to create compiler.");
				// We already logged an error
				return;
			}

			// Step 4:	Weave in aspect code
			int nTicksWeaveStart = Environment.TickCount;
			weaveAspects();
			int nTicksWeaveEnd = Environment.TickCount;
			MyTrace.Log(TraceLevel.Info, "Weave time: {0}ms.", 
				nTicksWeaveEnd - nTicksWeaveStart);

			// Step 5:	Log the compiler command
			logCompilerCommand();

			// Step 6:	Compile!

			const int nNUM_RETRIES = 5;
			const int nTIME_TO_SLEEP_BETWEEN_RETRIES = 200; // ms

			CompilerResults results = null;

			for(int i = 0; i < nNUM_RETRIES; ++i)
			{
				try
				{
					results = _compiler.CompileAssemblyFromFileBatch(	_compilerParams, 
																		_aryWeavedSourceFiles);
					break;
				}
				catch (IOException ex)
				{
					MyTrace.Log(TraceLevel.Warning, "Compiler had problems accessing file. Retrying. - {0}",
						ex.Message);
					
					Thread.Sleep(nTIME_TO_SLEEP_BETWEEN_RETRIES);
					continue;
				}
				catch (Exception ex)
				{
					MyTrace.Log(TraceLevel.Warning, "Compiler threw an exception. Retrying. - {0}",
						ex.Message);
					Thread.Sleep(nTIME_TO_SLEEP_BETWEEN_RETRIES);
					continue;
				}
			}

			int nTicksCompileEnd = Environment.TickCount;

			MyTrace.Log(TraceLevel.Info, "Compile elapsed time: {0}ms.", 
				nTicksCompileEnd - nTicksWeaveEnd);

			if (results == null)
			{
				MyTrace.Log(TraceLevel.Error, "Compile failed.");
				return;
			}
			else if (results.Errors.Count > 0)
			{
				foreach (CompilerError error in results.Errors)
				{
					MyTrace.Log(TraceLevel.Error, error.ToString());
				}
				return;
			}
			else if (results.Output.Count > 0)
			{
				foreach (string strOutput in results.Output)
				{
					MyTrace.Log(TraceLevel.Warning, strOutput.ToString());
				}
			}
			else
			{
				MyTrace.Log(TraceLevel.Info, "Compile succeeded.");
			}

			// Step 7:	Notify interested listeners

			if (_onModuleUpdatedDelegate != null)
			{
				_onModuleUpdatedDelegate(_moduleID, 
					_compilerParams.OutputAssembly, 
					nTicksStart,
					nTicksWeaveEnd,
					nTicksCompileEnd);
			}
		}

		/// <summary>
		/// Weaves the aspect code in with the source code according to the aspect rules.
		/// </summary>
		private void weaveAspects()
		{
			Helper helper = new Helper();

			for(int i = 0; i < _arySourceFiles.Length; ++i)
			{
				if (_aspectsFile == null || !_aspectsFile.Enabled)
				{
					_aryWeavedSourceFiles[i] = _arySourceFiles[i];
					continue;
				}

				IGraph ig = null;

				try
				{
					ig = helper.GetGraph(_arySourceFiles[i]);
				}
				catch(Exception ex)
				{
					MyTrace.Log(TraceLevel.Error, "Failed to parse source file '{0}': {1}",
						_arySourceFiles[i],
						ex.ToString());
					_aryWeavedSourceFiles[i] = _arySourceFiles[i];
					continue;
				}

				if (ig == null)
				{
					_aryWeavedSourceFiles[i] = _arySourceFiles[i];
					continue;
				}
				else
				{
					ArrayList aspectAssemblies = new ArrayList();

					_aspectsFile.instrument(ig, aspectAssemblies);

					foreach(object aspectAssemblyObj in aspectAssemblies)
					{
						string aspectLibrary = (string) aspectAssemblyObj;
						_compilerParams.ReferencedAssemblies.Add(aspectLibrary);
					}
				}

				string weavedPath = getWeavedPath(_arySourceFiles[i]);

				try
				{
					StreamWriter sr = File.CreateText(weavedPath);

					CSharpGen csg = new CSharpGen(sr);
					csg.Parse(ig);
					csg.Close();
					sr.Close();

					_aryWeavedSourceFiles[i] = weavedPath;
				}
				catch(Exception ex)
				{
					MyTrace.Log(TraceLevel.Error, "Failed to generate weaved file: {0}",
						ex.Message);

					_aryWeavedSourceFiles[i] = _arySourceFiles[i];
				}
			}
		}

		/// <summary>
		/// Creates a temporary filename for the weaved file.
		/// </summary>
		/// <param name="path">Original path.</param>
		/// <returns>Path to weaved file.</returns>
		private static string getWeavedPath(string path)
		{
			// "Hello.cs" -> "Hello.weaved.cs"
			return Path.Combine(Path.GetDirectoryName(path),
				Path.GetFileNameWithoutExtension(path) + ".weaved" + Path.GetExtension(path));
		}

		/// <summary>
		/// Creates the compiler used to compile the source files.
		/// </summary>
		/// <returns>true if the compiler could be created, false otherwise.</returns>
		private bool createCompiler()
		{
			if (_compiler != null)
			{
				return true;
			}
			
			Debug.Assert(_compilerName != null);

			CodeDomProvider provider = null;

			System.Configuration.ConfigurationSettings.GetConfig("supportedLanguages/language");

			AppConfigLanguageSectionHandler.CodeDomProviderDesc providerDesc = 
				AppConfigLanguageSectionHandler.getProviderForCompiler(_compilerName);

			// Special Case: For some reason the bugreport file emitted by the
			// VB.NET compiler (vbc) has the compiler name as "(null)"
			if (providerDesc == null && 
				_compilerName == "(null)" &&
				_arySourceFiles.Length > 0 &&
				Path.GetExtension(_arySourceFiles[0]) == ".vb")
			{
				// Note: "VB.NET" must match the supportedLanguages/language/name
				// attribute in App.config
				providerDesc = 
					AppConfigLanguageSectionHandler.getProviderForLanguage("VB.NET");
				_compilerName = providerDesc.compilerName;
			}
			
			if (providerDesc == null)
			{
				// Set to C# compiler by default
				MyTrace.Log(TraceLevel.Error, 
					"Unknown compiler name '{0}'. Defaulting to C# compiler.",
					_compilerName);

				provider = new Microsoft.CSharp.CSharpCodeProvider();
			}
			else
			{
				// Create an instance of the code provider
				provider = (CodeDomProvider) Helper.CreateInstance(providerDesc.assembly, providerDesc.codeProviderName);
			}

			if (provider == null)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to create compiler '{0}'.",
					_compilerName);
				return false;
			}

			_languageName = providerDesc.languageName;

			MyTrace.Log(TraceLevel.Info, "Using {0} compiler ('{1}').", _languageName, _compilerName);

			_compiler = provider.CreateCompiler();
			if (_compiler == null)
			{
				MyTrace.Log(TraceLevel.Error, "Failed to create '{0}' compiler.", _compilerName);
				return false;
			}

			return true;
		}

		/// <summary>
		/// Logs the compiler command used to build the files.
		/// </summary>
		private void logCompilerCommand()
		{
			if (TraceLevel.Info > MyTrace.level.Level)
				return;

			Debug.Assert(_compilerName != null && _compilerName.Length > 0);

			string strRefs = "";
			foreach (string strRef in _compilerParams.ReferencedAssemblies)
			{
				strRefs += " /r:" + strRef;
			}

			string strSources = "";
			foreach (string strSource in _aryWeavedSourceFiles)
			{
				strSources += " " + strSource;
			}

			MyTrace.Log(TraceLevel.Info,
				"Command-line: {0} /out:{1}{2}{3}{4}{5}",
				_compilerName,
				_compilerParams.OutputAssembly,
				_compilerParams.IncludeDebugInformation ? " /debug" : "",
				_compilerParams.CompilerOptions.Length > 0 ? " " + _compilerParams.CompilerOptions : "",
				strRefs,
				strSources);
		}
	}

	/// <summary>
	/// Holder of some helper functions.
	/// </summary>
	public class Helper
	{
		bool bErrorReported = false;

		/// <summary>
		/// Obtains the Type for a string. Type.GetType(string) is supposed to do 
		/// this exact thing but it doesn't work.
		/// </summary>
		/// <param name="typeName"></param>
		/// <returns></returns>
		static public Type GetType(string typeName)
		{
			Type type = Type.GetType(typeName);
			if (type != null)
				return type;

			// Scour the other loaded assemblies for the type. This is exactly what
			// Type.GetType() is supposed to do.
			AppDomain domain = AppDomain.CurrentDomain;
			Assembly[] assemblies = domain.GetAssemblies();
			foreach (Assembly assembly in assemblies)
			{
				Module[] modules2 = assembly.GetModules();
				foreach(Module module in modules2)
				{
					type = module.GetType(typeName);
					if (type != null)
						return type;
				}
			}

			return null;
		}

		/// <summary>
		/// Creates an instance of the type whose name is specified, using the named assembly and
		/// default constructor (same as Activator.CreateInstance). For some reason,
		/// Activator.CreateInstance() didn't work.
		/// </summary>
		/// <param name="assemblyName">Name of the assembly (without .dll) or full path to the assembly (with .dll).</param>
		/// <param name="typeName">Name of the type to create.</param>
		/// <returns></returns>
		public static Object CreateInstance(string assemblyName, string typeName)
		{
			Assembly assembly = null;

			try
			{
				if (System.IO.Path.GetDirectoryName(assemblyName) == "")
				{
					assembly = 
						System.Reflection.Assembly.LoadWithPartialName(assemblyName);	
				}
				else
				{
					assembly = 
						System.Reflection.Assembly.LoadFile(assemblyName);	
				}
			}
			catch(Exception ex)
			{
				MyTrace.Log(TraceLevel.Error, 
					"Failed to create type '{0}' because assembly load failed for '{1}'.",
					typeName,
					assemblyName);
				MyTrace.Log(TraceLevel.Error, ex.ToString());
				return false;
			}

			if (assembly == null)
			{
				MyTrace.Log(TraceLevel.Error, 
					"Failed to create type '{0}' because assembly load failed for '{1}' (unknown reason).",
					typeName,
					assemblyName);
				return null;
			}

			// Create an instance of the type using the default constructor
			return assembly.CreateInstance(typeName);
		}

		/// <summary>
		/// Parses a C# source file into an ANTLR IGraph abstract syntax tree.
		/// </summary>
		/// <param name="path">Path to source file.</param>
		/// <returns>ANTLR IGraph abstract syntax tree.</returns>
		public IGraph GetGraph(string path)
		{
			if (path.IndexOf("\\Node.cs") >= 0 ||
				path.IndexOf("\\dx") >= 0 ||
				path.IndexOf("\\Collision.cs") >= 0 ||
				path.IndexOf("\\DamagableNode.cs") >= 0 ||
				path.IndexOf("\\Bounds.cs") >= 0 ||
				path.IndexOf("\\MyMeshContainer.cs") >= 0 ||
				path.IndexOf("\\Engine.cs") >= 0 ||
				path.IndexOf("\\InterSense.cs") >= 0 ||
				path.IndexOf("\\InterSenseSocket.cs") >= 0 ||
				path.IndexOf("\\InputDeviceMapper.cs") >= 0 ||
				path.IndexOf("\\MyUtility.cs") >= 0 ||
				path.IndexOf("\\CameraNode.cs") >= 0 ||
				path.IndexOf("\\Transformer.cs") >= 0 ||
				path.IndexOf("\\CombinerTransformer.cs") >= 0 ||
				path.IndexOf("\\ChaseAITransformer.cs") >= 0 ||
				path.IndexOf("\\JGFInstrumentor.cs") >= 0 ||
				path.IndexOf("\\JGFTimer.cs") >= 0 ||
				path.IndexOf("\\JGFSerialBench.cs") >= 0 ||
				path.IndexOf("\\JGFArithBench.cs") >= 0 ||
				path.IndexOf("\\JGFMathBench.cs") >= 0)
			{
				MyTrace.Log(TraceLevel.Warning, "Skipping {0}.", path);

				return null;
			}

			FileStream stream = null;
			
			try
			{
				stream = new FileStream(path, FileMode.Open, FileAccess.Read);
			}
			catch(Exception ex)
			{
				MyTrace.Log(TraceLevel.Warning, "Failed to parse source file. {0}",
					ex.Message);
				return null;
			}

			// To fix a bug that occurs when parsing multiple source files, I added
			// the resetStatics() method which must be called before parsing each
			// source file.
			CSharpAST.resetStatics();

			CSharpLexer lexer = new CSharpLexer(stream);
			lexer.setFilename(path);

			CSharpParser parser = new CSharpParser(lexer);
			parser.setFilename(path);

			bErrorReported = false;

			lexer.WarningReported += new antlr.debug.MessageEventHandler(ErrorReported);
			lexer.ErrorReported += new antlr.debug.MessageEventHandler(ErrorReported);
			parser.MatchedNotToken += new antlr.debug.MatchEventHandler(MisMatchedToken);
			parser.MisMatchedNotToken += new antlr.debug.MatchEventHandler(MisMatchedToken);
			parser.MisMatchedToken += new antlr.debug.MatchEventHandler(MisMatchedToken); 
			parser.WarningReported += new antlr.debug.MessageEventHandler(ErrorReported); 
			parser.ErrorReported += new antlr.debug.MessageEventHandler(ErrorReported); 
			
			// Parse the C# source file
			parser.compilation_unit();

			stream.Close();

			CSharpAST antlrTree = (CSharpAST) parser.getAST();
			antlrTree.FileName = path;

			return bErrorReported ? null : antlrTree.GetGraph();
		}

		/// <summary>
		/// Calls the getter method for a Property of type string via Reflection.
		/// </summary>
		/// <param name="ig">IGraph element whose Property we want to retrieve.</param>
		/// <param name="propertyName">Name of the Property. Note: Property must be of type string.</param>
		/// <returns>Value returned by the Property getter method.</returns>
		static public string ReflectGetProperty(object ig, string propertyName)
		{
			if (ig == null)
			{
				return "";
			}

			Type type = ig.GetType();

			MemberInfo[] props = type.GetMember(propertyName, MemberTypes.Property, BindingFlags.Public | BindingFlags.Instance);
			if (props.Length == 1)
			{
				object objName = type.InvokeMember(	props[0].Name, 
						BindingFlags.Public | BindingFlags.GetProperty |  BindingFlags.Instance, 
						null, 
						ig, 
						new object[]{} );

				if (objName != null && 
					objName is string)
				{
					return (string) objName;
				}
			}

			return null;
		}

		private void ErrorReported(object sender, antlr.debug.MessageEventArgs e)
		{
			bErrorReported = true;

			MyTrace.Log(TraceLevel.Error, "Failed to parse source file 2: {0}",
				e.ToString());
		}

		private void MisMatchedToken(object sender, antlr.debug.MatchEventArgs e)
		{
			bErrorReported = true;

			MyTrace.Log(TraceLevel.Error, "Failed to parse source file 4: {0}",
				e.ToString());
		}
	}
}
