///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/EnCProcess.cs 11    12/16/05 8:43p Eaddy $
//
//    File name:	EnCProcess.cs
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/EnCProcess.cs $
 * 
 * 11    12/16/05 8:43p Eaddy
 * o Added a new command-line test
 * 
 * 10    12/10/05 12:43a Eaddy
 * o Now keeps track of elapsed times
 * 
 * 9     12/07/05 12:40a Eaddy
 * o Changed test params to test Stack use case
 * 
 * 8     10/12/05 5:42p Eaddy
 * o Now verifies that the user doesn't accidentally run Watcher with
 * Cor_Enable_Profiling set to 1!
 * 
 * 7     10/07/05 8:01p Eaddy
 * o Now catches exception when config file is missing
 * 
 * 6     5/28/05 1:23p Eaddy
 * o Changed copyright
 * 
 * 5     5/25/05 6:34p Eaddy
 * o Now uses EnC.Watcher namespace
 * 
 * 4     4/04/05 2:23p Eaddy
 * o Changed method name
 * 
 * 3     4/01/05 11:15a Eaddy
 * o Now supports rudimentary weaving
 * 
 * 2     3/22/05 6:08p Eaddy
 * o Now loads supported languages from app.config
 * Revision 1.7  2004/11/01 07:38:18  me133
 * o Fixed bug with heartbeat message
 *
 * Revision 1.6  2004/10/26 04:46:05  me133
 * o Doesn't check for a heartbeat if PATCHER_BREAK is true since this
 * probably means the heartbeat would be paused
 *
 * Revision 1.5  2004/10/17 08:45:25  me133
 * o Now uses DynModule which keeps track of all the compiler settings
 * and references (it gets these from Patcher)
 *
 * Revision 1.4  2004/10/11 05:57:53  me133
 * no message
 *
 * Revision 1.3  2004/10/10 07:40:06  me133
 * o Now outputs trace messages to log file usnig System.Diagnostics.Trace
 *
 * Revision 1.2  2004/10/10 06:07:24  me133
 * o Added header comment
 */

using System;
using System.Diagnostics;	// Process
using System.Collections;	// ArrayList
using System.IO;			// Path

namespace EnC.Watcher
{
	/// <summary>
	/// Summary description for EnCProcess.
	/// </summary>
	public class EnCProcess
	{
		private const string PATCHER_PROGID = "Patcher.Patcher";

		private ArrayList			_listWatchers;
		private EnCSocketServer		_encSocketServer;

		public EnCProcess(int nPortToUse)
		{ 
			bool bUseHeartbeat = true;

			if (IsEnvVarTrue("PATCHER_BREAK"))
				bUseHeartbeat = false;

			_listWatchers = new ArrayList();
			_encSocketServer = new EnCSocketServer(	nPortToUse, 
													bUseHeartbeat,
													new ModuleDelegate(OnGotModuleInfo));
			_encSocketServer.StartListening();
		}

		public void WaitUntilShutdownReceived()
		{
			_encSocketServer.WaitForShutdown();
		}

		public bool StartProcess(string exeName, string exeArgs)
		{
			Process process = new Process();

			process.StartInfo = new ProcessStartInfo( exeName, exeArgs );
			
			// Initialize environment variables
			
			// Required by the CLR in order to load our Profiler DLL.
			process.StartInfo.EnvironmentVariables[ "COR_ENABLE_PROFILING" ] = "0x1";
			process.StartInfo.EnvironmentVariables[ "COR_PROFILER" ] = PATCHER_PROGID;

			// Required by Rotor - NEED TO FIX THIS!!
			process.StartInfo.EnvironmentVariables[ "COR_PROFILER_DLL" ] = 
				@"C:\PROJECTS\Goblin\Patcher\debug\Patcher.dll";

			process.StartInfo.UseShellExecute = false;
			process.StartInfo.Arguments = exeArgs;
			//process.StartInfo.WorkingDirectory = pi.WorkingDirectory;
			process.EnableRaisingEvents = true;
			process.Exited += new EventHandler( OnProcessExited );

			bool bStart = process.Start();
			process.WaitForExit();

			_encSocketServer.StopAndWait();

			return true;
		}

		// Called by EnCSocketServer
		private void OnGotModuleInfo(DynModule dynModule)
		{
			dynModule.setModuleUpdatedCallback(
				new OnModuleUpdatedDelegate(_encSocketServer.ModuleUpdatedCallback));

			_listWatchers.Add(dynModule);
		}

		// Called by Process
		private void OnProcessExited( object oSender, EventArgs ea )
		{
			MyTrace.Log(TraceLevel.Info, "Process has exited");
		}

		static bool IsEnvVarTrue(string name)
		{
			return GetBoolEnvVar(name, false);
		}

		static bool GetBoolEnvVar(string name, bool defaultVal)
		{
			string val = (string) System.Environment.GetEnvironmentVariables()[name];

			MyTrace.Log(TraceLevel.Info, "Environment variable: {0}='{1}'", name, val);

			if (val != null && val.Length > 0)
			{
				if (val == "1" ||
					val.ToLower() == "true")
				{
					return true;
				}
				else if (val == "0" ||
					val.ToLower() == "false")
				{
					return false;
				}
				else
				{
					return defaultVal;
				}
			}
			else
			{
				return defaultVal;
			}
		}

		[STAThread]
		static void Main(string[] args)
		{
			int nPortToUse = 0;
			string exeName = null;

			// You have to modify Watcher.exe.config in order to change the log level.
			// Log messages are appended to Watcher.log.

			MyTrace.Log(TraceLevel.Info, "Watcher.exe started");

			// Load the supported CodeDomProviders for C#, VB.NET, etc. from the
			// configuration file
			try
			{
				System.Configuration.ConfigurationSettings.GetConfig("supportedLanguages/language");
			}
			catch(System.Configuration.ConfigurationException ex)
			{
				MyTrace.Log(TraceLevel.Error, 
					"Failed to load the list of supported languages:\r\n" + 
					ex);
			}

			if (IsEnvVarTrue("Cor_Enable_Profiling"))
			{
				string msg = "Profiling is enabled for Watcher! You should set Cor_Enable_Profiling=0.";

				MyTrace.Log(TraceLevel.Warning, msg);
				Console.WriteLine("WARNING: " + msg);
			}

			bool bTestDynModule = false;

			for(int i = 0; i < args.Length; ++i)
			{
				MyTrace.Log(TraceLevel.Info, "Arg[{0}]='{1}'", i, args[i]);

				if (args[i].ToLower() == "DynModule".ToLower())
				{
					bTestDynModule = true;
					continue;
				}

				if (args[i].IndexOf(".cs") >= 0)
				{
					Helper helper = new Helper();
					helper.GetGraph(args[i]);
					return;
				}
			}

			if (bTestDynModule)
			{
				DynModule dynModule = new DynModule();
				dynModule.init(	0, 
								@"C:\PROJECTS\Goblin\Patcher\Stack.exe",
								"csc.exe",
								@"/target:exe /d:TRACE /d:DEBUG /unsafe /debug /nologo",
								true,
								new string[] { "Stack.cs", "StackDBCAspect.cs", "StackDriver.cs" } );

				dynModule.weaveAndCompile(Environment.TickCount);

				Console.WriteLine("Press ENTER to quit.");
				Console.ReadLine();
				return;
			}

			if (args.Length > 0)
			{
				try
				{
					nPortToUse = Int32.Parse(args[0].Trim());
				}
				catch
				{
					exeName = args[0];
				}
			}

			if (nPortToUse >= 0 || exeName == null)
			{
				MyTrace.Log(TraceLevel.Info, "Running in listen mode. Listening for file changes.");

				EnCProcess encProcess = new EnCProcess(nPortToUse);
				encProcess.WaitUntilShutdownReceived();
				
				MyTrace.Log(TraceLevel.Info, "Exiting listen mode");
			}
			else
			{
				string exeArgs = "";
				for(int i = 1; i < args.Length; ++i)
				{
					if (i > 1)
					{
						exeArgs += " ";
					}

					exeArgs += args[i];
				}

				MyTrace.Log(TraceLevel.Info, "Starting process: " + exeName + " " + exeArgs);

				EnCProcess encProcess = new EnCProcess(0);
				encProcess.StartProcess(exeName, exeArgs);

				MyTrace.Log(TraceLevel.Info, "Process Exited: " + exeName + " " + exeArgs);
			}
		}
	}
}
