///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/HeartbeatWorker.cs 3     5/28/05 1:22p Eaddy $
//
//    File name:	HeartbeatWorker.cs
//   Created By:	Marc Eaddy
//   Created on:	10/22/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/HeartbeatWorker.cs $
 * 
 * 3     5/28/05 1:22p Eaddy
 * o Changed copyright
 * 
 * 2     5/25/05 6:34p Eaddy
 * o Now uses EnC.Watcher namespace
 * Revision 1.4  2005/01/14 23:33:51  me133
 * o Better log messages
 *
 * Revision 1.3  2004/11/01 07:38:19  me133
 * o Fixed bug with heartbeat message
 *
 * Revision 1.2  2004/10/26 04:46:05  me133
 * o Doesn't check for a heartbeat if PATCHER_BREAK is true since this
 * probably means the heartbeat would be paused
 *
 * Revision 1.1  2004/10/24 20:32:23  me133
 * no message
 */

using System;
using System.IO;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;

namespace EnC.Watcher
{
	public delegate void ShutdownDelegate();

	class HeartbeatWorker : SocketWorker
	{
		ShutdownDelegate _shutdownCallback = null;
		NetworkStream _networkStream = null;
		BinaryWriter _writer = null;

		public HeartbeatWorker()
			: base("HeartbeatWorker")
		{ }

		public void StartProcessing2(	Socket socket, 
										ShutdownDelegate shutdownCallback)
		{
			// Verify args
			Debug.Assert(shutdownCallback != null);

			// Only call StartProcessing() once!
			Debug.Assert(_shutdownCallback == null);

			_shutdownCallback = shutdownCallback;

			// Heartbeat messages are 1 second apart. If we don't receive
			// a heartbeat after 3 seconds we give up.
			const int nSendReceiveTimeout = 3000; // 3 seconds
			socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.ReceiveTimeout, nSendReceiveTimeout);
			socket.SetSocketOption(SocketOptionLevel.Socket, SocketOptionName.SendTimeout, nSendReceiveTimeout);

			_networkStream = new NetworkStream(socket);
			_writer = new BinaryWriter(_networkStream);

			StartProcessing(socket);
		}

		override protected void ThreadFuncInnerLoop()
		{
			try
			{
				Debug.Assert(_writer != null);
				// Respond with a heartbeat response
				_writer.Write((int) Globals.NetworkMessage.nTHUMP_THUMP);

				// Wait for the next heartbeat (or shutdown) message
				uint nPatcherRequest = Globals.ReadUINT32(_socket);
				if (nPatcherRequest == (int) Globals.NetworkMessage.nSHUTDOWN)
				{
					// Shutdown requested
					MyTrace.Log(TraceLevel.Info, "HeartbeatWorker - Shutdown requested from Patcher");
					_bShutdown = true;
					_shutdownCallback.BeginInvoke(null, null);
				}
				else if (nPatcherRequest != (int) Globals.NetworkMessage.nTHUMP_THUMP)
				{
					MyTrace.Log(TraceLevel.Error, "HeartbeatWorker - Unexpected message from Patcher: {0}", 
						nPatcherRequest);
				}
			}
			catch (SocketException ex)
			{
				if (ex.Message.IndexOf("forcibly") > -1)
				{
					// Patcher died (CTRL^C)
					MyTrace.Log(TraceLevel.Info, 
						"HeartbeatWorker - Patcher has left the building. Exiting...");
				}
				else if (ex.Message.IndexOf("period of time") > -1)
				{
					// Shutdown requested
					MyTrace.Log(TraceLevel.Info, 
						"HeartbeatWorker - Patcher flatlined (no heartbeat message). Exiting...");
				}
				else
				{
					MyTrace.Log(TraceLevel.Info, "{0} - Socket was closed: {1}", 
						_strDiagnosticName, ex.ToString());
				}

				_bShutdown = true;
				_shutdownCallback.BeginInvoke(null, null);
			}
		}

		override public void RequestStop()
		{
			_writer.Close();
			_writer = null;
			base.RequestStop();
		}
	}
}
