///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/ModuleSenderWorker.cs 5     12/15/05 2:41a Eaddy $
//
//    File name:	ModuleSenderWorker.cs
//   Created By:	Marc Eaddy
//   Created on:	10/22/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/ModuleSenderWorker.cs $
 * 
 * 5     12/15/05 2:41a Eaddy
 * o Fixed stats
 * 
 * 4     12/10/05 12:43a Eaddy
 * o Now keeps track of elapsed times
 * 
 * 3     5/28/05 1:22p Eaddy
 * o Changed copyright
 * 
 * 2     5/25/05 6:34p Eaddy
 * o Now uses EnC.Watcher namespace
 * Revision 1.1  2004/10/24 20:32:23  me133
 * no message
 *
 */

using System;
using System.IO;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using System.Collections;

namespace EnC.Watcher
{
	class ModuleSenderWorker : SocketWorker
	{
		ArrayList _aryModuleInfo = new ArrayList();
		ManualResetEvent _mreDataAvailable = new ManualResetEvent(false);
		BinaryWriter _writer = null;

		public ModuleSenderWorker()
			: base("ModuleSenderWorker")
		{ }

		/// <summary>
		/// Creates a thread that executes the ThreadFunc that is
		/// overridden by the subclasses
		/// </summary>
		/// <param name="socket"></param>
		override public void StartProcessing(Socket socket)
		{
			base.StartProcessing(socket);
			NetworkStream networkStream = new NetworkStream(_socket);
			_writer = new BinaryWriter(networkStream);
		}

		public void Send(ModuleAndPath moduleAndPath)
		{
			lock (_aryModuleInfo.SyncRoot)
			{
				_aryModuleInfo.Add(moduleAndPath);
			}

			_mreDataAvailable.Set();
		}

		override protected void ThreadFuncInnerLoop()
		{
			Debug.Assert(_socket != null);

			_mreDataAvailable.WaitOne();
			if (_bShutdown) 
				return;

			ModuleAndPath moduleAndPath = new ModuleAndPath();
			bool bGotData = false;
			lock (_aryModuleInfo.SyncRoot)
			{
				if (_aryModuleInfo.Count == 0)
				{
					_mreDataAvailable.Reset();
					bGotData = false;
				}
				else
				{
					moduleAndPath = (ModuleAndPath) _aryModuleInfo[0];
					_aryModuleInfo.RemoveAt(0);
					bGotData = true;
				}
			}

			//Debug.Assert(bGotData);
			if (!bGotData) 
				return;

			MyTrace.Log(TraceLevel.Info, "WATCHER Sending {0} {1} Time:{2}.", 
				moduleAndPath.moduleID, 
				moduleAndPath.modulePath, 
				moduleAndPath.startTime);

			_writer.Write(moduleAndPath.moduleID);
			_writer.Write(moduleAndPath.modulePath);
			_writer.Write(moduleAndPath.startTime);
			_writer.Write(moduleAndPath.weaveTimeEnd);
			_writer.Write(moduleAndPath.compileTimeEnd);

			MyTrace.Log(TraceLevel.Verbose, "Total elapsed time: {0}ms.",
				(Environment.TickCount - moduleAndPath.startTime));

		}

		override public void RequestStop()
		{
			_writer.Close();
			base.RequestStop();
			_mreDataAvailable.Set();
		}
	}
}
