///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/MyTrace.cs 3     5/28/05 1:21p Eaddy $
//
//    File name:	MyTrace.cs
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/MyTrace.cs $
 * 
 * 3     5/28/05 1:21p Eaddy
 * o Changed copyright
 * 
 * 2     5/25/05 6:34p Eaddy
 * o Now uses EnC.Watcher namespace
 * Revision 1.5  2005/01/14 23:34:19  me133
 * o Now displays milliseconds
 *
 * Revision 1.4  2004/10/24 19:31:03  me133
 * o Implemented heartbeat message
 * o Now handles more network error conditions
 *
 * Revision 1.3  2004/10/17 08:43:38  me133
 * o Now logs messages to the debug console
 *
 * Revision 1.2  2004/10/11 05:57:54  me133
 * no message
 *
 * Revision 1.1  2004/10/10 07:40:07  me133
 * o Now outputs trace messages to log file usnig System.Diagnostics.Trace
 */

using System;
using System.Diagnostics;
using System.Threading;

using System.Runtime.InteropServices;
using System.Security.Permissions;
using System.Collections;
using System.IO;
using System.Text;

namespace EnC.Watcher
{
	/// <summary>
	/// Summary description for MyTrace.
	/// </summary>
	//[System.Security.SuppressUnmanagedCodeSecurityAttribute()]
	public class MyTrace
	{
		[DllImport("kernel32.dll", CharSet=System.Runtime.InteropServices.CharSet.Auto)]
		public static extern void OutputDebugString(String message);
	      
		static bool bIsFirstMessage = true;
		static DateTime date = new DateTime();

		public static TraceSwitch level = new TraceSwitch("level", "Entire Application");

		public static void Log(TraceLevel levelMessage, string format, params object[] args)
		{
			if (levelMessage > level.Level)
				return;

			Log(levelMessage, String.Format(format, args));
		}

		public static void Log(TraceLevel levelMessage, string msg)
		{
			if (levelMessage > level.Level)
				return;

			if (bIsFirstMessage)
			{
				string strFileName = "Watcher.log";

				Trace.WriteLine("");
				Trace.WriteLine(String.Format("**  Started {0} on {1} **", strFileName, GetDate()) );
				Trace.WriteLine("**  Log File Format Version: 2.2 **");
			}

			// 10/10 20:56:01  3028 - INFO - Watcher: blah blah blah.

			string strMsg = String.Format("{0}  {1} - {2} - Watcher: {3}",
				GetDate(), 
				Thread.CurrentThread.GetHashCode(),
				GetSeverity(levelMessage),
				msg);

			Trace.WriteLine(strMsg);
			OutputDebugString(strMsg + "\n");

			bIsFirstMessage = false;
		}

		static string GetDate()
		{
			return DateTime.Now.ToString("MM/dd HH:mm:ss:fff");
		}

		static string GetSeverity(TraceLevel levelMessage)
		{
			if (levelMessage == TraceLevel.Verbose)
				return "SUCCESS";
			else if (levelMessage == TraceLevel.Info)
				return "INFO";
			else if (levelMessage == TraceLevel.Warning)
				return "WARNING";
			else if (levelMessage == TraceLevel.Error)
				return "ERROR";
			else
			{
				Debug.Assert(false, "Not supposed to happen");
				return levelMessage.ToString();
			}
		}
	}
}
