///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Watcher/SocketWorker.cs 3     5/28/05 1:21p Eaddy $
//
//    File name:	SocketWorker.cs
//   Created By:	Marc Eaddy
//   Created on:	10/22/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Watcher/SocketWorker.cs $
 * 
 * 3     5/28/05 1:21p Eaddy
 * o Changed copyright
 * 
 * 2     5/25/05 6:34p Eaddy
 * o Now uses EnC.Watcher namespace
 * Revision 1.2  2005/01/14 23:34:54  me133
 * o Better log messages
 *
 * Revision 1.1  2004/10/24 20:32:24  me133
 * no message
 */

using System;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;

namespace EnC.Watcher
{
	abstract class SocketWorker
	{
		// DATA
		protected Socket	_socket				= null;
		private Thread		_thread				= null;
		protected bool		_bShutdown			= false;
		protected string	_strDiagnosticName	= null;

		public SocketWorker(string strDiagnosticName)
		{ 
			_strDiagnosticName = strDiagnosticName;
		}

		/// <returns>true if StartProcessing() has been called.</returns>
		public bool IsStarted()
		{
			return _socket != null && _thread != null;
		}

		/// <summary>
		/// Creates a thread that executes the ThreadFunc that is
		/// overridden by the subclasses
		/// </summary>
		/// <param name="socket"></param>
		virtual public void StartProcessing(Socket socket)
		{
			// Pre-condition: Only call StartProcessing() once
			Debug.Assert(!IsStarted());
			_socket = socket;

			_thread = new Thread(new ThreadStart(ThreadFunc));
			_thread.Start();

			// Post-condition: We are started
			Debug.Assert(IsStarted());
		}

		/// <summary>
		/// Polite request for the thread to stop. The thread might not
		/// respond immediately.
		/// </summary>
		virtual public void RequestStop()
		{
			_bShutdown = true;
		}

		/// <summary>
		/// Requests for the thread to stop and waits for it to die.
		/// </summary>
		/// <param name="timeout">Max time to wait</param>
		public void WaitForThreadToDie(TimeSpan timeout)
		{
			RequestStop();
			Debug.Assert(_thread != null);
			if (_thread != null)
			{
				_thread.Join(timeout);

				if (_thread.IsAlive)
				{
					_thread.Abort(	_strDiagnosticName +
						" thread got sick and tired of waiting for you for '" + 
						timeout.ToString() + "'.");
					
					// The Join should return immeidately; Wait a few measly milliseconds
					_thread.Join(new TimeSpan(1000));
				}
			}
		}

		/// <summary>
		/// Calls ThreadFuncInnerLoop continuously until shutdown has been signaled.
		/// </summary>
		void ThreadFunc()
		{
			Thread.CurrentThread.Name = _strDiagnosticName;

			while (!_bShutdown)
			{
				if (_strDiagnosticName != "HeartbeatWorker")
				{
					MyTrace.Log(TraceLevel.Info, "{0} thread waiting for data",
						_strDiagnosticName);
				}

				try
				{
					ThreadFuncInnerLoop();
				}
				catch(SocketException ex)
				{
					// The help docs say that ex.ErrorCode will be the socket error code
					// but its wrong. The value should be WSAECONNRESET but instead its E_FAIL.
					if (ex.Message.IndexOf("forcibly") > -1)
					{
						// Patcher died (CTRL^C)
						MyTrace.Log(TraceLevel.Info, 
							"HeartbeatWorker - Patcher has left the building. Exiting...");
					}
					else
					{
						MyTrace.Log(TraceLevel.Info, "{0} - Socket was closed: {1}", 
							_strDiagnosticName, ex.ToString());
					}

					break;
				}
				catch(ThreadAbortException ex)
				{
					MyTrace.Log(TraceLevel.Error, "{0} - Thread was aborted: {1}", 
						_strDiagnosticName, ex.ExceptionState.ToString());
					break;
				}
				catch(Exception ex)
				{
					MyTrace.Log(TraceLevel.Error, "{0} socket error: {1}", 
						_strDiagnosticName, ex.ToString());
					break;
				}
			}

			MyTrace.Log(TraceLevel.Info, "{0} thread closing socket", _strDiagnosticName);

			_socket.Close();
		}

		/// <summary>
		/// Derived classes override this to perform whatever network
		/// communication they need on the socket.
		/// </summary>
		abstract protected void ThreadFuncInnerLoop();
	}
}
